(** Areas of different forms to describe shape of the torch *)
<* O2EXTENSIONS + *>
MODULE Area;

IMPORT
  Math:=mMath, SU:=msysUtils, COMPILER, Protocol;

CONST
  ModuleName = COMPILER.EQUATION('module');

  LimitOfDenominator = 1.0E-10;

TYPE
  Float = Math.Float;

  AreaDesc * = RECORD
  END;
  Area * = POINTER TO AreaDesc;

PROCEDURE (a: Area) Inside * (z, r: Float): BOOLEAN;
  (** Returns if the specified point is in the area *)
  BEGIN
    SU.RAISEError(ModuleName+'.Area.Inside: abstract method');
    RETURN z>r
  END Inside;

(* ============= Unions ============================ *)

TYPE
  UnionArea = POINTER TO RECORD (AreaDesc)
    a1,a2:Area;
  END;

PROCEDURE (a:UnionArea) Inside (z,r: Float): BOOLEAN;
  BEGIN
    RETURN a.a1.Inside(z,r) OR a.a2.Inside(z,r)
  END Inside;

PROCEDURE Union * (a1,a2:Area): Area;
  (** Returns union of the specified areas *)
  VAR
    u:UnionArea;
  BEGIN
    IF a1=NIL THEN
      RETURN a2
    ELSIF a2=NIL THEN
      RETURN a1
    ELSE
      NEW(u);
      u.a1:=a1;
      u.a2:=a2;
      RETURN u
    END;
  END Union;

(* ============= Miscellaneous procedures ========== *)

PROCEDURE sqr (x: Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x: Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

PROCEDURE SwapReal (VAR x, y: Float);
  VAR
    tmp: Float;
  BEGIN
    tmp:=x;
    x:=y;
    y:=tmp;
  END SwapReal;

PROCEDURE Min (x, y: Float): Float;
  BEGIN
    IF x<y THEN
      RETURN x
    ELSE
      RETURN y
    END;
  END Min;

PROCEDURE Max (x, y: Float): Float;
  BEGIN
    IF x<y THEN
      RETURN y
    ELSE
      RETURN x
    END;
  END Max;

PROCEDURE OutCone (z1, r1, z2, r2, z, r: Float): BOOLEAN;
  (* Returns if the point (z,r) is outside the cone specified by (z1,r1),(z2,r2) *)
  BEGIN
    IF z2<z1 THEN
      SwapReal(z1,z2);
      SwapReal(r1,r2);
    END;
    RETURN (r-r1)*(z2-z1)>=(z-z1)*(r2-r1)
  END OutCone;

PROCEDURE InCone (z1, r1, z2, r2, z, r: Float): BOOLEAN;
  (* Returns if the point (z,r) is inside the cone specified by (z1,r1),(z2,r2) *)
  BEGIN
    IF z2<z1 THEN
      SwapReal(z1,z2);
      SwapReal(r1,r2);
    END;
    RETURN (r-r1)*(z2-z1)<=(z-z1)*(r2-r1)
  END InCone;

PROCEDURE InTorus (z0, r0, rad, z, r: Float): BOOLEAN;
  (* Returns if the point (z,r) is inside the torus specified by (z0,r0,rad) *)
  BEGIN
    RETURN sqr(z-z0)+sqr(r-r0)<=sqr(rad)
  END InTorus;

PROCEDURE OutTorus (z0,r0,rad,z,r:Float): BOOLEAN;
  (* Returns if the point (z,r) is outide the torus specified by (z0,r0,rad) *)
  BEGIN
    RETURN sqr(z-z0)+sqr(r-r0)>=sqr(rad)
  END OutTorus;

PROCEDURE GetCenterOfSector (z1,r1,z2,r2,rad:Float; VAR z0,r0:Float);
  (* Returns center of a sector between the points (z1,r1) and (z2,r2) with the
    specified radius. If an observer is moved from (z1,r1) to (z2,r2) than
    the sector is drawn at the right. If rad>0 then the circle center is at
    the left, if rad<0 then the circle center is at right *)
  VAR
    SqrRad,SqrDist,alpha:Float;
  BEGIN
    SqrDist:=sqr(z1-z2)+sqr(r1-r2);
    ASSERT(SqrDist#0);
    SqrRad:=sqr(rad);
    ASSERT(SqrDist<=4*SqrRad);

    alpha:=sqrt((SqrRad-SqrDist/4)/SqrDist);
    IF rad>0 THEN
      alpha:=-alpha;
    END;
    z0:=(z1+z2)/2 + alpha*(r2-r1);
    r0:=(r1+r2)/2 - alpha*(z2-z1);
  END GetCenterOfSector;

PROCEDURE Fillet (z1,r1,z2,r2,z3,r3,rad:Float;
                  VAR SectorZ1,SectorR1,SectorZ2,SectorR2:Float);
  (* (z1,r1), (z2,r2), and (z3,r3) are edges of two lines joint in the point (z2,r2).
    The procedure returns coordinates of initial and final point of a sector,
    which forms the fillet between the lines with the specified radius.
    (SectorZ1,SectorR1) is situated on the line (z1,r1)-(z2,r2).
    (SectorZ2,SectorR2) is situated on the line (z2,r2)-(z3,r3).
  *)
  VAR
    dist12,dist23,denom,nz,nr,d,invcos2b:Float;
  BEGIN
    ASSERT(rad>=0);

    denom:=(r2-r1)*(z3-z2)-(z2-z1)*(r3-r2);
    IF ABS(denom)<=LimitOfDenominator THEN (* the lines are parellel *)
      SectorZ1:=z2;
      SectorR1:=r2;
      SectorZ2:=z2;
      SectorR2:=r2;
      RETURN
    END;
    dist12:=sqrt(sqr(z1-z2)+sqr(r1-r2));
    ASSERT(dist12#0);
    dist23:=sqrt(sqr(z2-z3)+sqr(r2-r3));
    ASSERT(dist23#0);

    nz:= dist12*(z3-z2) - dist23*(z2-z1);
    nr:= dist12*(r3-r2) - dist23*(r2-r1);
    invcos2b:=(sqr(nz)+sqr(nr))/sqr(denom);
    d:=rad*sqrt(invcos2b-1);

    SectorZ1:=z2+d*(z1-z2)/dist12;
    SectorR1:=r2+d*(r1-r2)/dist12;
    SectorZ2:=z2+d*(z3-z2)/dist23;
    SectorR2:=r2+d*(r3-r2)/dist23;
  END Fillet;

(* =============== Cylinders ================== *)

TYPE
  HollowCylinderAreaDesc = RECORD (AreaDesc)
    MinZ, MaxZ, MinR, MaxR: Float;
  END;
  HollowCylinderArea = POINTER TO HollowCylinderAreaDesc;

PROCEDURE (a:HollowCylinderArea) Inside (z,r: Float): BOOLEAN;
  BEGIN
    RETURN (z>=a.MinZ) & (z<=a.MaxZ) & (r<=a.MaxR) & (r>=a.MinR)
  END Inside;

PROCEDURE HollowCylinder * (MinZ, MinR, MaxZ, MaxR: Float): Area;
  (** Returns the specified hollow cylinder *)
  VAR
    a: HollowCylinderArea;
  BEGIN
    IF (MinZ=MaxZ) & (MinR=MaxR) THEN
      RETURN NIL
    END;

    IF MinZ>MaxZ THEN
      SwapReal(MinZ,MaxZ);
    END;
    IF MinR>MaxR THEN
      SwapReal(MinR,MaxR);
    END;
    NEW(a);
    a.MinZ:=MinZ;
    a.MaxZ:=MaxZ;
    a.MinR:=MinR;
    a.MaxR:=MaxR;
    RETURN a
  END HollowCylinder;

PROCEDURE Cylinder * (MinZ, MaxZ, Rad: Float): Area;
  (** Returns the soecified cylinder *)
  BEGIN
    RETURN HollowCylinder(MinZ, 0, MaxZ, Rad)
  END Cylinder;

(* ============== cones ========== *)

TYPE
  HollowConeAreaDesc = RECORD (AreaDesc)
    z1, r1, z2, r2: Float;
  END;
  HollowConeArea = POINTER TO HollowConeAreaDesc;

PROCEDURE InitHollowCone (a: HollowConeArea; z1, r1, z2, r2: Float);
  BEGIN
    ASSERT(a#NIL);

    IF z2<z1 THEN
      SwapReal(z1, z2);
      SwapReal(r1, r2);
    END;
    a.z1:=z1;
    a.z2:=z2;
    a.r1:=r1;
    a.r2:=r2;
  END InitHollowCone;

(* ========== Internal cone ==========*)

TYPE
  IntConeArea = POINTER TO RECORD (HollowConeAreaDesc)
  END;

PROCEDURE (a: IntConeArea) Inside (z, r: Float): BOOLEAN;
  BEGIN
    RETURN (z>=a.z1) & (z<=a.z2) & (r>=Min(a.r1,a.r2)) & InCone(a.z1,a.r1,a.z2,a.r2,z,r);
  END Inside;

PROCEDURE IntCone * (z1, r1, z2, r2: Float): Area;
  (** Returns the internal hollow cone *)
  VAR
    a: IntConeArea;
  BEGIN
		IF (z1=z2) OR (r1=r2) THEN
			RETURN NIL
		END;

    NEW(a);
    InitHollowCone(a, z1, r1, z2, r2);
    RETURN a
  END IntCone;

(* =========== external hollow cone ============== *)

TYPE
  ExtConeArea = POINTER TO RECORD (HollowConeAreaDesc)
  END;

PROCEDURE (a: ExtConeArea) Inside (z, r: Float): BOOLEAN;
  BEGIN
    RETURN (z>=a.z1) & (z<=a.z2) & (r<=Max(a.r1,a.r2)) & OutCone(a.z1,a.r1,a.z2,a.r2,z,r);
  END Inside;

PROCEDURE ExtCone * (z1, r1, z2, r2: Float): Area;
  (** Returns the external hollow cone *)
  VAR
    a: ExtConeArea;
  BEGIN
		IF (z1=z2) OR (r1=r2) THEN
			RETURN NIL
		END;

    NEW(a);
    InitHollowCone(a, z1, r1, z2, r2);
    RETURN a
  END ExtCone;

(* ========== Torus ===========*)

TYPE
  TorusArea = POINTER TO RECORD (AreaDesc)
    z0, r0, rad: Float;
  END;

PROCEDURE (a: TorusArea) Inside (z, r: Float): BOOLEAN;
  BEGIN
    RETURN InTorus( a.z0, a.r0, a.rad, z, r )
  END Inside;

PROCEDURE Torus * (z0, r0, rad: Float): Area;
  (** Returns the specified torus *)
  VAR
    a:TorusArea;
  BEGIN
		ASSERT(rad>=0);
		IF rad=0 THEN
			RETURN NIL
		END;

    NEW(a);
    a.z0:=z0;
    a.r0:=r0;
    a.rad:=rad;
    RETURN a
  END Torus;

(* ========== IntSectorArea ===========*)

TYPE
  IntSectorArea = POINTER TO RECORD (AreaDesc)
    z0, r0, z1, r1, z2, r2, rad: Float;
  END;

PROCEDURE (a: IntSectorArea) Inside (z, r: Float): BOOLEAN;
  (* z0,r0,z1,r1,z2,r2,rad:Float; *)
  BEGIN
    RETURN InTorus(a.z0, a.r0, a.rad, z, r)
            & (
                ((a.z2>=a.z1) & InCone(a.z1,a.r1,a.z2,a.r2,z,r))
                OR ((a.z2<a.z1) & OutCone(a.z1,a.r1,a.z2,a.r2,z,r))
              )
  END Inside;

PROCEDURE IntSector * (z1, r1, z2, r2, rad: Float): Area;
  (** Returns an internal sector. For meaning of the parameters see
      the procedure GetCenterOfSector *)
  VAR
    a: IntSectorArea;
  BEGIN
		IF (z1=z2) & (r1=r2) THEN
			RETURN NIL;
		ELSE
			ASSERT(rad#0);
      NEW(a);
      a.z1:=z1;
      a.z2:=z2;
      a.r1:=r1;
      a.r2:=r2;
      a.rad:=rad;
      GetCenterOfSector(z1,r1,z2,r2,rad,a.z0,a.r0);
      RETURN a;
    END;
  END IntSector;

(* ========== SectorOutAreaLimited ===========*)

TYPE
  ExtSectorAreaLimited = POINTER TO RECORD (AreaDesc)
    z0, r0, z1, r1, z2, r2, rad: Float;
  END;

PROCEDURE (a: ExtSectorAreaLimited) Inside (z, r: Float): BOOLEAN;
  BEGIN
    RETURN OutTorus( a.z0, a.r0, a.rad, z, r ) & (z>=Min(a.z1, a.z2))
           & (z<=Max(a.z1, a.z2)) & (r>=Min(a.r1, a.r2)) & (r<=Max(a.r1, a.r2));
  END Inside;

PROCEDURE ExtSector * (z1,r1,z2,r2,rad:Float): Area;
  (** Returns an internal sector. For meaning of the parameters see
      the procedure GetCenterOfSector *)
  VAR
    a:ExtSectorAreaLimited;
  BEGIN
		IF (z1=z2) & (r1=r2) THEN
			RETURN NIL
		ELSE
			ASSERT(rad#0);
      NEW(a);
      a.z1:=z1;
      a.z2:=z2;
      a.r1:=r1;
      a.r2:=r2;
      a.rad:=rad;
      GetCenterOfSector(z1,r1,z2,r2,rad,a.z0,a.r0);
      RETURN a;
    END;
  END ExtSector;

(* ========== Fillets ===================================*)

PROCEDURE IntExtFillet (z1,r1,z2,r2,z3,r3,rad:Float;
                        VAR SectorZ1,SectorR1,SectorZ2,SectorR2:Float;
                        SectorProc:PROCEDURE (z1,r1,z2,r2,rad:Float): Area): Area;
	BEGIN
		Fillet(z1,r1,z2,r2,z3,r3,rad,SectorZ1,SectorR1,SectorZ2,SectorR2);
		IF (r2-r1)*(z3-z2)-(z2-z1)*(r3-r2) > 0 THEN
			RETURN SectorProc(SectorZ2,SectorR2,SectorZ1,SectorR1,rad);
		ELSE
			RETURN SectorProc(SectorZ1,SectorR1,SectorZ2,SectorR2,rad);
		END;
	END IntExtFillet;

PROCEDURE IntFillet * (z1,r1,z2,r2,z3,r3,rad:Float;
                       VAR SectorZ1,SectorR1,SectorZ2,SectorR2:Float): Area;
	(** Returns an internal sector, which constitutes fillet for the given lines.
      Also returns coordinates of the edges of the sector.
      For meaning of the parameters see Fillet *)
	BEGIN
		RETURN IntExtFillet(z1,r1,z2,r2,z3,r3,rad,SectorZ1,SectorR1,SectorZ2,SectorR2,IntSector)
	END IntFillet;

PROCEDURE ExtFillet * (z1,r1,z2,r2,z3,r3,rad:Float;
                       VAR SectorZ1,SectorR1,SectorZ2,SectorR2:Float): Area;
	(** Returns an external sector, which constitutes fillet for the given lines.
      Also returns coordinates of the edges of the sector.
      For meaning of the parameters see Fillet *)
	BEGIN
		RETURN IntExtFillet(z1,r1,z2,r2,z3,r3,rad,SectorZ1,SectorR1,SectorZ2,SectorR2,ExtSector)
	END ExtFillet;

PROCEDURE FilletToAxis * (z1,r1,z2,r2,z3,r3,rad:Float;
                          VAR SectorZ1,SectorR1,SectorZ2,SectorR2:Float): Area;
  (** Depending on the specified lines, returns internal or external sector,
      which constitutes fillet for the given lines. The sector is situated towards the axis.
      If this is an internal sector, the returnes area includes additionally the inner cone,
      which is complimentary to the sector: IntCone( SectorZ1, SectorR1, SectorZ2, SectorR2 ).
      Thus the lower edge of the returned area is ever the horisontal line.
      Also returns coordinates of the edges of the sector. For meaning of the parameters see Fillet *)
  VAR
    a1,a2:Area;
  BEGIN
    IF (r2-r1)*(z3-z2)-(z2-z1)*(r3-r2) > 0 THEN
      a1:=IntFillet (z1,r1,z2,r2,z3,r3,rad,SectorZ1,SectorR1,SectorZ2,SectorR2);
      a2:=IntCone (SectorZ1,SectorR1,SectorZ2,SectorR2);
      RETURN Union(a1,a2)
    ELSE
      RETURN ExtFillet (z1,r1,z2,r2,z3,r3,rad,SectorZ1,SectorR1,SectorZ2,SectorR2)
    END;
  END FilletToAxis;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');

  Protocol.Real('LimitOfDenominator = ',LimitOfDenominator,'');

  Protocol.String('','','');
END Area.