<* O2EXTENSIONS + *>
<* O2ADDKWD + *>
MODULE Comsol;

IMPORT
  Area:=mArea2D, Out:=msysTextOut, PHC:=msciPhConst, Math:=mMath, In:=msysTextIn,
  Protocol, DStr:=msysDynStr, SF:=mhostSearchFile, COMPILER, HU:=mhostUtils,
  Geo:=Geometry, ComFunc:=mCommonFunctions, InOut, GeoArea;

TYPE
  Float = Math.Float;
  Array2 = Math.Array2;
  Array2Ptr = Math.Array2Ptr;
  String = DStr.String;

CONST
  Tab = Out.Tab;
  ModuleName = COMPILER.EQUATION('module');

  (** All the output operations with COMSOL are restricted to this box: *)
  MinZ * = Geo.PlasmaMinZ;
  MaxZ * = Geo.PlasmaMaxZ;
  Rad * = Geo.PlasmaRad;

  AreaTolerance = 1.0E-4; (* cm - Tolerance of areas at load operations *)

  OutputFile * = 'output.mph';
  EnvStr = 'COMSOLBATCH'; (* This environment string should contain the file
                             name, with the full path, of the COMSOL batch program. *)
  StatusFileName * = OutputFile+'.status';

VAR
  InputFile:String; (* Input file name *)
  exe:String; (* File name of the COMSOL batch program *)
  CmdLine:String;

CONST
  separator = ' ';

PROCEDURE SaveZRange (VAR f:Out.Formatter; area-:Area.Area);
  VAR
    iZ:LONGINT;
    z:Float;
  BEGIN
    FOR iZ:=0 TO area.LastX DO
      z:=Area.IndToX(area,iZ);
      IF (z>=MinZ) & (z<=MaxZ) THEN
        Out.Real(f,z,0);
        Out.Char(f,separator);
      END;
    END;
  END SaveZRange;

PROCEDURE SaveRRange (VAR f:Out.Formatter; area-:Area.Area);
  VAR
    iR:LONGINT;
    r:Float;
  BEGIN
    FOR iR:=0 TO area.LastY DO
      r:=Area.IndToY(area,iR);
      IF r<=Rad THEN
        Out.Real(f,r,0);
        Out.Char(f,separator);
      END;
    END;
  END SaveRRange;

TYPE
  Func * = ComFunc.Float2;

PROCEDURE SaveFunc * (area-:Area.Area; VAR f:Func;
                      FileName-:ARRAY OF CHAR; factor:Float);
  (** Saves the given function in the given area to the specified file so that
      the file can be imported later by COMSOL. *)
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
    fm:Out.Formatter;
  BEGIN
    ASSERT(InOut.OutDir#NIL);

    Out.OpenFile(fm,FileName,Out.Rewrite);

    Out.String(fm,'% Grid',0);
    Out.Ln(fm);
    SaveZRange(fm,area);
    Out.Ln(fm);
    SaveRRange(fm,area);
    Out.Ln(fm);
    Out.String(fm,'% Data',0);
    Out.Ln(fm);


    FOR iR:=0 TO area.LastY DO
      r:=Area.IndToY(area,iR);
      IF r<=Rad THEN
        FOR iZ:=0 TO area.LastX DO
          z:=Area.IndToX(area,iZ);
          IF (z>=MinZ) & (z<=MaxZ) THEN
            Out.Real(fm,f.f(z,r)*factor,0);
            Out.Char(fm,separator);
          END;
        END;
        Out.Ln(fm)
      END;
    END;

    Out.Close(fm);

    InOut.CopyFile(FileName);
  END SaveFunc;

PROCEDURE SaveSimpleFunc * (area-:Area.Area; f:ComFunc.Simple2;
                            FileName-:ARRAY OF CHAR; factor:Float);
  (** Saves the given function in the given area to the specified file so that
      the file can be imported later by COMSOL. *)
  VAR
    func:POINTER TO Func;
  BEGIN
    ASSERT(f#NIL);

    func:=ComFunc.FromSimple2(f);
    SaveFunc(area,func^,FileName,factor);
  END SaveSimpleFunc;

TYPE
  AFunc = RECORD (Func)
    area:Area.Area;
    field:Array2Ptr;
  END;

PROCEDURE (VAR f:AFunc) f (z,r:Float): Float;
  BEGIN
    ASSERT(f.field#NIL);

    RETURN GeoArea.Interpolate(f.area,f.field^,z,r);
  END f;

PROCEDURE SaveArray * (area-:Area.Area; field:Array2Ptr;
                            FileName-:ARRAY OF CHAR; factor:Float);
  (** Saves the given function in the given area to the specified file so that
      the file can be imported later by COMSOL. *)
  VAR
    func:AFunc;
  BEGIN
    ASSERT(Area.Correct(area));
    ASSERT(field#NIL);
    ASSERT(Area.CorrectField(area,field^));

    func.area:=area;
    func.field:=field;
    SaveFunc(area,func,FileName,factor);
  END SaveArray;

PROCEDURE InFloat (VAR s:In.Scanner; VAR x:Float; default:Float);
  (* If input is NaN, then the result is default *)
  VAR
    c:CHAR;
  BEGIN
    In.SkipSpaces(s);
    IF In.NextChar(s,c) & (c='N') THEN (* the 1st letter of 'NaN' *)
      x:=default;
      In.SkipUntilSpaces(s);
    ELSE
      In.Float(s,x);
    END;
  END InFloat;

PROCEDURE ReadRange (VAR s:In.Scanner; VAR min,max:Float; VAR len:LONGINT);
  BEGIN
    In.SkipSpaces(s);
    In.Float(s,min);
    len:=1;
    In.SkipSpacesUntilEOL(s);
    WHILE ~In.EndOfLine(s) DO
      In.Float(s,max);
      INC(len);
      In.SkipSpacesUntilEOL(s);
    END;
    min:=min*PHC.meter;
    max:=max*PHC.meter;
  END ReadRange;

PROCEDURE SkipCommentedLines (VAR s:In.Scanner);
  VAR
    c:CHAR;
  BEGIN
    In.SkipSpaces(s);
    WHILE In.NextChar(s,c) & (c='%') DO
      In.SkipLine(s);
      In.SkipSpaces(s);
    END;
  END SkipCommentedLines;

PROCEDURE ReadArea (VAR s:In.Scanner; VAR a:Area.Area);
  VAR
    MinZ,MaxZ,MinR,MaxR:Float;
    LenZ,LenR:LONGINT;
  BEGIN
    ReadRange(s,MinR,MaxR,LenR);
    ReadRange(s,MinZ,MaxZ,LenZ);
    Area.Init(a,LenZ-1,LenR-1,MinZ,MaxZ,MinR,MaxR);
  END ReadArea;

PROCEDURE ReadDirect (VAR s:In.Scanner; a-:Area.Area; VAR field:Array2;
                      factor,default:Float);
  VAR
    iZ,iR:LONGINT;
  BEGIN
    ASSERT(LEN(field)=a.LastX+1);
    ASSERT(LEN(field,1)=a.LastY+1);

    FOR iZ:=0 TO a.LastX DO
      FOR iR:=0 TO a.LastY DO
        InFloat(s,field[iZ,iR],default);
        field[iZ,iR]:=field[iZ,iR]*factor
      END;
    END;
  END ReadDirect;

PROCEDURE ReadInterpolated (VAR s:In.Scanner; a-:Area.Area;
                            VAR field:Array2; factor,default:Float;
                            FieldArea-:Area.Area);
  VAR
    fld:Array2Ptr;
    iZ,iR:LONGINT;
    z,r:Float;
  BEGIN
    NEW(fld,FieldArea.LastX+1,FieldArea.LastY+1);
    ReadDirect(s,FieldArea,fld^,factor,default);
    FOR iZ:=0 TO a.LastX DO
      z:=Area.IndToX(a,iZ);
      FOR iR:=0 TO a.LastY DO
        r:=Area.IndToY(a,iR);
        IF   (z>=FieldArea.MinX) & (z<=FieldArea.MaxX)
           & (r>=FieldArea.MinY) & (r<=FieldArea.MaxY)
        THEN
          field[iZ,iR]:=GeoArea.Interpolate(FieldArea,fld^,z,r);
        ELSE
          field[iZ,iR]:=default
        END;
      END;
    END;
  END ReadInterpolated;

PROCEDURE LoadArray * (area-:Area.Area; VAR field:Array2;
                       FileName-:ARRAY OF CHAR; factor,default:Float);
  (** Read the data from the specified file in COMSOL format to the given field
      in the given area. If the data in a point are not successfully read,
      then the default value is assigned to the point *)
  VAR
    s:In.Scanner;
    FieldArea:Area.Area;
  BEGIN
    In.OpenFile(s,FileName);

    SkipCommentedLines(s);
    ReadArea(s,FieldArea);
    SkipCommentedLines(s);
    IF   (FieldArea.LastX=area.LastX) & (FieldArea.LastY=area.LastY)
       & (ABS(FieldArea.MinX-area.MinX)<=AreaTolerance)
       & (ABS(FieldArea.MaxX-area.MaxX)<=AreaTolerance)
       & (ABS(FieldArea.MinY-area.MinY)<=AreaTolerance)
       & (ABS(FieldArea.MaxY-area.MaxY)<=AreaTolerance)
    THEN
      ReadDirect(s,area,field,factor,default);
    ELSE
      ReadInterpolated(s,area,field,factor,default,FieldArea);
    END;

    In.Close(s);
  END LoadArray;

PROCEDURE LoadFunc * (area-:Area.Area;
                      FileName-:ARRAY OF CHAR;
                      factor,default:Float): POINTER TO GeoArea.InterpolationProc;
  (** Read the data from the specified file in COMSOL format in the given area
      and returns the function, which interpolated the read data in the area.
      If the data in a point are not successfully read, then the default value
      is assigned to the point *)
  VAR
    field:Array2Ptr;
  BEGIN
    ASSERT(Area.Correct(area));

    NEW(field,area.LastX+1,area.LastY+1);
    LoadArray(area,field^,FileName,factor,default);
    RETURN GeoArea.Interpolation(area,field);
  END LoadFunc;

PROCEDURE Success (): BOOLEAN;
  VAR
    sc:In.Scanner;
    s:DStr.String;
  BEGIN
    In.OpenFile(sc,StatusFileName);

    s:=NIL;
    In.SkipSpaces(sc);
    WHILE ~In.EndOfInput(sc) DO
      In.DString(sc,s);
      In.SkipSpaces(sc);
    END;

    In.Close(sc);

    RETURN (s#NIL) & (s^='Done');
  END Success;

VAR
  pname,plist:String;

PROCEDURE AddParameter * (name-:ARRAY OF CHAR; val:Float; dimension-:ARRAY OF CHAR);
  (** Adds a parameter to the COMSOL command line.
      Units should be in a square brackets, if any.
      Example: AddParameter('r0',Geo.r0/PHC.cm,'[cm]')
               AddParameter('factor',factor,'') *)
  BEGIN
    ASSERT(pname#NIL);
    ASSERT(plist#NIL);

    IF pname^='' THEN
      ASSERT(plist^='');

      pname:=DStr.Concat2(' -pname ',name);
      plist:=DStr.Concat3(' -plist ',DStr.FromReal(val)^,dimension);
    ELSE
      ASSERT(plist^#'');

      pname:=DStr.Concat3(pname^,',',name);
      plist:=DStr.Concat4(plist^,',',DStr.FromReal(val)^,dimension);
    END;
  END AddParameter;

PROCEDURE Launch * (dir-:ARRAY OF CHAR);
  (** Launches the COMSOL and waits until it is finished.
      Dir is a current directory for the COMSOL. *)
  BEGIN
    IF pname#NIL THEN
      ASSERT(plist#NIL);
      Protocol.String('  COMSOL command line parameters:',pname^,plist^);
    END;
    IF ~HU.ExecAndWait(exe^,DStr.Concat3(CmdLine^,pname^,plist^)^,dir) THEN
      Protocol.AndOutString('','COMSOL can not run','');
      HALT(0)
    END;
    IF ~Success() THEN
      Protocol.AndOutString('','Error during COMSOL execution','');
      HALT(0)
    END;
    IF ~HU.DeleteFile(StatusFileName) THEN
      Protocol.String('',ModuleName+'.Launch: status file '+StatusFileName+' is not deleted','');
    END;
  END Launch;

PROCEDURE Init ();
  VAR
    info:SF.Info;
  PROCEDURE CloseAndHalt();
    BEGIN
      SF.Close(info);
      HALT(0);
    END CloseAndHalt;
  BEGIN
    pname:=DStr.Str('');
    plist:=pname;

    exe:=HU.GetEnvString(EnvStr);
    IF exe^='' THEN
      Protocol.AndOutString('','COMSOL is not found','');
      HALT(0);
    END;

    IF ~SF.First('*.mph',info) THEN
      Protocol.AndOutString('','no COMSOL file is found found','');
      CloseAndHalt;
    END;

    IF DStr.CompareCapitalized(SF.NameExt(info)^,OutputFile)=0 THEN
      IF ~SF.Next(info) THEN
        Protocol.AndOutString('only ',OutputFile,' is found');
        CloseAndHalt;
      END;
    END;

    InputFile:=SF.NameExt(info);
    IF SF.Next(info) THEN
      IF (DStr.CompareCapitalized(SF.NameExt(info)^,OutputFile)#0) OR SF.Next(info) THEN
        Protocol.AndOutString('','too many COMSOL files are found','');
        CloseAndHalt;
      END;
    END;

    SF.Close(info);
    CmdLine:=DStr.Concat3('-inputfile "',InputFile^,'" -outputfile "'+OutputFile+'" -job b1');
  END Init;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');
  Protocol.String('EnvStr = ',EnvStr,'');
  Init();
  Protocol.String('exe = ',exe^,'');
  Protocol.String('CmdLine = ',CmdLine^,'');
  Protocol.String('InputFile = ',InputFile^,'');
  Protocol.String('OutputFile = ',OutputFile,'');
  Protocol.Real('MinZ=',MinZ,' cm');
  Protocol.Real('MaxZ=',MaxZ,' cm');
  Protocol.Real('Rad=',Rad,' cm');
  Protocol.Real('AreaTolerance=',AreaTolerance,' cm');
  Protocol.String('','','');
END Comsol.