(** Calculates vector potential A from electric conductivity sigma and current
    in the coil using model of Mostaghimi.
    
    EMM = electro-magnetic module *)
<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE EMM;

IMPORT
  Math:=mMath, PHC:=msciPhConst, Geo:=Geometry, Protocol, COMPILER,
  Area:=mArea2DComplex, PEqu:=mParabolicEqu2DComplex, Out:=mhostStdOut,
  MathC:=mMathComplex, Arr:=mArraysComplex, Steps:=mEvolutionSteps,
  ComFunc:=mCommonFunctions, GeoArea;

CONST
  pi = Math.pi;
  cm = PHC.cm;

  omega = Geo.Frequency * 1.0E6 * 2*pi; (* Circular RF frequency in radians per second *)

  Z0 = Geo.CoilZ1;
  DeltaZ = 0.6*cm;
  DeltaR = DeltaZ;
  MinZ * = Geo.TotalMinZ - DeltaZ;
  MaxZ * = Geo.TotalMaxZ + DeltaZ;
  Rad * = Geo.TotalRad + DeltaR;

  dz = 0.01*cm;
  dr = dz;

  (** Dimensions of the used grid *)
  LastZ * = ENTIER((MaxZ-MinZ)/dz+0.5);
  LastR * = ENTIER(Rad/dr+0.5);

  (* Parameters of calculations *)
  RelErr = 1.0E-5;

  ModuleName = COMPILER.EQUATION('module');
  Tab = Protocol.Tab;

TYPE
  Float = Math.Float;
  Complex = MathC.Complex;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;
  ComplexArray2 = MathC.Array2;
  ComplexArray2Ptr = MathC.Array2Ptr;

VAR
  area - : Area.Area; (** Area for equations *)
  A-:ComplexArray2Ptr; (** the vector potential *)
  OldA:ComplexArray2Ptr; (* the vector potential in the previous iteration *)
  equ - :POINTER TO PEqu.Equ; (** The equation *)
  sigma - :FloatArray2Ptr; (** Electrical conductivity *)
  CurrentDensity:Float;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

TYPE
  AA = RECORD (PEqu.AreaFunc)
  END;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR a:AA) f (z,r:Float): Complex;
  BEGIN
    RETURN sqr(r)
  END f;
<* POP *>

TYPE
  D = RECORD (PEqu.AreaFunc)
  END;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR d:D) f (z,r:Float): Complex;
  BEGIN
    RETURN r
  END f;
<* POP *>

TYPE
  E = RECORD (PEqu.AreaFunc)
  END;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR e:E) f (z,r:Float): Complex;
  VAR
    iZ,iR:LONGINT;
  BEGIN
    ASSERT(sigma#NIL);
    ASSERT(LEN(sigma^)=area.LastX+1);
    ASSERT(LEN(sigma^,1)=area.LastY+1);

    iZ:=Area.XToInd(area,z);
    iR:=Area.YToInd(area,r);
    RETURN  -1 - 1.0i*sqr(r)*4*pi/sqr(PHC.c)*omega*sigma[iZ,iR]
  END f;
<* POP *>

TYPE
  F = RECORD (PEqu.AreaFunc)
  END;

PROCEDURE (VAR f:F) f (z,r:Float): Complex;
  BEGIN
    IF Geo.GetSpaceType(z,r)=Geo.Coil THEN
      RETURN sqr(r) * CurrentDensity * 4*pi/PHC.c;
    ELSE
      RETURN 0
    END;
  END f;

<* PUSH *>
<* WOFF301 + *>
TYPE
  MinZFunc = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:MinZFunc) f (r:Float): Complex;
  BEGIN
    RETURN 3*(MinZ-Z0)
  END f;

TYPE
  MinZDer = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:MinZDer) f (r:Float): Complex;
  BEGIN
    RETURN sqr(MinZ-Z0) + sqr(r)
  END f;

TYPE
  MaxZFunc = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:MaxZFunc) f (r:Float): Complex;
  BEGIN
    RETURN 3*(MaxZ-Z0)
  END f;
<* POP *>

TYPE
  MaxZDer = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:MaxZDer) f (r:Float): Complex;
  BEGIN
    RETURN sqr(MaxZ-Z0) + sqr(r)
  END f;

TYPE
  RadFunc = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:RadFunc) f (z:Float): Complex;
  BEGIN
    RETURN 3*sqr(Rad) - (sqr(z-Z0) + sqr(Rad))
  END f;

TYPE
  RadDer = RECORD (PEqu.BoundaryFunc)
  END;

PROCEDURE (VAR f:RadDer) f (z:Float): Complex;
  BEGIN
    RETURN Rad*(sqr(z-Z0) + sqr(Rad))
  END f;

TYPE
  ZeroBoundary = RECORD (PEqu.BoundaryFunc)
  END;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR f:ZeroBoundary) f (z:Float): Complex;
  BEGIN
    RETURN 0
  END f;
<* POP *>

TYPE
  Step = RECORD (Steps.Step)
    iteration:LONGINT;
  END;

PROCEDURE (VAR s:Step) Step (dt:Float);
  BEGIN
    INC(s.iteration);

    Arr.Swap2(A,OldA);
    PEqu.Step(equ,OldA,A,dt);
  END Step;

PROCEDURE (VAR s:Step) Norm (): Float;
  BEGIN
    RETURN Arr.Norm2(A^);
  END Norm;

PROCEDURE (VAR s:Step) NormOfDifference (): Float;
  BEGIN
    RETURN Arr.NormOfDifference2(A^,OldA^);
  END NormOfDifference;

PROCEDURE SetSigma * (VAR Sigma:ComFunc.Float2);
  (** Sets the electrical conductivity to the module structures *)
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
  BEGIN
    ASSERT(sigma#NIL);

    Out.String('EMM sigma... ',0);
    FOR iZ:=0 TO LastZ DO
      z:=Area.IndToX(area,iZ);
      FOR iR:=0 TO LastR DO
        r:=Area.IndToY(area,iR);
        sigma[iZ,iR]:=Sigma.f(z,r);
      END;
    END;
  END SetSigma;

PROCEDURE Calculate * (current:Float);
  (** Calculates electro-magnetic field generated by the specified current
      in the coil and specified electrical conductivity Sigma. The Sigma should
      be previously set by the procedure SetSigma *)
  VAR
    CurrRelErr:Float;
    e:E;
    f:F;
    dt,l_min,l_max:Float;
    s:Step;
    ev:Steps.Eigenvalues;
  BEGIN
    ASSERT(sigma#NIL);

    Out.String('EMM... ',0);

    CurrentDensity:=current/Geo.CoilCrossSectionArea;
    (* No initialization of A because the initial A is taken from the previous iteration *)
    PEqu.SetE(equ^,e);
    PEqu.SetF(equ^,f);

    PEqu.Eigenvalues(equ^,ev);
    dt:=Steps.OptimalDt(ev);
    Steps.LogEigenvalues(ev,l_min,l_max);
    l_min:=l_min*1;
    l_max:=l_max*3;

    s.iteration:=0;
    CurrRelErr:=Steps.GetRelErr(s,dt);
    WHILE CurrRelErr>RelErr DO
      Steps.Log(s,l_min,l_max,RelErr/CurrRelErr);
      CurrRelErr:=Steps.GetRelErr(s,dt);
    END;

    Out.Int(s.iteration,0); Out.String(' iterations',0);
    Out.String(', RHS_RelErr = ',0);
    Out.Real(PEqu.NormOfRHS(equ^,A^)/Arr.MaxAbs2(equ^.f^),0);
    Out.Ln;
  END Calculate;

PROCEDURE Init;
  VAR
    iZ,iR:LONGINT;
    a:AA;
    d:D;
    minZFunc:MinZFunc;
    minZDer:MinZDer;
    maxZFunc:MaxZFunc;
    maxZDer:MaxZDer;
    radFunc:RadFunc;
    radDer:RadDer;
    zb:ZeroBoundary;
  BEGIN
    Area.Init(area,LastZ,LastR,MinZ,MaxZ,0,Rad);

    NEW(A,LastZ+1,LastR+1);
    NEW(OldA,LastZ+1,LastR+1);
    NEW(sigma,LastZ+1,LastR+1);
    FOR iZ:=0 TO LastZ DO
      FOR iR:=0 TO LastR DO
        A[iZ,iR]:=0;
        OldA[iZ,iR]:=0;
        sigma[iZ,iR]:=0;
      END;
    END;

    NEW(equ);
    PEqu.Init(equ,area);
    PEqu.SetA(equ^,a);
    PEqu.SetB(equ^,a);
    PEqu.SetD(equ^,d);

    PEqu.SetMinX(equ^,minZFunc,minZDer,zb);
    PEqu.SetMaxX(equ^,maxZFunc,maxZDer,zb);
    PEqu.SetMaxY(equ^,radFunc,radDer,zb);
  END Init;

PROCEDURE InputOK (z,r:Float): BOOLEAN;
  BEGIN
    ASSERT(z>=MinZ);
    ASSERT(z<=MaxZ);
    ASSERT(r>=0);
    ASSERT(r<=Rad);

    RETURN TRUE
  END InputOK;

PROCEDURE Sigma * (z,r:Float): Float;
  (** Returns electrical conductivity *)
  BEGIN
    ASSERT(InputOK(z,r));

    IF Geo.GetSpaceType(z,r)=Geo.In THEN
      RETURN GeoArea.Interpolate(area,sigma^,z,r);
    ELSE
      RETURN 0
    END;
  END Sigma;

PROCEDURE GetA * (z,r:Float): Complex;
  (** Returns the vector potential *)
  BEGIN
    ASSERT(InputOK(z,r));

    RETURN Area.Interpolate(area,A^,z,r);
  END GetA;

PROCEDURE Ephi * (z,r:Float): Complex;
  (** Returns amplitude of tangential electrical field *)
  BEGIN
    ASSERT(InputOK(z,r));

    RETURN -1.0i*Area.Interpolate(area,A^,z,r)*omega/PHC.c;
  END Ephi;

PROCEDURE Bz * (z,r:Float): Complex;
  (** Returns z component of magnetic field *)
  BEGIN
    ASSERT(InputOK(z,r));

    IF r<area.dy THEN
      RETURN 2*Area.DiffY(area,A^,z,r)
    ELSE
      RETURN Area.Interpolate(area,A^,z,r)/r + Area.DiffY(area,A^,z,r)
    END;
  END Bz;

PROCEDURE Br * (z,r:Float): Complex;
  (** Returns r component of magnetic field. *)
  BEGIN
    ASSERT(InputOK(z,r));

    RETURN -Area.DiffX(area,A^,z,r);
  END Br;

PROCEDURE Power * (z,r:Float): Float;
  (** Returns density of electro-magnetic power dissipated in the plasma *)
  BEGIN
    ASSERT(InputOK(z,r));

    RETURN 0.5*Sigma(z,r)*MathC.SqrAbs(Ephi(z,r));
  END Power;

PROCEDURE Fz * (z,r:Float): Float;
  (** Returns z component of the Lorentz force *)
  VAR
    ephi,br:Complex;
  BEGIN
    ASSERT(InputOK(z,r));

    ephi:=Ephi(z,r);
    br:=Br(z,r);
    RETURN -0.5*Sigma(z,r)*( RE(ephi)*RE(br) + IM(ephi)*IM(br) ) / PHC.c;
  END Fz;

PROCEDURE Fr * (z,r:Float): Float;
  (** Returns r component of the Lorentz force *)
  VAR
    ephi,bz:Complex;
  BEGIN
    ASSERT(InputOK(z,r));

    ephi:=Ephi(z,r);
    bz:=Bz(z,r);
    RETURN 0.5*Sigma(z,r)*( RE(ephi)*RE(bz) + IM(ephi)*IM(bz) ) / PHC.c;
  END Fr;

PROCEDURE ElVol (iR:LONGINT): Float;
  VAR
    result:Float;
  BEGIN
    result:=2*pi*Area.IndToY(area,iR)*area.dy*area.dx;
    RETURN result;
  END ElVol;

PROCEDURE TotalPower * (): Float;
  (** Returns a total heat power generated by the current magnetic fields *)
  VAR
    iZ,iR:LONGINT;
    z,r,res:Float;
  BEGIN
    res:=0;
    FOR iZ:=0 TO LastZ DO
      z:=Area.IndToX(area,iZ);
      FOR iR:=0 TO LastR DO
        r:=Area.IndToY(area,iR);
        res:=res+Power(z,r)*ElVol(iR)
      END;
    END;
    RETURN res;
  END TotalPower;

TYPE
  DiscFunc = RECORD (Area.Func)
  END;

PROCEDURE (VAR f:DiscFunc) f (z,r:Float): Complex;
  VAR
    iZ,iR:LONGINT;
  BEGIN
    iZ:=Area.XToInd(area,z);
    iR:=Area.YToInd(area,r);
    RETURN PEqu.RHS(equ^,A^,iZ,iR);
  END f;

PROCEDURE SaveDiscrepancy * (FileName-:ARRAY OF CHAR; factor:Float);
  (** Saves the discrepancy of A with the specified factor *)
  VAR
    f:DiscFunc;
  BEGIN
    Area.SaveFunc(area,f,FileName,factor);
  END SaveDiscrepancy;

PROCEDURE Multiply * (factor:Float);
  (* A:=factor*A *)
  VAR
    iZ,iR:LONGINT;
  BEGIN
    FOR iZ:=0 TO LastZ DO
      FOR iR:=0 TO LastR DO
        A[iZ,iR]:=A[iZ,iR]*factor;
      END;
    END;
  END Multiply;

BEGIN
  Init;

  Protocol.String('MODULE ',ModuleName,':');
  Protocol.Real('Z0 = ',Z0/cm,' cm');
  Protocol.Real('DeltaZ = ',DeltaZ/cm,' cm');
  Protocol.Real('DeltaR = ',DeltaR/cm,' cm');
  Protocol.Real('MinZ = ',MinZ/cm,' cm');
  Protocol.Real('MaxZ = ',MaxZ/cm,' cm');
  Protocol.Real('Rad = ',Rad/cm,' cm');
  Protocol.Real('approximate dz = ',dz/cm,' cm');
  Protocol.Real('approximate dr = ',dr/cm,' cm');
  Protocol.Real('dz = ',area.dx/cm,' cm');
  Protocol.Real('dr = ',area.dy/cm,' cm');
  Protocol.Int('LastZ = ',LastZ,'');
  Protocol.Int('LastR = ',LastR,'');
  Protocol.Real('RelErr = ',RelErr,'');
  Protocol.String('','','');
END EMM.