(** Provides some operations with 3-dimensionsl vectors.

    Last change: 27.03.2015 *)

<* O2EXTENSIONS + *>

MODULE m3DVectors;

IMPORT
  Math:=mMath;

IMPORT
<* IF    Float="REAL" THEN*>         FMath:=RealMath;
<* ELSIF Float="LONGREAL" THEN*>     FMath:=LongMath;
<* ELSIF Float="LONGLONGREAL" THEN*> FMath:=msysLongLongMath;
<* END *>

TYPE
  Float * = Math.Float;

  Vector * = RECORD
    x*,y*,z*:Float;
  END;

  Transformation * = RECORD (** Linear transformation of vectors. *)
    m:ARRAY 3,3 OF Float;
  END;
  (* 1st index: index of the matrix line,
     2nd index: index of the matrix column.
     0 = x, 1 = y, 2 = z *)

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN FMath.sqrt(x);
  END sqrt;

PROCEDURE SqrAbs * (r-:Vector): Float;
  (** Returns r^2 *)
  BEGIN
    RETURN sqr(r.x)+sqr(r.y)+sqr(r.z)
  END SqrAbs;

PROCEDURE Abs * (r-:Vector): Float;
  (** Returns |r| *)
  BEGIN
    RETURN sqrt(SqrAbs(r))
  END Abs;

PROCEDURE SqrDist * (r1-,r2-:Vector): Float;
  (** Returns (r2-r1)^2 *)
  BEGIN
    RETURN sqr(r1.x-r2.x)+sqr(r1.y-r2.y)+sqr(r1.z-r2.z)
  END SqrDist;

PROCEDURE Dist * (r1-,r2-:Vector): Float;
  (** Returns |r2-r1| *)
  BEGIN
    RETURN sqrt(SqrDist(r1,r2))
  END Dist;

PROCEDURE SqrR * (r-:Vector): Float;
  (** Returns r.x^2+r.y^2 *)
  BEGIN
    RETURN sqr(r.x)+sqr(r.y)
  END SqrR;

PROCEDURE R * (r-:Vector): Float;
  (** Returns sqrt(r.x^2+r.y^2) *)
  BEGIN
    RETURN sqrt(SqrR(r))
  END R;

PROCEDURE ScalProd * (a-,b-:Vector): Float;
  (** Returns a scalar product of a and b. *)
  BEGIN
    RETURN a.x*b.x + a.y*b.y + a.z*b.z
  END ScalProd;

PROCEDURE VecProd * (a-,b-:Vector): Vector;
  (** Returns a vector product of a and b. *)
  VAR
    r:Vector;
  BEGIN
    r.x:=a.y*b.z - a.z*b.y;
    r.y:=a.z*b.x - a.x*b.z;
    r.z:=a.x*b.y - a.y*b.x;
    RETURN r
  END VecProd;

PROCEDURE Unit * (v-:Vector): Vector;
  (** Returns a vector in the same direction as v, but with Abs(result)=1. *)
  VAR
    tmp:Float;
    result:Vector;
  BEGIN
    tmp:=Abs(v);
    ASSERT(tmp#0);
    tmp:=1/tmp;
    result.x:=v.x*tmp;
    result.y:=v.y*tmp;
    result.z:=v.z*tmp;
    RETURN result
  END Unit;

PROCEDURE Transform * (t-:Transformation; v-:Vector): Vector;
  (** Applies a given linear transformation to the given vector and returns
      the result. *)
  VAR
    r:Vector;
  BEGIN
    r.x:=t.m[0,0]*v.x + t.m[0,1]*v.y + t.m[0,2]*v.z;
    r.y:=t.m[1,0]*v.x + t.m[1,1]*v.y + t.m[1,2]*v.z;
    r.z:=t.m[2,0]*v.x + t.m[2,1]*v.y + t.m[2,2]*v.z;
    RETURN r
  END Transform;

PROCEDURE ValidSinCos (sin,cos:Float): BOOLEAN;
  BEGIN
    RETURN ABS(sin*sin+cos*cos-1)<=Math.EpsFloat*10
  END ValidSinCos;

PROCEDURE AxialRotation * (a:Vector; sin,cos:Float; VAR t:Transformation);
  (** Returns a transformation, which rotates a vector around a specified axis.
      The axis direction is given by a.
      The rotation angle is given by it's sine and cosine. *)
  BEGIN
    ASSERT(ValidSinCos(sin,cos));
    ASSERT((a.x#0) OR (a.y#0) OR (a.z#0));

    a:=Unit(a);

    t.m[0,0]:= cos + (1-cos)*sqr(a.x);
    t.m[0,1]:= (1-cos)*a.x*a.y - sin*a.z;
    t.m[0,2]:= (1-cos)*a.x*a.z + sin*a.y;

    t.m[1,0]:= (1-cos)*a.x*a.y + sin*a.z;
    t.m[1,1]:= cos + (1-cos)*sqr(a.y);
    t.m[1,2]:= (1-cos)*a.y*a.z - sin*a.x;

    t.m[2,0]:= (1-cos)*a.z*a.x - sin*a.y;
    t.m[2,1]:= (1-cos)*a.z*a.y + sin*a.x;
    t.m[2,2]:= cos + (1-cos)*sqr(a.z);
  END AxialRotation;

PROCEDURE DirectRotation * (InitialVector,FinalVector:Vector; VAR t:Transformation);
  (** Returns a transformation, which is a direct rotation that makes the
      InitialVector codirectional to the FinalVector.
      The initial and final vectors should be not zero and not collinear. *)
  VAR
    a:Vector;
    sin,cos:Float;
  BEGIN
    ASSERT((InitialVector.x#0) OR (InitialVector.y#0) OR (InitialVector.z#0));
    ASSERT((FinalVector.x#0) OR (FinalVector.y#0) OR (FinalVector.z#0));

    a:=VecProd(InitialVector,FinalVector); (* a is the axis of the rotation *)
    ASSERT((a.x#0) OR (a.y#0) OR (a.z#0));

    cos:= ScalProd(InitialVector,FinalVector)
         /sqrt(SqrAbs(InitialVector)*SqrAbs(FinalVector));
    sin:=sqrt(1-sqr(cos));

    AxialRotation(a,sin,cos,t)
  END DirectRotation;

PROCEDURE RotateX * (VAR r:Vector; sin,cos:Float);
  (** Rotates the vector r around the axis x.
      The rotation angle is given by it's sine and cosine. *)
  VAR
    z:Float;
  BEGIN
    ASSERT(ValidSinCos(sin,cos));

    z:=r.z;
    r.z:=r.y*sin + r.z*cos;
    r.y:=r.y*cos - z*sin
  END RotateX;

PROCEDURE RotateY * (VAR r:Vector; sin,cos:Float);
  (** Rotates the vector r around the axis y.
      The rotation angle is given by it's sine and cosine. *)
  VAR
    z:Float;
  BEGIN
    ASSERT(ValidSinCos(sin,cos));

    z:=r.z;
    r.z:=-r.x*sin + r.z*cos;
    r.x:= r.x*cos + z*sin
  END RotateY;

PROCEDURE RotateZ * (VAR r:Vector; sin,cos:Float);
  (** Rotates the vector r around the axis z.
      The rotation angle is given by it's sine and cosine. *)
  VAR
    x:Float;
  BEGIN
    ASSERT(ValidSinCos(sin,cos));

    x:=r.x;
    r.x:=r.x*cos - r.y*sin;
    r.y:=  x*sin + r.y*cos
  END RotateZ;

END m3DVectors.