(** Direct and inverse Abel transformation.
    The data to transform are presented as array of Float.
    It is assumed that the points are homogeneously distribured in 0<=r<1

    The direct and simple inverse by matrix method transformations don't need
    any other assumptions.

    The polynomial inverse transformation is based on a polynomial
    f-interpolation: the entire r-range (0<=r<1) is divided
    into homogeneous intervals, in each interval the result of the Abel
    inversion is approximated by a polynom.
    Additional conditions for the polynoms can be added.

    Typical usege:
    mAbelPoly.Init(d,...)   initialization of the data structure for the inversion

    mAbelPoly.AddVal(d,...) Addition of conditions to the data structure
    mAbelPoly.AddEqu(d,...)

    mAbelPoly.Convert(d)    Conversion should be carried out before the inversion
                            No conditions can be added after conversion

    mAbelPoly.Inverse(d,x)  Inversion of x


    Last change: 03.06.2016 *)

<* O2EXTENSIONS + *>
MODULE mAbelPoly;

IMPORT
  Math:=mMath, Lagr:=mLagrangeMultiplierLinear, Arr:=mArrays, FFT:=mFFT;

IMPORT
<* IF    Float="REAL" THEN*>         FMath:=RealMath;
<* ELSIF Float="LONGREAL" THEN*>     FMath:=LongMath;
<* ELSIF Float="LONGLONGREAL" THEN*> FMath:=msysLongLongMath;
<* END *>

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN FMath.sqrt(x)
  END sqrt;

PROCEDURE ln (x:Float): Float;
  BEGIN
    RETURN FMath.ln(x)
  END ln;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE Int0 (a,b,Ri,yn:Float): Float;

  PROCEDURE Int (r:Float): Float;
    BEGIN
      RETURN sqrt((r-yn)*(r+yn))
    END Int;

  BEGIN
    RETURN Int(b)-Int(a)
  END Int0;
<* POP *>

PROCEDURE Int1 (a,b,Ri,yn:Float): Float;

  PROCEDURE Int (r:Float; VAR root,result:Float);
    BEGIN
      root:=sqrt((r-yn)*(r+yn));
      result:=(r/2-Ri)*root;
    END Int;

  VAR
    RootA,RootB,ResA,ResB,log:Float;
  BEGIN
    Int(a,RootA,ResA);
    Int(b,RootB,ResB);
    IF yn=0 THEN
      log:=0
    ELSE
      log:=ln((b+RootB)/(a+RootA))
    END;
    RETURN ResB-ResA + 1/2*sqr(yn)*log;
  END Int1;

PROCEDURE Int2 (a,b,Ri,yn:Float): Float;
  VAR
    SqrYn:Float;

  PROCEDURE Int (r:Float; VAR root,result:Float);
    BEGIN
      root:=sqrt((r-yn)*(r+yn));
      result:=( sqr(r)/3 + 2/3*SqrYn - Ri*r + sqr(Ri) )*root;
    END Int;

  VAR
    RootA,RootB,ResA,ResB,log:Float;
  BEGIN
    SqrYn:=sqr(yn);
    Int(a,RootA,ResA);
    Int(b,RootB,ResB);
    IF yn=0 THEN
      log:=0
    ELSE
      log:=ln((b+RootB)/(a+RootA))
    END;
    RETURN ResB-ResA - Ri*SqrYn*log;
  END Int2;

PROCEDURE Int3 (a,b,Ri,yn:Float): Float;
  VAR
    yn2,yn4,Ri2,Ri3:Float;

  PROCEDURE Int (r:Float; VAR root,result:Float);
    VAR
      r2,r3:Float;
    BEGIN
      r2:=sqr(r);
      r3:=r*r2;
      root:=sqrt((r-yn)*(r+yn));
      result:=( r3/4 + 3/8*yn2*r - r2*Ri - 2*Ri*yn2 + 3/2*Ri2*r - Ri3 )*root;
    END Int;

  VAR
    RootA,RootB,ResA,ResB,log:Float;
  BEGIN
    yn2:=sqr(yn);
    yn4:=sqr(yn2);
    Ri2:=sqr(Ri);
    Ri3:=Ri*Ri2;
    Int(a,RootA,ResA);
    Int(b,RootB,ResB);
    IF yn=0 THEN
      log:=0
    ELSE
      log:=ln((b+RootB)/(a+RootA))
    END;
    RETURN ResB-ResA + (3/8*yn4 + 3/2*Ri2*yn2)*log;
  END Int3;

TYPE
  IntProc = PROCEDURE (a,b,Ri,yn:Float): Float;
  IntProcArr = ARRAY OF IntProc;

CONST
  IProc = IntProcArr{ Int0, Int1, Int2, Int3 };
  MaxPower * = LEN(IProc)-1;
    (**  Maximal supported power of the polynoms*)

TYPE
  Data * = RECORD
    (** Data needed for polynomial inverse Abel transformation *)
    m:Lagr.Matrix;
    M:LONGINT; (* number of intervals *)
    K:LONGINT; (* K-1 is maxima polynom power *)
    N:LONGINT; (* number of points in the data arrays *)
    b:ArrayPtr;
  END;

PROCEDURE R (d-:Data; m:LONGINT): Float;
  (* The intervals are R(m)<=r<=R(m+1), 0<=m<=M.
     After initialization, all the procedures use R to determina the intervals,
     so inhomogeneous intervals can be implemented by the procedure R.
     At this moment homogeneous intervals are implemented.
     At that m=0 corresponds to r=0, m=M corresponds to maximal R=1 *)
  BEGIN
    RETURN m/d.M
  END R;

PROCEDURE Y (d-:Data; n:LONGINT): Float;
  (* The coordinate of points the measured and returned data, 0<=n<=N.
     After initialization, all the procedures use Y to determina the points,
     so inhomogeneous points can be implemented by the procedure Y.
     At this moment homogeneous points are implemented.
     At that n=0 corresponds to r=0, n=N corresponds to maximal r=1 *)
  BEGIN
    RETURN n/d.N
  END Y;

PROCEDURE I (d-:Data; m,k:LONGINT; r:Float): Float;
   VAR
     Rm,Rm1:Float;
   BEGIN
    ASSERT(m>=0);
    ASSERT(m<d.M);
    ASSERT(k>=0);
    ASSERT(k<d.K);
    ASSERT(r<=R(d,m+1)); (* optimization *)

    Rm:=R(d,m);
    Rm1:=R(d,m+1);
    IF r>=Rm THEN
      RETURN IProc[k] (r,Rm1,Rm,r)
    ELSE
      RETURN IProc[k] (Rm,Rm1,Rm,r)
    END;
  END I;

PROCEDURE Init * (VAR d:Data;
                  NumOfPoints, NumOfIntervals, PolynomPower:LONGINT);
  (** Initializes the data.
      NumOfPoints is number of the homogeneously distributed points
      in the arrays to convert.
      NumOfIntervals is number of intervals for polynomial f-interpolation.
      PolynomPower is maximal power of polynoms used in the f-interpolation. *)
  VAR
    rank,m1,k1,m2,k2,ind1,ind2,M,N,K,n:LONGINT;
    x,r,Rm1:Float;
  BEGIN
    ASSERT(NumOfPoints>0);
    ASSERT(NumOfIntervals>0);
    ASSERT(PolynomPower>=0);

    M:=NumOfIntervals;
    K:=PolynomPower+1;
    N:=NumOfPoints;
    d.M:=M;
    d.N:=N;
    d.K:=K;

    rank:=d.M*d.K;
    Lagr.Init(d.m,rank);
    NEW(d.b,rank);

    ind1:=0;
    FOR m1:=0 TO M-1 DO
      Rm1:=R(d,m1+1);
      FOR k1:=0 TO K-1 DO
        ind2:=m1*K;
        FOR m2:=m1 TO M-1 DO
          FOR k2:=0 TO K-1 DO
            x:=0;
            n:=0;
            r:=Y(d,0);
            WHILE (n<N) & (Rm1>r) DO
              x:= x + I(d,m1,k1,r) * I(d,m2,k2,r);
              INC(n);
              r:=Y(d,n);
            END;
            x:=2*x;
            Lagr.SetItem(d.m,ind1,ind2,x);
            Lagr.SetItem(d.m,ind2,ind1,x);
            INC(ind2);
          END;
        END;
        INC(ind1);
      END;
    END;
  END Init;

PROCEDURE Initialized * (d-:Data): BOOLEAN;
  (** Returns if d is initialized *)
  BEGIN
    RETURN d.b#NIL
  END Initialized;

PROCEDURE Converted * (d-:Data): BOOLEAN;
  (** Returns if d is converted *)
  BEGIN
    RETURN Lagr.Converted(d.m)
  END Converted;

PROCEDURE Prod (l,k:LONGINT): LONGINT;
  VAR
    i,res:LONGINT;
  BEGIN
    res:=1;
    FOR i:=0 TO l-1 DO
      res:=res*(k-i)
    END;
    RETURN res;
  END Prod;

PROCEDURE FillCoeffs (VAR d:Data; m,der:LONGINT; r:Float; neg:BOOLEAN);
  VAR
    dr,pow:Float;
    ind,k:LONGINT;
  BEGIN
    dr:=r-R(d,m);
    pow:=1;
    ind:=m*d.K+der;
    FOR k:=der TO d.K-1 DO
      IF neg THEN
        d.b[ind]:=-pow*Prod(der,k);
      ELSE
        d.b[ind]:=pow*Prod(der,k);
      END;
      pow:=pow*dr;
      INC(ind);
    END;
  END FillCoeffs;

PROCEDURE AddVal * (VAR d:Data; m,der:LONGINT; r,val:Float);
  (** Adds the followoing condition for the polynomial f-interpolation:
      der-th derivative of m-th polynim in point r is equal to val. *)
  BEGIN
    ASSERT(Lagr.Initialized(d.m));
    ASSERT(~Lagr.Converted(d.m));
    ASSERT(m>=0);
    ASSERT(m<d.M);
    ASSERT(der>=0);
    ASSERT(der<d.K);

    Arr.Fill(d.b^,0);
    FillCoeffs(d,m,der,r,FALSE);
    Lagr.AddConstraint(d.m,d.b^,val);
  END AddVal;

PROCEDURE AddEqu * (VAR d:Data; m1,m2,der:LONGINT; r:Float);
  (** Adds the followoing condition for the polynomial f-interpolation:
      der-th derivative of m1-th and m2-th polynim in point r are equal. *)
  BEGIN
    ASSERT(Lagr.Initialized(d.m));
    ASSERT(~Lagr.Converted(d.m));
    ASSERT(m1>=0);
    ASSERT(m1<d.M);
    ASSERT(m2>=0);
    ASSERT(m2<d.M);
    ASSERT(m1#m2);
    ASSERT(der>=0);
    ASSERT(der<d.K);

    Arr.Fill(d.b^,0);
    FillCoeffs(d,m1,der,r,FALSE);
    FillCoeffs(d,m2,der,r,TRUE);
    Lagr.AddConstraint(d.m,d.b^,0);
  END AddEqu;

PROCEDURE Convert * (VAR d:Data);
  (** Converts the d *)
  BEGIN
    ASSERT(Lagr.Initialized(d.m));
    ASSERT(~Lagr.Converted(d.m));

    Lagr.Convert(d.m);
  END Convert;

PROCEDURE InitSpline * (VAR d:Data; NumOfPoints, NumOfIntervals:LONGINT);
  (** Initiates and converts the data for spline f-interpolation
      as described in the article:
      H. Fulge, A. Knapp, C. Eichhorn, R. Wernitz, S. L\"ohle, S. Fasoulas,
      G.Herdrich, "Improved Abel Inversion Method for Analysis of Spectral
      and Photo-Optical Data of Magnetic Influenced Plasma Flows",
      42nd AIAA Plasmadynamics and Lasers Conference,
      Fluid Dynamics and Co-located Conferences,
      http://dx.doi.org/10.2514/6.2011-3456

      NumOfPoints is number of the homogeneously distributed points
      in the arrays to convert.
      NumOfIntervals is number of intervals for polynomial f-interpolation.  *)
  VAR
    m,M:LONGINT;
  BEGIN
    Init(d,NumOfPoints,NumOfIntervals,3);
    M:=d.M;
    FOR m:=1 TO M-1 DO
      AddEqu(d,m-1,m,0,R(d,m));
      AddEqu(d,m-1,m,1,R(d,m));
    END;
    AddVal(d,0,1,R(d,0),0);
    AddVal(d,M-1,0,R(d,M),0);
    AddVal(d,M-1,1,R(d,M),0);
    Convert(d);
  END InitSpline;

PROCEDURE Inverse * (d-:Data; VAR f:Array);
  (** Performs the inversion of f. LEN(f) must be equal to NumOfPoints
      at the initialization. *)
  VAR
    M,K,N:LONGINT;

  VAR
    b:ArrayPtr;
    m,k,ind,n:LONGINT;
    x,dr,pow,r,Rm,Rm1:Float;
  BEGIN
    ASSERT(Initialized(d));
    ASSERT(Converted(d));
    ASSERT(LEN(f)=d.N);

    M:=d.M;
    K:=d.K;
    N:=d.N;
    b:=d.b;
    ind:=0;
    FOR m:=0 TO M-1 DO
      Rm1:=R(d,m+1);
      FOR k:=0 TO K-1 DO
        x:=0;
        n:=0;
        r:=Y(d,0);
        WHILE (n<N) & (Rm1>r) DO
          x:=x+f[n]*I(d,m,k,r);
          INC(n);
          r:=Y(d,n);
        END;
        b[ind]:=x;
        INC(ind);
      END;
    END;

    Lagr.Solve(d.m,b^);

    m:=0;
    Rm:=R(d,0);
    Rm1:=R(d,1);
    FOR n:=0 TO N-1 DO
      r:=Y(d,n);
      IF r>Rm1 THEN
        INC(m);
        Rm:=Rm1;
        Rm1:=R(d,m+1)
      END;
      x:=0;
      dr:=r-Rm;
      pow:=1;
      FOR k:=0 TO K-1 DO
        x:=x+b[m*K+k]*pow;
        pow:=pow*dr;
      END;
      f[n]:=x
    END;
  END Inverse;

PROCEDURE d (x1,x2:LONGINT): Float;
  BEGIN
    ASSERT(x1>=x2);
    ASSERT(x2>=0);

    RETURN sqrt(1.0*(x1-x2)*(x1+x2));
  END d;

PROCEDURE Direct * (VAR f:Array);
  (** Performs direct Abel transformation of f and returns it in f. *)
  VAR
    i,j,N:LONGINT;
    sum,factor:Float;
  BEGIN
    ASSERT(LEN(f)>0);

    N:=LEN(f)-1;
    factor:=2/LEN(f);
    FOR j:=0 TO N DO
      sum:=0;
      FOR i:=j TO N DO
        sum:=sum + f[i] * ( d(i+1,j) - d(i,j) );
      END;
      f[j]:=factor*sum;
    END;
  END Direct;

PROCEDURE InverseMatrix * (VAR f:Array);
  (** Performs simple inverse Abel transformation by matrix method *)
  VAR
    i,j,N:LONGINT;
    sum,factor:Float;
  BEGIN
    ASSERT(LEN(f)>0);

    N:=LEN(f)-1;
    factor:=LEN(f)/2;
    f[N]:=factor*f[N]/d(N+1,N);
    FOR j:=N-1 TO 0 BY -1 DO
      sum:=0;
      FOR i:=j+1 TO N DO
        sum:= sum + f[i] * ( d(i+1,j) - d(i,j) )
      END;
      f[j]:= ( factor*f[j] - sum ) / d(j+1,j)
    END;
  END InverseMatrix;

PROCEDURE IndOfMax (f-:Array): LONGINT;
  VAR
    i,ind:LONGINT;
    max:Float;
  BEGIN
    max:=f[0];
    ind:=0;
    FOR i:=1 TO LEN(f)-1 DO
      IF f[i]>max THEN
        ind:=i;
        max:=f[i];
      END;
    END;
    RETURN ind;
  END IndOfMax;

PROCEDURE SymmetryCenter * (f-:Array): LONGINT;
  (** Returns index of f, which corresponds probably to center of symmetry. *)
  VAR
    ImF,ConvRe,ConvIm:ArrayPtr;
    N,i1,i2,MaxInd:LONGINT;
  BEGIN
    N:=LEN(f);
    ImF:=Arr.Filled(N,0);
    NEW(ConvRe,N);
    NEW(ConvIm,N);
    FFT.Convolution(f,ImF^,f,ImF^,ConvRe^,ConvIm^);

    (* There are 2 candidates for the symmetry point: IndOfMax/2 and (IndOfMax+N)/2 *)
    MaxInd:=IndOfMax(ConvRe^);
    i1:=MaxInd DIV 2;
    i2:=(MaxInd+N) DIV 2;
    IF f[i1]<f[i2] THEN
      RETURN i2
    ELSE
      RETURN i1
    END;
  END SymmetryCenter;

PROCEDURE WeightedCenter * (f-:Array): LONGINT;
  (** Simpler method to estimate index of the probable center of symmetry *)
  VAR
    num,denom:Float;
    i:LONGINT;
  BEGIN
    num:=0;
    denom:=0;
    FOR i:=0 TO LEN(f)-1 DO
      num:=num+f[i]*i;
      denom:=denom+f[i];
    END;
    RETURN ENTIER(num/denom + 0.5)
  END WeightedCenter;

END mAbelPoly.