(** This module provides toos for solution of nonlinear algebraic
    system of equations.

    Last change: 03.06.2016 *)
<* O2EXTENSIONS + *>

MODULE mAlgEqu;

IMPORT
  Math:=mMath, Matrix:=mSquareMatrix, SU:=msysUtils, COMPILER;

CONST
  ModuleName = COMPILER.EQUATION('module');

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

  BiFunc * = RECORD
  END;

PROCEDURE (VAR f:BiFunc) f * (x:Float): Float;
  BEGIN
    SU.RAISEError(ModuleName+'.BiFunc.f: abstract method');
    RETURN x
  END f;

PROCEDURE Bisection * (VAR f:BiFunc; x1,x2,RelErr,AbsErr:Float): Float;
  (** Solves the equation f.f=0 in [x1,x2] by bisection method and returns
      the solution. *)
  VAR
    x,f1:Float;
  BEGIN
    f1:=f.f(x1);
    ASSERT(f1*f.f(x2)<=0);

    REPEAT
      x:=(x1+x2)/2;
      IF f.f(x)*f1<=0 THEN
        x2:=x;
      ELSE
        x1:=x;
      END;
    UNTIL ((ABS(x1)<=AbsErr) & (ABS(x2)<=AbsErr)) OR (ABS(x1-x2)<=RelErr*ABS(x));
    RETURN x
  END Bisection;

TYPE
  NewtonFunc * = RECORD
  END;
  (* Function for method of Newton. It includes both the function and its
     derivative *)

PROCEDURE (VAR f:NewtonFunc) Func * (var-:Array; index:LONGINT): Float;
  BEGIN
    SU.RAISEError(ModuleName+'.NewtonFunc.Func: abstract method');
    RETURN var[index]
  END Func;

PROCEDURE (VAR f:NewtonFunc) Der * (var-:Array; FuncIndex,VarIndex:LONGINT): Float;
  BEGIN
    SU.RAISEError(ModuleName+'.NewtonFunc.Der: abstract method');
    RETURN var[FuncIndex-VarIndex]
  END Der;

TYPE
  NewtonStopDesc * = RECORD
  END;
  NewtonStop * = POINTER TO NewtonStopDesc;

PROCEDURE (s:NewtonStop) Stop * (var-:Array): BOOLEAN;
  BEGIN
    SU.RAISEError(ModuleName+'.NewtonStop.Stop: call of abstract method');
    RETURN LEN(var)>1
  END Stop;

PROCEDURE Newton * (VAR var:Array; VAR func:NewtonFunc; RelErr,AbsErr:Float;
                    stop:NewtonStop);
  (** Solves the equation func.Func=0. Proceudre func.Func should return left
      hand side of the index'th equation as function of var. Func.Der defines
      matrix of derivatives: it should return derivative of FuncIndex'th function
      on VarIndex'th variable as function of var. *)
  VAR
    a:Matrix.Matrix;
    b:POINTER TO ARRAY OF Float;
    i,j,k,count:LONGINT;
  BEGIN
    count:=LEN(var);
    NEW(b,count);
    REPEAT
      Matrix.Init(a,count);
      FOR i:=0 TO count-1 DO
        b[i]:=-func.Func(var,i);
      END;
      FOR i:=0 TO count-1 DO
        FOR j:=0 TO count-1 DO
          Matrix.SetItem(a,i,j,func.Der(var,i,j));
        END;
      END;
      Matrix.ConvertAndSolve(a,b^,count-1);
      FOR k:=0 TO count-1 DO
        var[k]:=var[k]+b[k]
      END;

      k:=0;
      WHILE (k<count) &
            ((ABS(var[k])<=AbsErr) OR
             (ABS(b[k])<=RelErr*ABS(var[k])))
      DO
        INC(k)
      END;
    UNTIL ((stop#NIL) & stop.Stop(var)) OR (k>=count);
  END Newton;

TYPE
  NewtonFuncEx * = RECORD (NewtonFunc)
    h:Float;
    var:ArrayPtr;
  END;

PROCEDURE (VAR f:NewtonFuncEx) Der * (var-:Array; FuncIndex,VarIndex:LONGINT): Float;
  VAR
    x,result:Float;
  BEGIN
    ASSERT(FuncIndex>=0);
    ASSERT(FuncIndex<LEN(var));
    ASSERT(VarIndex>=0);
    ASSERT(VarIndex<LEN(var));
    ASSERT(f.var#NIL);
    ASSERT(LEN(f.var^) = LEN(var));
    ASSERT(f.h#0);

    x:=var[VarIndex];
    f.var[VarIndex]:=x+f.h;
    result:=f.Func(f.var^,FuncIndex);
    f.var[VarIndex]:=x-f.h;
    result:=(result-f.Func(f.var^,FuncIndex))/(2*f.h);
    f.var[VarIndex]:=x;

    RETURN result
  END Der;

PROCEDURE NewtonEx * (VAR var:Array; VAR func:NewtonFuncEx;
                     h,RelErr,AbsErr:Float;
                     stop:NewtonStop);
  (** Solves the equation func.Func=0. Proceudre func.Func returns left hand
      side of the index'th equation as function of equ.var. Matrix of
      derivatives is calculated automatically using derivative step h.
      The derivetives are calculated with accuracy O(h^2).
      Thus the user should not override func.Der. *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(h#0);

    func.h:=h;
    NEW(func.var,LEN(var));
    FOR i:=0 TO LEN(var)-1 DO
      func.var[i]:=var[i]
    END;

    Newton(func.var^,func,RelErr,AbsErr,stop);

    FOR i:=0 TO LEN(var)-1 DO
      var[i]:=func.var[i]
    END;
    func.var:=NIL;
  END NewtonEx;

END mAlgEqu.