(** Provides tools to manipulate a rectangle 2-dimensional areas and
    a homogeneous grids on it.

    Last change: 04.01.2016 *)
<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE mArea2DComplex;

IMPORT
  Math:=mMath, MathComplex:=mMathComplex, Out:=msysTextOut, In:=msysTextIn,
  AreaFloat:=mArea2D, TextData:=msysTextData, ComFunc:=mCommonFunctionsComplex,
  Reg:=mLinearRegression;

TYPE
  Float * = Math.Float;
  Complex * = MathComplex.Complex;
  Array2 = MathComplex.Array2;
  Array2Ptr = MathComplex.Array2Ptr;

  Area * = AreaFloat.Area;
(** Area*  = RECORD
      MinX-,MaxX-,MinY-,MaxY-: Float;         (** Boundaries of the area *)
      LastX-,LastY-          : LONGINT; (** Grid in the area is indexed from 0 to LastX, LastY *)
      dx-,dy-                : Float; (** x-size and y-size of frid cells *)
      InvDx-,InvDy-          : Float; (** 1/dx and 1/dy *)
      Inv2dx-,Inv2dy-        : Float; (** 1/(2*dx) and (1/(2*dy)) *)
      InvSqrDx-,InvSqrDy-    : Float; (** 1/sqr(dx) and 1/sqr(dy) *)
    END; *)
  (** The area. *)

CONST
  Tab = Out.Tab;

PROCEDURE round(x:Float): LONGINT;
  BEGIN
    RETURN ENTIER(x+0.5);
  END round;

PROCEDURE Init * (VAR a:Area; LastX,LastY:LONGINT; MinX,MaxX,MinY,MaxY:Float);
  (** Initiates the area and the grid. MinX,MaxX,MinY,MaxY are the area
      boundaries, points of the grid are indexed in [0..LastX,0..LastY].
      Outermost points are in the area boundaries. *)
  BEGIN
    AreaFloat.Init(a,LastX,LastY,MinX,MaxX,MinY,MaxY);
  END Init;

PROCEDURE IndToX * (a-:Area; iX:LONGINT): Float;
  (** Returns coordinate x corresponding to index iX. *)
  BEGIN
    IF iX=a.LastX THEN (* To improve accuracy *)
      RETURN a.MaxX
    ELSE
      RETURN a.MinX+iX*a.dx
    END;
  END IndToX;

PROCEDURE XToInd * (a-:Area; x:Float): LONGINT;
  (** Returns index, nearest to coordinate x. *)
  BEGIN
    RETURN round((x-a.MinX)*a.InvDx);
  END XToInd;

PROCEDURE IndToY * (a-:Area; iY:LONGINT): Float;
  (** Returns coordinate y corresponding to index iY. *)
  BEGIN
    IF iY=a.LastY THEN (* To improve accuracy *)
      RETURN a.MaxY
    ELSE
      RETURN a.MinY+iY*a.dy
    END;
  END IndToY;

PROCEDURE YToInd * (a-:Area; y:Float): LONGINT;
  (** Returns index, nearest to coordinate x. *)
  BEGIN
    RETURN round((y-a.MinY)*a.InvDy);
  END YToInd;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE Correct* (a-:Area): BOOLEAN;
  (* Asserts that the area is valid for differential operations
     and returns TRUE *)
  BEGIN
    ASSERT(a.LastX>=2);
    ASSERT(a.LastY>=2);

    RETURN TRUE
  END Correct;

PROCEDURE CorrectField * (a-:Area; field-:Array2): BOOLEAN;
  (** Asserts that the field dimensions correspond to the area
      and returns TRUE *)
  BEGIN
    ASSERT(LEN(field)=a.LastX+1);
    ASSERT(LEN(field,1)=a.LastY+1);

    RETURN TRUE
  END CorrectField;
<* POP *>

PROCEDURE iXIn (a-:Area; iX:LONGINT): BOOLEAN;
  (* Asserts that x-index is in the grid and returns TRUE *)
  BEGIN
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);

    RETURN TRUE
  END iXIn;

PROCEDURE iYIn (a-:Area; iY:LONGINT): BOOLEAN;
  (* Asserts that y-index is in the grid and returns TRUE *)
  BEGIN
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    RETURN TRUE
  END iYIn;

PROCEDURE iXYIn (a-:Area; iX,iY:LONGINT): BOOLEAN;
  (* Asserts that the indexes are in the grid and returns TRUE *)
  BEGIN
    ASSERT(iXIn(a,iX));
    ASSERT(iYIn(a,iY));

    RETURN TRUE
  END iXYIn;

PROCEDURE DiffXIndLeft * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on x in point [iX,iY], which is
      calculates using points [iX,iY], [iX-1,iY], and [iX-2,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=2);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    RETURN -(-3*Field[iX,iY]+4*Field[iX-1,iY]-Field[iX-2,iY])*a.Inv2dx
  END DiffXIndLeft;

PROCEDURE DiffXIndRight * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on x in point [iX,iY], which is
      calculates using points [iX,iY], [iX+1,iY], and [iX+2,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX-2);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    RETURN (-3*Field[iX,iY]+4*Field[iX+1,iY]-Field[iX+2,iY])*a.Inv2dx
  END DiffXIndRight;

PROCEDURE DiffXInd * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on x in point [iX,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));

    IF iX=0 THEN
      RETURN DiffXIndRight(a,Field,0,iY);
    ELSIF iX=a.LastX THEN
      RETURN DiffXIndLeft(a,Field,a.LastX,iY);
    ELSE
      RETURN (Field[iX+1,iY]-Field[iX-1,iY])*a.Inv2dx
    END;
  END DiffXInd;

PROCEDURE DiffYIndLeft * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on y in point [iX,iY], which is
      calculates using points [iX,iY], [iX,iY-1], and [iX,iY-2].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=2);
    ASSERT(iY<=a.LastY);

    RETURN -(-3*Field[iX,iY]+4*Field[iX,iY-1]-Field[iX,iY-2])*a.Inv2dy
  END DiffYIndLeft;

PROCEDURE DiffYIndRight * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on y in point [iX,iY], which is
      calculates using points [iX,iY], [iX,iY+1], and [iX,iY+2].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY-2);

    RETURN (-3*Field[iX,iY]+4*Field[iX,iY+1]-Field[iX,iY+2])*a.Inv2dy
  END DiffYIndRight;

PROCEDURE DiffYInd * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the first derivative of field on y in point [iX,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));

    IF iY=0 THEN
      RETURN DiffYIndRight(a,Field,iX,0);
    ELSIF iY=a.LastY THEN
      RETURN DiffYIndLeft(a,Field,iX,a.LastY);
    ELSE
      RETURN (Field[iX,iY+1]-Field[iX,iY-1])*a.Inv2dy
    END;
  END DiffYInd;

PROCEDURE DiffXIndSmoothed * (a-:Area; Field-:Array2;
                              iX,iY,delta:LONGINT): Complex;
  (** Returns the first derivative of field on x in point [iX,iY].
      The result is smoothed within [iX-delta..iX+delta].
      If delta=1, the procedure has the second order of accuracy. *)
  VAR
    LastX:LONGINT;

<* PUSH *>
<* WOFF901 + *>
  PROCEDURE Extension (ind:LONGINT): Complex;
    BEGIN
      IF ind<0 THEN
        ind:=-ind;
        ASSERT(ind<=LastX);
        RETURN 2*Field[0,iY]-Field[ind,iY];
      ELSIF ind>LastX THEN
        ind:=2*LastX-ind;
        ASSERT(ind>=0);
        RETURN 2*Field[LastX,iY]-Field[ind,iY];
      ELSE
        RETURN Field[ind,iY]
      END;
    END Extension;
<* POP *>

  VAR
    i:LONGINT;
    d:Reg.Data;
    SlopeRe,SlopeIm,x:Float;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));
    ASSERT(delta>=1);

    LastX:=a.LastX;

    Reg.Init(d);
    FOR i:=iX-delta TO iX+delta DO
      x:=IndToX(a,i);
      Reg.AddPoint(d,x,RE(Extension(i)));
    END;
    SlopeRe:=Reg.Slope(d);

    Reg.Init(d);
    FOR i:=iX-delta TO iX+delta DO
      x:=IndToX(a,i);
      Reg.AddPoint(d,x,IM(Extension(i)));
    END;
    SlopeIm:=Reg.Slope(d);

    RETURN SlopeRe + 1.0i*SlopeIm
  END DiffXIndSmoothed;

PROCEDURE DiffYIndSmoothed * (a-:Area; Field-:Array2;
                              iX,iY,delta:LONGINT): Complex;
  (** Returns the first derivative of field on y in point [iX,iY].
      The result is smoothed within [iY-delta..iY+delta].
      If delta=1, the procedure has the second order of accuracy. *)
  VAR
    LastY:LONGINT;

<* PUSH *>
<* WOFF901 + *>
  PROCEDURE Extension (ind:LONGINT): Complex;
    BEGIN
      IF ind<0 THEN
        ind:=-ind;
        ASSERT(ind<=LastY);
        RETURN 2*Field[iX,0]-Field[iX,ind];
      ELSIF ind>LastY THEN
        ind:=2*LastY-ind;
        ASSERT(ind>=0);
        RETURN 2*Field[iX,LastY]-Field[iX,ind];
      ELSE
        RETURN Field[iX,ind]
      END;
    END Extension;
<* POP *>

  VAR
    i:LONGINT;
    d:Reg.Data;
    SlopeRe,SlopeIm,y:Float;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));
    ASSERT(delta>=1);

    LastY:=a.LastY;

    Reg.Init(d);
    FOR i:=iY-delta TO iY+delta DO
      y:=IndToY(a,i);
      Reg.AddPoint(d,y,RE(Extension(i)));
    END;
    SlopeRe:=Reg.Slope(d);

    Reg.Init(d);
    FOR i:=iY-delta TO iY+delta DO
      y:=IndToY(a,i);
      Reg.AddPoint(d,y,IM(Extension(i)));
    END;
    SlopeIm:=Reg.Slope(d);

    RETURN SlopeRe + 1.0i*SlopeIm
  END DiffYIndSmoothed;

PROCEDURE Diff2XIndLeft * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on x in point [iX,iY], which is
      calculates using points [iX,iY], [iX-1,iY], [iX-2,iY], [iX-3,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=3);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    RETURN (2*Field[iX,iY]-5*Field[iX-1,iY]+4*Field[iX-2,iY]-Field[iX-3,iY])*a.InvSqrDx
  END Diff2XIndLeft;

PROCEDURE Diff2XIndRight * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on x in point [iX,iY], which is
      calculates using points [iX,iY], [iX+1,iY], [iX+2,iY], [iX+3,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX-3);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    RETURN (2*Field[iX,iY]-5*Field[iX+1,iY]+4*Field[iX+2,iY]-Field[iX+3,iY])*a.InvSqrDx
  END Diff2XIndRight;

PROCEDURE Diff2XInd * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on x in point [iX,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));

    IF iX=0 THEN
      RETURN Diff2XIndRight(a,Field,0,iY)
    ELSIF iX=a.LastX THEN
      RETURN Diff2XIndLeft(a,Field,a.LastX,iY)
    ELSE
      RETURN (Field[iX+1,iY]+Field[iX-1,iY]-2*Field[iX,iY])*a.InvSqrDx
    END;
  END Diff2XInd;

PROCEDURE Diff2YIndLeft * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on y in point [iX,iY], which is
      calculates using points [iX,iY], [iX,iY-1], [iX,iY-2], [iX,iY-3].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=3);
    ASSERT(iY<=a.LastY);

    RETURN (2*Field[iX,iY]-5*Field[iX,iY-1]+4*Field[iX,iY-2]-Field[iX,iY-3])*a.InvSqrDy
  END Diff2YIndLeft;

PROCEDURE Diff2YIndRight * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on y in point [iX,iY], which is
      calculates using points [iX,iY], [iX,iY+1], [iX,iY+2], [iX,iY+3].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY-3);

    RETURN (2*Field[iX,iY]-5*Field[iX,iY+1]+4*Field[iX,iY+2]-Field[iX,iY+3])*a.InvSqrDy
  END Diff2YIndRight;

PROCEDURE Diff2YInd * (a-:Area; Field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns the second derivative of field on y in point [iX,iY].
      The second order of accuracy. *)
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(iXYIn(a,iX,iY));

    IF iY=0 THEN
      RETURN Diff2YIndRight(a,Field,iX,0)
    ELSIF iY=a.LastY THEN
      RETURN Diff2YIndLeft(a,Field,iX,a.LastY)
    ELSE
      RETURN (Field[iX,iY+1]+Field[iX,iY-1]-2*Field[iX,iY])*a.InvSqrDy
    END;
  END Diff2YInd;

<* PUSH *>
<* WOFF901 + *> (* In XDS 2.60 beta here is a wrong message *)

PROCEDURE CoordToGrid(a-:Area; VAR x,y:Float; VAR iX,iY:LONGINT);
  BEGIN
    x:=(x-a.MinX)*a.InvDx;
    y:=(y-a.MinY)*a.InvDy;
    iX:=ENTIER(x);
    iY:=ENTIER(y);
    x:=x-iX;
    y:=y-iY;
  END CoordToGrid;

<* POP *>

PROCEDURE XIn (a-:Area; x:Float): BOOLEAN;
  (* Asserts that x is in the area and returns TRUE *)
  BEGIN
    ASSERT(x>=a.MinX);
    ASSERT(x<=a.MaxX);

    RETURN TRUE
  END XIn;

PROCEDURE YIn (a-:Area; y:Float): BOOLEAN;
  (* Asserts that y is in the area and returns TRUE *)
  BEGIN
    ASSERT(y>=a.MinY);
    ASSERT(y<=a.MaxY);

    RETURN TRUE
  END YIn;

PROCEDURE XYIn (a-:Area; x,y:Float): BOOLEAN;
  (* Asserts that (x,y) is in the area and returns TRUE *)
  BEGIN
    ASSERT(XIn(a,x));
    ASSERT(YIn(a,y));

    RETURN TRUE
  END XYIn;

PROCEDURE Interpolate * (a-:Area; Field-:Array2; x,y:Float): Complex;
  (** Returns interpolation of function, given in points by field, at point (x,y).
      The second order of accuracy. *)
  VAR
    iX,iY:LONGINT;
    z0,z1:Complex;
  BEGIN
    ASSERT(CorrectField(a,Field));
    ASSERT(XYIn(a,x,y));

    CoordToGrid(a,x,y,iX,iY);
    IF iX=a.LastX THEN
      iX:=a.LastX-1;
      x:=1;
    END;
    IF iY=a.LastY THEN
      iY:=a.LastY-1;
      y:=1;
    END;

    z0:=Field[iX,iY] + x*(Field[iX+1,iY] - Field[iX,iY]);
    z1:=Field[iX,iY+1] + x*(Field[iX+1,iY+1] - Field[iX,iY+1]);
    RETURN z0 + y*(z1-z0)
  END Interpolate;

PROCEDURE DetailedInterpolate (Field-:Array2; iX,iY:LONGINT; x,y:Float;
                               dx,dy:LONGINT): Complex;
  VAR
    a,b,c,d,e,z0:Complex;
  BEGIN
    ASSERT(ABS(dx)=1);
    ASSERT(ABS(dy)=1);
    ASSERT(iX>=0);
    ASSERT(iX<LEN(Field));
    ASSERT(iX+dx+dx>=0);
    ASSERT(iX+dx+dx<LEN(Field));
    ASSERT(iY>=0);
    ASSERT(iY<LEN(Field,1));
    ASSERT(iY+dy+dy>=0);
    ASSERT(iY+dy+dy<LEN(Field,1));

    z0:=Field[iX,iY];
    a:=Field[iX+dx,iY]-z0;
    b:=Field[iX+dx+dx,iY]-z0-2*a;
    c:=Field[iX,iY+dy]-z0;
    d:=Field[iX,iY+dy+dy]-z0-2*c;
    e:=Field[iX+dx,iY+dy]-z0-a-c;
    RETURN z0 + a*x + b*x*(x-1) + c*y + d*y*(y-1) + e*x*y
  END DetailedInterpolate;

PROCEDURE Interpolate3 * (a-:Area; Field-:Array2; x,y: Float): Complex;
  (** Returns interpolation of function, given in points by field, at point (x,y).
      The 3rd order of accuracy. *)
  VAR
    iX,iY:LONGINT;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(XYIn(a,x,y));

    CoordToGrid(a,x,y,iX,iY);
    IF iX=a.LastX THEN
      iX:=a.LastX-1;
      x:=1;
    END;
    IF iY=a.LastY THEN
      iY:=a.LastY-1;
      y:=1;
    END;

    IF ((x<0.5) OR (iX=a.LastX-1)) & (iX#0) THEN
      IF ((y<0.5) OR (iY=a.LastY-1)) & (iY#0) THEN
        RETURN DetailedInterpolate(Field,iX+1,iY+1,1-x,1-y,-1,-1);
      ELSE
        RETURN DetailedInterpolate(Field,iX+1,iY,1-x,y,-1,1);
      END;
    ELSE
      IF ((y<0.5) OR (iY=a.LastY-1)) & (iY#0) THEN
        RETURN DetailedInterpolate(Field,iX,iY+1,x,1-y,1,-1);
      ELSE
        RETURN DetailedInterpolate(Field,iX,iY,x,y,1,1);
      END;
    END;
  END Interpolate3;

PROCEDURE DiffX * (a-:Area; Field-:Array2; x,y:Float): Complex;
  (** Returns interpolation of the first derivative on x of function,
      given in points by field, at point (x,y).
      The first order of accuracy. *)
  VAR
    iX,iY:LONGINT;
    b,d:Complex;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(XYIn(a,x,y));

    CoordToGrid(a,x,y,iX,iY);
    IF iX=a.LastX THEN
      iX:=a.LastX-1;
    END;
    IF iY=a.LastY THEN
      iY:=a.LastY-1;
      y:=1;
    END;

    b:=Field[iX+1,iY]-Field[iX,iY];
    d:=Field[iX,iY]+Field[iX+1,iY+1]-Field[iX,iY+1]-Field[iX+1,iY];
    RETURN (b+d*y)*a.InvDx;
  END DiffX;

PROCEDURE DiffY * (a-:Area; Field-:Array2; x,y:Float): Complex;
  (** Returns interpolation of the first derivative on y of function,
      given in points by field, at point (x,y).
      The first order of accuracy. *)
  VAR
    iX,iY:LONGINT;
    c,d:Complex;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,Field));
    ASSERT(XYIn(a,x,y));

    CoordToGrid(a,x,y,iX,iY);
    IF iX=a.LastX THEN
      iX:=a.LastX-1;
      x:=1;
    END;
    IF iY=a.LastY THEN
      iY:=a.LastY-1;
    END;

    c:=Field[iX,iY+1]-Field[iX,iY];
    d:=Field[iX,iY]+Field[iX+1,iY+1]-Field[iX,iY+1]-Field[iX+1,iY];
    RETURN (c+d*x)*a.InvDy;
  END DiffY;

TYPE
  Func * = ComFunc.Float2;

PROCEDURE SaveFunc * (a-:Area; VAR f:Func;
                      FileName-:ARRAY OF CHAR; multiplier:Complex);
  (** Saves f.f(x,y)*multiplier in points of the grid
      in file FileName in text 3-columns format.
      The file is overwritten if exists. *)
  VAR
    iX,iY:LONGINT;
    x,y:Float;
    fm:Out.Formatter;
    res:Complex;
  BEGIN
    ASSERT(Correct(a));

    Out.OpenFile(fm,FileName,Out.Rewrite);
    FOR iX:=0 TO a.LastX DO
      x:=IndToX(a,iX);
      FOR iY:=0 TO a.LastY DO
        y:=IndToY(a,iY);
        res:=f.f(x,y)*multiplier;
        Out.Real(fm,x,0);
        Out.Char(fm,Tab);
        Out.Real(fm,y,0);
        Out.Char(fm,Tab);
        Out.Real(fm,RE(res),0);
        Out.Char(fm,Tab);
        Out.Real(fm,IM(res),0);
        Out.Ln(fm);
      END;
    END;
    Out.Close(fm)
  END SaveFunc;

TYPE
  SaveFuncProc = RECORD (ComFunc.Float2)
    func:ComFunc.Simple2;
  END;

PROCEDURE (VAR f:SaveFuncProc) f (z,r:Float): Complex;
  BEGIN
    RETURN f.func(z,r)
  END f;

PROCEDURE SaveSimpleFunc * (a-:Area; f:ComFunc.Simple2;
                            FileName-:ARRAY OF CHAR; multiplier:Complex);
  (** Saves f(x,y)*multiplier in points of the grid
      in file FileName in text 3-columns format.
      The file is overwritten if exists. *)
  VAR
    sf:SaveFuncProc;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(f#NIL);

    sf.func:=f;
    SaveFunc(a,sf,FileName,multiplier);
  END SaveSimpleFunc;

PROCEDURE SaveArray * (a-:Area; field-:Array2;
                       FileName-:ARRAY OF CHAR; multiplier:Complex);
  (** Saves field[iX,iY]*multiplier in file FileName in text
      3-columns format. The file is overwritten if exists. *)
  VAR
    iX,iY:LONGINT;
    x,y:Float;
    f:Out.Formatter;
    res:Complex;
  BEGIN
    ASSERT(Correct(a));
    ASSERT(CorrectField(a,field));

    Out.OpenFile(f,FileName,Out.Rewrite);
    FOR iX:=0 TO a.LastX DO
      x:=IndToX(a,iX);
      FOR iY:=0 TO a.LastY DO
        y:=IndToY(a,iY);
        res:=field[iX,iY]*multiplier;
        Out.Real(f,x,0);
        Out.Char(f,Tab);
        Out.Real(f,y,0);
        Out.Char(f,Tab);
        Out.Real(f,RE(res),0);
        Out.Char(f,Tab);
        Out.Real(f,IM(res),0);
        Out.Ln(f);
      END;
    END;
    Out.Close(f)
  END SaveArray;

PROCEDURE LoadArray * (a-:Area; VAR field:Array2;
                       FileName-:ARRAY OF CHAR; multiplier:Complex);
  (** Loads values, multiplied by multiplier, from file FileName to the field.
      Format of the file is the same one used by SaveXXXX procedures. *)
  VAR
    iX,iY:LONGINT;
    s:In.Scanner;
    x,y:Float;
  BEGIN
    ASSERT(CorrectField(a,field));

    In.OpenFile(s,FileName);
    FOR iX:=0 TO a.LastX DO
      FOR iY:=0 TO a.LastY DO
        In.Float(s,x);
        In.Float(s,x);
        In.Float(s,x);
        In.Float(s,y);
        field[iX,iY]:=(x+1.0i*y)*multiplier;
      END;
    END;
    In.Close(s);
  END LoadArray;

PROCEDURE NumberOfFirstSameValues (FileName-:ARRAY OF CHAR): LONGINT;
  (* Returns number of first lines with similar x *)
  VAR
    s:In.Scanner;
  PROCEDURE Next (VAR x:Float): BOOLEAN;
    BEGIN
      In.SkipSpaces(s);
      IF In.EndOfInput(s) THEN
        x:=0;
        RETURN FALSE
      ELSE
        In.Float(s,x);
        In.SkipLine(s);
        RETURN TRUE;
      END;
    END Next;
  VAR
    OK:BOOLEAN;
    i:LONGINT;
    a,x:Float;
  BEGIN
    In.OpenFile(s,FileName);

<* PUSH *>
<* WOFF900 + *>
    OK:=Next(a);
<* POP *>
    ASSERT(OK);
    i:=1;
    WHILE Next(x) & (x=a) DO
      INC(i);
    END;

    In.Close(s);
    RETURN i;
  END NumberOfFirstSameValues;

PROCEDURE LoadArrayAndArea * (VAR a:Area; VAR res:Array2Ptr;
                              FileName-:ARRAY OF CHAR; multiplier:Complex);
  (** Loads values, multiplied by multiplier, from file FileName to the field,
      as well as the corresponding area.
      Format of the file is the same one used by SaveXXXX procedures. *)
  VAR
    s:In.Scanner;
    iX,iY:LONGINT;
  PROCEDURE ReadVal();
    VAR
      re,im:Float;
    BEGIN
      In.Float(s,re);
      In.Float(s,im);
      res[iX,iY]:=multiplier*(re + 1.0i*im);
      In.SkipLine(s);
    END ReadVal;
  VAR
    len,lenx,leny:LONGINT;
    MinX,MinY,MaxX,MaxY:Float;
  BEGIN
    leny:=NumberOfFirstSameValues(FileName);
    len:=TextData.LinesCount(FileName);
    ASSERT(len MOD leny = 0);
    lenx:=len DIV leny;
    ASSERT(lenx>0);
    ASSERT(leny>0);

    In.OpenFile(s,FileName);

    NEW(res,lenx,leny);

<* PUSH *>
<* WOFF900 + *>
    iX:=0;
    iY:=0;
<* POP *>
    In.Float(s,MinX);
    In.Float(s,MinY);
    ReadVal();
    iX:=0;
    iY:=1;
    WHILE (iX#lenx-1) OR (iY#leny-1) DO
      In.SkipUntilSpaces(s); (* x *)
      In.SkipSpaces(s);
      In.SkipUntilSpaces(s); (* y *)
      ReadVal();

      INC(iY);
      IF iY=leny THEN
        INC(iX);
        iY:=0;
      END;
    END;
    In.Float(s,MaxX);
    In.Float(s,MaxY);
    ReadVal();

    In.Close(s);
    Init(a,lenx-1,leny-1,MinX,MaxX,MinY,MaxY);
  END LoadArrayAndArea;

TYPE
  InterpolationFunc = RECORD (Func)
    area:Area;
    field:Array2Ptr;
  END;

PROCEDURE (VAR f:InterpolationFunc) f (x,y:Float): Complex;
  BEGIN
    RETURN Interpolate(f.area,f.field^,x,y);
  END f;

PROCEDURE Interpolation * (area-:Area; field:Array2Ptr): POINTER TO Func;
  (** Returns a function, which is interpolation of the specified field in the
      specified area. Dimensions of the field shold correspond to the grid
      of the area. *)
  VAR
    f:POINTER TO InterpolationFunc;
  BEGIN
    ASSERT(Correct(area));
    ASSERT(field#NIL);
    ASSERT(CorrectField(area,field^));

    NEW(f);
    f.area:=area;
    f.field:=field;
    RETURN f
  END Interpolation;

PROCEDURE Fill * (VAR f:ComFunc.Float2; area-:Area; VAR field:Array2);
  (** Fills the specified field in the specified area by result of the
      specified function f *)
  VAR
    iX,iY:LONGINT;
    x,y:Float;
  BEGIN
    ASSERT(Correct(area));
    ASSERT(CorrectField(area,field));

    FOR iX:=0 TO area.LastX DO
      x:=IndToX(area,iX);
      FOR iY:=0 TO area.LastY DO
        y:=IndToY(area,iY);
        field[iX,iY]:=f.f(x,y)
      END;
    END;
  END Fill;

PROCEDURE Reduced * (InitialArea-:Area; VAR ReducedArea:Area);
  (** Initiates the reduced area so, that it presents the initial area
      excluding outer points of the grid *)
  VAR
    dx,dy:Float;
  BEGIN
    ASSERT(Correct(InitialArea));

    dx:=InitialArea.dx;
    dy:=InitialArea.dy;
    Init(ReducedArea, InitialArea.LastX-2, InitialArea.LastY-2,
         InitialArea.MinX+dx, InitialArea.MaxX-dx,
         InitialArea.MinY+dy, InitialArea.MaxY-dy);
  END Reduced;

END mArea2DComplex.