(** This module provides common operations with complex arrays.

    Last chenged: 20.11.2015 *)
<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE mArraysComplex;

IMPORT
  MathComplex:=mMathComplex, Math:=mMath;

<* IF Float="REAL" THEN *>              IMPORT FMath:=RealMath, CMath:=ComplexMath;
<* ELSIF Float="LONGREAL" THEN *>       IMPORT FMath:=LongMath, CMath:=LongComplexMath;
<* ELSIF Float="LONGLONGREAL" THEN *>
  'ERROR: LONGLONG is not supported
<* ELSE *>
  'ERROR: equation Float is not defined
<* END *>

TYPE
  Float * = MathComplex.Float;
  Complex * = MathComplex.Complex;
  Array * = MathComplex.Array;
  ArrayPtr * = MathComplex.ArrayPtr;
  Array2 * = MathComplex.Array2;
  Array2Ptr * = MathComplex.Array2Ptr;

PROCEDURE Abs (x:Complex): Float;
  BEGIN
    RETURN MathComplex.abs(x)
  END Abs;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN FMath.sqrt(x)
  END sqrt;

PROCEDURE Fill * (VAR x:Array; value:Complex);
  (** x[i]:=value *)
  VAR
    i:LONGINT;
  BEGIN
    FOR i:=0 TO LEN(x)-1 DO
      x[i]:=value
    END;
  END Fill;

PROCEDURE Filled * (len:LONGINT; value:Complex): ArrayPtr;
  (** Creates and returns an array with specified dimensions and filled by
      the specified value *)
  VAR
    res:ArrayPtr;
    i:LONGINT;
  BEGIN
    ASSERT(len>0);

    NEW(res,len);
    FOR i:=0 TO len-1 DO
      res[i]:=value
    END;
    RETURN res;
  END Filled;

PROCEDURE Assign * (VAR dest:Array; source-:Array);
  (** dest:=source *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(LEN(dest)=LEN(source));

    FOR i:=0 TO LEN(dest)-1 DO
      dest[i]:=source[i]
    END;
  END Assign;

PROCEDURE Copy * (source-:Array): ArrayPtr;
  (** Returns copy of source *)
  VAR
    res:ArrayPtr;
  BEGIN
    NEW(res,LEN(source));
    Assign(res^,source);
    RETURN res;
  END Copy;

PROCEDURE Swap * (VAR x1,x2:ArrayPtr);
  (** tmp:=x1; x1:=x2; x2:=tmp *)
  VAR
    tmp:ArrayPtr;
  BEGIN
    tmp:=x1;
    x1:=x2;
    x2:=tmp;
  END Swap;

PROCEDURE Prod * (x1-,x2-:Array): Complex;
  (** Returns scalar product of x1 and x2 *)
  VAR
    i:LONGINT;
    sum:Complex;
  BEGIN
    ASSERT(LEN(x1)=LEN(x2));

    sum:=0;
    FOR i:=0 TO LEN(x1)-1 DO
      sum:=sum+x1[i]*CMath.conj(x2[i])
    END;
    RETURN sum;
  END Prod;

PROCEDURE SqrNorm * (x-:Array): Float;
  (** Returns ABS(x)^2 *)
  VAR
    i:LONGINT;
    sum:Float;
  BEGIN
    sum:=0;
    FOR i:=0 TO LEN(x)-1 DO
      sum:=sum + sqr(RE(x[i])) + sqr(IM(x[i]))
    END;
    RETURN sum;
  END SqrNorm;

PROCEDURE Norm * (x-:Array): Float;
  (** Returns ABS(x) *)
  BEGIN
    RETURN sqrt(SqrNorm(x));
  END Norm;

PROCEDURE Normalize * (VAR x:Array): Float;
  (** IF x#0 THEN normalize x to 1 END;
      RETURN norm of initial x  *)
  VAR
    norm,InvNorm:Float;
    i:LONGINT;
  BEGIN
    norm:=Norm(x);
    IF norm#0 THEN
      InvNorm:=1/norm;
      FOR i:=0 TO LEN(x)-1 DO
        x[i]:=x[i]*InvNorm;
      END;
    END;
    RETURN norm;
  END Normalize;

PROCEDURE SqrNormOfDifference * (x1-,x2-:Array): Float;
  (** Returns ABS(x1-x2)^2 *)
  VAR
    i:LONGINT;
    sum:Float;
    diff:Complex;
  BEGIN
    ASSERT(LEN(x1)=LEN(x2));

    sum:=0;
    FOR i:=0 TO LEN(x1)-1 DO
      diff:=x1[i]-x2[i];
      sum:=sum + sqr(RE(diff)) + sqr(IM(diff))
    END;
    RETURN sum;
  END SqrNormOfDifference;

PROCEDURE NormOfDifference * (x1-,x2-:Array): Float;
  (** Returns ABS(x1-x2) *)
  BEGIN
    RETURN sqrt(SqrNormOfDifference(x1,x2))
  END NormOfDifference;

PROCEDURE Fill2 * (VAR x:Array2; value:Complex);
  (** x[i,j]:=value *)
  VAR
    i,j:LONGINT;
  BEGIN
    FOR i:=0 TO LEN(x)-1 DO
      FOR j:=0 TO LEN(x,1)-1 DO
        x[i,j]:=value
      END;
    END;
  END Fill2;

PROCEDURE Filled2 * (len0,len1:LONGINT; value:Complex): Array2Ptr;
  (** Creates and returns an array with specified dimensions and filled by
      the specified value *)
  VAR
    res:Array2Ptr;
    i,j:LONGINT;
  BEGIN
    ASSERT(len0>0);
    ASSERT(len1>0);

    NEW(res,len0,len1);
    FOR i:=0 TO len0-1 DO
      FOR j:=0 TO len1-1 DO
        res[i,j]:=value
      END;
    END;
    RETURN res;
  END Filled2;

PROCEDURE Assign2 * (VAR dest:Array2; source-:Array2);
  (** dest:=source *)
  VAR
    i,j:LONGINT;
  BEGIN
    ASSERT(LEN(dest)=LEN(source));
    ASSERT(LEN(dest,1)=LEN(source,1));

    FOR i:=0 TO LEN(dest)-1 DO
      FOR j:=0 TO LEN(dest,1)-1 DO
        dest[i,j]:=source[i,j]
      END;
    END;
  END Assign2;

PROCEDURE Copy2 * (source-:Array2): Array2Ptr;
  (** Returns copy of source *)
  VAR
    res:Array2Ptr;
  BEGIN
    NEW(res,LEN(source),LEN(source,1));
    Assign2(res^,source);
    RETURN res;
  END Copy2;

PROCEDURE Swap2 * (VAR x1,x2:Array2Ptr);
  (** tmp:=x1; x1:=x2; x2:=tmp *)
  VAR
    tmp:Array2Ptr;
  BEGIN
    tmp:=x1;
    x1:=x2;
    x2:=tmp;
  END Swap2;

PROCEDURE Prod2 * (x1-,x2-:Array2): Complex;
  (** Returns scalar product of x1 and x2 *)
  VAR
    i,j:LONGINT;
    sum:Complex;
  BEGIN
    ASSERT(LEN(x1)=LEN(x2));
    ASSERT(LEN(x1,1)=LEN(x2,1));

    sum:=0;
    FOR i:=0 TO LEN(x1)-1 DO
      FOR j:=0 TO LEN(x1,1)-1 DO
        sum:=sum+x1[i,j]*CMath.conj(x2[i,j])
      END
    END;
    RETURN sum;
  END Prod2;

PROCEDURE SqrNorm2 * (x-:Array2): Float;
  (** Returns ABS(x)^2 *)
  VAR
    sum:Float;
    i,j:LONGINT;
  BEGIN
    sum:=0;
    FOR i:=0 TO LEN(x)-1 DO
      FOR j:=0 TO LEN(x,1)-1 DO
        sum:=sum + sqr(RE(x[i,j])) + sqr(IM(x[i,j]))
      END;
    END;
    RETURN sum;
  END SqrNorm2;

PROCEDURE Norm2 * (x-:Array2): Float;
  (** Returns ABS(x) *)
  BEGIN
    RETURN sqrt(SqrNorm2(x));
  END Norm2;

PROCEDURE SqrNormOfDifference2 * (x1-,x2-:Array2): Float;
  (** Returns ABS(x1-x2)^2 *)
  VAR
    sum:Float;
    i,j:LONGINT;
    diff:Complex;
  BEGIN
    ASSERT(LEN(x1)=LEN(x2));
    ASSERT(LEN(x1,1)=LEN(x2,1));

    sum:=0;
    FOR i:=0 TO LEN(x1)-1 DO
      FOR j:=0 TO LEN(x1,1)-1 DO
        diff:=x1[i,j]-x2[i,j];
        sum:=sum + sqr(RE(diff)) + sqr(IM(diff))
      END;
    END;
    RETURN sum;
  END SqrNormOfDifference2;

PROCEDURE NormOfDifference2 * (x1-,x2-:Array2): Float;
  (** Returns ABS(x1-x2) *)
  BEGIN
    RETURN sqrt(SqrNormOfDifference2(x1,x2))
  END NormOfDifference2;

PROCEDURE MaxAbs2 * (field-:Array2): Float;
  (** Returns maximum of module value of the aray elements *)
  VAR
    iZ,iR:LONGINT;
    max,x:Float;
  BEGIN
    max:=-Math.MaxFloat;
    FOR iZ:=0 TO LEN(field)-1 DO
      FOR iR:=0 TO LEN(field,1)-1 DO
        x:=Abs(field[iZ,iR]);
        IF max<x THEN
          max:=x;
        END;
      END;
    END;
    RETURN max;
  END MaxAbs2;

END mArraysComplex.