(** This module provides the same tools as mArea2D,
    but for 3D case with cylindrical coordinates and symmetry,
    with axis at R=Y=0. Thus the area is a half of a cross-section of the
    cylindrical area.

    X correcponds to the axial direction, Y - to the radial one.
    In this module, for convenience, X and Y are called Z and R.

    There is a feature in the grid: line r=0 lies in middle of
    points lines. For solution of different equations an additional
    line of points r=-dr/2 is created, r-index of this line is 0. The first
    real line of points r=dr/2 have r-index 1.

    Last change: 04.01.2016 *)

<* O2EXTENSIONS + *>
MODULE mCylArea;

IMPORT
  Math:=mMath, BasicArea:=mArea2D, ComFunc:=mCommonFunctions;

TYPE
  Float * = Math.Float;
  Array2 = Math.Array2;
  Array2Ptr = Math.Array2Ptr;

  Area * = BasicArea.Area;
(**  Area*  = RECORD
       MinX-,MaxX-,MinY-,MaxY-: Float;         (** Boundaries of the area *)
       LastX-,LastY-          : LONGINT; (** Grid in the area is indexed from 0 to LastX, LastY *)
       dx-,dy-                : Float; (** x-size and y-size of frid cells *)
       InvDx-,InvDy-          : Float; (** 1/dx and 1/dy *)
       Inv2dx-,Inv2dy-        : Float; (** 1/(2*dx) and (1/(2*dy)) *)
       InvSqrDx-,InvSqrDy-    : Float; (** 1/sqr(dx) and 1/sqr(dy) *)
     END; *)
  (** The area. x corresponds to z, y corresponds to r. *)

CONST
  pi = Math.pi;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x;
  END sqr;

PROCEDURE Init * (VAR a:Area; LastZ,LastR:LONGINT; MinZ,MaxZ,Rad:Float);
  (** Initiates the area and the grid. MinZ,MaxZ,Rad are the area
      boundaries, points of the grid are indexed in [0..LastZ,0..LastR].
      Outermost points are in the area boundaries.
      About zeroth points line see the very first comment. *)
  BEGIN
    BasicArea.Init(a,LastZ,LastR,MinZ,MaxZ,-Rad/(2*LastR-1),Rad);
  END Init;

PROCEDURE Correct* (a-:Area): BOOLEAN;
  (* Asserts that the area is valid for differential operations
     and returns TRUE *)
  BEGIN
    ASSERT(a.MinY<0);
    ASSERT(a.MinY+a.dy>0);
    RETURN BasicArea.Correct(a);
  END Correct;

PROCEDURE MinZ * (a-:Area): Float;
  (** Returns minimal z-coordinate of the area. *)
  BEGIN
    RETURN a.MinX
  END MinZ;

PROCEDURE MaxZ * (a-:Area): Float;
  (** Returns maximal z-coordinate of the area. *)
  BEGIN
    RETURN a.MaxX
  END MaxZ;

PROCEDURE Rad * (a-:Area): Float;
  (** Returns radius of the area. *)
  BEGIN
    RETURN a.MaxY
  END Rad;

PROCEDURE MinR * (a-:Area): Float;
  (** Returns a negative r-coordinate of the 0th r-line. *)
  BEGIN
    RETURN a.MinY
  END MinR;

PROCEDURE LastZ * (a-:Area): LONGINT;
  (** The grid points are indexed form 0 to LastZ in z dimension. *)
  BEGIN
    RETURN a.LastX
  END LastZ;

PROCEDURE LastR * (a-:Area): LONGINT;
  (** The grid points are indexed form 0 to LastR in r dimension. *)
  BEGIN
    RETURN a.LastY
  END LastR;

PROCEDURE Dz * (a-:Area): Float;
  (** Returns z-size of the grid cell. *)
  BEGIN
    RETURN a.dx
  END Dz;

PROCEDURE Dr * (a-:Area): Float;
  (** Returns r-size of the grid cell. *)
  BEGIN
    RETURN a.dy
  END Dr;

PROCEDURE IndToZ * (a-:Area; iZ:LONGINT): Float;
  (** Returns coordinate z corresponding to index iZ. *)
  BEGIN
    RETURN BasicArea.IndToX(a,iZ)
  END IndToZ;

PROCEDURE ZToInd * (a-:Area; z:Float): LONGINT;
  (** Returns index, nearest to coordinate z. *)
  BEGIN
    RETURN BasicArea.XToInd(a,z);
  END ZToInd;

PROCEDURE IndToR * (a-:Area; iR:LONGINT): Float;
  (** Returns coordinate r corresponding to index iR. *)
  BEGIN
    RETURN BasicArea.IndToY(a,iR)
  END IndToR;

PROCEDURE RToInd * (a-:Area; r:Float): LONGINT;
  (** Returns index, nearest to coordinate r. *)
  BEGIN
    RETURN BasicArea.YToInd(a,r);
  END RToInd;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE CorrectField * (a-:Area; field-:Array2): BOOLEAN;
  (** Asserts that the field dimensions correspond to the area
      and returns TRUE *)
  BEGIN
    ASSERT(BasicArea.CorrectField(a,field));

    RETURN TRUE
  END CorrectField;
<* POP *>

PROCEDURE DiffZIndLeft * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on z in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ-1,iR], and [iZ-2,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffXIndLeft(a,Field,iZ,iR)
  END DiffZIndLeft;

PROCEDURE DiffZIndRight * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on z in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ+1,iR], and [iZ+2,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffXIndRight(a,Field,iZ,iR)
  END DiffZIndRight;

PROCEDURE DiffZInd * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on z in point [iZ,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffXInd(a,Field,iZ,iR)
  END DiffZInd;

PROCEDURE DiffRIndLeft * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on r in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ,iR-1], and [iZ,iR-2].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffYIndLeft(a,Field,iZ,iR)
  END DiffRIndLeft;

PROCEDURE DiffRIndRight * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on r in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ,iR+1], and [iZ,iR+2].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffYIndRight(a,Field,iZ,iR)
  END DiffRIndRight;

PROCEDURE DiffRInd * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the first derivative of field on r in point [iZ,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffYInd(a,Field,iZ,iR)
  END DiffRInd;

PROCEDURE DiffZIndSmoothed * (a-:Area; Field-:Array2;
                              iZ,iR,delta:LONGINT): Float;
  (** Returns the first derivative of field on z in point [iZ,iR].
      The result is smoothed within [iZ-delta..iZ+delta].
      If delta=1, the procedure has the second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffXIndSmoothed(a,Field,iZ,iR,delta)
  END DiffZIndSmoothed;

PROCEDURE DiffRIndSmoothed * (a-:Area; Field-:Array2;
                              iZ,iR,delta:LONGINT): Float;
  (** Returns the first derivative of field on r in point [iZ,iR].
      The result is smoothed within [iR-delta..iR+delta].
      If delta=1, the procedure has the second order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffYIndSmoothed(a,Field,iZ,iR,delta)
  END DiffRIndSmoothed;

PROCEDURE Diff2ZIndLeft * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on z in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ-1,iR], [iZ-2,iR], [iZ-3,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2XIndLeft(a,Field,iZ,iR)
  END Diff2ZIndLeft;

PROCEDURE Diff2ZIndRight * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on z in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ+1,iR], [iZ+2,iR], [iZ+3,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2XIndRight(a,Field,iZ,iR)
  END Diff2ZIndRight;

PROCEDURE Diff2ZInd * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on z in point [iZ,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2XInd(a,Field,iZ,iR)
  END Diff2ZInd;

PROCEDURE Diff2RIndLeft * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on r in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ,iR-1], [iZ,iR-2], [iZ,iR-3].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2YIndLeft(a,Field,iZ,iR)
  END Diff2RIndLeft;

PROCEDURE Diff2RIndRight * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on r in point [iZ,iR], which is
      calculates using points [iZ,iR], [iZ,iR+1], [iZ,iR+2], [iZ,iR+3].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2YIndRight(a,Field,iZ,iR)
  END Diff2RIndRight;

PROCEDURE Diff2RInd * (a-:Area; Field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns the second derivative of field on r in point [iZ,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Diff2YInd(a,Field,iZ,iR)
  END Diff2RInd;

PROCEDURE DivergenceRInd * (a-:Area; field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns radial part of divergence of field at point [iZ,iR], that is
      d/dr(field) + field/r.
      The second order of accuracy. *)
  BEGIN
    RETURN DiffRInd(a,field,iZ,iR)+field[iZ,iR]/IndToR(a,iR)
  END DivergenceRInd;

PROCEDURE LaplaceRInd * (a-:Area; field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns radial part of laplasian of field at point [iZ,iR], that is
      d^2/dr^2(field) + d/dr(field)/r.
      The second order of accuracy. *)
  BEGIN
    RETURN Diff2RInd(a,field,iZ,iR)+DiffRInd(a,field,iZ,iR)/IndToR(a,iR)
  END LaplaceRInd;

PROCEDURE LaplaceInd * (a-:Area; field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns laplasian of field at point [iZ,iR].
      The second order of accuracy. *)
  BEGIN
    RETURN LaplaceRInd(a,field,iZ,iR)+Diff2ZInd(a,field,iZ,iR)
  END LaplaceInd;

PROCEDURE ElVol * (a-:Area; iZ,iR:LONGINT): Float;
  (** Returns value of 3-dimensional elementary volume with center
      at point [iZ,iR]. If the point lies on a boundary, then
      only volume of internal part is returned. *)
  VAR
    result,dr:Float;
  BEGIN
    IF iR=LastR(a) THEN
      result:=Rad(a);
      dr:=Dr(a);
      result:=pi*(result*dr - sqr(dr)/4)*Dz(a);
    ELSE
      result:=2*pi*IndToR(a,iR)*Dr(a)*Dz(a);
    END;

    IF (iZ=0) OR (iZ=LastZ(a)) THEN
      result:=result/2;
    END;

    RETURN result;
  END ElVol;

PROCEDURE ElRadArea * (a-:Area; iR:LONGINT): Float;
  (** Returns value of 2-dimensional elementary ring area at a cross section,
      which is orthagonal to the axis, with center at point [iR].
      If the point lies on a boundary, then only area of internal part
      is returned. *)
  VAR
    result,rad,dr:Float;
    lastR:LONGINT;
  BEGIN
    lastR:=LastR(a);

    dr:=Dr(a);
    IF iR=lastR THEN
      rad:=Rad(a);
      result:=pi*(rad*dr - sqr(dr)/4);
    ELSE
      result:=2*pi*IndToR(a,iR)*dr;
    END;

    RETURN result
  END ElRadArea;

PROCEDURE Interpolate * (a-:Area; Field-:Array2; z,r: Float): Float;
  (** Returns interpolation of function, given in points by field, at point (z,r).
      The second order of accuracy. *)
  BEGIN
    RETURN BasicArea.Interpolate(a,Field,z,r);
  END Interpolate;

PROCEDURE Interpolate3 * (a-:Area; Field-:Array2; z,r: Float): Float;
  (** Returns interpolation of function, given in points by field, at point (z,r).
      The 3rd order of accuracy. *)
  BEGIN
    RETURN BasicArea.Interpolate3(a,Field,z,r);
  END Interpolate3;

PROCEDURE DiffZ * (a-:Area; Field-:Array2; z,r: Float): Float;
  (** Returns interpolation of the first derivative on z of function,
      given in points by field, at point (z,r).
      The first order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffX(a,Field,z,r);
  END DiffZ;

PROCEDURE DiffR * (a-:Area; Field-:Array2; z,r: Float): Float;
  (** Returns interpolation of the first derivative on r of function,
      given in points by field, at point (z,r).
      The first order of accuracy. *)
  BEGIN
    RETURN BasicArea.DiffY(a,Field,z,r);
  END DiffR;

TYPE
  Func * = ComFunc.Float2;

PROCEDURE SaveFunc * (a-:Area; VAR f:Func;
                      FileName-:ARRAY OF CHAR; multiplier:Float);
  (** Saves f.f(z,r)*multiplier in points of the grid, including zeroth line,
      in file FileName in text 3-columns format.
      The file is overwritten if exists. *)
  BEGIN
    BasicArea.SaveFunc(a,f,FileName,multiplier);
  END SaveFunc;

PROCEDURE SaveSimpleFunc * (a-:Area; f:ComFunc.Simple2;
                            FileName-:ARRAY OF CHAR; multiplier:Float);
  (** Saves f(x,y)*multiplier in points of the grid
      in file FileName in text 3-columns format.
      The file is overwritten if exists. *)
  BEGIN
    BasicArea.SaveSimpleFunc(a,f,FileName,multiplier);
  END SaveSimpleFunc;

PROCEDURE SaveArray * (a:Area; field-:Array2;
                       FileName-:ARRAY OF CHAR; multiplier:Float);
  (** Saves field[iZ,iR]*multiplier in file FileName in text
      3-columns format. The file is overwritten if exists. *)
  BEGIN
    BasicArea.SaveArray(a,field,FileName,multiplier);
  END SaveArray;

PROCEDURE LoadArray * (a:Area; VAR field:Array2;
                       FileName-:ARRAY OF CHAR; multiplier:Float);
  (** Loads values, multiplied by multiplier, from file FileName to field.
      Format of the fiel is the same one used by SaveXXXX procedures. *)

  BEGIN
    BasicArea.LoadArray(a,field,FileName,multiplier);
  END LoadArray;

PROCEDURE Interpolation * (area-:Area; field:Array2Ptr): POINTER TO Func;
  (** Returns a function, which is interpolation of the specified field in the
      specified area. Dimensions of the field shold correspond to the grid
      of the area. *)
  BEGIN
    RETURN BasicArea.Interpolation(area,field);
  END Interpolation;

PROCEDURE Fill * (VAR f:ComFunc.Float2; area-:Area; VAR field:Array2);
  (** Fills the specified field in the specified area by result of the
      specified function f *)
  BEGIN
    BasicArea.Fill(f,area,field);
  END Fill;

PROCEDURE Reduced * (InitialArea-:Area; VAR ReducedArea:Area);
  (** Initiates the reduced area so, that it presents the initial area
      excluding outer points of the grid *)
  BEGIN
    BasicArea.Reduced(InitialArea,ReducedArea);
  END Reduced;

END mCylArea.