(** This module provides solution of 3-dimentional cylindrically
    symmetrical parabolic equation in partial derivatives in a cylinder.
    The equation is the following:

    dn/dt = -div j + E(z,r)*n + F(z,r),
    j = -D(z,r)*grad n + V(z,r)*n,

    where n=n(z,r) is the solution.

    There are 3 sides of the area: z=MinZ, z=MaxZ and r=Rad. The fourth
    side is r=0 - the axis.

    Boundary conditions at each side of the area are specified by
    three functions, which represent coefficients at n, derivative of
    n in normal direction, and the result. For example, at min(z) the
    following conditions are taken:

    MinZFunc(r)*n(MinZ,r) + MinZDer(r)dn/dz = MinZRes(r).

    Coefficients D, V, E and F should be specified by procedures SetCoeffs
    and SetF before the solution. By default they are zeros. Because V is
    a vector, is it presented by 2 functions Vz and Vr, corresponding to
    projections of the V to axial and radial directions.

    Boundary conditions are specified by procedures SetMinZ, SetMaxZ
    and SetBoundRad. This specification may be missed, then the zeroth boundary
    conditions (n=0) are taken default.

    The task is in specified rectangular area, which corresponds to
    3-dimentional cylindrical area, and on a grid on it.

    Last change: 04.01.2016 *)

<* O2EXTENSIONS + *>
MODULE mCylParabolicEqu;

IMPORT
  Math:=mMath, Area:=mCylArea, BasicEqu:=mParabolicEqu2D,
  Steps:=mEvolutionSteps, SU:=msysUtils, COMPILER;

CONST
  ModuleName = COMPILER.EQUATION('module');

TYPE
  Float * = Math.Float;
  Array2 = Math.Array2;
  Array2Ptr = Math.Array2Ptr;

TYPE
  DerFunc * = RECORD
  END;

PROCEDURE (VAR d:DerFunc) f (iZ,iR:LONGINT; field-:Array2): Float;
  BEGIN
    SU.RAISEError(ModuleName+'.DerFunc.f: abstract method');
    RETURN field[iZ,iR]
  END f;

TYPE
  Equ * = RECORD (BasicEqu.Equ)
    DVzDz * , DVrDr * :POINTER TO DerFunc; (** Used at the procedure SetCoeffs *)
    DDDz * , DDDr * : POINTER TO DerFunc; (** Used at the procedure SetCoeffs *)
  END;
  (** The equation *)

  AreaFunc * = BasicEqu.AreaFunc;
(**  AreaFunc* = RECORD
       PROCEDURE (VAR f: AreaFunc) f*(x,y: Float): Float;
     END; *)

  BoundaryFunc * = BasicEqu.BoundaryFunc;
(**  BoundaryFunc* = RECORD
       PROCEDURE (VAR f: BoundaryFunc) f*(x: Float): Float;
     END; *)

PROCEDURE Init * (equ:POINTER TO Equ; area-:Area.Area);
  (** Initiates the equation. area is an existing area. *)
  BEGIN
    ASSERT(equ#NIL);
    ASSERT(Area.Correct(area));

    BasicEqu.Init(equ,area);
    BasicEqu.SetMinYSpecial(equ^);

    equ.DVzDz:=NIL;
    equ.DVrDr:=NIL;
    equ.DDDz:=NIL;
    equ.DDDr:=NIL;
  END Init;

PROCEDURE Initialized * (equ-:Equ): BOOLEAN;
  (** Asserts that the equation is already initialized and returns TRUE. *)
  BEGIN
    RETURN BasicEqu.Initialized(equ)
  END Initialized;

PROCEDURE GetVal (a-:Array2; iX,iY:LONGINT): Float;
  BEGIN
    RETURN a[iX-1,iY-1]
  END GetVal;

PROCEDURE SetVal (VAR a:Array2; iX,iY:LONGINT; value:Float);
  BEGIN
    a[iX-1,iY-1]:=value;
  END SetVal;

PROCEDURE DivZ (field-:Array2; VAR func:AreaFunc; area-:Area.Area; iZ,iR:LONGINT): Float;
  (* used in SetCoeffs *)
  VAR
    z,r:Float;
    LastZ:Float;
  BEGIN
    LastZ:=area.LastX;

    ASSERT(iZ>0);
    ASSERT(iZ<LastZ);
    ASSERT(iR>0);
    ASSERT(iR<area.LastY);

    IF iZ=1 THEN
      z:=area.MinX;
      r:=Area.IndToR(area,iR);
      RETURN ( GetVal(field,iZ+1,iR) - func.f(z,r) ) *area.Inv2dx;
    ELSIF iZ=LastZ-1 THEN
      z:=area.MaxX;
      r:=Area.IndToR(area,iR);
      RETURN ( func.f(z,r) - GetVal(field,iZ-1,iR) ) *area.Inv2dx;
    ELSE
      RETURN ( GetVal(field,iZ+1,iR) - GetVal(field,iZ-1,iR) ) *area.Inv2dx;
    END;
  END DivZ;

PROCEDURE DivR (field-:Array2; VAR func:AreaFunc; area-:Area.Area; iZ,iR:LONGINT): Float;
  (* used in SetCoeffs *)
  VAR
    z,r:Float;
    LastR:LONGINT;
  BEGIN
    LastR:=area.LastY;

    ASSERT(iZ>0);
    ASSERT(iZ<area.LastX);
    ASSERT(iR>0);
    ASSERT(iR<LastR);

    IF iR=1 THEN (* near axis *)
      RETURN ( GetVal(field,iZ,iR+1) - GetVal(field,iZ,iR) ) *area.Inv2dy;
    ELSIF iR=LastR-1 THEN
      z:=Area.IndToZ(area,iZ);
      r:=area.MaxY;
      RETURN ( func.f(z,r) - GetVal(field,iZ,iR-1) ) *area.Inv2dy;
    ELSE
      RETURN ( GetVal(field,iZ,iR+1) - GetVal(field,iZ,iR-1) ) *area.Inv2dy;
    END;
  END DivR;

PROCEDURE SetCoeffs * (VAR equ:Equ; VAR D,Vz,Vr,E:AreaFunc);
  (** Sets D, Vz, Vr and E.

      Procedures equ.DVzDz and DVrDr are used to get partial derivatives of
      Vz by z and Vr by r.
      Procedures equ.DDDz and DDDr are used to get partial derivatives of
      D by z and by r.
      By default (if the mentioned procedures are not set or LINed)
      a 2nd-order numerical approximation is used, that is not ever good.
      When calling these procedures, the parameter field contains filled
      array of the valued to be differenciated: at calling DDDz, the field
      contains D, at callinf DVrDr the field contains Vr, and so on.
      Dimension and contents of the field corresponds to the internal areas
      (see Area.Reduced): LEN(field)=LastZ-1, LEN(field,1)=LastR-1 *)
  VAR
    LastZ,LastR:LONGINT;

  PROCEDURE DVzDz (iZ,iR:LONGINT): Float;
    BEGIN
      IF equ.DVzDz=NIL THEN
        RETURN DivZ(equ.c^,Vz,equ.area,iZ,iR);
      ELSE
        RETURN equ.DVzDz.f(iZ,iR,equ.c^)
      END;
    END DVzDz;

  PROCEDURE DVrDr (iZ,iR:LONGINT): Float;
    BEGIN
      IF equ.DVrDr=NIL THEN
        RETURN DivR(equ.d^,Vr,equ.area,iZ,iR);
      ELSE
        RETURN equ.DVrDr.f(iZ,iR,equ.d^)
      END;
    END DVrDr;

  PROCEDURE DDDz (iZ,iR:LONGINT): Float;
    BEGIN
      IF equ.DDDz=NIL THEN
        RETURN DivZ(equ.a^,D,equ.area,iZ,iR);
      ELSE
        RETURN equ.DDDz.f(iZ,iR,equ.a^)
      END;
    END DDDz;

  PROCEDURE DDDr (iZ,iR:LONGINT): Float;
    BEGIN
      IF equ.DDDr=NIL THEN
        RETURN DivR(equ.a^,D,equ.area,iZ,iR);
      ELSE
        RETURN equ.DDDr.f(iZ,iR,equ.a^)
      END;
    END DDDr;

  VAR
    iZ,iR:LONGINT;
    InvR:Float;
  BEGIN (* SetCoeffs *)
    ASSERT(Initialized(equ));

    BasicEqu.SetA(equ,D);
    BasicEqu.SetB(equ,D);
    BasicEqu.SetC(equ,Vz);
    BasicEqu.SetD(equ,Vr);
    BasicEqu.SetE(equ,E);

    LastZ:=equ.area.LastX;
    LastR:=equ.area.LastY;
    FOR iR:=1 TO LastR-1 DO
      InvR:=1/Area.IndToR(equ.area,iR);
      FOR iZ:=1 TO LastZ-1 DO
        SetVal(equ.e^,iZ,iR, GetVal(equ.e^,iZ,iR)
                            -DVzDz(iZ,iR)
                            -DVrDr(iZ,iR)
                            -GetVal(equ.d^,iZ,iR)*InvR
              );
      END;
    END;

    FOR iR:=1 TO LastR-1 DO
      InvR:=1/Area.IndToR(equ.area,iR);
      FOR iZ:=1 TO LastZ-1 DO
        SetVal(equ.c^,iZ,iR, DDDz(iZ,iR)
                            -GetVal(equ.c^,iZ,iR)
              );
        SetVal(equ.d^,iZ,iR, DDDr(iZ,iR)
                            +GetVal(equ.a^,iZ,iR)*InvR
                            -GetVal(equ.d^,iZ,iR)
              );
      END;
    END;

  END SetCoeffs;

PROCEDURE SetF * (VAR equ:Equ; VAR F:AreaFunc);
  (** Sets F *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetF(equ,F);
  END SetF;

PROCEDURE SetMinZ * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for z=MinZ:
      func(r)*n(MinZ,r) + der(r)*d/dz(n)(MinZ,r) = res(r) *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMinX(equ,func,der,res);
  END SetMinZ;

PROCEDURE SetMaxZ * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for z=MaxZ:
      func(r)*n(MaxZ,r) + der(r)*d/dz(n)(MaxZ,r) = res(r) *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMaxX(equ,func,der,res);
  END SetMaxZ;

PROCEDURE SetRad * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for r=Rad:
      func(z)*n(z,Rad) + der(z)*d/dr(n)(z,Rad) = res(z) *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMaxY(equ,func,der,res);
  END SetRad;

PROCEDURE SetMinZSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at z=MinZ: n[0,iR]=n[1,iR] for all iR *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMinXSpecial(equ)
  END SetMinZSpecial;

PROCEDURE SetMaxZSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at z=MaxZ: n[LastZ,iR]=n[LastZ-1,iR] for all iR *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMaxXSpecial(equ)
  END SetMaxZSpecial;

PROCEDURE SetRadSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at r=Rad: n[iZ,LastR]=n[iZ,LastR-11] for all iZ *)
  BEGIN
    ASSERT(Initialized(equ));

    BasicEqu.SetMaxYSpecial(equ)
  END SetRadSpecial;

PROCEDURE RHS * (equ-:Equ; field-:Array2; iZ,iR:LONGINT): Float;
  (** Returns right hand side of the equation (see the header of the module)
      with the specified field. At the boundary the procedure returns residual
      of the boundary conditions. *)
  BEGIN
    RETURN BasicEqu.RHS(equ,field,iZ,iR);
  END RHS;

PROCEDURE NormOfRHS * (equ-:Equ; field-:Array2): Float;
  (** Return mean norm of RHS wuth the specified field *)
  BEGIN
    RETURN BasicEqu.NormOfRHS(equ,field)
  END NormOfRHS;

PROCEDURE Step * (equ:POINTER TO Equ; OldField,NewField:Array2Ptr; dt:Float);
  (** Performs one step of solution with specified time step dt
      by "parallel-normal" method. The method has accuracy O(dz^2+dr^2+dt^2).
      OldField should contain initial conditions, the result is returned
      in NewField. OldField=NewField is accepted. *)
  BEGIN
    BasicEqu.Step(equ,OldField,NewField,dt);
  END Step;

PROCEDURE Eigenvalues * (equ-:Equ; VAR ev:Steps.Eigenvalues);
  (** Returns an estimation of the boundaries of the spectrum of the RHS
      of the equation *)
  BEGIN
    BasicEqu.Eigenvalues(equ,ev);
  END Eigenvalues;

PROCEDURE EigenvaluesExacter * (equ-:Equ; VAR ev:Steps.Eigenvalues);
  (** The same as Eigenvalues, but exacter, but significantly slower *)
  BEGIN
    BasicEqu.EigenvaluesExacter(equ,ev);
  END EigenvaluesExacter;

PROCEDURE SaveCoeffs * (equ-:Equ; prefix-,suffix-:ARRAY OF CHAR);
  (** Saves all the coefficients of the equ to files. The file names are
      prefix+name+suffix, where name is 'A','B','C','D','E','F', respectively.
      The equation should be already initialized. *)
  BEGIN
    BasicEqu.SaveCoeffs(equ,prefix,suffix)
  END SaveCoeffs;

PROCEDURE StabilizeByDiffusion * (VAR equ:Equ; parameter:Float);
  (** Restricts equ.a and equ.b from low for better stability of the equation.
      Introduces an error of O(dx,dy).
      If parameter=0, no changes happen.
      If parameter=1, full stabilization changes happen.
      Based on inconsistent stabilization of COMSOL. *)
  BEGIN
    BasicEqu.StabilizeByDiffusion(equ,parameter)
  END StabilizeByDiffusion;

PROCEDURE StabilizeByMigration * (VAR equ:Equ; parameter:Float);
  (** Restricts equ.c and equ.d from up for better stability of the equation.
      Introduces an error of O(dx,dy).
      If parameter=0, no changes happen.
      If parameter=1, full stabilization changes happen.
      Based on inconsistent stabilization of COMSOL. *)
  BEGIN
    BasicEqu.StabilizeByMigration(equ,parameter)
  END StabilizeByMigration;

END mCylParabolicEqu.