(** Organization of evolution of equations with different time steps to reach
    the steady-state result. Often l_min and l_max are needed, which are minimal
    and maximal eigenvalues of the equation operator.

    Last changed: 20.11.2015 *)

<* O2EXTENSIONS + *>
MODULE mEvolutionSteps;

IMPORT
  Math:=mMath, COMPILER, SU:=msysUtils, mMisc;

CONST
  ModuleName = COMPILER.EQUATION('Module');
  pi = Math.pi;

TYPE
  Float * = Math.Float;
  ArrayPtr = Math.ArrayPtr;

  Step * = RECORD (** This is an interface to the equations *)
  END;

<* PUSH *>
<* WOFF301+ *>
PROCEDURE (VAR s:Step) Step * (dt:Float);
  (** Should perform one step with specified time *)
  BEGIN
    SU.RAISEError(ModuleName+'.Step.Step: abstract method');
  END Step;
<* POP *>

PROCEDURE (VAR s:Step) Norm * (): Float;
  (** Should return norm of the last result *)
  BEGIN
    SU.RAISEError(ModuleName+'.Step.Norm: abstract method');
    RETURN 0;
  END Norm;

PROCEDURE (VAR s:Step) NormOfDifference * (): Float;
  (** Should return norm of difference of two last iterations *)
  BEGIN
    SU.RAISEError(ModuleName+'.Step.NormOfDifference: abstract method');
    RETURN 0;
  END NormOfDifference;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

PROCEDURE ln (x:Float): Float;
  BEGIN
    RETURN Math.ln(x)
  END ln;

PROCEDURE RelErr * (PrevDiff,diff,norm:Float): Float;
  (** Returns a relative error of a field.
      It is based on 2 steps (i.e. 3 time layers) with the same time step.
      Let index be the three succeedeing time layers 0,1,2.
      PrevDiff is norm of difference of layers 0 and 1.
      diff is norm of difference of layers 1 and 2.
      norm is a norm of the layer 2. *)
  VAR
    nu:Float;
  BEGIN
    IF diff=0 THEN (* it means that the iterations have already converged *)
      RETURN 0;
    END;

    ASSERT(norm#0);

    nu:=ABS(PrevDiff/diff);
    IF nu=1 THEN (* this is not an exponential convergence yet *)
      RETURN 1
    ELSE
      RETURN ABS(diff/((1-nu)*norm));
    END;
  END RelErr;

PROCEDURE GetRelErr * (VAR s:Step; dt:Float): Float;
  (** Performs two steps with the specified time and returns the resulting
      relative error, as desctibed in the procedure RelErr *)
  VAR
    PrevDiff,diff,norm:Float;
  BEGIN
    s.Step(dt);
    PrevDiff:=s.NormOfDifference();
    s.Step(dt);
    diff:=s.NormOfDifference();
    norm:=s.Norm();
    RETURN RelErr(PrevDiff,diff,norm);
  END GetRelErr;

(** == Data for evolution from the 'parallel-perpendicular' method of solution
    == of 2D parabolic equations *)

TYPE
  Eigenvalues * = RECORD
    (** minimal and maximal modules of eigenvalues at x and y directions *)
    MinX * , MaxX * , MinY * , MaxY * : Float;
  END;

PROCEDURE EVCorrect (ev-:Eigenvalues): BOOLEAN;
  BEGIN
    ASSERT(ev.MinX>0);
    ASSERT(ev.MaxX>ev.MinX);
    ASSERT(ev.MinY>0);
    ASSERT(ev.MaxY>ev.MinY);

    RETURN TRUE
  END EVCorrect;

PROCEDURE OptimalDt * (ev-:Eigenvalues): Float;
  (** Returns an optimal time step for evolution with constnt steps *)
  BEGIN
    ASSERT(EVCorrect(ev));

    RETURN 2*sqrt( (ev.MaxX-ev.MinX+ev.MaxY-ev.MinY) / ( ev.MinX*ev.MaxX*(ev.MaxY-ev.MinY) + ev.MinX*ev.MinY*(ev.MaxX-ev.MinX) ) )
  END OptimalDt;

PROCEDURE ChebyshevEigenvalues * (ev-:Eigenvalues; VAR l_min,l_max:Float);
  (** Returns minimal and maximal eigenvalues, which can be used for evolution
      with Chebyshev steps *)
  VAR
    dt:Float;
  PROCEDURE InBrackets (x:Float): Float;
    BEGIN
      RETURN 1+dt*x/2
    END InBrackets;
  VAR
    rho:Float;
  BEGIN
    ASSERT(EVCorrect(ev));

    dt:=OptimalDt(ev);
    rho:=InBrackets(-ev.MaxX)*InBrackets(-ev.MaxY)/(InBrackets(ev.MaxX)*InBrackets(ev.MaxY));
    l_min:=(1-rho)/dt;
    l_max:=(1+rho)/dt;
  END ChebyshevEigenvalues;

PROCEDURE LogEigenvalues * (ev-:Eigenvalues; VAR l_min,l_max:Float);
  (** Returns minimal and maximal eigenvalues, which can be used for evolution
      with logarithmical steps *)
  BEGIN
    ASSERT(EVCorrect(ev));

    l_min:=mMisc.Min(ev.MinX,ev.MinY);
    l_max:=mMisc.Max(ev.MaxX,ev.MaxY);
  END LogEigenvalues;

(** ========================= Chebyshev steps =============================== *)

PROCEDURE CalculateChebyshevOrder (VAR ind:ARRAY OF LONGINT);
  (* This procedure returns a right order of Chebyshev steps, as explained in
     the book .. , .. ,    ,
      "" 1978, page 280.
     Number of Chebyshev steps is LEN(ind), the result is returned in the ind.
     The result is 2*k-1, where k is index of the step, 1<=k<=LEN(ind) *)
  VAR
    TransInd:LONGINT;
  PROCEDURE OddTransition (): BOOLEAN;
    BEGIN
      RETURN ind[TransInd]=0
    END OddTransition;
  PROCEDURE LastOrDoubleTransition (): BOOLEAN;
    BEGIN
      RETURN (TransInd=LEN(ind)-1) OR (ind[TransInd+1]=1)
    END LastOrDoubleTransition;
  VAR
    i,m:LONGINT;
  BEGIN
    i:=LEN(ind);
    TransInd:=LEN(ind);
    WHILE i>1 DO
      DEC(TransInd);
      IF ODD(i) THEN
        ind[TransInd]:=0;
        DEC(i)
      ELSE
        ind[TransInd]:=1;
        i:=i DIV 2;
      END;
    END;

    ind[0]:=1;
    m:=1;
    WHILE TransInd<LEN(ind) DO
      ASSERT(m<=TransInd);
      IF OddTransition() THEN
        ASSERT(~ODD(m));
        ind[m]:=m+1;
        INC(m);
      ELSE
        FOR i:=m TO 1 BY -1 DO
          ind[2*i-2]:=ind[i-1];
        END;
        IF LastOrDoubleTransition() THEN
          FOR i:=m TO 1 BY -1 DO
            ind[2*i-1]:=4*m-ind[2*i-2]
          END;
        ELSE
          FOR i:=m TO 1 BY -1 DO
            ind[2*i-1]:=4*m+2-ind[2*i-2]
          END;
        END;
        m:=m+m;
      END;
      INC(TransInd)
    END;
  END CalculateChebyshevOrder;

PROCEDURE NumberOfChebyshevSteps * (l_min,l_max,eps:Float): LONGINT;
  (** l_min and l_max are bounds of the spectrum of the equation.
      The procedure returns an estimation of number of Chebyshev steps which are
      needed to reduce the error in 1/eps times *)
  (* By book of Kalitkin 1978 *)
  VAR
    rho:Float;
  BEGIN
    ASSERT(eps>0);
    ASSERT(l_min>0);
    ASSERT(l_max>=l_min);

    IF (eps>=1) OR (l_min=l_max) THEN
      RETURN 0
    ELSE
      rho:=(l_max-l_min)/(l_max+l_min);
      RETURN Math.GreaterInt(ln(eps/2)/ln(rho));
    END;
  END NumberOfChebyshevSteps;

PROCEDURE ChebyshevDt * (l_min,l_max:Float; NumberOfSteps:LONGINT): ArrayPtr;
  (** l_min and l_max are bounds of the spectrum of the equation.
      Returns array of correctly ordered Chebyshev time steps, which are based
      on the l_min, l_max and number of the Chebyshev steps *)
  (* By book of Kalitkin 1978 *)
  VAR
    ind:POINTER TO ARRAY OF LONGINT;
    res:ArrayPtr;
    s:LONGINT;
  BEGIN
    ASSERT(l_min>0);
    ASSERT(l_max>l_min);
    ASSERT(NumberOfSteps>0);

    NEW(ind,NumberOfSteps);
    CalculateChebyshevOrder(ind^);
    NEW(res,NumberOfSteps);
    FOR s:=0 TO NumberOfSteps-1 DO
      res[s]:=2/(l_max+l_min + (l_max-l_min)*Math.cos(pi*ind[s]/(2*NumberOfSteps)));
    END;
    RETURN res;
  END ChebyshevDt;

PROCEDURE Chebyshev * (VAR s:Step; l_min,l_max,eps:Float);
  (** l_min and l_max are bounds of the spectrum of the equation.
      The procedure performs iterations with Chebyshev time steps, which are
      specified by l_min,l_max,eps.

      Example:

      VAR
        ev:Eigenvalues;

      dt:=OptimalDt(ev);
      ChebyshevEigenvalues(ev,l_min,l_max);
      CurrRelErr:=GetRelErr(s,dt);
      WHILE CurrRelErr>DesiredRelErr DO
        Chebyshev(s,l_min,l_max,DesiredRelErr/CurrRelErr);
        CurrRelErr:=GetRelErr(s,dt);
      END;
 *)
  (* By book of Kalitkin 1978 *)
  VAR
    k,K:LONGINT;
    dt:ArrayPtr;
  BEGIN
    K:=NumberOfChebyshevSteps(l_min,l_max,eps);
    IF K>0 THEN
      dt:=ChebyshevDt(l_min,l_max,K);
      FOR k:=0 TO K-1 DO
        s.Step(dt[k]);
      END;
    END;
  END Chebyshev;

(** ========================= Logarithmical steps =========================== *)

CONST
  ExtensionFactor = 3;

PROCEDURE NumberOfLogSteps * (l_min,l_max,eps:Float): LONGINT;
  (** l_min and l_max are bounds of the spectrum of the equation.
      The procedure returns an estimation of number of logarithmical steps which
      are needed to reduce the error in 1/eps times *)
  (* By book of Kalitkin 2013 *)
  BEGIN
    ASSERT(eps>0);
    ASSERT(l_min>0);
    ASSERT(l_max>l_min);

    l_min:=l_min/ExtensionFactor;
    l_max:=l_max*ExtensionFactor;

    IF eps>=1 THEN
      RETURN 0
    ELSE
      RETURN Math.GreaterInt( -Math.ln(eps) * Math.ln(l_max/l_min) /4)+1;
    END;
  END NumberOfLogSteps;

PROCEDURE LogDt * (l_min,l_max:Float; NumberOfSteps:LONGINT): ArrayPtr;
  (** l_min and l_max are bounds of the spectrum of the equation.
      Returns array of logarithmical time steps, which are based
      on the l_min, l_max and number of the logarithmic steps *)
  (* By book of Kalitkin 2013 *)
  VAR
    factor:Float;
    i:LONGINT;
    res:ArrayPtr;
  BEGIN
    ASSERT(l_min>0);
    ASSERT(l_max>l_min);
    ASSERT(NumberOfSteps>1);

    l_min:=l_min/ExtensionFactor;
    l_max:=l_max*ExtensionFactor;

    NEW(res,NumberOfSteps);
    factor:=Math.power((l_max/l_min),1/(NumberOfSteps-1));
    res[0]:=2/l_max;
    FOR i:=1 TO NumberOfSteps-1 DO
      res[i]:=res[i-1]*factor
    END;
    RETURN res;
  END LogDt;

PROCEDURE Log * (VAR s:Step; l_min,l_max,eps:Float);
  (** l_min and l_max are bounds of the spectrum of the equation.
      The procedure performs iterations with logarithmical time steps, which are
      specified by l_min,l_max,eps.

      Example:
      VAR
        ev:Eigenvalues;

      dt:=OptimalDt(ev);
      LogEigenvalues(ev,l_min,l_max);
      CurrRelErr:=GetRelErr(s,dt);
      WHILE CurrRelErr>DesiredRelErr DO
        Log(s,l_min,l_max,DesiredRelErr/CurrRelErr);
        CurrRelErr:=GetRelErr(s,dt);
      END;
 *)
  (* By book of Kalitkin 2013 *)
  VAR
    k,K:LONGINT;
    dt:ArrayPtr;
  BEGIN
    K:=NumberOfLogSteps(l_min,l_max,eps);
    IF K>0 THEN
      dt:=LogDt(l_min,l_max,K);
      FOR k:=0 TO K-1 DO
        s.Step(dt[k]);
      END;
    END;
  END Log;

END mEvolutionSteps.