(** This module provides tools for fast Fourier transformation and complementary
    conversions.

    Last changed: 24.07.2014 *)
<* O2EXTENSIONS + *>
MODULE mFFT;

IMPORT
  Math:=mMath, LowMath:=mLowMath, Arr:=mArrays;

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

CONST
  pi * = Math.pi;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE Core (InRe-,InIm-:Array; VAR OutRe,OutIm:Array; reverse:BOOLEAN);
  VAR
    FactInd:LONGINT;
  PROCEDURE Factorization (n:LONGINT; VAR K,L:LONGINT);
    BEGIN
      WHILE (FactInd*FactInd<=n) & (n MOD FactInd #0) DO
        INC(FactInd);
      END;
      IF FactInd*FactInd>n THEN
        K:=n;
        L:=1;
      ELSE
        ASSERT(n MOD FactInd =0);
        K:=n DIV FactInd;
        L:=FactInd;
      END;
    END Factorization;
  VAR
    ReB,ImB:ArrayPtr;
    ArraySize:LONGINT;
    SinArr,CosArr:ArrayPtr;
  PROCEDURE InitSinCos;
    VAR
      arg:Float;
      i:LONGINT;
    BEGIN
      NEW(SinArr,ArraySize);
      NEW(CosArr,ArraySize);
      IF reverse THEN
        arg:=2*pi/ArraySize;
      ELSE
        arg:=-2*pi/ArraySize;
      END;
      FOR i:=0 TO ArraySize-1 DO
        LowMath.SinCos(i*arg,SinArr[i],CosArr[i]);
      END;
    END InitSinCos;
  PROCEDURE MulExp (FactorRe,FactorIm:Float; ind:LONGINT; VAR ResRe,ResIm:Float);
    VAR
      sin,cos:Float;
    BEGIN
      ASSERT(ind>=0);
      ASSERT(ind<ArraySize);

      sin:=SinArr[ind];
      cos:=CosArr[ind];

      ResRe:=FactorRe*cos - FactorIm*sin;
      ResIm:=FactorRe*sin + FactorIm*cos;
    END MulExp;
  PROCEDURE IncMod (VAR i:LONGINT; j:LONGINT);
    BEGIN
      INC(i,j);
      IF i>=ArraySize THEN
        DEC(i,ArraySize);
      END;
    END IncMod;
  PROCEDURE DirectB (M,N:LONGINT);
    VAR
      p,m1,ind,pM:LONGINT;
    PROCEDURE Sums (VAR re,im:Float);
      VAR
        m2,pow:LONGINT;
        re1,im1:Float;
      BEGIN
        re:=0;
        im:=0;
        m2:=m1;
        pow:=m1*p;
        WHILE m2-m1<ArraySize DO
          MulExp(InRe[m2],InIm[m2],pow,re1,im1);
          re:=re+re1;
          im:=im+im1;
          INC(m2,M);
          IncMod(pow,pM);
        END;
      END Sums;
    BEGIN (* DirectB *)
      ASSERT(M*N=ArraySize);
      pM:=0;
      FOR p:=0 TO N-1 DO
        ind:=p;
        m1:=0;
        WHILE ind-p<ArraySize DO
          Sums(OutRe[ind],OutIm[ind]);
          INC(ind,N);
          INC(m1);
        END;
        INC(pM,M);
      END;
    END DirectB;
  PROCEDURE b (M,N:LONGINT);
    VAR
      K,L,p1,p2,m1,ind,KM,p2KM:LONGINT;
    PROCEDURE Sums (VAR re,im:Float);
      VAR
        l1,pow,ind:LONGINT;
        re1,im1:Float;
      BEGIN
        re:=0;
        im:=0;
        l1:=m1;
        pow:=m1*p2;
        WHILE l1-m1<ArraySize DO
          ind:=p1+l1;
          MulExp(OutRe[ind],OutIm[ind],pow,re1,im1);
          re:=re+re1;
          im:=im+im1;
          INC(l1,KM);
          IncMod(pow,p2KM);
        END;
      END Sums;
    BEGIN (* b *)
      ASSERT(M*N=ArraySize);
      Factorization(N,K,L);
      IF L=1 THEN
        DirectB(M,N);
      ELSE
        KM:=K*M;
        b(M*L,K);
        FOR p1:=0 TO K-1 DO
          p2KM:=0;
          FOR p2:=0 TO L-1 DO
            m1:=0;
            WHILE m1<KM DO
              ind:=p1+K*p2 + m1*L;
              Sums(ReB[ind],ImB[ind]);
              INC(m1,K);
            END;
            INC(p2KM,KM);
          END;
        END;
        Arr.Assign(OutRe,ReB^);
        Arr.Assign(OutIm,ImB^);
      END;
    END b;
  BEGIN (* Core *)
    ASSERT(LEN(InIm)=LEN(InRe));
    ASSERT(LEN(OutRe)=LEN(InRe));
    ASSERT(LEN(OutIm)=LEN(InRe));

    FactInd:=2;
    ArraySize:=LEN(InRe);
    NEW(ReB,ArraySize);
    NEW(ImB,ArraySize);
    InitSinCos();
    b(1,ArraySize);
  END Core;

PROCEDURE Direct * (SignalRe-,SignalIm-:Array; VAR SpectrumRe,SpectrumIm:Array);
  (** Performs Fourier transformation, fast, if possible. Signal is given in
      SignalRe, SignalIm, spectrum is returned in SpectrumRe, SpectrumIm.

      The Fourier transformation is defined as follows:

      Spectrum[k]=1/ArraySize*Sum(Signal[n]*exp(-2*pi*i*n*k/ArraySize)),
      ArraySize = LEN(Signal/Spectrum), n=0..ArraySize-1, i*i=-1.
      Thus Spectrum[k] is a periodical function with period ArraySize.
      Value of the spectrum is amplitude of the exponent in the signal.

      Elements k in [0..ArraySize DIV 2] in the resulting spectrum correspond
      to frequencies k.
      Elements k in [ArraySize DIV 2 +1 .. ArraySize-1] in the resulting
      spectrum correspond to negative frequencies k-ArraySize.

      Result can not be in the same arrays as input. *)

  VAR
    ArraySize:LONGINT;
    factor:Float;
    i:LONGINT;
  BEGIN
    ArraySize:=LEN(SignalRe);
    ASSERT(LEN(SignalIm)=ArraySize);
    ASSERT(LEN(SpectrumRe)=ArraySize);
    ASSERT(LEN(SpectrumIm)=ArraySize);

    Core(SignalRe,SignalIm,SpectrumRe,SpectrumIm,FALSE);
    factor:=1/ArraySize;
    FOR i:=0 TO ArraySize-1 DO
      SpectrumRe[i]:=factor*SpectrumRe[i];
      SpectrumIm[i]:=factor*SpectrumIm[i];
    END;
  END Direct;

PROCEDURE Reverse * (SpectrumRe-,SpectrumIm-:Array; VAR SignalRe,SignalIm:Array);
  (** Performs reverse Fourier transformation, which is:
      Signal[n]=Sum(Spectrum[k]*exp(2*pi*i*n*k/ArraySize)).
      Result can not be in the same arrays as input. *)
  BEGIN
    ASSERT(LEN(SignalIm)=LEN(SignalRe));
    ASSERT(LEN(SpectrumRe)=LEN(SignalRe));
    ASSERT(LEN(SpectrumIm)=LEN(SignalRe));

    Core(SpectrumRe,SpectrumIm,SignalRe,SignalIm,TRUE);
  END Reverse;

PROCEDURE Frequency * (ArraySize,index:LONGINT): LONGINT;
  (** Returns frequency which correcponds to the specified index in a spectrum
      with ArraySize elements *)
  BEGIN
    ASSERT(ArraySize>0);

    IF index<=ArraySize DIV 2 THEN
      RETURN index
    ELSE
      RETURN index-ArraySize
    END;
  END Frequency;

PROCEDURE Index * (ArraySize,frequency:LONGINT): LONGINT;
  (** Returns index of the specified frequency in a spectrum
      with ArraySize elements *)
  BEGIN
    ASSERT(ArraySize>0);

    IF frequency>=0 THEN
      RETURN frequency
    ELSE
      RETURN frequency+ArraySize
    END;
  END Index;

PROCEDURE BandPass2 * (VAR Re,Im:Array; low,high:Float);
  (** Performs bandpass fourier filtering: only frequencies
      low<=freq<=high are left. *)
  VAR
    SpeRe,SpeIm:ArrayPtr;
    ArraySize,i:LONGINT;
    freq:Float;
  BEGIN
    ASSERT(LEN(Re)=LEN(Im));

    ArraySize:=LEN(Re);
    NEW(SpeRe,ArraySize);
    NEW(SpeIm,ArraySize);

    Direct(Re,Im,SpeRe^,SpeIm^);

    FOR i:=0 TO ArraySize-1 DO
      freq:=ABS(Frequency(ArraySize,i));
      IF (freq<low) OR (freq>high) THEN
        SpeRe[i]:=0;
        SpeIm[i]:=0;
      END;
    END;

    Reverse(SpeRe^,SpeIm^,Re,Im);
  END BandPass2;

PROCEDURE BandPass * (VAR x:Array; low,high:Float);
  (** Uses BandPass2 for x *)
  VAR
    im:ArrayPtr;
    i:LONGINT;
  BEGIN
    NEW(im,LEN(x));
    FOR i:=0 TO LEN(x)-1 DO
      im[i]:=0;
    END;
    BandPass2(x,im^,low,high);
  END BandPass;

PROCEDURE LowPass2 * (VAR Re,Im:Array; threshold:Float);
  (** Only frequencies freq<=threshold are left. *)
  BEGIN
    ASSERT(LEN(Re)=LEN(Im));

    BandPass2(Re,Im,-1,threshold);
  END LowPass2;

PROCEDURE LowPass * (VAR x:Array; threshold:Float);
  (** Only frequencies freq<=threshold are left. *)
  BEGIN
    BandPass(x,-1,threshold);
  END LowPass;

PROCEDURE HighPass2 * (VAR Re,Im:Array; threshold:Float);
  (** Only frequencies freq>=threshold are left. *)
  BEGIN
    ASSERT(LEN(Re)=LEN(Im));

    BandPass2(Re,Im,threshold,LEN(Re));
  END HighPass2;

PROCEDURE HighPass * (VAR x:Array; threshold:Float);
  (** Only frequencies freq>=threshold are left. *)
  BEGIN
    BandPass(x,threshold,LEN(x));
  END HighPass;

PROCEDURE Smooth2 * (VAR Re,Im:Array; points:Float);
  (** Smooths the Re and Im using FFT transformations.
      points is (approximately) number of neighbour points to smooth.
      Uses the same algorithm as FFT-smoothing in Origin. *)
  VAR
    SpeRe,SpeIm:ArrayPtr;
    ArraySize,i:LONGINT;
    lim,factor,freq:Float;
    MinRe,SlopeRe,MinIm,SlopeIm:Float;
  BEGIN
    ASSERT(LEN(Re)=LEN(Im));

    ArraySize:=LEN(Re);
    NEW(SpeRe,ArraySize);
    NEW(SpeIm,ArraySize);

    MinRe:=Re[0];
    SlopeRe:=(Re[ArraySize-1]-Re[0])/(ArraySize-1);
    MinIm:=Im[0];
    SlopeIm:=(Im[ArraySize-1]-Im[0])/(ArraySize-1);;

    FOR i:=0 TO ArraySize-1 DO
      Re[i]:=Re[i]-MinRe-SlopeRe*i;
      Im[i]:=Im[i]-MinIm-SlopeIm*i;
    END;

    Direct(Re,Im,SpeRe^,SpeIm^);

    lim:=ArraySize/(2*points);
    FOR i:=0 TO ArraySize-1 DO
      freq:=ABS(Frequency(ArraySize,i));
      IF freq<=lim THEN
        factor:=1-sqr(freq/lim);
        SpeRe[i]:=SpeRe[i]*factor;
        SpeIm[i]:=SpeIm[i]*factor;
      ELSE
        SpeRe[i]:=0;
        SpeIm[i]:=0;
      END;
    END;

    Reverse(SpeRe^,SpeIm^,Re,Im);

    FOR i:=0 TO ArraySize-1 DO
      Re[i]:=Re[i]+MinRe+SlopeRe*i;
      Im[i]:=Im[i]+MinIm+SlopeIm*i;
    END;
  END Smooth2;

PROCEDURE Smooth * (VAR x:Array; points:Float);
  (** Uses Smooth2 to smooth the given array. *)
  VAR
    im:ArrayPtr;
    i:LONGINT;
  BEGIN
    NEW(im,LEN(x));
    FOR i:=0 TO LEN(x)-1 DO
      im[i]:=0;
    END;
    Smooth2(x,im^,points);
  END Smooth;

PROCEDURE Mul (re1,im1,re2,im2:Float; VAR ResRe,ResIm:Float);
  BEGIN
    ResRe:= re1*re2 - im1*im2;
    ResIm:= re1*im2 + re2*im1;
  END Mul;

PROCEDURE ConjMul (re1,im1,re2,im2:Float; VAR ResRe,ResIm:Float);
  BEGIN
    ResRe:= re1*re2 + im1*im2;
    ResIm:= re1*im2 - re2*im1;
  END ConjMul;

TYPE
  MulProc = PROCEDURE (re1,im1,re2,im2:Float; VAR ResRe,ResIm:Float);

PROCEDURE Conv (FRe-,FIm-,GRe-,GIm-:Array; VAR ResRe,ResIm:Array; mul:MulProc);
  VAR
    n,N:LONGINT;
    SpeFRe,SpeFIm:ArrayPtr;
    re,im:Float;
  BEGIN
    ASSERT(mul#NIL);

    N:=LEN(FRe);
    NEW(SpeFRe,N);
    NEW(SpeFIm,N);
    Direct(FRe,FIm,SpeFRe^,SpeFIm^);
    Direct(GRe,GIm,ResRe,ResIm);
    FOR n:=0 TO N-1 DO
      mul(SpeFRe[n],SpeFIm[n],ResRe[n],ResIm[n],re,im);
      SpeFRe[n]:=re*N;
      SpeFIm[n]:=im*N;
    END;
    Reverse(SpeFRe^,SpeFIm^,ResRe,ResIm);
  END Conv;

PROCEDURE Convolution * (FRe-,FIm-,GRe-,GIm-:Array; VAR ResRe,ResIm:Array);
  (** Perofms a convolution (a fast algorithm based on FFT):
      Res[n] = Sum(F[m]*g[n-m]).
      Here f is a function specified by FRe, FIm,
      g is a periodic extension of function specified by GRe, GIm.
      Length of all the arrays should be equal.
      Result can not be in the same arrays as input. *)
  BEGIN
    ASSERT(LEN(FRe)=LEN(FIm));
    ASSERT(LEN(FRe)=LEN(GRe));
    ASSERT(LEN(FRe)=LEN(GIm));
    ASSERT(LEN(FRe)=LEN(ResRe));
    ASSERT(LEN(FRe)=LEN(ResIm));

    Conv(FRe,FIm,GRe,GIm,ResRe,ResIm,Mul);
  END Convolution;

PROCEDURE CrossCorrelation * (FRe-,FIm-,GRe-,GIm-:Array; VAR ResRe,ResIm:Array);
  (** Perofms a cross-correlation (a fast algorithm based on FFT):
      Res[n] = Sum(Conj(F[m])*g[m+n]).
      Here f is a function specified by FRe, FIm,
      g is a periodic extension of function specified by GRe, GIm.
      Length of all the arrays should be equal.
      Result can not be in the same arrays as input. *)
  BEGIN
    ASSERT(LEN(FRe)=LEN(FIm));
    ASSERT(LEN(FRe)=LEN(GRe));
    ASSERT(LEN(FRe)=LEN(GIm));
    ASSERT(LEN(FRe)=LEN(ResRe));
    ASSERT(LEN(FRe)=LEN(ResIm));

    Conv(FRe,FIm,GRe,GIm,ResRe,ResIm,ConjMul);
  END CrossCorrelation;

END mFFT.