(** This module provides Lagrange multipliers method for minimization of
    quadratic forms with additional linear constraints. The quadratic form is
    set by a system of linear equations, which correspond to extremum of
    the form.
    First the matrix should me initiated.
    Than elements of the matrix and the constrains should be set.
    Than the matrix should be converted before solution of the equations.
    Than the equations should be solved, resulting in minimum of the form
    with the constraints.
    
    Last change: 03.06.2016 *)

<* O2EXTENSIONS + *>
MODULE mLagrangeMultiplierLinear;

IMPORT
  Math:=mMath, SM:=mSquareMatrix, Arr:=mArrays;

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

  Constraint = POINTER TO RECORD
    alpha:ArrayPtr;
    beta:Float;
    next:Constraint;
  END;

  Matrix * = RECORD
    m:SM.Matrix;
    rank:LONGINT; (* rank of the initial matrix *)
    NoC:LONGINT;  (* number of constraints. *)
    c:Constraint; (* the constraints *)
    beta:ArrayPtr; (* Array, which contains betas - RHS for the big linear system *)
    x:ArrayPtr;   (* Buffer with length of rank+NoC for solution of equations *)
  END;
  (** The matrix *)

PROCEDURE Init * (VAR m:Matrix; rank:LONGINT);
  (** Initiates the matrix with the specified rank *)
  BEGIN
    ASSERT(rank>0);

    SM.Init(m.m,rank);
    m.rank:=rank;
    m.NoC:=0;
    m.c:=NIL;
    m.x:=NIL;
    m.beta:=NIL;
  END Init;

PROCEDURE Initialized * (m-:Matrix): BOOLEAN;
  (** Returns if the matrix is initialized *)
  BEGIN
    RETURN SM.Initialized(m.m)
  END Initialized;

PROCEDURE Converted * (m-:Matrix): BOOLEAN;
  (** Returns if the matrix is converted *)
  BEGIN
    RETURN SM.Converted(m.m)
  END Converted;

PROCEDURE Rank * (m-:Matrix): LONGINT;
  (** Returns rank of the matrix *)
  BEGIN
    ASSERT(Initialized(m));

    RETURN m.rank
  END Rank;

PROCEDURE GetItem * (m-:Matrix; i,j:LONGINT): Float;
  (** Returns (i,j)-th item of the matrix. The matrix must not be converted. *)
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(i>=0);
    ASSERT(i<m.rank);
    ASSERT(j>=0);
    ASSERT(j<m.rank);

    RETURN SM.GetItem(m.m,i,j)
  END GetItem;

PROCEDURE SetItem * (VAR m:Matrix; i,j:LONGINT; val:Float);
  (** Sets (i,j)-th item of the matrix. The matrix must not be converted. *)
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(i>=0);
    ASSERT(i<m.rank);
    ASSERT(j>=0);
    ASSERT(j<m.rank);

    SM.SetItem(m.m,i,j,val);
  END SetItem;

PROCEDURE AddConstraint * (VAR m:Matrix; alpha-:Array; beta:Float);
  (** Adds a linear constraint to the minimization of the corresponding form.
      LEN(alpha) should be equal ti Rank(m).
      The constraint: Sum(alpha[i]*x[i])=beta, where x is the variable *)
  VAR
    c:Constraint;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(m.NoC<m.rank-1);
    ASSERT(LEN(alpha)=m.rank);

    NEW(c);
    c.alpha:=Arr.Copy(alpha);
    c.beta:=beta;
    c.next:=m.c;
    m.c:=c;
    INC(m.NoC);
  END AddConstraint;

PROCEDURE Convert * (VAR m:Matrix);
  (** Converts the matrix and the previously added constraints
      before the solution *)
  VAR
    i,j,rank,RANK,NoC,ind:LONGINT;
    mm:SM.Matrix;
    c:Constraint;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));

    NoC:=m.NoC;
    IF NoC#0 THEN
      rank:=m.rank;
      RANK:=rank+NoC;

      SM.Init(mm,RANK);
      NEW(m.beta,NoC);
      NEW(m.x,RANK);

      FOR i:=0 TO rank-1 DO
        FOR j:=0 TO rank-1 DO
          SM.SetItem(mm,i,j,SM.GetItem(m.m,i,j));
        END;
      END;

      FOR i:=rank TO RANK-1 DO
        FOR j:=rank TO RANK-1 DO
          SM.SetItem(mm,i,j,0);
        END;
      END;

      c:=m.c;
      ind:=rank;
      WHILE c#NIL DO
        FOR i:=0 TO rank-1 DO
          SM.SetItem(mm,ind,i,c.alpha[i]);
          SM.SetItem(mm,i,ind,c.alpha[i]);
        END;
        m.beta[ind-rank]:=c.beta;
        c:=c.next;
        INC(ind);
      END;

      m.m:=mm;
      m.c:=NIL;
    END;

    SM.Convert(m.m,m.m.rank-1);
  END Convert;

PROCEDURE Solve * (m-:Matrix; VAR x:Array);
  (** Results in minimum of the form m*x=b with the previously
      added constraints. *)
  VAR
    i,rank:LONGINT;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(Converted(m));
    ASSERT(LEN(x)=m.rank);

    IF m.NoC=0 THEN
      SM.Solve(m.m,x);
    ELSE
      rank:=m.rank;
      FOR i:=0 TO rank-1 DO
        m.x[i]:=x[i];
      END;
      FOR i:=rank TO LEN(m.x^)-1 DO
        m.x[i]:=m.beta[i-rank]
      END;
      SM.Solve(m.m,m.x^);
      FOR i:=0 TO rank-1 DO
        x[i]:=m.x[i];
      END;
    END;
  END Solve;

END mLagrangeMultiplierLinear.