(** This module presents a linear interpolation of a function of 1
    variable given by points.
    Last change: 20.11.2015 *)

<* O2EXTENSIONS + *>
MODULE mLinearInterpolation;

IMPORT
  Math:=mMath, mMisc, Arrays:=mArrays;

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

  Interpolation * = RECORD (** The interpolation *)
    x-,y-:ArrayPtr; (** Given points *)
    slope:ArrayPtr;
  END;

PROCEDURE (VAR ipn:Interpolation) FindIndex * (x:Float): LONGINT;
  (** Should return an index of the array ipn.x with the following condition:
      ipn.x[i]<=x<=ipn.x[i+1].
      Precondition: ipn.x[0]<=x<=ipn.x[LEN(ipn.x^)-1]
      This procesure may be override if a better search algorithm may be applied.
      Default a binary search takes splace. *)
  VAR
    index:LONGINT;
  BEGIN
    index:=mMisc.FindArrayIndex(ipn.x^,x,LEN(ipn.x^))-1;
    IF index=-1 THEN (* it means that x=ipn.x[0] *)
      index:=0
    END;
    RETURN index;
  END FindIndex;

PROCEDURE Init * (VAR ipn:Interpolation; x-,y-:Array);
  (** Initializes the interpolation by arrays of given points.
      Each point is specified by (x[i],y[i]).
      x[i] should be ordered by growing (i.e. x[i]<x[i+1]),
      x-coordinated should not be repeated. *)
  VAR
    i,len:LONGINT;
  BEGIN
    ASSERT(LEN(x)=LEN(y));
    ASSERT(LEN(x)>=2);
    ASSERT(Arrays.IsIncreasing(x));

    len:=LEN(x);

    NEW(ipn.x,len);
    NEW(ipn.y,len);
    NEW(ipn.slope,len-1);

    FOR i:=0 TO len-1 DO
      ipn.x[i]:=x[i];
      ipn.y[i]:=y[i];
    END;
    FOR i:=0 TO len-2 DO
      ipn.slope[i]:=(ipn.y[i+1]-ipn.y[i])/(ipn.x[i+1]-ipn.x[i])
    END;
  END Init;

PROCEDURE Clear * (VAR ipn:Interpolation);
  (** Clears the interpolation *)
  BEGIN
    ipn.x:=NIL;
    ipn.y:=NIL;
    ipn.slope:=NIL;
  END Clear;

<* PUSH *>
<* WOFF900 + *>

PROCEDURE Calculate * (ipn-:Interpolation; x:Float): Float;
  (** Returns the linearly interpolated y at specified point x.
      x should be within boubs of the array which was specified in
      the procedure Init: ipn.x[0]<=x<=ipn.x[LEN(ipn.x^)-1] *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(ipn.x#NIL);
    ASSERT(ipn.y#NIL);
    ASSERT(LEN(ipn.x^)=LEN(ipn.y^));

    ASSERT(x>=ipn.x[0]);
    ASSERT(x<=ipn.x^[LEN(ipn.x^)-1]);

    i:=ipn.FindIndex(x);
    ASSERT(i>=0);
    ASSERT(i<LEN(ipn.x^)-1);
    RETURN ipn.y[i]+(x-ipn.x[i])*ipn.slope[i]
  END Calculate;

<* POP *>

END mLinearInterpolation.