(** This module presents a bilinear interpolation of a function of two arguments
    given by points in a rectangular not necessary homogeneous grid.

    Last change: 20.11.2015 *)

<* O2EXTENSIONS + *>
MODULE mLinearInterpolation2D;

IMPORT
  Math:=mMath, mMisc, Arrays:=mArrays;

TYPE
  Float * = Math.Float;
  FloatArray = Math.Array;
  FloatArrayPtr = Math.ArrayPtr;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;

  Interpolation* = RECORD  (** The interpolation *)
    x - , y- : FloatArrayPtr; (** x- and y-points of the grid *)
    z - : FloatArray2Ptr; (** Values of the function in the points *)
  END;

PROCEDURE Init * (VAR ipn: Interpolation; x-,y-:FloatArray; z-:FloatArray2);
  (** Initializes the interpolation by arrays of given points.
      Each point is specified by (x[i],y[j],z[i,j]).
      x[i] and y[j] should be ordered by increasing, i.e. x[i]<x[i+1],
      y[j]<y[j+1]. *)
  BEGIN
    ASSERT(LEN(x)=LEN(z,0));
    ASSERT(LEN(y)>=LEN(z,1));
    ASSERT(LEN(x)>=2);
    ASSERT(LEN(y)>=2);
    ASSERT(Arrays.IsIncreasing(x));
    ASSERT(Arrays.IsIncreasing(y));

    ipn.x:=Arrays.Copy(x);
    ipn.y:=Arrays.Copy(y);
    ipn.z:=Arrays.Copy2(z);
  END Init;

PROCEDURE Clear * (VAR ipn: Interpolation);
  (** Clears the interpolation *)
  BEGIN
    ipn.x:=NIL;
    ipn.y:=NIL;
    ipn.z:=NIL;
  END Clear;

PROCEDURE FindIndex (arr-:FloatArray; x:Float): LONGINT;
  VAR
    index:LONGINT;
  BEGIN
    index:=mMisc.FindArrayIndex(arr,x,LEN(arr))-1;
    IF index=-1 THEN (* it means that x=ipn.x[0] *)
      index:=0
    END;
    RETURN index;
  END FindIndex;

PROCEDURE Calculate * (ipn-:Interpolation; x,y:Float):Float;
  (** Returns the bilinearly interpolated z at specified point (x,y).
      (x,y) should be within boubs of the arrays which were specified in
      the procedure Init:
      ipn.x[0]<=x<=ipn.x[LEN(ipn.x^)-1],
      ipn.y[0]<=y<=ipn.y[LEN(ipn.y^)-1] *)
  VAR
    iX,iY:LONGINT;
    z0,z1:Float;
  BEGIN
    ASSERT(ipn.x#NIL);
    ASSERT(ipn.y#NIL);
    ASSERT(ipn.z#NIL);
    ASSERT(LEN(ipn.x^)=LEN(ipn.z^));
    ASSERT(LEN(ipn.y^)=LEN(ipn.z^,1));

    ASSERT(x>=ipn.x[0]);
    ASSERT(x<=ipn.x^[LEN(ipn.x^)-1]);
    ASSERT(y>=ipn.y[0]);
    ASSERT(y<=ipn.y^[LEN(ipn.y^)-1]);

    iX:=FindIndex(ipn.x^,x);
    iY:=FindIndex(ipn.y^,y);
    z0:= ipn.z[iX,iY] + (x-ipn.x[iX]) * (ipn.z[iX+1,iY]-ipn.z[iX,iY]) / (ipn.x[iX+1]-ipn.x[iX]);
    z1:= ipn.z[iX,iY+1] + (x-ipn.x[iX]) * (ipn.z[iX+1,iY+1]-ipn.z[iX,iY+1]) / (ipn.x[iX+1]-ipn.x[iX]);
    RETURN z0 + (y-ipn.y[iY]) / (ipn.y[iY+1]-ipn.y[iY]) * (z1-z0);
  END Calculate;

END mLinearInterpolation2D.