(** Linear regression of statistical data.

    Last change: 20.11.2015 *)

<* O2EXTENSIONS + *>
MODULE mLinearRegression;

IMPORT
  Math:=mMath;

TYPE
  Float * = Math.Float;

  Data * = RECORD (** Collector of the statistical data *)
    SumX,SumY,SumSqrX,SumSqrY,SumXY:Float;
    N:LONGINT;
  END;

  StatData * = RECORD   (** Statistical data of linear regression. *)
    RSS * ,             (** Residual sum of squares = Sum(sqr(y[i]-intercept-slope*x[i])) *)
    SqrDevIntercept * , (** Standard deviation of intercept *)
    SqrDevSlope * ,     (** Standard deviation of slope *)
    SqrAdjR * :Float;   (** Adjusted R^2 *)
    (** Relation between the statistical data and results of Origin's linear regression:
        RSS = Origin's Residual sum of squares
        SqrDevIntercept^0.5 = Origin's Standard Error of Intercept
        SqrDevSlope^0.5 = Origin's Standard Error of Slope
        SqrAdjR = Origin's Adj. R-Square *)
  END;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE Init * (VAR d:Data);
  (** Initiates the collector *)
  BEGIN
    d.SumX:=0;
    d.SumY:=0;
    d.SumSqrX:=0;
    d.SumSqrY:=0;
    d.SumXY:=0;
    d.N:=0;
  END Init;

PROCEDURE AddPoint * (VAR d:Data; x,y:Float);
  (** Adds the specified point (x,y) to the collector *)
  BEGIN
    d.SumX:=d.SumX+x;
    d.SumY:=d.SumY+y;
    d.SumSqrX:=d.SumSqrX+sqr(x);
    d.SumSqrY:=d.SumSqrY+sqr(y);
    d.SumXY:=d.SumXY+x*y;
    INC(d.N);
  END AddPoint;

PROCEDURE Get * (d-:Data; VAR intercept,slope:Float; VAR stat:StatData);
  (** Performs a simple linear regression algorithm to find the linear
      approximation of points {x[i],y[i]}, which were previously added to d:
      y = intercept + slope * x.
      intercept and slope are returned in the corresponding variables.
      Returns some statistics in stat. *)
  VAR
    SumX,SumSqrX,SumY,SumSqrY,SumXY,Z,InvZ,RSS,TSS,a,b,S,InvN:Float;
    N:LONGINT;
  BEGIN
    ASSERT(d.N>1);

    N:=d.N;
    InvN:=1/N;
    SumX:=d.SumX;
    SumSqrX:=d.SumSqrX;
    SumY:=d.SumY;
    SumSqrY:=d.SumSqrY;
    SumXY:=d.SumXY;

    Z:=N*SumSqrX-sqr(SumX);
    ASSERT(Z#0);
    InvZ:=1/Z;

    b:=(N*SumXY-SumX*SumY)*InvZ;
    a:=(SumY-b*SumX)*InvN;
    RSS:=(N*SumSqrY-sqr(SumY)-sqr(b)*Z)*InvN;

    intercept:=a;
    slope:=b;
    stat.RSS:=RSS;
    IF N=2 THEN
      S:=0
    ELSE
      S:=RSS/(N-2)
    END;
    stat.SqrDevIntercept:=S*SumSqrX*InvZ;
    stat.SqrDevSlope:=S*N*InvZ;

    TSS:=SumSqrY - sqr(SumY)*InvN;
    IF TSS=0 THEN (* all the y-values are similar *)
      stat.SqrAdjR:=1
    ELSE
      stat.SqrAdjR:=1 - S/TSS*(N-1);
    END;
  END Get;

PROCEDURE Slope * (d-:Data): Float;
  (** Returns slope of the linear approximation, see Get *)
  VAR
    Z:Float;
  BEGIN
    ASSERT(d.N>1);

    Z:=d.N*d.SumSqrX-sqr(d.SumX);
    ASSERT(Z#0);

    RETURN (d.N*d.SumXY-d.SumX*d.SumY)/Z;
  END Slope;

PROCEDURE SlopeAndIntercept * (d-:Data; VAR slope,intercept:Float);
  (** Returns slope and intercept of the linear approximation, see Get *)
  BEGIN
    ASSERT(d.N>1);

    slope:=Slope(d);
    intercept:=(d.SumY-slope*d.SumX)/d.N;
  END SlopeAndIntercept;

PROCEDURE Intercept * (d-:Data): Float;
  (** Returns intercept of the linear approximation, see Get *)
  VAR
    slope,intercept:Float;
  BEGIN
    ASSERT(d.N>1);

    SlopeAndIntercept(d,slope,intercept);
    RETURN intercept
  END Intercept;

END mLinearRegression.