(** This module defines the basic numerical type Float and some constants and
    procedures for it, including all constants and procedures from
    modules RealMath/LongMath and MathR/MathL.
    The type Float may be REAL or LONGREAL. This is set not here
    but in the project file via equation "Float".

    Last change: 26.03.2015 *)

<* O2EXTENSIONS + *>

MODULE mMath;

<* IF Float="REAL" THEN *>

  IMPORT
    MathF:=MathR, FMath:=RealMath;

  TYPE
    Float * = REAL;

<* ELSIF Float="LONGREAL" THEN *>

  IMPORT
    MathF:=MathL, FMath:=LongMath;

  TYPE
    Float * = LONGREAL;

<* ELSIF Float="LONGLONGREAL" THEN *>

  IMPORT
    MathF:=msysMathLL, FMath:=msysLongLongMath;

  TYPE
    Float * = LONGLONGREAL;

<* ELSE *>

  'ERROR: equation Float is not defined

<* END *>

TYPE
  Array * = ARRAY OF Float;
  ArrayPtr * = POINTER TO Array;
  Array2 * = ARRAY OF ARRAY OF Float;
  Array2Ptr * = POINTER TO Array2;

CONST
  pi *   = 3.1415926535897932384626433832795028841972;
  exp1 * = 2.7182818284590452353602874713526624977572;
  SqrtPi * = 1.772453850905516027298167;     (** sqrt(pi) *)
  Sqrt2 * = 1.414213562373095048801689;      (** sqrt(2) *)
  Sqrt3 * = 1.732050807568877293527446;      (** sqrt(3) *)
  Ln2 * = 6.93147180559945309417232121458E-1;(** ln(2) *)

VAR

  MaxREAL-:REAL;
    (** Maximal value of REAL varable. *)
  MaxLONGREAL-:LONGREAL;
    (** Maximal value of LONGREAL varable. *)
  MaxLONGLONGREAL-:LONGLONGREAL;
    (** Maximal value of LONGLONGREAL varable. *)
  MaxFloat-:Float;
    (** Maximal value of Float varable. *)

  MinREAL-:REAL;
    (** Minimal positive value of REAL varable. *)
  MinLONGREAL-:LONGREAL;
    (** Minimal positive value of LONGREAL varable. *)
  MinLONGLONGREAL-:LONGLONGREAL;
    (** Minimal positive value of LONGLONGREAL varable. *)
  MinFloat-:Float;
    (** Minimal positive value of Float varable. *)

CONST
  EpsREAL * = 1.1920928955078125E-0007;
    (** Value x: 1.0<x<1.0+EpsREAL can not be presented by REAL variable. *)
  EpsLONGREAL * = 2.22044604925031308E-0016;
    (** Value x: 1.0<x<1.0+EpsLONGREAL can not be presented by LONGREAL variable. *)
  EpsLONGLONGREAL * = 1.08420217248550443E-0019;
    (** Value x: 1.0<x<1.0+EpsLONGLONGREAL can not be presented by LONGLONGREAL variable. *)
<* IF Float="REAL" THEN*>
  EpsFloat * = EpsREAL;
    (** Value x: 1.0<x<1.0+EpsFloat can not be presented by Float variable. *)
<* ELSIF Float="LONGREAL" THEN*>
  EpsFloat * = EpsLONGREAL;
    (** Value x: 1.0<x<1.0+EpsFloat can not be presented by Float variable. *)
<* ELSE *>
  EpsFloat * = EpsLONGLONGREAL;
    (** Value x: 1.0<x<1.0+EpsFloat can not be presented by Float variable. *)
<* END *>

PROCEDURE^ arctan2 (y,x:Float): Float;

PROCEDURE sqr * (x:Float): Float;
  (** Returns x*x *)
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt * (x:Float): Float;
  BEGIN
    RETURN FMath.sqrt(x);
  END sqrt;

PROCEDURE exp * (x:Float): Float;
  BEGIN
    RETURN FMath.exp(x);
  END exp;

PROCEDURE ln * (x:Float): Float;
  BEGIN
    RETURN FMath.ln(x);
  END ln;

PROCEDURE sin * (x:Float): Float;
  BEGIN
    RETURN FMath.sin(x);
  END sin;

PROCEDURE cos * (x:Float): Float;
  BEGIN
    RETURN FMath.cos(x);
  END cos;

PROCEDURE tan * (x:Float): Float;
  BEGIN
    RETURN FMath.tan(x);
  END tan;

PROCEDURE cotan * (x:Float): Float;
  BEGIN
    RETURN cos(x)/sin(x);
  END cotan;

PROCEDURE sec * (x:Float): Float;
  BEGIN
    RETURN 1/cos(x);
  END sec;

PROCEDURE cosec * (x:Float): Float;
  BEGIN
    RETURN 1/sin(x);
  END cosec;

PROCEDURE arcsin * (x:Float): Float;
  BEGIN
    ASSERT(x>=-1);
    ASSERT(x<=1);

    RETURN arctan2(x,sqrt((1-x)*(1+x)));
  END arcsin;

PROCEDURE arccos * (x:Float): Float;
  BEGIN
    ASSERT(x>=-1);
    ASSERT(x<=1);

    RETURN arctan2(sqrt((1-x)*(1+x)),x);
  END arccos;

PROCEDURE arctan * (x:Float): Float;
  BEGIN
    RETURN FMath.arctan(x);
  END arctan;

PROCEDURE power * (base,exponent:Float): Float;
  BEGIN
    RETURN FMath.power(base,exponent)
  END power;

PROCEDURE IntPower * (base:Float; exponent:LONGINT): Float;
  (** Returns base**exponent. *)

  VAR
    result:Float;
  BEGIN
    ASSERT((base # 0) OR (exponent>=0));

    IF exponent<0 THEN
      exponent:=-exponent;
      base:=1/base;
    END;

    result:=1;

    WHILE exponent<>0 DO
      IF ODD(exponent) THEN
        DEC(exponent);
        result:=base*result;
      ELSE
        exponent:=exponent DIV 2;
        base:=base*base;
      END;
    END;

    RETURN result

  END IntPower;

PROCEDURE HalfIntPower * (base:Float; exponent2:LONGINT): Float;
  (** Returns base^(exponent2/2) *)
  BEGIN
    ASSERT(base>0);

    IF ODD(exponent2) THEN
      RETURN IntPower(base,exponent2 DIV 2)*sqrt(base)
    ELSE
      RETURN IntPower(base,exponent2 DIV 2)
    END;
  END HalfIntPower;

PROCEDURE LongENTIER * (x:LONGLONGREAL): LONGLONGINT;
  (** Returns integer part of x rounded to -infinity *)
  VAR
    res:LONGLONGINT;
    cw,cw0:INTEGER;
  BEGIN
    ASM
      fstcw cw0
    END;
    cw:=(cw0 & 0F3FFh) OR 0400h; (* Rounding to minus infinity *)
    ASM
      fldcw cw
      fld x
      fistp res
      fldcw cw0
    END;
    RETURN res;
  END LongENTIER;

PROCEDURE round * (x:Float): LONGINT;
  (** Returns a nearest to x integer *)
  BEGIN
    RETURN ENTIER(x+0.5)
  END round;

PROCEDURE frac * (x:Float): Float;
  (** Returns fractional part of x *)
  BEGIN
    x:=x-ENTIER(x);
    IF x<0 THEN (* Because of possible round errors *)
      x:=0
    END;
    RETURN x
  END frac;

PROCEDURE GreaterInt * (x:Float): LONGINT;
  (** Returns minimal integer greater or equal to x *)
  BEGIN
    RETURN -ENTIER(-x)
  END GreaterInt;

PROCEDURE log * (x,base:Float): Float;
  BEGIN
    ASSERT(x>0);
    ASSERT(base>0);

    RETURN ln(x)/ln(base)
  END log;

PROCEDURE arctan2 * (y,x:Float): Float;
  (** Returns the quadrant-correct arctan(y/x).

      Attention: In modules MathR/MathL there is an error in this procedure:
      it returns arctan(x/y) rather than arctan(y/x), as it should be
      according to OakWood guidelines.
      This procedure returns correct value arctan(y/x). *)
  BEGIN
    RETURN MathF.arctan2(x,y);
  END arctan2;

PROCEDURE sinh * (x:Float): Float;
  BEGIN
    x:=exp(x);
    RETURN (x-1/x)/2
  END sinh;

PROCEDURE cosh * (x:Float): Float;
  BEGIN
    x:=exp(x);
    RETURN (x+1/x)/2
  END cosh;

PROCEDURE tanh * (x:Float): Float;
  BEGIN
    IF x>0 THEN
      x:=exp(-2*x);
      RETURN (1-x)/(1+x)
    ELSE
      x:=exp(2*x);
      RETURN (x-1)/(x+1)
    END;
  END tanh;

PROCEDURE cotanh * (x:Float): Float;
  BEGIN
    ASSERT(x#0);

    RETURN 1/tanh(x)
  END cotanh;

PROCEDURE sech * (x:Float): Float;
  BEGIN
    RETURN 1/cosh(x);
  END sech;

PROCEDURE cosech * (x:Float): Float;
  BEGIN
    RETURN 1/sinh(x);
  END cosech;

PROCEDURE arcsinh * (x:Float): Float;
  BEGIN
    IF x>0 THEN
      RETURN ln(x+sqrt(x*x+1))
    ELSE
      RETURN -ln(-x+sqrt(x*x+1));
    END;
  END arcsinh;

PROCEDURE arccosh * (x:Float): Float;
  BEGIN
    ASSERT(x>=1);

    RETURN ln(x+sqrt((x-1)*(x+1)))
  END arccosh;

PROCEDURE arctanh * (x:Float): Float;
  BEGIN
    ASSERT(x>-1);
    ASSERT(x<1);

    RETURN ln((1+x)/(1-x))/2;
  END arctanh;

PROCEDURE arccotanh * (x:Float): Float;
  BEGIN
    ASSERT(x>-1);
    ASSERT(x<1);

    RETURN ln((x+1)/(x-1))/2;
  END arccotanh;

<* PUSH *>
<* ALIGNMENT = '1' *>
TYPE
  RealBytes = ARRAY SIZE(REAL) OF SHORTINT;
  LongRealBytes = ARRAY SIZE(LONGREAL) OF SHORTINT;
  LongLongRealBytes = ARRAY SIZE(LONGLONGREAL) OF SHORTINT;
<* POP *>

PROCEDURE SetLongLongReal(bytes-:LongLongRealBytes): LONGLONGREAL;
  VAR
    x:RECORD ["Modula"]
      CASE :INTEGER OF
        0:
          bytes:LongLongRealBytes;
      | 1:
          result:LONGLONGREAL;
      END;
    END;
  BEGIN
    x.bytes:=bytes;
    RETURN x.result
  END SetLongLongReal;

PROCEDURE SetLongReal(bytes-:LongRealBytes): LONGREAL;
  VAR
    x:RECORD ["Modula"]
      CASE :INTEGER OF
        0:
          bytes:LongRealBytes;
      | 1:
          result:LONGREAL;
      END;
    END;
  BEGIN
    x.bytes:=bytes;
    RETURN x.result
  END SetLongReal;

PROCEDURE SetReal(bytes-:RealBytes): REAL;
  VAR
    x:RECORD ["Modula"]
      CASE :INTEGER OF
        0:
          bytes:RealBytes;
      | 1:
          result:REAL;
      END;
    END;
  BEGIN
    x.bytes:=bytes;
    RETURN x.result
  END SetReal;

BEGIN
  MaxLONGLONGREAL:=SetLongLongReal(LongLongRealBytes{-1,-1,-1,-1,-1,-1,-1,-1,-2,127});
  MinLONGLONGREAL:=SetLongLongReal(LongLongRealBytes{1,0,0,0,0,0,0,0,0,0});
  MaxLONGREAL:=SetLongReal(LongRealBytes{-1,-1,-1,-1,-1,-1,-17,127});
  MinLONGREAL:=SetLongReal(LongRealBytes{1,0,0,0,0,0,0,0});
  MaxREAL:=SetReal(RealBytes{-1,-1,127,127});
  MinREAL:=SetReal(RealBytes{1,0,0,0});

<* IF Float="REAL" THEN*>
  MaxFloat:=MaxREAL;
  MinFloat:=MinREAL;
<* ELSIF Float="LONGREAL" THEN*>
  MaxFloat:=MaxLONGREAL;
  MinFloat:=MinLONGREAL;
<* ELSE *>
  MaxFloat:=MaxLONGLONGREAL;
  MinFloat:=MinLONGLONGREAL;
<* END *>
END mMath.