(** This module works with three-diagonal matrix and corresponding
    systems of linear equations.
    Last change: 05.07.2014 *)
<* O2EXTENSIONS + *>
MODULE mMatrix3;

IMPORT
  Math:=mMath, Arr:=mArrays;

TYPE
  Float * = Math.Float;
  Complex = Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;

  Matrix * = RECORD
    a-,b-,c-:ArrayPtr;
    item02 *, item20 * :Complex;
    LastIndex-:LONGINT;
  END;
  (** The three-diagonal matrix. Presentation of the
      fields corresponds to the following equation:
      a[i]*x[i-1] + b[i]*x[i] + c[i]*x[i+1] = d[i].
      a,b,c are arrays [0..LastIndex] of Float. They
      contain the mentioned above coefficients.
      Hence, a[0] and c[LastIndex] makes no sence
      and are not used.
      Additionally, 2 more items are presented and taken
      into account: item[0,2] in the variable item02
      and item[LastIndex,LastIndex-1] in the variable item20. *)

PROCEDURE Init * (VAR m:Matrix; rank:LONGINT);
  (** Initiates the matrix with given rank. *)
  BEGIN
    ASSERT(rank>=3);

    NEW(m.a,rank);
    NEW(m.b,rank);
    NEW(m.c,rank);
    m.LastIndex:=rank-1;
    m.item20:=0;
    m.item02:=0;
  END Init;

PROCEDURE GetDiag (m-:Matrix; i,j:LONGINT): ArrayPtr;
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT((i>=0) & (i<=m.LastIndex));
    ASSERT((j>=0) & (j<=m.LastIndex));
    ASSERT(ABS(i-j)<=1);

    CASE j-i OF
      -1:
        RETURN m.a
    | 0:
        RETURN m.b
    | 1:
        RETURN m.c
    END;
  END GetDiag;

PROCEDURE GetItem * (m-:Matrix; i,j:LONGINT): Complex;
  (** Returns value of item of the matrix specified by indexes i,j.
      Works only with elements in a,b, or c, doesn't work
      with additional elements item20 and item02.
      ABS(i-j) should be <=1 *)
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT((i>=0) & (i<=m.LastIndex));
    ASSERT((j>=0) & (j<=m.LastIndex));
    ASSERT(ABS(i-j)<=1);

    RETURN GetDiag(m,i,j)^[i];
  END GetItem;

PROCEDURE SetItem * (m-:Matrix; i,j:LONGINT; value:Complex);
  (** Sets value of item of the matrix specified by indexes i,j.
      Works only with elements in a,b, or c, doesn't work
      with additional elements item20 and item02.
      ABS(i-j) should be <=1. *)
  VAR
    p:ArrayPtr;
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT((i>=0) & (i<=m.LastIndex));
    ASSERT((j>=0) & (j<=m.LastIndex));
    ASSERT(ABS(i-j)<=1);

    p:=GetDiag(m,i,j);
    p^[i]:=value
  END SetItem;

PROCEDURE Rank * (m-:Matrix): LONGINT;
  (** Returns rank of the matrix. *)
  BEGIN
    ASSERT(m.a#NIL);

    RETURN m.LastIndex+1;
  END Rank;

PROCEDURE Assign * (VAR m,m1:Matrix);
  (** m:=copy of m1 *)
  BEGIN
    IF m1.a=NIL THEN
      m.a:=NIL;
      m.b:=NIL;
      m.c:=NIL;
      m.LastIndex:=-1;
      RETURN
    END;

    IF (m.a=NIL) OR (m.LastIndex#m1.LastIndex) THEN
      Init(m,m1.LastIndex+1);
    END;

    Arr.Assign(m.a^,m1.a^);
    Arr.Assign(m.b^,m1.b^);
    Arr.Assign(m.c^,m1.c^);
    m.item02:=m1.item02;
    m.item20:=m1.item20;
  END Assign;

PROCEDURE MulVector * (VAR m:Matrix; VAR b:Array);
  (** b:=m*b *)
  VAR
    i:LONGINT;
    x,y,b2:Complex;
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT(LEN(b)=m.LastIndex+1);

    b2:=b[LEN(b)-3];

    x:=b[0];
    b[0]:= m.b[0]*b[0] + m.c[0]*b[1] + m.item02*b[2];
    FOR i:=1 TO LEN(b)-2 DO
      y:=b[i];
      b[i]:= m.a[i]*x + m.b[i]*b[i] + m.c[i]*b[i+1];
      x:=y;
    END;
    b[LEN(b)-1]:= m.a[LEN(b)-1]*x +m.b[LEN(b)-1]*b[LEN(b)-1] +m.item20*b2;

  END MulVector;

PROCEDURE Solve * (VAR m:Matrix; VAR d:Array);
  (** Solves the system of equations m*x=d and returns the result in d.
      At that elements item20 and item02 are assumed to be 0,
      the real values are ignored. Matrix m is changed. *)
  VAR
    i:LONGINT;
    tmp:Complex;
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT(LEN(d)=m.LastIndex+1);

    tmp:=1/m.b[0];
    m.a[0]:=-m.c[0]*tmp;
    m.b[0]:=d[0]*tmp;

    FOR i:=1 TO m.LastIndex-1 DO
      tmp:=-1/(m.b[i]+m.a[i]*m.a[i-1]);
      m.b[i]:=(m.a[i]*m.b[i-1]-d[i])*tmp;
      m.a[i]:=m.c[i]*tmp;
    END;

    i:=m.LastIndex;
    d[i]:=(m.a[i]*m.b[i-1]-d[i])/(-m.b[i]-m.a[i]*m.a[i-1]);

    FOR i:=m.LastIndex-1 TO 0 BY -1 DO
      d[i]:=d[i+1]*m.a[i]+m.b[i]
    END;
  END Solve;

PROCEDURE SolveEx * (VAR m:Matrix; VAR d:Array);
  (** Solves the system of equations m*x=d and returns the result in d.
      At that elements item20 and item02 are taken into account.
      Matrix m is changed. *)
  VAR
    i:LONGINT;
    l1,l2:Complex;
  BEGIN
    ASSERT(m.a#NIL);
    ASSERT(LEN(d)=m.LastIndex+1);

    l1:=-m.item02/m.b[0];
    m.c[1]:=m.c[1]+l1*m.a[1];

    i:=m.LastIndex;
    l2:=-m.item20/m.b[i];
    m.a[i-1]:=m.a[i-1]+l2*m.c[i-1];

    Solve(m,d);

    d[0]:=d[0]+l1*d[2];
    d[i]:=d[i]+l2*d[i-2]

  END SolveEx;

END mMatrix3.