(** This module contains miscellaneous mathematical routines.

    If an option mNeglectGrowingAssert is defined, then the procedure
    FindArrayIndex doesn't assert that the input array is not decreasing.

    Last change: 20.11.2015 *)

(* <* NEW mNeglectGrowingAssert *> *)

<* O2EXTENSIONS + *>
MODULE mMisc;

IMPORT
  Math:=mMath, AlgEqu:=mAlgEqu;

<* IF ~DEFINED(mNeglectGrowingAssert) THEN *>
IMPORT
  Arrays:=mArrays;
<* END *>

IMPORT
<* IF    Float="REAL" THEN*>         FMath:=RealMath;
<* ELSIF Float="LONGREAL" THEN*>     FMath:=LongMath;
<* ELSIF Float="LONGLONGREAL" THEN*> FMath:=msysLongLongMath;
<* END *>

TYPE
  Float * = Math.Float;

  Array = Math.Array;

CONST
  pi = Math.pi;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN FMath.sqrt(x)
  END sqrt;

PROCEDURE Min * (x,y:Float): Float;
  (** Returns minimal of two values *)
  BEGIN
    IF x<y THEN
      RETURN x
    ELSE
      RETURN y
    END;
  END Min;

PROCEDURE Max * (x,y:Float): Float;
  (** Returns maximal of two values *)
  BEGIN
    IF x>y THEN
      RETURN x
    ELSE
      RETURN y
    END;
  END Max;

PROCEDURE IntFunc1px2PowN2 * (x:Float; n:LONGINT): Float;
  (** Returns integral (1+t*t)**(-n/2) dt from 0 to x *)
  VAR
    a,b,d,result:Float;
    k:LONGINT;
  BEGIN
    IF ODD(n) THEN
      IF n=1 THEN
        result:=Math.arcsinh(x)
      ELSE
        n:=n DIV 2;
        a:=1/(1+x*x);
        b:=1;
        d:=1;
        result:=1;
        FOR k:=0 TO n-2 DO
          b:=b*(2*k+1)/(2*k+2);
          d:=d*a;
          result:=result+b*d
        END;
        result:=result/(b*(2*n-1))*x*Math.sqrt(a)
      END
    ELSE
      CASE n OF
        0:
          result:=x;
      | 2:
          result:=Math.arctan(x);
      ELSE
        n:=n DIV 2;
        a:=1/(1+x*x);
        b:=1;
        d:=a;
        result:=a;
        FOR k:=1 TO n-2 DO
          b:=b*2*k/(2*k+1);
          d:=d*a;
          result:=result+b*d
        END;
        result:=(x*result+Math.arctan(x))/(b*2*(n-1))
      END;
    END;
    RETURN result
  END IntFunc1px2PowN2;

TYPE
  StudentFunc = RECORD (AlgEqu.NewtonFunc)
    Const:Float;
    N:LONGINT;
  END;

 <* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR f:StudentFunc) Func (var-:Array; index:LONGINT): Float;
  BEGIN
    RETURN IntFunc1px2PowN2(var[0],f.N+1)-f.Const
  END Func;

PROCEDURE (VAR f:StudentFunc) Der (var-:Array; FuncIndex,VarIndex:LONGINT): Float;
  BEGIN
    RETURN Math.HalfIntPower(1+sqr(var[0]),-(f.N+1))
  END Der;
<* POP *>

PROCEDURE StudentT * (n:LONGINT; p,RelErr,AbsErr:Float): Float;
  (** Returns coefficient of Student, corresponding to n degrees of freedom
      and confidence probability p, with accuracy defined b RelErr and AbsErr.
      At a simplest case n is equal to count of measurements minus 1. *)
  VAR
    i:LONGINT;
    factorial,SqrtN:Float;
    f:StudentFunc;
    var:Float;
  BEGIN
    ASSERT(n>0);
    ASSERT(p>=0);
    ASSERT(p<1);

    SqrtN:=Math.sqrt(n);

    f.N:=n;
    f.Const:=p*0.5;
    IF ODD(n) THEN
      factorial:=1;
      FOR i:=1 TO n DIV 2 DO
        f.Const:=f.Const*(i-0.5);
        factorial:=factorial*i
      END;
      f.Const:=f.Const*pi/factorial
    ELSE
      n:=n DIV 2;
      factorial:=1;
      FOR i:=1 TO n DO
        factorial:=factorial*(i-0.5);
      END;
      FOR i:=1 TO n-1 DO
        f.Const:=f.Const*i;
      END;
      f.Const:=f.Const/factorial
    END;
    var:=0;
    AlgEqu.Newton(var,f,RelErr,AbsErr,NIL);
    RETURN var*SqrtN;
  END StudentT;

PROCEDURE FindArrayIndex * (array-:ARRAY OF Float; x:Float;
                            count:LONGINT): LONGINT;
  (** Array should be sorted by increase. (i.e. array[i]<=array[i+1]).
      The procedure returns an index of the first element, which is
      greater or equal to x, or count if all elements are smaller.
      A binary search algorithm is used.
      Only first count elements of the array are processed.
      If count=0 then 0 is returned. *)
  VAR
    min,max,i:LONGINT;
  BEGIN
    ASSERT(count>=0);
    ASSERT(count<=LEN(array));

<* IF ~DEFINED(mNeglectGrowingAssert) THEN *>
    ASSERT(Arrays.IsNotDecreasing(array));
<* END *>

    IF count=0 THEN
      RETURN 0
    END;

    IF x<=array[0] THEN
      RETURN 0
    END;

    IF x>array[count-1] THEN
      RETURN count
    END;

    min:=0;
    max:=count-1;
    WHILE max-min>1 DO
      i:=(min+max) DIV 2;
      IF x>array[i] THEN
        min:=i
      ELSE
        max:=i
      END;
    END;

    RETURN max
  END FindArrayIndex;

PROCEDURE SolveReducedQuadraticEqu * (p,q:Float; VAR x1,x2:Float): LONGINT;
  (** Searchs roots of equation x*x + p*x + q = 0
      Returns number of real roots. Possible values: 0, 2.
      If there are real roots, they are returned in x1,x2 *)
  VAR
    D:Float;
  BEGIN
    D:=sqr(p)-4*q;
    IF D<0 THEN
      RETURN 0
    ELSE
      IF p>0 THEN
        x1:=(-p-sqrt(D))/2
      ELSE
        x1:=(-p+sqrt(D))/2
      END;
      x2:=q/x1;
      RETURN 2;
    END;
  END SolveReducedQuadraticEqu;

PROCEDURE SolveQuadraticEqu * (a,b,c:Float; VAR x1,x2:Float): LONGINT;
  (** Searchs roots of equation a*x*x + b*x + c = 0
      Returns number of real roots. Possible values: 0, 1, 2.
      If there are real roots, they are returned in x1,x2.
      If only one root exists, it is returned in both x1,x2 *)
  BEGIN
    ASSERT((a#0) OR (b#0) OR (c#0));

    IF a=0 THEN
      IF b=0 THEN
        RETURN 0
      ELSE
        x1:=-c/b;
        x2:=x1;
        RETURN 1
      END
    ELSE
      RETURN SolveReducedQuadraticEqu(b/a,c/a,x1,x2);
    END;
  END SolveQuadraticEqu;

END mMisc.