(** This module provides solution of 2-dimentional parabolic equation
    in partial derivatives in a rectangular area with a homogeneous grid.
    The equation is the following:

    dn/dt = A(x,y) * d/dx(dn/dx) + B(x,y) * d/dy(dn/dy)
           +C(x,y) * dn/dx + D(x,y) * dn/dy + E(x,y)*n + F(x,y),
    where n=n(x,y) - is the solution.

    Boundary conditions at each side of the are are specified by
    three functions, which represent coefficients at n, derivative of
    n in normal direction, and the result. For example, at min(x) the
    following conditions are taken:

    MinXFunc(y)*n(MinX,y) + MinXDer(y)dn/dx = MinXRes(y).

    Coefficients A .. F should be specified by procedures SetA .. SetF
    before the solution. By default they all are zeros.

    Boundary conditions are specified by procedures SetMinX, SetMaxX,
    SerMinY, SetMaxY. This specification may be missed, then the zero boundary
    conditions (n=0) are taken default.

    The task is in specified rectangular area and a grid on it.

    Last changed: 04.01.2016 *)

<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE mParabolicEqu2DComplex;

IMPORT

  Ar:=mArea2DComplex, Matrix:=mMatrix3Complex, MathComplex:=mMathComplex,
  Math:=mMath, Arr:=mArraysComplex, Thr:=mthrSym, Steps:=mEvolutionSteps,
  ComFunc:=mCommonFunctionsComplex, LowMath:=mLowMath, DStr:=msysDynStr;

TYPE
  Float * = Math.Float;
  Complex * = MathComplex.Complex;

  Array2 = MathComplex.Array2;
  ArrayPtr = MathComplex.ArrayPtr;
  Array2Ptr = MathComplex.Array2Ptr;
  Area = Ar.Area;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

PROCEDURE Abs (x:Complex): Float;
  BEGIN
    RETURN MathComplex.abs(x)
  END Abs;

PROCEDURE SqrAbs (x:Complex): Float;
  BEGIN
    RETURN sqr(RE(x)) + sqr(IM(x))
  END SqrAbs;

CONST
  pi = Math.pi;

TYPE
  Step1Proc = RECORD (Thr.Proc)
    e:POINTER TO Equ;
    InvDt:Float;
    old:Array2Ptr;
  END;

  Step2Proc = RECORD (Thr.Proc)
    e:POINTER TO Equ;
    InvDt:Float;
    new:Array2Ptr;
  END;

  Equ * = RECORD (** The equation *)
    area - : Area; (** Area *)
    a-,b-,c-,d-,e-,f-:Array2Ptr;
    (** the fields of the coefficients. Because they are needed only in
        internal areas, their dimensions are (LastX-1,LastY-1), and indexes
        [iZ,iR] correspond to the grid point [iZ+1,iR+1] *)
    NullArea:Array2Ptr;
    MinXF,MinXD,MinXR,MaxXF,MaxXD,MaxXR,MinYF,MinYD,MinYR,MaxYF,MaxYD,MaxYR:ArrayPtr;
    middle:Array2Ptr;
    mx,my:POINTER TO ARRAY OF Matrix.Matrix;
    xd,yd:POINTER TO ARRAY OF ArrayPtr;
    SpecialMinXBoundary,SpecialMinYBoundary,SpecialMaxXBoundary,SpecialMaxYBoundary:BOOLEAN;
    p1:POINTER TO Step1Proc;
    p2:POINTER TO Step2Proc;
  END;

PROCEDURE Init * (equ:POINTER TO Equ; area-:Area);
  (** Initialize the equation by the specified area *)
  VAR
    i,LastX,LastY:LONGINT;
  BEGIN
    ASSERT(equ#NIL);
    ASSERT(Ar.Correct(area));

    equ.NullArea:=Arr.Filled2(area.LastX-1,area.LastY-1,0);
    LastX:=area.LastX;
    LastY:=area.LastY;
    equ.area:=area;
    equ.a:=equ.NullArea;
    equ.b:=equ.NullArea;
    equ.c:=equ.NullArea;
    equ.d:=equ.NullArea;
    equ.e:=equ.NullArea;
    equ.f:=equ.NullArea;

    equ.MinXF:=Arr.Filled(LastY+1,1);
    equ.MaxXF:=equ.MinXF;
    equ.MinXD:=Arr.Filled(LastY+1,0);
    equ.MaxXD:=equ.MinXD;
    equ.MinXR:=equ.MinXD;
    equ.MaxXR:=equ.MinXD;

    equ.MinYF:=Arr.Filled(LastX+1,1);
    equ.MaxYF:=equ.MinYF;
    equ.MinYD:=Arr.Filled(LastX+1,0);
    equ.MaxYD:=equ.MinYD;
    equ.MinYR:=equ.MinYD;
    equ.MaxYR:=equ.MinYD;
    equ.SpecialMinXBoundary:=FALSE;
    equ.SpecialMaxXBoundary:=FALSE;
    equ.SpecialMinYBoundary:=FALSE;
    equ.SpecialMaxYBoundary:=FALSE;

    NEW(equ.middle,LastX+1,LastY-1);
    NEW(equ.mx,Thr.Count);
    NEW(equ.my,Thr.Count);
    NEW(equ.xd,Thr.Count);
    NEW(equ.yd,Thr.Count);
    FOR i:=0 TO Thr.Count-1 DO
     Matrix.Init(equ.mx[i],LastX+1);
     Matrix.Init(equ.my[i],LastY+1);
     NEW(equ.xd[i],LastX+1);
     NEW(equ.yd[i],LastY+1);
   END;

    NEW(equ.p1);
    equ.p1.e:=equ;
    NEW(equ.p2);
    equ.p2.e:=equ;
  END Init;

PROCEDURE Initialized * (equ-:Equ): BOOLEAN;
  (** Returns if the equation is initialized *)
  BEGIN
    RETURN equ.a#NIL
  END Initialized;

PROCEDURE GetVal (VAR a:Array2; iX,iY:LONGINT): Complex;
  BEGIN
    RETURN a[iX-1,iY-1]
  END GetVal;

PROCEDURE SetVal (VAR a:Array2; iX,iY:LONGINT; value:Complex);
  BEGIN
    a[iX-1,iY-1]:=value;
  END SetVal;

TYPE
  AreaFunc * = ComFunc.Float2;

PROCEDURE SetAreaArray (VAR arr:Array2Ptr; VAR equ:Equ; VAR f:AreaFunc);
  VAR
    iX,iY:LONGINT;
    x,y:Float;
    LastX,LastY:LONGINT;
    NewArea:BOOLEAN;
  BEGIN
    LastX:=equ.area.LastX;
    LastY:=equ.area.LastY;
    NewArea:=(arr=NIL) OR (arr=equ.NullArea) OR (LEN(arr^)#LastX-1) OR (LEN(arr^,1)#LastY-1);
    IF NewArea THEN (* IF - to reduce load of the memory manager *)
      NEW(arr,LastX-1,LastY-1);
    END;
    FOR iX:=1 TO LastX-1 DO
      x:=Ar.IndToX(equ.area,iX);
      FOR iY:=1 TO LastY-1 DO
        y:=Ar.IndToY(equ.area,iY);
        SetVal(arr^,iX,iY,f.f(x,y));
      END;
    END;
    IF  NewArea & (equ.NullArea#equ.a) & (equ.NullArea#equ.b) & (equ.NullArea#equ.c)
      & (equ.NullArea#equ.d) & (equ.NullArea#equ.e) & (equ.NullArea#equ.f)
    THEN (* If all the coefficients are set then NullArea is not needed anymore *)
      equ.NullArea:=NIL
    END;
  END SetAreaArray;

PROCEDURE SetA * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets A *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.a,equ,f);
  END SetA;

PROCEDURE SetB * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets B *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.b,equ,f);
  END SetB;

PROCEDURE SetC * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets C *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.c,equ,f);
  END SetC;

PROCEDURE SetD * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets D *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.d,equ,f);
  END SetD;

PROCEDURE SetE * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets E *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.e,equ,f);
  END SetE;

PROCEDURE SetF * (VAR equ:Equ; VAR f:AreaFunc);
  (** Sets F *)
  BEGIN
    ASSERT(Initialized(equ));

    SetAreaArray(equ.f,equ,f);
  END SetF;

TYPE
  BoundaryFunc * = ComFunc.Float1;

PROCEDURE SetArrayAlongX (VAR arr:ArrayPtr; a-:Area; VAR f:BoundaryFunc);
  VAR
    iX:LONGINT;
    x:Float;
  BEGIN
    ASSERT(Ar.Correct(a));

    NEW(arr,a.LastX+1);
    FOR iX:=0 TO a.LastX DO
      x:=Ar.IndToX(a,iX);
      arr[iX]:=f.f(x);
    END;
  END SetArrayAlongX;

PROCEDURE SetArrayAlongY (VAR arr:ArrayPtr; a-:Area; VAR f:BoundaryFunc);
  VAR
    iY:LONGINT;
    y:Float;
  BEGIN
    ASSERT(Ar.Correct(a));

    NEW(arr,a.LastY+1);
    FOR iY:=0 TO a.LastY DO
      y:=Ar.IndToY(a,iY);
      arr[iY]:=f.f(y);
    END;
  END SetArrayAlongY;

PROCEDURE SetMinX * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for x=MinX:
      func(y)*n(MinX,y) + der(y)*d/dx(n)(MinX,y) = res(y) *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMinXBoundary:=FALSE;
    SetArrayAlongY(equ.MinXF,equ.area,func);
    SetArrayAlongY(equ.MinXD,equ.area,der);
    SetArrayAlongY(equ.MinXR,equ.area,res);
  END SetMinX;

PROCEDURE SetMaxX * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for x=MaxX:
      func(y)*n(MaxX,y) + der(y)*d/dx(n)(MaxX,y) = res(y) *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMaxXBoundary:=FALSE;
    SetArrayAlongY(equ.MaxXF,equ.area,func);
    SetArrayAlongY(equ.MaxXD,equ.area,der);
    SetArrayAlongY(equ.MaxXR,equ.area,res);
  END SetMaxX;

PROCEDURE SetMinY * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for y=MinY:
      func(x)*n(x,MinY) + der(x)*d/dy(n)(x,MinY) = res(x) *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMinYBoundary:=FALSE;
    SetArrayAlongX(equ.MinYF,equ.area,func);
    SetArrayAlongX(equ.MinYD,equ.area,der);
    SetArrayAlongX(equ.MinYR,equ.area,res);
  END SetMinY;

PROCEDURE SetMaxY * (VAR equ:Equ; VAR func,der,res:BoundaryFunc);
  (** Sets the boundary condition for y=MaxY:
      func(x)*n(x,MaxY) + der(x)*d/dy(n)(x,MaxY) = res(x) *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMaxYBoundary:=FALSE;
    SetArrayAlongX(equ.MaxYF,equ.area,func);
    SetArrayAlongX(equ.MaxYD,equ.area,der);
    SetArrayAlongX(equ.MaxYR,equ.area,res);
  END SetMaxY;

PROCEDURE SetMinXSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at x=MinX: n[0,iY]=n[1,iY] for all iY *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMinXBoundary:=TRUE;
  END SetMinXSpecial;

PROCEDURE SetMaxXSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at x=MaxX: n[LastX,iY]=n[LastX-1,iY] for all iY *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMaxXBoundary:=TRUE;
  END SetMaxXSpecial;

PROCEDURE SetMinYSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at y=MinY: n[iX,0]=n[iX,1] for all iX *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMinYBoundary:=TRUE;
  END SetMinYSpecial;

PROCEDURE SetMaxYSpecial * (VAR equ:Equ);
  (** Sets a special boundary condition at y=MaxY: n[iX,LastY]=n[iX,LastY-1] for all iX *)
  BEGIN
    ASSERT(Initialized(equ));

    equ.SpecialMaxYBoundary:=TRUE;
  END SetMaxYSpecial;

PROCEDURE (VAR p:Step1Proc) Do ();
  VAR
    e:POINTER TO Equ;
    ThrInd,iX,iY,LastX,LastY:LONGINT;
    Inv2dx,Inv2dy,InvSqrDx,InvSqrDy:Float;
    aa,bb,cc,dd,ee,ff:Complex;
    m:Matrix.Matrix;
    InvDt:Float;
    d:ArrayPtr;
    old,new:Array2Ptr;
  BEGIN
    ASSERT(p.e#NIL);
    ASSERT(p.old#NIL);

    e:=p.e;
    ThrInd:=Thr.Index();
    m:=e.mx[ThrInd];
    d:=e.xd[ThrInd];
    InvDt:=p.InvDt;
    old:=p.old;
    new:=p.e.middle;

    Inv2dx:=e.area.Inv2dx;
    Inv2dy:=e.area.Inv2dy;
    InvSqrDx:=e.area.InvSqrDx;
    InvSqrDy:=e.area.InvSqrDy;
    LastX:=e.area.LastX;
    LastY:=e.area.LastY;

    iY:=1+ThrInd;
    WHILE iY<LastY DO
      IF e.SpecialMinXBoundary THEN
        m.b[0]:=-InvDt;
        m.c[0]:=InvDt;
        m.item02:=0;
        d[0]:=0;
      ELSE
        m.b[0]:=(e.MinXF[iY]-3*e.MinXD[iY]*Inv2dx)*InvDt;
        m.c[0]:=4*e.MinXD[iY]*Inv2dx*InvDt;
        m.item02:=-e.MinXD[iY]*Inv2dx*InvDt;
        d[0]:=e.MinXR[iY]*InvDt;
      END;
      FOR iX:=1 TO LastX-1 DO
        aa:=GetVal(e.a^,iX,iY);
        bb:=GetVal(e.b^,iX,iY);
        cc:=GetVal(e.c^,iX,iY);
        dd:=GetVal(e.d^,iX,iY);
        ee:=GetVal(e.e^,iX,iY);
        ff:=GetVal(e.f^,iX,iY);
        m.a[iX]:=aa*InvSqrDx - cc*Inv2dx;
        m.b[iX]:= (* ee *) - 2*aa*InvSqrDx - InvDt;
        m.c[iX]:=aa*InvSqrDx + cc*Inv2dx;
        d[iX]:=-old[iX,iY]*InvDt
               -bb*(old[iX,iY+1]+old[iX,iY-1]-2*old[iX,iY])*InvSqrDy
               -dd*(old[iX,iY+1]-old[iX,iY-1])*Inv2dy
               -ff
               -ee*old[iX,iY]
        (* Here ee belongs to the old layer to provide the 2nd order of accuracy.
           In old times this caused sometimes worse stability.
           Then transfer of ee to the new layer helped, but this is only the 1st
           order of accuracy. *)
      END;
      IF e.SpecialMaxXBoundary THEN
        m.b[LastX]:=InvDt;
        m.a[LastX]:=-InvDt;
        m.item20:=0;
        d[LastX]:=0;
      ELSE
        m.b[LastX]:=(e.MaxXF[iY]+3*e.MaxXD[iY]*Inv2dx)*InvDt;
        m.a[LastX]:=-4*e.MaxXD[iY]*Inv2dx*InvDt;
        m.item20:=e.MaxXD[iY]*Inv2dx*InvDt;
        d[LastX]:=e.MaxXR[iY]*InvDt;
      END;
      Matrix.SolveEx(m,d^);
      FOR iX:=0 TO LastX DO
        new[iX,iY-1]:=d[iX]; (* iY-1 in the middle = iY in other 2D fields *)
      END;
      INC(iY,Thr.Count);
    END;
  END Do;

PROCEDURE Step1 (e:POINTER TO Equ; old:Array2Ptr; InvDt:Float);
  BEGIN
    e.p1.InvDt:=InvDt;
    e.p1.old:=old;
    Thr.Exec(e.p1);
  END Step1;

PROCEDURE (VAR p:Step2Proc) Do ();
  VAR
    ThrInd,iX,iY,LastX,LastY:LONGINT;
    Inv2dx,Inv2dy,InvSqrDx,InvSqrDy:Float;
    aa,bb,cc,dd,ee,ff:Complex;
    e:POINTER TO Equ;
    m:Matrix.Matrix;
    d:ArrayPtr;
    InvDt:Float;
    old,new:Array2Ptr;
  BEGIN
    ASSERT(p.e#NIL);
    ASSERT(p.new#NIL);

    e:=p.e;
    ThrInd:=Thr.Index();
    m:=e.my[ThrInd];
    d:=e.yd[ThrInd];
    InvDt:=p.InvDt;
    old:=e.middle;
    new:=p.new;

    Inv2dx:=e.area.Inv2dx;
    Inv2dy:=e.area.Inv2dy;
    InvSqrDx:=e.area.InvSqrDx;
    InvSqrDy:=e.area.InvSqrDy;
    LastX:=e.area.LastX;
    LastY:=e.area.LastY;

    iX:=1+ThrInd;
    WHILE iX<LastX DO
      IF e.SpecialMinYBoundary THEN
        m.b[0]:=-InvDt;
        m.c[0]:=InvDt;
        m.item02:=0;
        d[0]:=0;
      ELSE
        m.b[0]:=(e.MinYF[iX]-3*e.MinYD[iX]*Inv2dy)*InvDt;
        m.c[0]:=4*e.MinYD[iX]*Inv2dy*InvDt;
        m.item02:=-e.MinYD[iX]*Inv2dy*InvDt;
        d[0]:=e.MinYR[iX]*InvDt;
      END;
      FOR iY:=1 TO LastY-1 DO
        aa:=GetVal(e.a^,iX,iY);
        bb:=GetVal(e.b^,iX,iY);
        cc:=GetVal(e.c^,iX,iY);
        dd:=GetVal(e.d^,iX,iY);
        ee:=GetVal(e.e^,iX,iY);
        ff:=GetVal(e.f^,iX,iY);
        m.a[iY]:=bb*InvSqrDy - dd*Inv2dy;
        m.b[iY]:= ee - 2*bb*InvSqrDy - InvDt;
        m.c[iY]:=bb*InvSqrDy + dd*Inv2dy;
        d[iY]:=-old[iX,iY-1]*InvDt      (* iY-1 in the middle = iY in other 2D fields *)
               -aa*(old[iX+1,iY-1]+old[iX-1,iY-1]-2*old[iX,iY-1])*InvSqrDx
               -cc*(old[iX+1,iY-1]-old[iX-1,iY-1])*Inv2dx
               -ff
      END;
      IF e.SpecialMaxYBoundary THEN
        m.b[LastY]:=InvDt;
        m.a[LastY]:=-InvDt;
        m.item20:=0;
        d[LastY]:=0;
      ELSE
        m.b[LastY]:=(e.MaxYF[iX]+3*e.MaxYD[iX]*Inv2dy)*InvDt;
        m.a[LastY]:=-4*e.MaxYD[iX]*Inv2dy*InvDt;
        m.item20:=e.MaxYD[iX]*Inv2dy*InvDt;
        d[LastY]:=e.MaxYR[iX]*InvDt;
      END;
      Matrix.SolveEx(m,d^);
      FOR iY:=0 TO LastY DO
        new[iX,iY]:=d[iY];
      END;
      INC(iX,Thr.Count);
    END;
  END Do;

PROCEDURE Step2 (e:POINTER TO Equ; new:Array2Ptr; InvDt:Float);
  VAR
    iY,LastX,LastY:LONGINT;
    Inv2dx:Float;
  BEGIN
    e.p2.InvDt:=InvDt;
    e.p2.new:=new;
    Thr.Exec(e.p2);

    Inv2dx:=e.area.Inv2dx;
    LastX:=e.area.LastX;
    LastY:=e.area.LastY;
    IF e.SpecialMinXBoundary THEN
      FOR iY:=0 TO LastY DO
        new[0,iY]:=new[1,iY]
      END;
    ELSE
      FOR iY:=0 TO LastY DO
        new[0,iY]:=  (  e.MinXR[iY]
                       -new[1,iY]*4*e.MinXD[iY]*Inv2dx
                       +new[2,iY]*e.MinXD[iY]*Inv2dx)
                    / (e.MinXF[iY]-3*e.MinXD[iY]*Inv2dx);
      END;
    END;
    IF e.SpecialMaxXBoundary THEN
      FOR iY:=0 TO LastY DO
        new[LastX,iY]:=new[LastX-1,iY]
      END;
    ELSE
      FOR iY:=0 TO LastY DO
        new[LastX,iY]:=(  e.MaxXR[iY]
                          +new[LastX-1,iY]*4*e.MaxXD[iY]*Inv2dx
                          -new[LastX-2,iY]*e.MaxXD[iY]*Inv2dx)
                       / (e.MaxXF[iY]+3*e.MaxXD[iY]*Inv2dx);
      END;
    END;
  END Step2;

PROCEDURE Step * (equ:POINTER TO Equ; OldField,NewField:Array2Ptr; dt:Float);
  (** Performs one step of solution with specified time step dt
      by "parallel-normal" method. The method has accuracy O(dx^2+dy^2+dt^2).
      OldField should contain initial conditions, the result is returned
      in NewField. OldField=NewField is accepted. *)
  VAR
    InvDt:Float;
  BEGIN
    ASSERT(equ#NIL);
    ASSERT(Initialized(equ^));
    ASSERT(OldField#NIL);
    ASSERT(NewField#NIL);
    ASSERT(LEN(OldField^)=equ.area.LastX+1);
    ASSERT(LEN(OldField^,1)=equ.area.LastY+1);
    ASSERT(LEN(NewField^)=equ.area.LastX+1);
    ASSERT(LEN(NewField^,1)=equ.area.LastY+1);

    InvDt:=2/dt;
    Step1(equ,OldField,InvDt);
    Step2(equ,NewField,InvDt);
  END Step;

PROCEDURE RHS * (equ-:Equ; field-:Array2; iX,iY:LONGINT): Complex;
  (** Returns right hand side of the equation (see the header of the module)
      with the specified field. At the boundary the procedure returns residual
      of the boundary conditions. *)
  VAR
    LastX,LastY:LONGINT;
  BEGIN
    ASSERT(Initialized(equ));

    LastX:=equ.area.LastX;
    LastY:=equ.area.LastY;
    ASSERT(LEN(field)=LastX+1);
    ASSERT(LEN(field,1)=LastY+1);
    ASSERT(iX>=0);
    ASSERT(iX<=LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=LastY);

    IF iX=0 THEN
      IF equ.SpecialMinXBoundary THEN
        RETURN field[1,iY]-field[0,iY]
      ELSE
        RETURN equ.MinXF[iY]*field[iX,iY]+Ar.DiffXInd(equ.area,field,iX,iY)*equ.MinXD[iY]-equ.MinXR[iY]
      END;
    ELSIF iX=LastX THEN
      IF equ.SpecialMaxXBoundary THEN
        RETURN field[LastX-1,iY]-field[LastX,iY]
      ELSE
        RETURN equ.MaxXF[iY]*field[iX,iY]+Ar.DiffXInd(equ.area,field,iX,iY)*equ.MaxXD[iY]-equ.MaxXR[iY]
      END;
    ELSIF iY=0 THEN
      IF equ.SpecialMinYBoundary THEN
        RETURN field[iX,1]-field[iX,0]
      ELSE
        RETURN equ.MinYF[iX]*field[iX,iY]+Ar.DiffYInd(equ.area,field,iX,iY)*equ.MinYD[iX]-equ.MinYR[iX]
      END
    ELSIF iY=LastY THEN
      IF equ.SpecialMaxYBoundary THEN
        RETURN field[iX,LastY-1]-field[iX,LastY]
      ELSE
        RETURN equ.MaxYF[iX]*field[iX,iY]+Ar.DiffYInd(equ.area,field,iX,iY)*equ.MaxYD[iX]-equ.MaxYR[iX]
      END;
    ELSE
      RETURN   GetVal(equ.a^,iX,iY)*Ar.Diff2XInd(equ.area,field,iX,iY)
             + GetVal(equ.b^,iX,iY)*Ar.Diff2YInd(equ.area,field,iX,iY)
             + GetVal(equ.c^,iX,iY)*Ar.DiffXInd(equ.area,field,iX,iY)
             + GetVal(equ.d^,iX,iY)*Ar.DiffYInd(equ.area,field,iX,iY)
             + GetVal(equ.e^,iX,iY)*field[iX,iY]
             + GetVal(equ.f^,iX,iY)
    END;
  END RHS;

PROCEDURE NormOfRHS * (equ-:Equ; field-:Array2): Float;
  (** Return mean norm of RHS wuth the specified field *)
  VAR
    iX,iY,LastX,LastY:LONGINT;
    sum:Float;
    x:Complex;
  BEGIN
    ASSERT(Initialized(equ));
    ASSERT(LEN(field)=equ.area.LastX+1);
    ASSERT(LEN(field,1)=equ.area.LastY+1);

    LastX:=equ.area.LastX;
    LastY:=equ.area.LastY;
    sum:=0;
    FOR iX:=0 TO LastX DO
      FOR iY:=0 TO LastY DO
        x:=RHS(equ,field,iX,iY);
        sum:=sum + SqrAbs(x)
      END;
    END;
    RETURN sqrt(sum/((LastX+1)*(LastY+1)))
  END NormOfRHS;

PROCEDURE MinMaxA (MinA,MaxA,h,a:Float; VAR min,max:Float);
  (* h is spatial step, a is spatial range *)
  VAR
    factor:Float;
  BEGIN
    ASSERT(MinA>0);
    ASSERT(MaxA>=MinA);

    factor:=4/sqr(h);
    max:=factor*MaxA;
    min:=factor*MinA*sqr(pi*h/(2*a))
  END MinMaxA;

PROCEDURE MinMaxC (MinC,MaxC,h,a:Float; VAR min,max:Float);
  BEGIN
    ASSERT(MinC>=0);
    ASSERT(MaxC>=MinC);

    max:=MaxC/h;
    min:=MinC*pi/a;
  END MinMaxC;

PROCEDURE FindMinMax (field-:Array2; VAR min,max:Float);
  VAR
    iX,iY:LONGINT;
    x:Float;

  BEGIN
    min:=Math.MaxFloat;
    max:=0;
    FOR iX:=0 TO LEN(field)-1 DO
      FOR iY:=0 TO LEN(field,1)-1 DO
        x:=Abs(field[iX,iY]);
        IF x#0 THEN
          IF min>x THEN
            min:=x
          END;
          IF max<x THEN
            max:=x
          END;
        END;
      END;
    END;
    IF min>max THEN (* all values are zeros *)
      ASSERT(max=0);
      min:=0;
    END;
  END FindMinMax;

PROCEDURE Eigenvalues * (equ-:Equ; VAR ev:Steps.Eigenvalues);
  (** Returns an estimation of the boundaries of the spectrum of the RHS
      of the equation *)
  VAR
    hx,hy,a,b,min,max:Float;
    MinA,MaxA,MinB,MaxB,MinC,MaxC,MinD,MaxD,MinE,MaxE:Float;
  BEGIN
    ASSERT(Initialized(equ));

    FindMinMax(equ.a^,MinA,MaxA);
    FindMinMax(equ.b^,MinB,MaxB);
    FindMinMax(equ.c^,MinC,MaxC);
    FindMinMax(equ.d^,MinD,MaxD);
    FindMinMax(equ.e^,MinE,MaxE);

    hx:=equ.area.dx;
    hy:=equ.area.dy;
    a:=equ.area.MaxX-equ.area.MinX;
    b:=equ.area.MaxY-equ.area.MinY;

    MinMaxA(MinA,MaxA,hx,a,min,max);
    ev.MinX:=min;
    ev.MaxX:=max;
    MinMaxA(MinB,MaxB,hy,b,min,max);
    ev.MinY:=min;
    ev.MaxY:=max;

    MinMaxC(MinC,MaxC,hx,a,min,max);
    ev.MinX:=ev.MinX+min;
    ev.MaxX:=ev.MaxX+max;
    MinMaxC(MinD,MaxD,hy,b,min,max);
    ev.MinY:=ev.MinY+min;
    ev.MaxY:=ev.MaxY+max;

    ev.MinX:=ev.MinX+MinE;
    ev.MaxX:=ev.MaxX+MaxE;
    ev.MinY:=ev.MinY+MinE;
    ev.MaxY:=ev.MaxY+MaxE;
  END Eigenvalues;

PROCEDURE AbsEigenvalue (A,C,E:Complex; a,h:Float; index:LONGINT): Float;
  VAR
    arg,sin,cos:Float;
  BEGIN
    arg:=pi*index*h/(2*a);
    LowMath.SinCos(arg,sin,cos);
    RETURN Abs( -4*A/sqr(h)*sqr(sin) + C/h*2*sin*cos + E);
  END AbsEigenvalue;

PROCEDURE EigenvaluesExacter * (equ-:Equ; VAR ev:Steps.Eigenvalues);
  (** The same as Eigenvalues, but exacter, but significantly slower *)
  VAR
    iX,iY,LastX,LastY,i:LONGINT;
    hx,hy,a,b:Float;
    l,MinX,MaxX,MinY,MaxY:Float;
  BEGIN
    ASSERT(Initialized(equ));

    hx:=equ.area.dx;
    hy:=equ.area.dy;
    a:=equ.area.MaxX-equ.area.MinX;
    b:=equ.area.MaxY-equ.area.MinY;
    LastX:=equ.area.LastX;
    LastY:=equ.area.LastY;

    MinX:=Math.MaxFloat;
    MaxX:=0;
    MinY:=Math.MaxFloat;
    MaxY:=0;
    FOR iX:=0 TO LEN(equ.a^)-1 DO
      FOR iY:=0 TO LEN(equ.a^,1)-1 DO
        FOR i:=1 TO LastX-1 DO
          l:=AbsEigenvalue(equ.a[iX,iY],equ.c[iX,iY],equ.e[iX,iY],a,hx,i);
          IF l#0 THEN
            IF MinX>l THEN
              MinX:=l;
            END;
            IF MaxX<l THEN
              MaxX:=l;
            END;
          END;
        END;
        FOR i:=1 TO LastY-1 DO
          l:=AbsEigenvalue(equ.b[iX,iY],equ.d[iX,iY],equ.e[iX,iY],b,hy,i);
          IF l#0 THEN
            IF MinY>l THEN
              MinY:=l;
            END;
            IF MaxY<l THEN
              MaxY:=l;
            END;
          END;
        END;
      END;
    END;

    (* not all the eigenvalues are zeros *)
    ASSERT(MinX<MaxX);
    ASSERT(MinY<MaxY);

    ev.MinX:=MinX;
    ev.MaxX:=MaxX;
    ev.MinY:=MinY;
    ev.MaxY:=MaxY;
  END EigenvaluesExacter;

PROCEDURE SaveCoeffs * (equ-:Equ; prefix-,suffix-:ARRAY OF CHAR);
  (** Saves all the coefficients of the equ to files. The file names are
      prefix+name+suffix, where name is 'A','B','C','D','E','F', respectively.
      The equation should be already initialized. *)
  VAR
    a:Area;

  PROCEDURE Save (field-:Array2; name-:ARRAY OF CHAR);
    BEGIN
      Ar.SaveArray(a,field,DStr.Concat3(prefix,name,suffix)^,1);
    END Save;

  BEGIN
    ASSERT(Initialized(equ));

    Ar.Reduced(equ.area,a);
    Save(equ.a^,'A');
    Save(equ.b^,'B');
    Save(equ.c^,'C');
    Save(equ.d^,'D');
    Save(equ.e^,'E');
    Save(equ.f^,'F');
  END SaveCoeffs;

PROCEDURE AddStabilizationByDiffusion (VAR c:Complex; b:Complex; h,parameter:Float);
  (* c is diffusion coefficient, b is migration coefficient *)
  VAR
    AbsB,AbsC,NewAbsC,threshold:Float;
  BEGIN
    ASSERT(h>0);

    IF c=0 THEN (* sometimes all the coeffs of the equation are zero *)
      ASSERT(b=0);
      RETURN
    END;

    AbsB:=Abs(b);
    AbsC:=Abs(c);
    threshold:=h*AbsB/2;
    IF AbsC<threshold THEN
      NewAbsC:=AbsC + threshold*parameter;
      IF NewAbsC>threshold THEN
        NewAbsC:=threshold;
      END;
      c:=c/AbsC*NewAbsC;
    END;
  END AddStabilizationByDiffusion;

PROCEDURE StabilizeByDiffusion * (VAR equ:Equ; parameter:Float);
  (** Restricts equ.a and equ.b from low for better stability of the equation.
      Introduces an error of O(dx,dy).
      If parameter=0, no changes happen.
      If parameter=1, full stabilization changes happen.
      Based on inconsistent stabilization of COMSOL. *)
  VAR
    iX,iY:LONGINT;
  BEGIN
    ASSERT(Initialized(equ));

    FOR iX:=1 TO equ.area.LastX-1 DO
      FOR iY:=1 TO equ.area.LastY-1 DO
        AddStabilizationByDiffusion(equ.a[iX-1,iY-1],equ.c[iX-1,iY-1],
                                    equ.area.dx,parameter);
        AddStabilizationByDiffusion(equ.b[iX-1,iY-1],equ.d[iX-1,iY-1],
                                    equ.area.dy,parameter);
      END;
    END;
  END StabilizeByDiffusion;

PROCEDURE AddStabilizationByMigration (c:Complex; VAR b:Complex; h,parameter:Float);
  (* c is diffusion coefficient, b is migration coefficient *)
  VAR
    AbsB,AbsC,NewAbsB,threshold:Float;
  BEGIN
    ASSERT(h>0);

    IF c=0 THEN (* sometimes all the coeffs of the equation are zero *)
      ASSERT(b=0);
      RETURN
    END;

    AbsB:=Abs(b);
    AbsC:=Abs(c);
    threshold:=2*AbsC/h;
    IF AbsB>threshold THEN
      NewAbsB:=AbsB + parameter*(threshold-AbsB);
      b:=b/AbsB*NewAbsB;
    END;
  END AddStabilizationByMigration;

PROCEDURE StabilizeByMigration * (VAR equ:Equ; parameter:Float);
  (** Restricts equ.c and equ.d from up for better stability of the equation.
      Introduces an error of O(dx,dy).
      If parameter=0, no changes happen.
      If parameter=1, full stabilization changes happen.
      Based on inconsistent stabilization of COMSOL. *)
  VAR
    iX,iY:LONGINT;
  BEGIN
    ASSERT(Initialized(equ));

    FOR iX:=1 TO equ.area.LastX-1 DO
      FOR iY:=1 TO equ.area.LastY-1 DO
        AddStabilizationByMigration(equ.a[iX-1,iY-1],equ.c[iX-1,iY-1],
                                    equ.area.dx,parameter);
        AddStabilizationByMigration(equ.b[iX-1,iY-1],equ.d[iX-1,iY-1],
                                    equ.area.dy,parameter);
      END;
    END;
  END StabilizeByMigration;

END mParabolicEqu2DComplex.