(** This module provides square matrices and solution of the corresponding
    system of linear eqiations.

    Before solution of the system of linear equations, the matrix should be
    converted. All the solutions are possible only after the conversion.
    Setting and getting of the matrix elements and algebraic operations are
    only possible before the conversion
    
    Last change: 03.06.2016 *)

<* O2EXTENSIONS + *>
MODULE mSquareMatrix;

IMPORT
  Math:=mMath, Arr:=mArrays;

TYPE
  Float * = Math.Float;
  Array = Math.Array;
  ArrayPtr = Math.ArrayPtr;
  Array2Ptr = Math.Array2Ptr;

  Matrix * = RECORD
    rank - : LONGINT; (** Rank of the matrix *)
    field : Array2Ptr;
    ind: POINTER TO ARRAY OF LONGINT;
    AdditionalLines:LONGINT;
    NegativeDeterminant:BOOLEAN;
  END;
  (** The matrix *)

PROCEDURE InitWithoutField (VAR m:Matrix; rank:LONGINT);
  BEGIN
    m.rank:=rank;
    m.ind:=NIL;
    m.NegativeDeterminant:=FALSE;
  END InitWithoutField;

PROCEDURE Init * (VAR m:Matrix; rank:LONGINT);
  (** Initializes the matrix with the specified rank *)
  BEGIN
    ASSERT(rank>0);

    IF (m.field=NIL) OR (LEN(m.field^)#rank) THEN
      NEW(m.field,rank,rank);
    END;
    InitWithoutField(m,rank);
  END Init;

PROCEDURE Initialized * (m-:Matrix): BOOLEAN;
  (** Returns if the matrix is initialized *)
  BEGIN
    RETURN m.field#NIL
  END Initialized;

PROCEDURE Converted * (m-:Matrix): BOOLEAN;
  (** Returns if the matrix is converted for solution of systems of linear
      equations *)
  BEGIN
    RETURN m.ind#NIL
  END Converted;

PROCEDURE GetItem * (m-:Matrix; i,j:LONGINT): Float;
  (** Returns (i,j)-th element of the matrix.
      The matrix must not be converted. *)
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(i>=0);
    ASSERT(i<m.rank);
    ASSERT(j>=0);
    ASSERT(j<m.rank);

    RETURN m.field[i,j]
  END GetItem;

PROCEDURE SetItem * (m-:Matrix; i,j:LONGINT; val:Float);
  (** Sets (i,j)-th element of the matrix.
      The matrix must not be converted. *)
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(i>=0);
    ASSERT(i<m.rank);
    ASSERT(j>=0);
    ASSERT(j<m.rank);

    m.field[i,j]:=val
  END SetItem;

PROCEDURE Field * (m-:Matrix): Array2Ptr;
  (** Returns a pointer to field of elements: result[i,j] contains
      [i,j]-th element of the matrix. Writeble.
      The matrix should not be converted. *)
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));

    RETURN m.field
  END Field;

PROCEDURE Assign * (VAR dest:Matrix; source-:Matrix);
  (** dest:=source *)
  VAR
    i:LONGINT;
  BEGIN
    IF source.field#dest.field THEN
      (* source.field=dest.field means that source=dest or both are not initialized *)
      IF source.field=NIL THEN
        dest.field:=NIL
      ELSE
        IF (dest.field=NIL) OR (LEN(dest.field^)#LEN(source.field^)) THEN
          dest.field:=Arr.Copy2(source.field^)
        ELSE
          Arr.Assign2(dest.field^,source.field^);
        END;
      END;
      IF source.ind=NIL THEN
        dest.ind:=NIL
      ELSE
        IF (dest.ind=NIL) OR (LEN(dest.ind^)#LEN(source.ind^)) THEN
          NEW(dest.ind,LEN(source.ind^))
        END;
        FOR i:=0 TO LEN(source.ind^)-1 DO
          dest.ind[i]:=source.ind[i]
        END;
      END;
      dest.rank:=source.rank;
      dest.AdditionalLines:=source.AdditionalLines;
      dest.NegativeDeterminant:=source.NegativeDeterminant;
    END;
  END Assign;

PROCEDURE Add * (VAR m:Matrix; addition-:Matrix);
  (** m:=m+addition. The matrices must not be converted. *)
  VAR
    i,j,last:LONGINT;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(Initialized(addition));
    ASSERT(~Converted(addition));
    ASSERT(m.rank=addition.rank);

    last:=m.rank-1;
    FOR i:=0 TO last DO
      FOR j:=0 TO last DO
        m.field[i,j]:=m.field[i,j]+addition.field[i,j]
      END;
    END;
  END Add;

PROCEDURE MulVector * (m-:Matrix; b-:Array): ArrayPtr;
  (** Returns m*b. The matrix must not be converted. *)
  VAR
    res:ArrayPtr;
    i:LONGINT;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(m.rank=LEN(b));

    NEW(res,m.rank);
    FOR i:=0 TO m.rank-1 DO
      res[i]:=Arr.Prod(m.field[i],b);
    END;
    RETURN res;
  END MulVector;

PROCEDURE TMulVector * (m-:Matrix; b-:Array): ArrayPtr;
  (** Returns Transposed(m)*b. The matrix must not be converted. *)
  VAR
    res:ArrayPtr;
    i,k,last:LONGINT;
    sum:Float;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(m.rank=LEN(b));

    last:=m.rank-1;
    NEW(res,m.rank);
    FOR i:=0 TO last DO
      sum:=0;
      FOR k:=0 TO last DO
        sum:=sum+m.field[k,i]*b[k];
      END;
      res[i]:=sum;
    END;
    RETURN res;
  END TMulVector;

PROCEDURE AtA * (m-:Matrix; VAR result:Matrix);
  (** result:= Transpose(m)*m. m must not be converted. *)
  VAR
    i,j,k,last:LONGINT;
    sum:Float;
    new:Matrix;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));

    IF ~Initialized(result) OR (result.rank#m.rank) OR (result.field=m.field) THEN
      Init(new,m.rank)
    ELSE
      InitWithoutField(new,m.rank);
      new.field:=result.field;
    END;

    last:=m.rank-1;
    FOR i:=0 TO last DO
      FOR j:=0 TO i DO
        sum:=0;
        FOR k:=0 TO last DO
          sum:=sum+m.field[k,i]*m.field[k,j]
        END;
        new.field[i,j]:=sum;
        new.field[j,i]:=sum;
      END;
    END;
    result:=new
  END AtA;

PROCEDURE AddWeakRegularization * (VAR m:Matrix; alpha:Float);
  (** Adds weak regularization term alpha to m. m must not be converted. *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));

    FOR i:=0 TO m.rank-1 DO
      m.field[i,i]:=m.field[i,i] + alpha;
    END;
  END AddWeakRegularization;

PROCEDURE AddStrongRegularization * (VAR m:Matrix; alpha:Float);
  (** Adds strong regularization term alpha to m. m must not be converted. *)
  VAR
    i,last:LONGINT;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));

    last:=m.rank-1;
    FOR i:=0 TO last DO
      m.field[i,i]:=m.field[i,i] + 3*alpha;
    END;
    FOR i:=0 TO last DO
      m.field[i,i+1]:=m.field[i,i+1] - alpha;
    END;
    FOR i:=1 TO last DO
      m.field[i,i-1]:=m.field[i,i-1] - alpha;
    END;
  END AddStrongRegularization;

PROCEDURE Min (i,j:LONGINT): LONGINT;
  BEGIN
    IF i<j THEN
      RETURN i
    ELSE
      RETURN j
    END;
  END Min;

PROCEDURE SwapFloat (VAR x,y:Float);
  VAR
    tmp:Float;
  BEGIN
    tmp:=x;
    x:=y;
    y:=tmp;
  END SwapFloat;

PROCEDURE SwapLines (VAR line1,line2:Array; StartIndex:LONGINT);
  (* line1 <-> line2 from StartIndex to the end of the lines *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(LEN(line1)=LEN(line2));
    ASSERT(StartIndex>=0);
    ASSERT(StartIndex<LEN(line1));

    FOR i:=StartIndex TO LEN(line1)-1 DO
      SwapFloat(line1[i],line2[i])
    END;
  END SwapLines;

PROCEDURE SubLine (VAR line1,line2:Array; StartIndex:LONGINT; factor:Float);
  (* line1:=line1-factor*line2 from StartIndex to the end of the lines *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(LEN(line1)=LEN(line2));
    ASSERT(StartIndex>=0);
    ASSERT(StartIndex<LEN(line1));

    FOR i:=StartIndex TO LEN(line1)-1 DO
      line1[i]:=line1[i]-factor*line2[i]
    END;
  END SubLine;

PROCEDURE Convert * (VAR m:Matrix; AdditionalLines:LONGINT);
  (** Converts the matrix for solution of system of linear equations.
      The initial matrix is assumed to be up-triangle plus AdditionalLines
      additional lines in the low part.
      For example, a matrix, which is sum of a up-triangle matrix and
      three-diagonal band matrix, corresponds to AdditionalLines=1.
      If AdditionalLines=rank-1, then this is general matrix. *)
  VAR
    i,j,last,MaxInd,IndOfMax:LONGINT;
    c,max,x:Float;
    a:Array2Ptr;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(~Converted(m));
    ASSERT(AdditionalLines>=0);
    ASSERT(AdditionalLines<m.rank);

    last:=m.rank-1;
    a:=m.field;
    NEW(m.ind,m.rank);
    FOR i:=0 TO last DO
      m.ind[i]:=i
    END;

    FOR j:=0 TO last-1 DO
      (* Selecting line with maximal ABS(a[i,j]) *)
      MaxInd:=Min(j+AdditionalLines,last);
      IndOfMax:=j;
      max:=ABS(a[j,j]);
      FOR i:=j+1 TO MaxInd DO
        x:=ABS(a[i,j]);
        IF x>max THEN
          max:=x;
          IndOfMax:=i
        END;
      END;
      IF IndOfMax#j THEN
        SwapLines(a[j],a[IndOfMax],j);
        m.ind[j]:=IndOfMax;
        IF ODD(j-IndOfMax) THEN
          m.NegativeDeterminant:=~m.NegativeDeterminant;
        END;
      END;

      (* Eliminating elements of lower half of the matrix *)
      FOR i:=j+1 TO MaxInd DO
        IF a[i,j]#0 THEN
          c:=a[i,j]/a[j,j];
          SubLine(a[i],a[j],j+1,c);
          a[i,j]:=c
        END;
      END;
    END;

    m.AdditionalLines:=AdditionalLines;
  END Convert;

PROCEDURE ProdLine (VAR line1,line2:Array; StartIndex:LONGINT): Float;
  (* Returns line1*line2 from StartIndex to the end of the lines *)
  VAR
    i:LONGINT;
    sum:Float;
  BEGIN
    ASSERT(LEN(line1)=LEN(line2));
    ASSERT(StartIndex>=0);
    ASSERT(StartIndex<LEN(line1));

    sum:=0;
    FOR i:=StartIndex TO LEN(line1)-1 DO
      sum:=sum+line1[i]*line2[i]
    END;
    RETURN sum
  END ProdLine;

PROCEDURE Solve * (m-:Matrix; VAR b:Array);
  (** Solves system of equations m*x=b. m should be preliminary converted.
      The result is returned in b.
      The matrix m is not changed by this procedure. *)
  VAR
    c:Float;
    i,j,last,AdditionalLines:LONGINT;
    a:Array2Ptr;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(Converted(m));
    ASSERT(LEN(b)=m.rank);

    a:=m.field;
    last:=m.rank-1;
    AdditionalLines:=m.AdditionalLines;
    FOR j:=0 TO last-1 DO
      i:=m.ind[j];
      IF i#j THEN
        SwapFloat(b[j],b[i]);
      END;
      FOR i:=j+1 TO Min(j+AdditionalLines,last) DO
        c:=a[i,j];
        b[i]:=b[i]-c*b[j]
      END;
    END;

    b[last]:=b[last]/a[last,last];
    FOR i:=last-1 TO 0 BY -1 DO
      b[i]:=(b[i]-ProdLine(a[i],b,i+1))/a[i,i];
    END;
  END Solve;

PROCEDURE ConvertAndSolve * (VAR m:Matrix; VAR b:ARRAY OF Float; AdditionalLines:LONGINT);
  (** equivalent to Convert(...); Solve(...); *)
  BEGIN
    Convert(m,AdditionalLines);
    Solve(m,b);
  END ConvertAndSolve;

PROCEDURE Determinant * (m-:Matrix): Float;
  (** Returns determinant of m. m should be preliminary converted. *)
  VAR
    i:LONGINT;
    result:Float;
  BEGIN
    ASSERT(Initialized(m));
    ASSERT(Converted(m));

    result:=1;
    FOR i:=0 TO m.rank-1 DO
      result:=result*m.field[i,i];
    END;
    IF m.NegativeDeterminant THEN
      result:=-result
    END;
    RETURN result;
  END Determinant;

END mSquareMatrix.