(** This module contains routines for direct output to text console

    Last change: 26.03.2015 *)

<* O2EXTENSIONS + *>
<* O2ADDKWD + *>
MODULE mhostDisplay;

IMPORT
  Win:=Windows, SYS:=SYSTEM, DStr:=msysDynStr,
  IntConv:=msysLongLongIntConv, RealConv:=msysLongLongRealConv;

VAR
  HOut:Win.HANDLE;

TYPE
  Color * = (Black,Blue,Green,Cyan,Red,Magenta,Brown,LightGray,DarkGray,LightBlue,
            LightGreen,LightCyan,LightRed,LightMagenta,Yellow,White);

  COORD * = RECORD
    x*,y*:LONGINT;
  END;

CONST
  TextForegroundSet=Win.FOREGROUND_BLUE+Win.FOREGROUND_GREEN+Win.FOREGROUND_RED+Win.FOREGROUND_INTENSITY;
  TextBackgroundSet=Win.BACKGROUND_BLUE+Win.BACKGROUND_GREEN+Win.BACKGROUND_RED+Win.BACKGROUND_INTENSITY;

PROCEDURE GetInfo (VAR info:Win.CONSOLE_SCREEN_BUFFER_INFO);
  BEGIN
    Win.GetConsoleScreenBufferInfo(HOut,info);
  END GetInfo;

PROCEDURE Attr (x:Color): Win.CHAR_ATTRIBUTES_SET;
  BEGIN
    RETURN SYS.VAL(Win.CHAR_ATTRIBUTES_SET,x)
  END Attr;

PROCEDURE TextColor * (color:Color);
  (** Sets color of text output *)
  VAR
    info:Win.CONSOLE_SCREEN_BUFFER_INFO;
  BEGIN
    GetInfo(info);
    Win.SetConsoleTextAttribute(HOut,(info.wAttributes*TextBackgroundSet) + (Attr(color)*TextForegroundSet))
  END TextColor;

PROCEDURE BkColor * (color:Color);
  (** Sets background of text output *)
  VAR
    info:Win.CONSOLE_SCREEN_BUFFER_INFO;
  BEGIN
    GetInfo(info);
    Win.SetConsoleTextAttribute(HOut,(info.wAttributes*TextForegroundSet) + (SYS.SHIFT(Attr(color),4)*TextBackgroundSet))
  END BkColor;

PROCEDURE GetCursorPos * (): COORD;
  (** Returns position of the text cursor *)
  VAR
    info:Win.CONSOLE_SCREEN_BUFFER_INFO;
    result:COORD;
  BEGIN
    GetInfo(info);
    result.x:=info.dwCursorPosition.X;
    result.y:=info.dwCursorPosition.Y;
    RETURN result
  END GetCursorPos;

PROCEDURE SetCursorPos * (Coord:COORD);
  (** Sets position of the text cursor *)
  VAR
    c:Win.COORD;
  BEGIN
    c.X:=SHORT(Coord.x);
    c.Y:=SHORT(Coord.y);
    Win.SetConsoleCursorPosition(HOut,c);
  END SetCursorPos;

PROCEDURE GoToXY * (x,y:LONGINT);
  (** Sets position of the text cursor *)
  VAR
    coord:Win.COORD;
  BEGIN
    coord.X:=SHORT(x);
    coord.Y:=SHORT(y);
    Win.SetConsoleCursorPosition(HOut,coord);
  END GoToXY;

PROCEDURE Size * (): COORD;
  (** Resurns size of the text console *)
  VAR
    info:Win.CONSOLE_SCREEN_BUFFER_INFO;
    result:COORD;
  BEGIN
    GetInfo(info);
    result.x:=info.dwSize.X;
    result.y:=info.dwSize.Y;
    RETURN result;
  END Size;

<* PUSH *>
<* CHECKRANGE - *>
(* Or else there is an error if the standard output is redirected to file *)

PROCEDURE Clear * ();
  (** Clears text console *)
  VAR
    width:INTEGER;
    res:Win.DWORD;
    info:Win.CONSOLE_SCREEN_BUFFER_INFO;
  BEGIN
    GetInfo(info);
    width:=info.dwSize.X;
    info.dwSize.X:=0;
    WHILE info.dwSize.Y>0 DO
      DEC(info.dwSize.Y);
      Win.FillConsoleOutputCharacter(HOut,' ',width,info.dwSize,res);
      Win.FillConsoleOutputAttribute(HOut,info.wAttributes,width,info.dwSize,res)
    END;
    GoToXY(0,0);
  END Clear;

<* POP *>

PROCEDURE Read * (x,y,length:LONGINT): DStr.String;
  (** Reads length characters from the screen starting from coordinates (x,y),
      and returns the result. *)
  VAR
    c:Win.COORD;
    i:SYS.CARD32;
    result:DStr.String;
  BEGIN
    ASSERT(length>=0);

    result:=DStr.New(length);
    c.X:=SHORT(x);
    c.Y:=SHORT(y);
    Win.ReadConsoleOutputCharacter(HOut,result^,length,c,i);
    IF VAL(LONGINT,i)#length THEN
      RETURN DStr.Copy(result^,0,i)
    ELSE
      RETURN result
    END;
  END Read;

PROCEDURE Char * (x,y:LONGINT; c:CHAR; count:LONGINT);
  (** Writes c count times starting at coordinates (x,y).
      This procedure doesn't change the cursor position. *)
  VAR
    crd:Win.COORD;
    res:Win.DWORD;
  BEGIN
    crd.X:=SHORT(x);
    crd.Y:=SHORT(y);
    Win.FillConsoleOutputCharacter(HOut,c,count,crd,res);
  END Char;

PROCEDURE String * (x,y:LONGINT; s-:ARRAY OF CHAR; MinLength:LONGINT);
  (** Writes a null-terminaed string s starting at coordinates (x,y).
      MinLength is minimal displayed length; leading spaces are added if
      necessary. This procedure doesn't change the cursor position. *)
  VAR
    len:LONGINT;
    c:Win.COORD;
    res:Win.DWORD;
  BEGIN
    len:=0;
    WHILE (len<LEN(s)) & (s[len]#0X) DO
      INC(len);
    END;

    c.Y:=SHORT(y);
    IF len<MinLength THEN
      Char(x,y,' ',MinLength-len);
      c.X:=SHORT(x+MinLength-len);
    ELSE
      c.X:=SHORT(x);
    END;
    Win.WriteConsoleOutputCharacter(HOut,s,len,c,res);
  END String;

PROCEDURE Int * (x,y:LONGINT; value:LONGLONGINT; MinLength:LONGINT);
  (** Writes integer value starting at coordinates (x,y). MinLength is minimal
      displayed length; leading spaces are added if necessary.
      This procedure doesn't change the cursor position. *)
  VAR
    buf:IntConv.String;
    len:LONGINT;
  BEGIN
    len:=IntConv.ToString(value,buf);
    buf[len]:=0X;
    String(x,y,buf,MinLength);
  END Int;

PROCEDURE Real * (x,y:LONGINT; value:LONGLONGREAL; MinLength:LONGINT);
  (** Writes numerical value at coordinates (x,y). MinLength is minimal
      displayed length; leading spaces are added if necessary.
      This procedure doesn't change the cursor position. *)
  VAR
    buf:RealConv.String;
    len:LONGINT;
  BEGIN
    len:=RealConv.ToString(value,DStr.DefaultDigits,buf);
    buf[len]:=0X;
    String(x,y,buf,MinLength);
  END Real;

BEGIN
  HOut:=Win.CreateFile('CONOUT$',Win.GENERIC_READ + Win.GENERIC_WRITE,Win.FILE_SHARE_WRITE,NIL,
                      Win.OPEN_EXISTING,SYS.VAL(Win.FILE_BIT_SET,{}),NIL);
FINALLY
  Win.CloseHandle(HOut);
END mhostDisplay.