(** This module provides files search.

    Last changed: 27.03.2015 *)

<* O2EXTENSIONS + *>
MODULE mhostSearchFile;

IMPORT
  Win:=Windows, DStr:=msysDynStr, SYS:=SYSTEM, FileName, SU:=msysUtils,
  COMPILER;

TYPE
  String = DStr.String;

CONST
  ModuleName = COMPILER.EQUATION('module');

  (** File Attributes *)
  ReadOnly * = Win.FILE_ATTRIBUTE__READONLY;
  Hidden * = Win.FILE_ATTRIBUTE__HIDDEN;
  System * = Win.FILE_ATTRIBUTE__SYSTEM;
  Directory * = Win.FILE_ATTRIBUTE__DIRECTORY;
  Archive * = Win.FILE_ATTRIBUTE__ARCHIVE;
  Normal * = Win.FILE_ATTRIBUTE__NORMAL;
  Temporary * = Win.FILE_ATTRIBUTE__TEMPORARY;
  Compressed * = Win.FILE_ATTRIBUTE__COMPRESSED;
  Offline * = Win.FILE_ATTRIBUTE__OFFLINE;
  NotContentIndexed * = Win.F_13;
  Encrypted * = Win.F_14;
  ReparsePoint * = Win.F_10;
  SparseFile * = Win.F_9;

  EmptySet = Win.FILE_ATTRIBUTE_READONLY-Win.FILE_ATTRIBUTE_READONLY;

TYPE
  Attributes * = Win.FILE_ATTRIBUTE_SET; (** Set of file attributes *)

  Info * = RECORD
    attr - : Attributes; (** File attributes *)
    handle:Win.HANDLE;
    data:Win.WIN32_FIND_DATA;
    dir:String;
  END;
  (** Information on a found files. *)

TYPE
  Format = RECORD
    dirPos,dirLen,namePos,nameLen,extPos,extLen:LONGINT;
  END;

PROCEDURE GetFormat (s-:ARRAY OF CHAR; VAR f:Format);
  VAR
    i:LONGINT;
    LastDot,LastSlash,len:LONGINT;
  BEGIN
    i:=0;
    LastDot:=-1;
    LastSlash:=-1;
    WHILE (i<LEN(s)) & (s[i]#0X) DO
      CASE s[i] OF
        '.':
          LastDot:=i;
      | '\':
          LastSlash:=i
      ELSE
      END;
      INC(i)
    END;
    len:=i;

    IF LastDot<=LastSlash THEN
      LastDot:=len
    END;

    f.dirPos:=0;
    f.dirLen:=LastSlash;
    IF f.dirLen<0 THEN
      f.dirLen:=0
    END;

    IF (len>0) & (s[len-1]='.') THEN
      IF (len>1) & (s[len-2]='.') THEN
        f.nameLen:=2
      ELSE
        f.nameLen:=1;
      END;
      f.namePos:=len-f.nameLen;
      f.extPos:=len;
      f.extLen:=0;
    ELSE
      f.namePos:=LastSlash+1;
      f.nameLen:=LastDot-LastSlash-1;

      f.extPos:=LastDot+1;
      f.extLen:=len-LastDot-1;
      IF f.extLen<0 THEN
        f.extLen:=0
      END;
    END;

  END GetFormat;

PROCEDURE GetFormatFromInfo (info-:Info; VAR f:Format);
  BEGIN
    GetFormat(info.data.cFileName,f);
  END GetFormatFromInfo;

PROCEDURE ConcatWithSlash (s1-,s2-:ARRAY OF CHAR): String;
  VAR
    len:LONGINT;
  BEGIN
    len:=DStr.Length(s1);
    IF (len=0) OR (s1[len-1]='\') THEN
      RETURN DStr.Concat2(s1,s2)
    ELSE
      RETURN DStr.Concat3(s1,'\',s2);
    END
  END ConcatWithSlash;

PROCEDURE PathNameExt * (info-:Info): String;
  (** Returns path, name and extension of the file *)
  BEGIN
    ASSERT(info.dir#NIL);

    RETURN ConcatWithSlash(info.dir^,DStr.Till0(info.data.cFileName)^)
  END PathNameExt;

PROCEDURE PathName * (info-:Info): String;
  (** Returns path and name the file without extension *)
  VAR
    f:Format;
  BEGIN
    ASSERT(info.dir#NIL);

    GetFormatFromInfo(info,f);
    RETURN ConcatWithSlash(info.dir^,DStr.Copy(info.data.cFileName,f.namePos,f.nameLen)^)
  END PathName;

PROCEDURE Path * (info-:Info): String;
  (** Returns path of the file *)
  BEGIN
    ASSERT(info.dir#NIL);

    RETURN info.dir
  END Path;

PROCEDURE NameExt * (info-:Info): String;
  (** Returns name and extension of the file without directory *)
  BEGIN
    ASSERT(info.dir#NIL);

    RETURN DStr.Till0(info.data.cFileName);
  END NameExt;

PROCEDURE Name * (info-:Info): String;
  (** Returns name of the file without directory and extension *)
  VAR
    f:Format;
  BEGIN
    ASSERT(info.dir#NIL);

    GetFormatFromInfo(info,f);
    RETURN DStr.Copy(info.data.cFileName,f.namePos,f.nameLen);
  END Name;

PROCEDURE Ext * (info-:Info): String;
  (** Returns extension of the file *)
  VAR
    f:Format;
  BEGIN
    ASSERT(info.dir#NIL);

    GetFormatFromInfo(info,f);
    RETURN DStr.Copy(info.data.cFileName,f.extPos,f.extLen);
  END Ext;

PROCEDURE Size * (info-:Info): LONGLONGINT;
  (** Size of the found file *)
  VAR
    result:LONGLONGCARD;
  BEGIN
    ASSERT(info.dir#NIL);
    result:=info.data.nFileSizeHigh;
    result:=result*10000H*10000H+info.data.nFileSizeLow;
    RETURN SYS.VAL(LONGLONGINT,result);
  END Size;

PROCEDURE GetDir (filename-:ARRAY OF CHAR): String;
  VAR
    f:FileName.Format;
  BEGIN
    FileName.GetFormat(filename,f);
    IF f.ok THEN
      RETURN DStr.Copy(filename,f.dirPos,f.dirLen);
    ELSE
      RETURN DStr.Empty()
    END;
  END GetDir;

PROCEDURE First * (FileMask-:ARRAY OF CHAR; VAR info:Info): BOOLEAN;
  (** Searches the first file matching the mask.
      Returns TRUE if the file is found.
      When not more needed, the info should be closed by Close. *)
  VAR
    b:BOOLEAN;
  BEGIN
    ASSERT(info.dir=NIL);
    info.handle:=Win.FindFirstFile(FileMask,info.data);
    b:=(info.handle#Win.INVALID_HANDLE_VALUE);
    IF b THEN
      info.dir:=GetDir(FileMask);
      info.attr:=info.data.dwFileAttributes
    END;
    RETURN b;
  END First;

PROCEDURE Next * (VAR info:Info): BOOLEAN;
  (** Searches the next file matching the mask, which was defined by First.
      Returns TRUE if the file is found. *)
  VAR
    b:BOOLEAN;
  BEGIN
    ASSERT(info.dir#NIL);
    b:=Win.FindNextFile(info.handle,info.data);
    IF b THEN
      info.attr:=info.data.dwFileAttributes
    END;
    RETURN b
  END Next;

PROCEDURE Close * (VAR info:Info);
  (** Closes the file info. *)
  BEGIN
    Win.FindClose(info.handle);
    info.attr:=EmptySet;
    info.dir:=NIL;
  END Close;

TYPE
  Process * = RECORD
  END;
  (** A general procedure, which accepts the found files.
      Should return if the search should be continued *)

PROCEDURE (VAR p:Process) Do * (info-:Info): BOOLEAN;
  BEGIN
    SU.RAISEError(ModuleName+'.Process.Do: abstract method.');
    RETURN info.dir#NIL
  END Do;

PROCEDURE Shallow * (FileMask-:ARRAY OF CHAR; VAR p:Process): BOOLEAN;
  (** Searchs all files matching the mask and processes them
      by p one by one. Returns if all files are processed. *)
  VAR
    FoundNext,continue:BOOLEAN;
    info:Info;
  BEGIN
    continue:=TRUE;
    FoundNext:=First(FileMask,info);
    WHILE FoundNext DO
      continue:=p.Do(info);
      FoundNext:=continue & Next(info)
    END;
    Close(info);

    RETURN continue
  END Shallow;

TYPE
  ProcessDir = RECORD (Process)
    ProcessFile:POINTER TO Process;
    FileMask:String;
  END;

PROCEDURE LookAtDir (directory-:ARRAY OF CHAR; VAR dir:ProcessDir): BOOLEAN;
  BEGIN
    IF directory='' THEN
      RETURN Shallow('*.*',dir) & Shallow(dir.FileMask^,dir.ProcessFile^)
    ELSE
      RETURN   Shallow(DStr.Concat2(directory,'\*.*')^,dir)
             & Shallow(DStr.Concat3(directory,'\',dir.FileMask^)^,dir.ProcessFile^);
    END;
  END LookAtDir;

PROCEDURE (VAR p:ProcessDir) Do (info-:Info): BOOLEAN;
  VAR
    DirName:String;
  BEGIN
    ASSERT(p.ProcessFile#NIL);
    ASSERT(p.FileMask#NIL);

    IF (Directory IN info.attr) THEN
      DirName:=NameExt(info);
      IF (DirName^#'.') & (DirName^#'..') THEN
        RETURN LookAtDir(PathNameExt(info)^,p);
      END;
    END;
    RETURN TRUE
  END Do;

PROCEDURE Deep * (directory, FileMask-:ARRAY OF CHAR; p:POINTER TO Process): BOOLEAN;
  (** Searchs all files matching the mask in the specified directory and
      all it's subdirectories, and processes them
      by p one by one. Returns if all files are processed. *)
  VAR
    proc:ProcessDir;
  BEGIN
    ASSERT(p#NIL);

    proc.ProcessFile:=p;
    proc.FileMask:=DStr.Str(FileMask);

    RETURN LookAtDir(directory,proc);
  END Deep;

END mhostSearchFile.
