(** This module contain different system routines collection.

    Last change: 20.11.2015 *)

<* O2EXTENSIONS + *>

MODULE mhostUtils;

IMPORT
  Win:=Windows, DStr:=msysDynStr, ProgEnv, dllRTS, SYS:=SYSTEM, FileSys;

TYPE
  String = DStr.String;

VAR
  ConsoleWindow: Win.HWND; (* Handle of the console window *)

PROCEDURE ProgArgNumber * (): LONGINT;
  (** Returns the number of the program arguments, 0 if there is no arguments. *)
  BEGIN
    RETURN ProgEnv.ArgNumber()
  END ProgArgNumber;

PROCEDURE ProgArg * (n:LONGINT): String;
  (** Returns n-th program argument or empty string if number of
      program arguments is less than n. The n is zero-based. *)
  VAR
    result:String;
    len:LONGINT;
  BEGIN
    ASSERT(n>=0);

    len:=ProgEnv.ArgLength(n);
    IF len=0 THEN
      RETURN DStr.Empty()
    ELSE
      result:=DStr.New(len);
      ProgEnv.GetArg(n,result^);
      RETURN result
    END;
  END ProgArg;

PROCEDURE GetEnvString * (name-:ARRAY OF CHAR): DStr.String;
  (** Returns specified environment strind, or an empty string
      if it is not found. *)
  VAR
    result:String;
    len:LONGINT;
  BEGIN
    len:=ProgEnv.StringLength(name);
    IF len=0 THEN
      RETURN DStr.Empty()
    ELSE
      result:=DStr.New(len);
      ProgEnv.String(name,result^);
      RETURN result
    END;
  END GetEnvString;

PROCEDURE Ms * ():LONGINT;
  (** Returns number of milliseconds passed since some certain event *)
  BEGIN
    RETURN Win.GetTickCount();
  END Ms;

PROCEDURE SetLowestPriority * ();
  (** Sets lowest priority of current process and thread *)
  BEGIN
    Win.SetPriorityClass(Win.GetCurrentProcess(),Win.IDLE_PRIORITY_CLASS);
    Win.SetThreadPriority(Win.GetCurrentThread(),Win.THREAD_PRIORITY_IDLE);
  END SetLowestPriority;

PROCEDURE SetHighestPriority * ();
  (** Sets highest priority of current process and thread *)
  BEGIN
    Win.SetPriorityClass(Win.GetCurrentProcess(),Win.REALTIME_PRIORITY_CLASS);
    Win.SetThreadPriority(Win.GetCurrentThread(),Win.THREAD_PRIORITY_TIME_CRITICAL);
  END SetHighestPriority;

PROCEDURE GetConsoleWindow(): Win.HWND;
  (* Returns window handle of the console *)
  TYPE
    NoParamWinHandle = PROCEDURE ["StdCall"] (): Win.HWND;
  CONST
    TempTitle = "^%@#*(%^@#$%#$&^*(^2_wrqegkpjkjz'zs sp";
    BufLen = 256;
  VAR
    result:Win.HWND;
    GetConWin:NoParamWinHandle;
    kernel32:dllRTS.HMOD;
    b:BOOLEAN;
    buffer:POINTER TO ARRAY OF CHAR;
    len,ErrorCode:LONGINT;
  BEGIN
    kernel32:=dllRTS.LoadModule("Kernel32.dll");
    IF kernel32#NIL THEN
      GetConWin:=SYS.VAL(NoParamWinHandle,dllRTS.GetProcAdr(kernel32,"GetConsoleWindow"));
    ELSE
      GetConWin:=NIL
    END;

    IF GetConWin#NIL THEN
      result:=GetConWin()
    ELSE (* "GetConsoleWindow" is not found in "Kernel32.dll". *)
      NEW(buffer,BufLen);

      ErrorCode:=Win.ERROR_SUCCESS;
      REPEAT
        len:=Win.GetConsoleTitle(buffer,LEN(buffer^));
        IF len=0 THEN
          ErrorCode:=Win.GetLastError();
          IF ErrorCode=Win.ERROR_SUCCESS THEN
            NEW(buffer,2*LEN(buffer^));
          END;
        END;
      UNTIL (len>0) OR (ErrorCode#Win.ERROR_SUCCESS);
      ASSERT(len>0);

      b:=Win.SetConsoleTitle(TempTitle);
      ASSERT(b);

      Win.Sleep(40);

      result:=Win.FindWindow(NIL,TempTitle);
      ASSERT(result#NIL);

      b:=Win.SetConsoleTitle(buffer^);
      ASSERT(b);
    END;

    IF kernel32#NIL THEN
      dllRTS.FreeModule(kernel32,b);
    END;

    RETURN result;
  END GetConsoleWindow;

PROCEDURE IsConsoleFocused * (): BOOLEAN;
  (** Returns if the console is focused *)
  BEGIN
    RETURN Win.GetForegroundWindow()=ConsoleWindow
  END IsConsoleFocused;

PROCEDURE CopyFile * (FileName,NewFileName:ARRAY OF CHAR; overwrite:BOOLEAN): BOOLEAN;
  (** Copies the specified file to a new name. Overwrite specifies if the
      existing file hsould be overwritten. Returns if the file was copied. *)
  BEGIN
    RETURN Win.CopyFile(FileName,NewFileName,~overwrite);
  END CopyFile;

PROCEDURE MoveFile * (OldFileName,NewFileName:ARRAY OF CHAR): BOOLEAN;
  (** Moves (or renames) the specified file to a new name. Returns if the
      operation was successfull. *)
  BEGIN
    RETURN Win.MoveFile(OldFileName,NewFileName);
  END MoveFile;

PROCEDURE DeleteFile * (FileName-:ARRAY OF CHAR): BOOLEAN;
  (** Delets the specified file, if it exists.
      Returns if the file does't exist after the procedure. *)
  VAR
    done:BOOLEAN;
  BEGIN
    IF FileSys.Exists(FileName) THEN
      FileSys.Remove(FileName,done);
      RETURN done;
    ELSE
      RETURN TRUE;
    END;
  END DeleteFile;

<* PUSH *>
<* WOFF900 + *>
PROCEDURE MakeDir * (DirName-:ARRAY OF CHAR);
  (** Creates the specified directory, if it did not exist yet. *)
  VAR
    b:BOOLEAN;
  BEGIN
    b:=Win.CreateDirectory(DirName,NIL);
    IF ~b THEN
      ASSERT(Win.GetLastError()=Win.ERROR_ALREADY_EXISTS);
    END;
  END MakeDir;
<* POP *>

PROCEDURE FileSize * (FileName-:ARRAY OF CHAR): LONGLONGINT;
  (** Returns size of the specified file. *)
  VAR
    FindData:Win.WIN32_FIND_DATA;
    h:Win.HANDLE;
    int:RECORD ["Modula"]
      CASE :LONGINT OF
        0:
          i:LONGLONGINT;
      | 1:
          lo,hi:Win.DWORD
      END;
    END;
  BEGIN
    h:=Win.FindFirstFile(FileName,FindData);
    ASSERT(h#Win.INVALID_HANDLE_VALUE);

    int.lo:=FindData.nFileSizeLow;
    int.hi:=FindData.nFileSizeHigh;
    Win.FindClose(h);
    RETURN int.i;
  END FileSize;

<* PUSH *>
<* WOFF900 + *>
PROCEDURE GetCurrentDir * (): String;
  (** Returns current directory *)
  VAR
    len:LONGINT;
    res:String;
    NULL:CHAR;
  BEGIN
    len:=Win.GetCurrentDirectory(0,NULL)-1; (* -1 to exclude terminating 0X *)
    ASSERT(len>0);
    res:=DStr.New(len);
    len:=Win.GetCurrentDirectory(len+1,res^); (* +1 to include terminating 0X *)
    ASSERT(len#0);
    RETURN res;
  END GetCurrentDir;

PROCEDURE SetCurrentDir * (dir-:ARRAY OF CHAR);
  (** Sets current directory *)
  VAR
    b:BOOLEAN;
  BEGIN
    b:=Win.SetCurrentDirectory(dir);
    ASSERT(b);
  END SetCurrentDir;
<* POP *>

PROCEDURE GetThreadNumber * (): LONGINT;
  (** Returns a number of threads available on the computer *)
  VAR
    si:Win.SYSTEM_INFO;
  BEGIN
    Win.GetSystemInfo(si);
    RETURN si.dwNumberOfProcessors;
  END GetThreadNumber;

PROCEDURE ExecAndWait * (ProgramName-,CommandLine-,CurrentDir-:ARRAY OF CHAR): BOOLEAN;
  (** Executes the given program with specified command line and current
      directory. Waits until the program is executed. Returns if the program
      was successfully launched. If the program is a console application,
      then it does not create a new console. *)
  VAR
    si:Win.STARTUPINFO;
    pi:Win.PROCESS_INFORMATION;
    cmd:String;
  PROCEDURE Exec (): BOOLEAN;
    BEGIN
      IF CurrentDir='' THEN
        RETURN Win.CreateProcess(ProgramName,cmd^,NIL,NIL,FALSE,
                                 SYS.VAL(Win.CREATE_SET,{}),NIL,NIL,si,pi)
      ELSE
        RETURN Win.CreateProcess(ProgramName,cmd^,NIL,NIL,FALSE,
                                 SYS.VAL(Win.CREATE_SET,{}),NIL,CurrentDir,si,pi)
      END;
    END Exec;
  BEGIN
    cmd:=DStr.Concat2(' ',CommandLine);

    si.cb:=SIZE(si);
    si.lpReserved:=NIL;
    si.lpDesktop:=NIL;
    si.lpTitle:=NIL;
    si.dwX:=0;
    si.dwY:=0;
    si.dwXSize:=0;
    si.dwYSize:=0;
    si.dwXCountChars:=0;
    si.dwYCountChars:=0;
    si.dwFillAttribute:=0;
    si.dwFlags:=SYS.VAL(Win.STARTF_SET,0);
    si.wShowWindow:=0;
    si.cbReserved2:=0;
    si.lpReserved2:=NIL;
    si.hStdInput:=NIL;
    si.hStdOutput:=NIL;
    si.hStdError:=NIL;

    IF Exec()
    THEN
      Win.WaitForSingleObject(pi.hProcess,Win.INFINITE);
      Win.CloseHandle(pi.hThread);
      Win.CloseHandle(pi.hProcess);
      RETURN TRUE
    ELSE
      RETURN FALSE
    END;
  END ExecAndWait;

PROCEDURE ExecCmd * (cmd-,CurrentDir-:ARRAY OF CHAR): BOOLEAN;
  (** Executes a specified command line in the specified current directory. *)
  BEGIN
    RETURN ExecAndWait(GetEnvString('COMSPEC')^,DStr.Concat2('/C ',cmd)^,CurrentDir);
  END ExecCmd;

PROCEDURE GetSystemDirectory * (): DStr.String;
  (** Returns system directory. *)
  VAR
    c:CHAR;
    size:LONGINT;
    res:String;
  BEGIN
    size:=Win.GetSystemDirectory(c,0);
    ASSERT(size>0);
    NEW(res,size);
    Win.GetSystemDirectory(res^,size);
    IF res[DStr.Length(res^)-1]#'\' THEN
      res:=DStr.Concat2(res^,'\');
    END;
    RETURN res;
  END GetSystemDirectory;

TYPE
  <* PUSH *>
  <* ALIGNMENT = "1" *>
  Color * = RECORD (** A color of a pixel in a bitmap *)
    B * , G * , R * , A: SYS.CARD8; (** R,G,B components of the pixel *)
  END;
  <* POP *>
  Pixels * = POINTER TO ARRAY OF ARRAY OF Color;
  (** Pixels[iY,iX] Represents pixels of a bitmap.
      Pixels[0,0] is upper-left corner of the bitmap. *)

TYPE
  BitmapInfo = RECORD
    w,h,BitCount:LONGINT;
  END;

PROCEDURE GetBitmapInfo (bmp:Win.HBITMAP; VAR info:BitmapInfo);
  VAR
    i:LONGINT;
    b:Win.BITMAP;
  BEGIN
    i:=Win.GetObject(bmp,SIZE(b),b);
    ASSERT(i=SIZE(b));
    info.w:=b.bmWidth;
    info.h:=b.bmHeight;
    info.BitCount:=b.bmBitsPixel;
  END GetBitmapInfo;

PROCEDURE LoadBitmap * (FileName-:ARRAY OF CHAR): Pixels;
  (** Loads a bitmap from the specified file name and converts it to the pixels.
      Returns NIL if the file is not loaded. *)
  VAR
    res:Pixels;
    hbmp:Win.HANDLE;
    bmp:Win.HBITMAP;
    i:LONGINT;
    info:Win.BITMAPINFOHEADER;
    BInfo:BitmapInfo;
    dc:Win.HDC;
  BEGIN
    hbmp:=Win.LoadImage(NIL,FileName,Win.IMAGE_BITMAP,0,0,
                        Win.LR_LOADFROMFILE+Win.LR_CREATEDIBSECTION);
                        (* with LR_CREATEDIBSECTION the bitmap format
                           is not mapped for the current screen settings *)
    IF hbmp=NIL THEN
      RETURN NIL
    END;
    bmp:=SYS.VAL(Win.HBITMAP,hbmp);
    GetBitmapInfo(bmp,BInfo);
    NEW(res,BInfo.h,BInfo.w);

    info.biSize:=SIZE(info);
    info.biWidth:=BInfo.w;
    info.biHeight:=-BInfo.h;
    info.biPlanes:=1;
    info.biBitCount:=32;
    info.biCompression:=Win.BI_RGB;
    dc:=Win.GetDC(NIL);
    i:=Win.GetDIBits(dc,bmp,0,BInfo.h,SYS.M2ADR(res^),info,Win.DIB_RGB_COLORS);
    Win.ReleaseDC(NIL,dc);
    ASSERT(i=BInfo.h);

    Win.DeleteObject(bmp);
    RETURN res;
  END LoadBitmap;

BEGIN
  ConsoleWindow:=GetConsoleWindow();
END mhostUtils.