(** Collection of basic physical constants.

    Last change 27.03.2015 *)

(* One of the following equations should be defined to specify system of units.
   If CGS is specified then all electrio-magnetic units are specified in a
   symmetric CGSsystem, i.e. in Gaussian one. *)

<* NEW Units = "CGS"*>
(* <* NEW Units = "SI"*> *)

<* O2NUMEXT + *>
MODULE msciPhConst;

IMPORT
  LongMath;

CONST
  pi * =  LongMath.pi;

(** Basic units and constants: *)

  c_SI = 2.99792458E8; (* speed of light in meters per second *)

<* IF Units = "SI" THEN *>

  meter * = 1;                                  (** meter in meters *)
  gramm * = 1.0E-3;                             (** gramm in kilogramms *)
  second * = 1;                                 (** second in seconds *)

  Amper * = 1;                                  (** Amper in ampers *)
  Tesla * = 1;                                  (** Tesla (magnetic B field in SI) *)
  AmperPerMeter * = 1;                          (** Amper per Meter (magnetic H field in SI) *)

<* ELSIF Units = "CGS" THEN *>

  meter * = 100;                                (** meter in centimeters *)
  gramm * = 1;                                  (** gramm in gramms *)
  second * = 1;                                 (** second in seconds *)

  Amper * = c_SI*meter/second/10;               (** Amper in CGS units *)
  Tesla * = 1.0E4;                              (** Tesla (magnetic B field in SI) *)
  AmperPerMeter * = 4*pi*1.0E-3;                (** Amper per Meter (magnetic H field in SI) *)

<* ELSE *>
  'Units ate not specified!
<* END *>

  mole * = 1;                                   (** mole *)
  Kelvin * = 1;                                 (** Kelvin (temperature) *)

(** Derived units: *)

  cm * = meter/100;                             (** centimeter *)
  mm * = meter/1000;                            (** millimeter *)
  micrometer * = meter*1.0E-6;                  (** micrometer *)
  nm * = meter*1.0E-9;                          (** nanometer *)
  Angstrom * = meter*1.0E-10;                   (** Angstrom *)
  inch * = 2.54*cm;                             (** inch *)
  kg * = 1000*gramm;                            (** kilogramm *)
  min * = 60*second;                            (** minute *)
  Newton * = kg*meter/(second*second);          (** Newton (force) *)
  Pascal * = Newton/meter**2;                   (** Pascal (pressure) *)
  Joule * = Newton*meter;                       (** Joule (energy) *)
  Watt * = Joule/second;                        (** Watt (power) *)

  Atmosphere * = 101325 * Pascal;               (** Atmosphere (pressure) *)
  Torr * = 133.332 * Pascal;                    (** Torr (pressure) *)
  bar * = 1.0E5 * Pascal;                       (** bar (pressure) *)
  psi * = 6894.75729 * Pascal;                  (** psi (pressure) *)
  Calorie * = 4.1868 * Joule;                   (** Calorie (energy) *)

  Coulomb * = Amper * second;                   (** Coulomb (charge in SI) *)
  Volt * = Watt/Amper;                          (** Volt (voltage in SI) *)
  Ohm * = Volt/Amper;                           (** Ohm (resistance in SI) *)
  Farad * = Coulomb/Volt;                       (** Farad (capacitance in SI) *)
  Henry * = Volt*second/Amper;                  (** Henry (inductance in SI) *)


(** Physical constants: *)

  mu0 * = 4*pi*1.0E-7;                          (** H/m - magnetic constant - this constant doesn't depend on units system (the Units equation) *)
  eps0 * = 1.0/(mu0*c_SI*c_SI);                 (** F/m - electric constant - this constant doesn't depend on units system (the Units equation) *)
  c * = c_SI * meter / second;                  (** speed of light in vacuum *)
  G * = 6.6731E-11 * Newton*meter**2/kg**2;     (** gravitational constant *)
  h * = 6.6260687652E-34 * Joule*second;        (** Planck constant *)
  amu * = 1.660540210E-27 * kg;                 (** atomic mass unit *)
  e * = 1.60217646263E-19 * Coulomb;            (** absolute value of electron charge *)
  R * = 8.31447215 * Joule/(mole*Kelvin);       (** gas constant *)
  Vm * = 22.41399639E-3 * meter**3/mole;        (** volume of 1 mol of ideal gas
                                                    at normal conditions (1 atm., 273.15 K) *)
  Rinf * = 10973731.56854983 / meter;           (** Rydberg constant *)
  g * = 9.80665 * meter/second**2;              (** standard gravitational acceleration *)
  SonicSpeed * = 331.46 *meter/second;          (** sonic speed in air (n.c.) *)
  AudibilityThreshold0 * = 2.0E-5 * Pascal;     (** reference value of audibility threshold *)
  WaterDensity * = 999.973 * kg/meter**3;       (** water density (n.c.) *)
  AirDensity * = 1.293 * kg/meter**3;           (** dencity of dry air (n.c.) *)
  NormalPressure * = Atmosphere;                (** normal atmospheric pressure *)
  NormalTemperature * = 273.15 * Kelvin;        (** normal temperature *)
  hbar * = h/(2*pi);                            (** reduced Planck constant *)
  Me * = 5.48579911012E-4*amu;                  (** rest mass of electron *)
  Mp * = 1.0072764668813*amu;                   (** rest mass of proton *)
  Mn * = 1.0086649157855*amu;                   (** rest mass of neutron *)
  ElEnergy * = 0.51099890221;                   (** MeV - rest energy of electron *)
  PEnergy * = 938.27199838;                     (** MeV - rest energy of proton *)
  NEnergy * = 939.56533038;                     (** MeV - rest energy of neutron *)
  MpMe * = 1836.152667539;                      (** Mp/Me ratio *)
  MeMp * = 1.0/MpMe;                            (** Me/Mp ratio *)
  Na * = 6.0221407818E23;                       (** Avogadro constant *)
  Nl * = Na/Vm;                                 (** Loshmidt's constant *)
  F * = Na*e;                                   (** Faraday constant *)
  k * = R/Na;                                   (** Boltzmann constant *)
  alpha * = 7.29735253327E-3;                   (** fine structure constant *)
  MuB * = e*hbar*0.5/Me;                        (** Bohr magneton *)
  MuN * = e*hbar*0.5/Mp;                        (** nuclear magneton *)
  MuP * = 2.79276330*MuN;                       (** magnetic moment of proton *)
  MuE * = 1.00116*MuB;                          (** absolute value of magnetic moment of electron *)
  a0 * = alpha/(4*pi*Rinf);                     (** radius of the 1st Bohr orbit *)
  Re * = alpha*alpha*a0;                        (** classical radius of electron *)

  Electronvolt * = e*Volt;                      (** Electronvolt (energy) *)

(** Constants of translation. For example, to translate value in degrees to radians,
    it should be multiplied by DegToRad. *)

  DegToRad * = pi/180;
  RadToDeg * = 1/DegToRad;

  SCCMToCPS * = NormalPressure/(k*NormalTemperature)*cm**3/min; (** Translates SCCM to particles per second *)
  CPSToSCCM * = 1/SCCMToCPS;                                    (** Translates particles per second to SCCM *)

(** Synonims for units: *)
  Pa * = Pascal;                                (** Pascal (pressure) *)
  atm * = Atmosphere;                           (** Atmosphere (pressure) *)
  cal * = Calorie;                              (** Calorie (energy) *)
  Da * = amu;                                   (** Dalton = atomic mass unit *)
  eV * = Electronvolt;                          (** Electronvolt (energy) *)
  micron * = micrometer;                        (** micrometer *)

END msciPhConst.