(** This module presents a sequence of chars, dynamically allocated in heap.

    Last change: 26.03.2015 *)

<* O2EXTENSIONS + *>

MODULE msysDynCharSeq;

CONST
  BufSize * = 32;
  (** First BufSize chars are allocated in buffer inside the Sequence variable,
      other chars are allocated in heap. *)
  InitLen = 16;
  AddLen = InitLen;
  StringLen = 128;

TYPE
  DynStr * = POINTER TO ARRAY OF CHAR;
  StringArrayPtr = POINTER TO ARRAY OF DynStr;
  Sequence * = RECORD
    buf:ARRAY BufSize OF CHAR;
    s:StringArrayPtr;
    StrInd,CharInd:LONGINT;
  END;
  (** The sequence type. *)

PROCEDURE New (Length:LONGINT): DynStr;
  (* Returns a new created string with specified Length *)
  VAR
    result:DynStr;
  BEGIN
    ASSERT(Length>=0);

    NEW(result,Length+1);
    result^[Length]:=0X;
    RETURN result
  END New;

PROCEDURE Init * (VAR sq:Sequence);
  (** Initiates the sequence variable. Deinitiation is not needed because of
      garbage collection. *)
  BEGIN
    sq.s:=NIL;
    sq.StrInd:=-1;
    sq.CharInd:=0;
  END Init;

PROCEDURE AddChar * (VAR sq:Sequence; c:CHAR);
  (** Adds char c to sequence sq. *)
  VAR
    NewS:StringArrayPtr;
    i:LONGINT;
  BEGIN

    IF sq.StrInd<0 THEN
      sq.buf[sq.CharInd]:=c;
      INC(sq.CharInd);
      IF sq.CharInd>=BufSize THEN
        sq.StrInd:=0;
        sq.CharInd:=0;
      END;
      RETURN
    END;

    IF (sq.StrInd=0) & (sq.CharInd=0) THEN
      NEW(sq.s,InitLen);
    ELSIF sq.StrInd>=LEN(sq.s^) THEN
      NEW(NewS,LEN(sq.s^)+AddLen);
      FOR i:=0 TO LEN(sq.s^)-1 DO
        NewS[i]:=sq.s[i]
      END;
      sq.s:=NewS;
    END;

    IF sq.CharInd=0 THEN
      NEW(sq.s[sq.StrInd],StringLen);
    END;

    sq.s[sq.StrInd,sq.CharInd]:=c;
    INC(sq.CharInd);

    IF sq.CharInd>=StringLen THEN
      INC(sq.StrInd);
      sq.CharInd:=0;
    END;
  END AddChar;

PROCEDURE Count * (VAR sq:Sequence): LONGINT;
  (** Returns count of characters sq. *)
  BEGIN
    IF sq.StrInd<0 THEN
      RETURN sq.CharInd
    ELSE
      RETURN BufSize + sq.StrInd*StringLen + sq.CharInd;
    END;
  END Count;

PROCEDURE String * (VAR sq:Sequence): DynStr;
  (** Returns all content of the sequence as a dynamical string. *)
  VAR
    result:DynStr;
    i,j,ind:LONGINT;
  BEGIN
    IF sq.StrInd<0 THEN
      result:=New(sq.CharInd);
      FOR i:=0 TO sq.CharInd-1 DO
        result[i]:=sq.buf[i]
      END;
      RETURN result;
    END;

    result:=New(BufSize+sq.StrInd*StringLen+sq.CharInd);

    FOR i:=0 TO BufSize-1 DO
      result[i]:=sq.buf[i];
    END;

    ind:=BufSize;

    FOR i:=0 TO sq.StrInd-1 DO
      FOR j:=0 TO StringLen-1 DO
        result[ind]:=sq.s[i,j];
        INC(ind);
      END;
    END;
    FOR j:=0 TO sq.CharInd-1 DO
      result[ind]:=sq.s[sq.StrInd,j];
      INC(ind);
    END;
    RETURN result
  END String;

TYPE
  Proc * = RECORD
  END;
  (** See FormString *)

<* PUSH *>
<* WOFF301 + *>
PROCEDURE (VAR p:Proc) Do * (s-:ARRAY OF CHAR; len:LONGINT);
  BEGIN
    (* Call of n abstract method *)
    HALT(255);
  END Do;
<* POP *>

PROCEDURE FormString * (VAR sq:Sequence; VAR p:Proc);
  (** Forms all the contents of the sequence as a null-terminated string
      and calls p.Do with the formed string and it's length (without
      a terminating 0X) parameters.
      Doesn't use heap if not necessary. *)
  VAR
    s:DynStr;
  BEGIN
    IF sq.StrInd<0 THEN
      sq.buf[sq.CharInd]:=0X;
      p.Do(sq.buf,sq.CharInd)
    ELSE
      s:=String(sq);
      p.Do(s^,LEN(s^)-1);
    END;
  END FormString;

END msysDynCharSeq.