(** This module provides storage for array of dynamical strings
    with dynamically changing length.
    
    Last change: 27.03.2015 *)

<* O2EXTENSIONS + *>
MODULE msysDynStrList;

IMPORT
  DStr:=msysDynStr, In:=msysTextIn, Out:=msysTextOut, Lst:=msysList, COMPILER,
  SU:=msysUtils;

CONST
  InitialCapacity = 32;

CONST
  ModuleName = COMPILER.EQUATION('module');

PROCEDURE RAISE (message-:ARRAY OF CHAR);
  BEGIN
    SU.RAISEError(message);
  END RAISE;

TYPE
  String = DStr.String;
  Array * = POINTER TO ARRAY OF String;

  List * = RECORD (Lst.AnyRec)
    count-:LONGINT;  (** strings count *)
    str-:Array; (** strings *)
  END;
  (** String list type *)

<* PUSH *>
<* WOFF301 + *>
PROCEDURE AssertCorrect * (l-:List);
  (** Asserts the correct structure of the list *)
  BEGIN
    ASSERT((l.str#NIL) OR (l.count=0))
  END AssertCorrect;
<* POP *>

PROCEDURE SetCapacity * (VAR L:List; NewCapacity:LONGINT);
  (** Sets the allocated length of the array of strings maintained by the list.
      Capacity may not be less than count of items and increased if necessary. *)
  VAR
    tmp:Array;
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF NewCapacity<L.count THEN
      NewCapacity:=L.count
    END;
    IF NewCapacity=0 THEN
      L.str:=NIL;
      RETURN
    END;
    IF (L.str#NIL) & (NewCapacity=LEN(L.str^)) THEN
      RETURN
    END;

    NEW(tmp,NewCapacity);
    FOR i:=0 TO L.count-1 DO
      tmp^[i]:=L.str^[i]
    END;
    L.str:=tmp
  END SetCapacity;

PROCEDURE GrowIfNeeded (VAR L:List);
  BEGIN
    IF L.str=NIL THEN
      SetCapacity(L,InitialCapacity);
    ELSIF L.count=LEN(L.str^) THEN
      SetCapacity(L,2*LEN(L.str^));
    END;
  END GrowIfNeeded;

PROCEDURE AddDynStr * (VAR L:List; s:String);
  (** Inserts a new dynamical string at the end of the list.
      s may not be NIL. *)
  BEGIN
    AssertCorrect(L);
    ASSERT(s#NIL);

    GrowIfNeeded(L);
    L.str^[L.count]:=s;
    INC(L.count);
  END AddDynStr;

PROCEDURE Add * (VAR L:List; s-:ARRAY OF CHAR);
  (** Inserts a new string at the end of the list. *)
  BEGIN
    AssertCorrect(L);

    AddDynStr(L,DStr.Str(s))
  END Add;

PROCEDURE Assign * (VAR L:List; Source-:List);
  (** Copies all strings of Source to L. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(Source);

    L.count:=Source.count;
    IF L.count>0 THEN
      NEW(L.str,L.count);
      FOR i:=0 TO L.count-1 DO
        L.str[i]:=Source.str[i]
      END;
    ELSE
      L.str:=NIL
    END
  END Assign;

PROCEDURE Clear * (VAR L:List);
  (** Deletes all strings from the list. *)
  BEGIN
    L.str:=NIL;
    L.count:=0
  END Clear;

PROCEDURE Delete * (VAR L:List; index:LONGINT);
  (** Removes the string at the position given by the index parameter.
      Calling Delete moves up all strings in the strings array that follow
      the deleted string, and reduces the count. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((index>=0) & (index<L.count)) THEN
      RAISE(ModuleName+'.Delete: index out of range.');
    END;

    DEC(L.count);
    FOR i:=index TO L.count-1 DO
      L.str[i]:=L.str[i+1]
    END;
  END Delete;

PROCEDURE Exchange * (VAR L:List; index1,index2:LONGINT);
  (** Swaps the position of two strings in the list. *)
  VAR
    tmp:String;
  BEGIN
    AssertCorrect(L);

    IF ~((index1>=0) & (index1<L.count)) THEN
      RAISE(ModuleName+'.Exchange: index1 out of range.');
    END;
    IF ~((index2>=0) & (index2<L.count)) THEN
      RAISE(ModuleName+'.Exchange: index2 out of range.');
    END;

    tmp:=L.str[index1];
    L.str[index1]:=L.str[index2];
    L.str[index2]:=tmp
  END Exchange;

PROCEDURE IndexOf * (VAR L:List; s-:ARRAY OF CHAR): LONGINT;
  (** Returns the index of the first entry in the list with a specified value. *)
  VAR
    result:LONGINT;
  BEGIN
    AssertCorrect(L);

    result:=0;
    WHILE (result<L.count) & (DStr.Compare(L.str[result]^,s)#0) DO
      INC(result);
    END;
    IF result=L.count THEN
      RETURN -1
    ELSE
      RETURN result
    END;
  END IndexOf;

PROCEDURE IndexOfDynStr * (VAR L:List; s:String): LONGINT;
  (** Returns the index of the first entry in the list with a specified
      dynamical string. *)
  VAR
    result:LONGINT;
  BEGIN
    AssertCorrect(L);
    ASSERT(s#NIL);

    result:=0;
    WHILE (result<L.count) & (L.str[result]#s) DO
      INC(result);
    END;
    IF result=L.count THEN
      RETURN -1
    ELSE
      RETURN result
    END;
  END IndexOfDynStr;

PROCEDURE InsertDynStr * (VAR L:List; index:LONGINT; s:String);
  (** Adds a string to the list at the position specified by index.
      Insert adds the string at the indicated position, shifting the string
      that previously occupied that position, and all subsequent strings, up.
      Insert increments count and, if necessary, allocates memory. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((index>=0) & (index<L.count)) THEN
      RAISE(ModuleName+'.InsertDynStr: index out of range.');
    END;

    GrowIfNeeded(L);
    FOR i:=L.count TO index+1 BY -1 DO
      L.str[i]:=L.str[i-1]
    END;
    INC(L.count);
    L.str[index]:=s
  END InsertDynStr;

PROCEDURE Insert * (VAR L:List; index:LONGINT; s-:ARRAY OF CHAR);
  (** Adds a string to the list at the position specified by index.
      Insert adds the string at the indicated position, shifting the string
      that previously occupied that position, and all subsequent strings, up.
      Insert increments count and, if necessary, allocates memory. *)
  BEGIN
    AssertCorrect(L);

    InsertDynStr(L,index,DStr.Str(s));
  END Insert;

PROCEDURE Move * (VAR L:List; CurIndex,NewIndex:LONGINT);
  (** Changes the position of a string in the list. *)
  VAR
    tmp:String;
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((CurIndex>=0) & (CurIndex<L.count)) THEN
      RAISE(ModuleName+'.Move: CurIndex out of range.');
    END;
    IF ~((NewIndex>=0) & (NewIndex<L.count)) THEN
      RAISE(ModuleName+'.Move: NewIndex out of range.');
    END;

    tmp:=L.str[CurIndex];
    IF NewIndex>CurIndex THEN
      FOR i:=CurIndex TO NewIndex-1 DO
        L.str[i]:=L.str[i+1]
      END;
    ELSE
      FOR i:=CurIndex TO NewIndex+1 BY -1 DO
        L.str[i]:=L.str[i-1]
      END;
    END;
    L.str[NewIndex]:=tmp
  END Move;

TYPE
  Compare * = PROCEDURE (s1-,s2-:ARRAY OF CHAR): LONGINT;
  (** Compare is a type of comparison procedure that should return a positive
      value if s1 is greater than s2, 0 if they are equal, and a negative
      value if s1 is less than s2. This type is used by Sort. *)

PROCEDURE Sort * (VAR L:List; compare:Compare);
  (** Performs a QuickSort on the list based on the comparison function compare *)
  PROCEDURE QuickSort(Left,Right:LONGINT);
    VAR
      I,J:LONGINT;
      P,T:String;
    BEGIN
      REPEAT
        I:=Left;
        J:=Right;
        P:=L.str[(Left+Right) DIV 2];
        REPEAT
          WHILE compare(L.str[I]^,P^)<0 DO
            INC(I);
          END;
          WHILE compare(L.str[J]^,P^)>0 DO
            DEC(J);
          END;
          IF I<=J THEN
            T:=L.str[I];
            L.str[I]:=L.str[J];
            L.str[J]:=T;
            INC(I);
            DEC(J);
          END;
        UNTIL I>J;
        IF Left<J THEN
          QuickSort(Left,J);
        END;
        Left:=I;
      UNTIL I>=Right;
    END QuickSort;
  BEGIN
    AssertCorrect(L);

    IF L.count>0 THEN
      QuickSort(0,L.count-1)
    END;
  END Sort;

CONST
  NewLine * = ''+0DX+0AX; (** New line separator *)

PROCEDURE GetText * (L-:List; delimiter-:ARRAY OF CHAR): String;
  (** Returns all strings of the list as one string.
      Strings are separated by the delimiter. *)
  VAR
    result:String;
    len,i,index:LONGINT;
    DelLen:LONGINT;
    s:String;
  PROCEDURE CopyStr (s-:ARRAY OF CHAR; count:LONGINT);
    VAR
      i:LONGINT;
    BEGIN
      ASSERT(count<=LEN(s));
      FOR i:=0 TO count-1 DO
        result[index]:=s[i];
        INC(index);
      END;
    END CopyStr;
  BEGIN
    AssertCorrect(L);

    DelLen:=DStr.Length(delimiter);

    len:=(L.count-1)*DelLen;
    FOR i:=0 TO L.count-1 DO
      ASSERT(L.str[i]#NIL);
      len:=len+DStr.Length(L.str[i]^);
    END;

    result:=DStr.New(len);

    index:=0;
    s:=L.str[0];
    CopyStr(s^,DStr.Length(s^));
    FOR i:=1 TO L.count-1 DO
      CopyStr(delimiter,DelLen);
      s:=L.str[i];
      CopyStr(s^,DStr.Length(s^));
    END;

    RETURN result;
  END GetText;

TYPE
  Mode * = Out.Mode; (** Files open mode *)

CONST
  Rewrite * = Out.Rewrite;
  Append * = Out.Append;
  (** Constants of file open mode *)

PROCEDURE LoadFromFile * (VAR L:List; FileName-:ARRAY OF CHAR; mode:Mode);
  (** Loads the list from a text file. If mode=Rewrite then the list is cleared
      before the loading. *)
  VAR
    sc:In.Scanner;
    s:String;
  BEGIN
    IF mode=Rewrite THEN
      Clear(L)
    ELSE
      AssertCorrect(L);
    END;
    In.OpenFile(sc,FileName);
    WHILE ~In.EndOfInput(sc) DO
      In.DString(sc,s);
      AddDynStr(L,s);
    END;
    In.Close(sc);
  END LoadFromFile;

PROCEDURE SaveToFile * (VAR L:List; FileName-:ARRAY OF CHAR; mode:Mode);
  (** Saves the list to a text file. If mode=Apend then lines are added to the
      file, if it exists. *)
  VAR
    f:Out.Formatter;
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    Out.OpenFile(f,FileName,mode);
    FOR i:=0 TO L.count-1 DO
      Out.DString(f,L.str[i]^,0);
      Out.Ln(f);
    END;
    Out.Close(f);
  END SaveToFile;

END msysDynStrList.