(** This module provides storage for array of pointers
    with dynamically changing length.
    
    Last change: 28.01.2015 *)

<* O2EXTENSIONS + *>
MODULE msysList;

IMPORT
  Math:=mMath, mMisc, SU:=msysUtils, COMPILER;

TYPE
  Float = Math.Float;

CONST
  InitialCapacity = 32;

CONST
  ModuleName = COMPILER.EQUATION('module');

PROCEDURE RAISE (message-:ARRAY OF CHAR);
  BEGIN
    SU.RAISEError(message);
  END RAISE;

TYPE
  AnyRec * = RECORD END;
  (** Base for any record, *)
  (** Extend other records from AnyRec to use with List, *)

  AnyPtr * = POINTER TO AnyRec; (** Pointer to any record *)
  AnyArray * = POINTER TO ARRAY OF AnyPtr; (** Array of any records *)

  List * = RECORD (AnyRec)
    count-:LONGINT;  (** items count *)
    items-:AnyArray; (** items *)
  END;
  (** List type *)

  Compare * = PROCEDURE (item1,item2:AnyPtr): LONGINT;
  (** Compare is a type of comparison procedure that should return a positive
      value if item1 is greater than item2, 0 if they are equal, and a negative
      value if item1 is less than item2. *)

<* PUSH *>
<* WOFF301 + *>
PROCEDURE AssertCorrect * (l-:List);
  (** Asserts the correct structure of the list *)
  BEGIN
    ASSERT((l.items#NIL) OR (l.count=0));
  END AssertCorrect;
<* POP *>

PROCEDURE SetCapacity * (VAR L:List; NewCapacity:LONGINT);
  (** Sets the allocated length of the array of pointers maintained by the list.
      Capacity may not be less than count of items and increased if necessary. *)
  VAR
    tmp:AnyArray;
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF NewCapacity<L.count THEN
      NewCapacity:=L.count
    END;
    IF NewCapacity=0 THEN
      L.items:=NIL;
      RETURN
    END;
    IF (L.items#NIL) & (NewCapacity=LEN(L.items^)) THEN
      RETURN
    END;

    NEW(tmp,NewCapacity);
    FOR i:=0 TO L.count-1 DO
      tmp^[i]:=L.items^[i]
    END;
    L.items:=tmp
  END SetCapacity;

PROCEDURE GrowIfNeeded (VAR L:List);
  BEGIN
    IF L.items=NIL THEN
      SetCapacity(L,InitialCapacity);
    ELSIF L.count=LEN(L.items^) THEN
      SetCapacity(L,2*LEN(L.items^));
    END;
  END GrowIfNeeded;

PROCEDURE Add * (VAR L:List; item:AnyPtr);
  (** Inserts a new item at the end of the list. *)
  BEGIN
    AssertCorrect(L);

    GrowIfNeeded(L);
    L.items^[L.count]:=item;
    INC(L.count);
  END Add;

PROCEDURE Assign * (VAR L:List; Source-:List);
  (** Copies elements of Source to L. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(Source);

    L.count:=Source.count;
    IF L.count>0 THEN
      NEW(L.items,L.count);
      FOR i:=0 TO L.count-1 DO
        L.items^[i]:=Source.items^[i]
      END;
    ELSE
      L.items:=NIL
    END
  END Assign;

PROCEDURE Clear * (VAR L:List);
  (** Deletes all items from the list. *)
  BEGIN
    L.items:=NIL;
    L.count:=0
  END Clear;

PROCEDURE Delete * (VAR L:List; index:LONGINT);
  (** Removes the item at the position given by the index parameter.
      Calling Delete moves up all items in the items array that follow
      the deleted item, and reduces the count. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((index>=0) & (index<L.count)) THEN
      RAISE(ModuleName+'.Delete: index out of range.');
    END;

    DEC(L.count);
    FOR i:=index TO L.count-1 DO
      L.items^[i]:=L.items^[i+1]
    END;
  END Delete;

PROCEDURE Exchange * (VAR L:List; index1,index2:LONGINT);
  (** Swaps the position of two items in the Items array. *)
  VAR
    tmp:AnyPtr;
  BEGIN
    AssertCorrect(L);

    IF ~((index1>=0) & (index1<L.count)) THEN
      RAISE(ModuleName+'.Exchange: index1 out of range.');
    END;
    IF ~((index2>=0) & (index2<L.count)) THEN
      RAISE(ModuleName+'.Exchange: index2 out of range.');
    END;

    tmp:=L.items^[index1];
    L.items^[index1]:=L.items^[index2];
    L.items^[index2]:=tmp
  END Exchange;

PROCEDURE IndexOf * (L-:List; item:AnyPtr): LONGINT;
  (** Returns the index of the first entry in the items array
      with a specified value, or -1 if the item is not found. *)
  VAR
    result:LONGINT;
  BEGIN
    AssertCorrect(L);

    result:=0;
    WHILE (result<L.count) & (L.items^[result]#item) DO
      INC(result);
    END;
    IF result=L.count THEN
      RETURN -1
    ELSE
      RETURN result
    END;
  END IndexOf;

PROCEDURE Sorted (l-:List; compare:Compare): BOOLEAN;
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(compare#NIL);

    i:=0;
    WHILE (i<l.count-1) & (compare(l.items[i],l.items[i+1])<=0) DO
      INC(i);
    END;
    RETURN i=l.count-1
  END Sorted;

PROCEDURE SortedIndexOf * (l-:List; item:AnyPtr; compare:Compare): LONGINT;
  (** The list should be sorted by increase. (i.e. compare(l.items[i],l.items[i+1]<=0).
      The procedure returns an index of the first element in the list, which is
      greater or equal to the specified item, or count if all elements are smaller.
      A binary search algorithm is used.
      If l.count=0 then 0 is returned. *)
  VAR
    min,max,i:LONGINT;
  BEGIN
    AssertCorrect(l);

    ASSERT(compare#NIL);
    ASSERT(Sorted(l,compare));

    IF l.count=0 THEN
      RETURN 0
    END;

    IF compare(item,l.items[0])<=0 THEN
      RETURN 0
    END;

    IF compare(item,l.items[l.count-1])>0 THEN
      RETURN l.count
    END;

    min:=0;
    max:=l.count-1;
    WHILE max-min>1 DO
      i:=(min+max) DIV 2;
      IF compare(item,l.items[i])>0 THEN
        min:=i
      ELSE
        max:=i
      END;
    END;

    RETURN max
  END SortedIndexOf;

PROCEDURE Extract * (VAR L:List; item:AnyPtr);
  (** Removes a specified item from the list.
      After the item is removed, all the objects that follow it are moved up
      in index position and Count is decremented. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    i:=IndexOf(L,item);
    IF i>=0 THEN
      Delete(L,i)
    END;
  END Extract;

PROCEDURE Insert * (VAR L:List; index:LONGINT; item:AnyPtr);
  (** Adds an object to the Items array at the position specified by Index.
      Insert adds the item at the indicated position, shifting the item
      that previously occupied that position, and all subsequent items, up.
      Insert increments count and, if necessary, allocates memory. *)
  VAR
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((index>=0) & (index<=L.count)) THEN
      RAISE(ModuleName+'.Insert: index out of range.');
    END;

    GrowIfNeeded(L);
    FOR i:=L.count TO index+1 BY -1 DO
      L.items^[i]:=L.items^[i-1]
    END;
    INC(L.count);
    L.items^[index]:=item
  END Insert;

PROCEDURE Move * (VAR L:List; CurIndex,NewIndex:LONGINT);
  (** Changes the position of an item in the items array. *)
  VAR
    tmp:AnyPtr;
    i:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF ~((CurIndex>=0) & (CurIndex<L.count)) THEN
      RAISE(ModuleName+'.Move: CurIndex out of range.');
    END;
    IF ~((NewIndex>=0) & (NewIndex<L.count)) THEN
      RAISE(ModuleName+'.Move: NewIndex out of range.');
    END;

    tmp:=L.items^[CurIndex];
    IF NewIndex>CurIndex THEN
      FOR i:=CurIndex TO NewIndex-1 DO
        L.items^[i]:=L.items^[i+1]
      END;
    ELSE
      FOR i:=CurIndex TO NewIndex+1 BY -1 DO
        L.items^[i]:=L.items^[i-1]
      END;
    END;
    L.items^[NewIndex]:=tmp
  END Move;

PROCEDURE Pack * (VAR L:List);
  (** Deletes all NIL items from the items array.
      Call Pack to move all non-NIL items to the front of the items array
      and reduce the count to the number of items actually used. *)
  VAR
    i,index:LONGINT;
  BEGIN
    AssertCorrect(L);

    IF L.count>0 THEN
      index:=0;
      FOR i:=0 TO L.count-1 DO
        IF L.items^[i]#NIL THEN
          L.items^[index]:=L.items^[i];
          INC(index)
        END;
      END;
      L.count:=index;
      FOR i:=index TO LEN(L.items^)-1 DO
        L.items[i]:=NIL
      END;
    END
  END Pack;

PROCEDURE Sort * (VAR L:List; compare:Compare);
  (** Performs a QuickSort on the list based on the comparison function compare *)
  PROCEDURE QuickSort(Left,Right:LONGINT);
    VAR
      I,J:LONGINT;
      P,T:AnyPtr;
    BEGIN
      REPEAT
        I:=Left;
        J:=Right;
        P:=L.items^[(Left+Right) DIV 2];
        REPEAT
          WHILE compare(L.items^[I],P)<0 DO
            INC(I);
          END;
          WHILE compare(L.items^[J],P)>0 DO
            DEC(J);
          END;
          IF I<=J THEN
            T:=L.items^[I];
            L.items^[I]:=L.items^[J];
            L.items^[J]:=T;
            INC(I);
            DEC(J);
          END;
        UNTIL I>J;
        IF Left<J THEN
          QuickSort(Left,J);
        END;
        Left:=I;
      UNTIL I>=Right;
    END QuickSort;
  BEGIN
    AssertCorrect(L);

    IF L.count>0 THEN
      QuickSort(0,L.count-1)
    END;
  END Sort;

TYPE
  RecFunc * = PROCEDURE (VAR rec: AnyRec): Float;
  IntArrayPtr * = POINTER TO ARRAY OF LONGINT;

PROCEDURE FindMinima * (list-:List; func:RecFunc; count:LONGINT): IntArrayPtr;
  (** Searches count items of the list, which have lesser func(list.items[i])
      than other items, and returns their indexes. NILs are ignored.
      The returned indexes are ordered so that func(result[i])<=func(result[i+1]).
      The length of the result corresponds to number of found items, even if
      less than count items are found. The result is NIL if there are no items
      in the list. *)
  VAR
    result,tmp:IntArrayPtr;
    levels:POINTER TO ARRAY OF Float;
    found,i,k,ind:LONGINT;
    level:Float;
  BEGIN
    AssertCorrect(list);
    ASSERT(count>0);
    ASSERT(func#NIL);

    NEW(result,count);
    NEW(levels,count);
    found:=0;

    FOR i:=0 TO list.count-1 DO
      IF list.items[i]#NIL THEN

        level:=func(list.items[i]^);
        ind:=mMisc.FindArrayIndex(levels^,level,found);
        IF ind<count THEN
          IF found<count THEN
            INC(found);
          END;
          FOR k:=found-2 TO ind BY -1 DO
            result[k+1]:=result[k];
            levels[k+1]:=levels[k];
          END;
          result[ind]:=i;
          levels[ind]:=level;
        END;

      END;
    END;

    IF found=0 THEN
      RETURN NIL
    END;

    IF found<count THEN
      NEW(tmp,found);
      FOR i:=0 TO found-1 DO
        tmp[i]:=result[i]
      END;
      result:=tmp
    END;

    RETURN result;
  END FindMinima;

END msysList.