(** This module provides tool to convert integer to string and back.

    Last changed: 26.03.2015 *)

<* O2EXTENSIONS + *>
<* NOOPTIMIZE - *>
MODULE msysLongLongIntConv;

IMPORT
  SYS:=SYSTEM;

CONST
  MaxStringSize * = 21;
  (** Maximal size of a string, which is enough to present
      any LONGLONGINT and LONGLONGCARD including a terminating 0X *)

TYPE
  CARD = SYS.CARD32;
  String * = ARRAY MaxStringSize OF CHAR;
    (** This string is enough big to store any converted LONGLONGINT
        and LONGLONGCARD, including terminating 0X *)

  Section = (beg,sign,numbers);

  Scanner * = RECORD
    s:Section;
  END;
  (** Scanner of chars. It doesn't any conversions, it only tells if a number
      formed char by char is correct. *)

  Status * = (OK,OutOfRange,NotValid); (** Result of conversion procedures *)

CONST
  MaxLimitShort = MAX(CARD) DIV 10;
  MaxLimitLong = MAX(LONGLONGCARD) DIV 10;
  ORD0 = ORD('0');
  MaxLongLongInt = VAL(LONGLONGCARD,MAX(LONGLONGINT));
  NegMinLongLongInt = VAL(LONGLONGCARD,MaxLongLongInt+1); (* -MIN(LONGLONGINT) *)

PROCEDURE IsNumber (c:CHAR): BOOLEAN;
  BEGIN
    RETURN (c>='0') & (c<='9')
  END IsNumber;

PROCEDURE Init * (VAR s:Scanner);
  (** Initiates the scanner. *)
  BEGIN
    s.s:=beg;
  END Init;

PROCEDURE AddChar * (VAR s:Scanner; c:CHAR): BOOLEAN;
  (** If c is a valid next char of a number, the procedure
      changes the scanner correspondingly. Otherwise s is not changed.
      Returns if c is the valid next char of a number *)
  VAR
    res:BOOLEAN;
  BEGIN
    CASE s.s OF
      beg:
        IF (c='-') OR (c='+') THEN
          s.s:=sign;
          res:=TRUE;
        ELSIF IsNumber(c) THEN
          s.s:=numbers;
          res:=TRUE;
        ELSE
          res:=FALSE;
        END;
    | sign:
        IF IsNumber(c) THEN
          s.s:=numbers;
          res:=TRUE;
        ELSE
          res:=FALSE;
        END;
    | numbers:
        res:=IsNumber(c)
    END;

    RETURN res;
  END AddChar;

PROCEDURE AddCharShort (VAR i:CARD; c:CHAR): BOOLEAN;
  (* i:=10*i+ORD(c)-ORD('0'). Returns if it was successfully done.
     If the operation results in too high result, i is not changed. *)
  VAR
    shift,res:CARD;
  BEGIN
    IF i<=MaxLimitShort THEN
      res:=i*10;
      shift:=ORD(c)-ORD0;
      IF res<=MAX(CARD)-shift THEN
        i:=res+shift;
        RETURN TRUE
      ELSE
        RETURN FALSE
      END;
    ELSE
      RETURN FALSE;
    END;
  END AddCharShort;

PROCEDURE AddCharLong (VAR i:LONGLONGCARD; c:CHAR): BOOLEAN;
  VAR
    shift:CARD;
    res:LONGLONGCARD;
  BEGIN
    IF i<=VAL(LONGLONGCARD,MaxLimitLong) THEN
      res:=i*10;
      shift:=ORD(c)-ORD0;
      IF res<=MAX(LONGLONGCARD)-shift THEN
        i:=res+shift;
        RETURN TRUE
      ELSE
        RETURN FALSE
      END;
    ELSE
      RETURN FALSE;
    END;
  END AddCharLong;

PROCEDURE CardFromString * (str-:ARRAY OF CHAR; ind,len:LONGINT;
                            VAR i:LONGLONGCARD): Status;
  (** Converts the string[ind..ind+len-1] to integer value.
      Returns status of the conversion.
      Sign before the number is not allowed. *)
  VAR
    ShortRes:CARD;
    next:LONGINT;
    res:LONGLONGCARD;
  BEGIN
    ASSERT(ind>=0);
    ASSERT(ind<LEN(str));
    ASSERT(len>0);
    next:=ind+len;
    ASSERT(next<=LEN(str));

    ShortRes:=0;
    WHILE (ind<next) & IsNumber(str[ind]) & AddCharShort(ShortRes,str[ind]) DO
      INC(ind);
    END;

    res:=ShortRes;
    WHILE (ind<next) & IsNumber(str[ind]) & AddCharLong(res,str[ind]) DO
      INC(ind);
    END;

    IF ind=next THEN
      i:=res;
      RETURN OK
    ELSIF ~IsNumber(str[ind]) THEN
      RETURN NotValid
    ELSE
      RETURN OutOfRange
    END;
  END CardFromString;

PROCEDURE FromString * (str-:ARRAY OF CHAR; ind,len:LONGINT;
                        VAR i:LONGLONGINT): Status;
  (** Converts the string[ind..ind+len-1] to the integer value.
      Returns status of the conversion.
      '+' or '-' before the number is allowed. *)
  VAR
    res:LONGLONGCARD;
    neg:BOOLEAN;
    status:Status;
  BEGIN
    ASSERT(ind>=0);
    ASSERT(ind<LEN(str));
    ASSERT(len>0);
    ASSERT(ind+len<=LEN(str));

    CASE str[ind] OF
      '+':
        neg:=FALSE;
        INC(ind);
        DEC(len);
        IF len=0 THEN
          RETURN NotValid
        END;
    | '-':
        neg:=TRUE;
        INC(ind);
        DEC(len);
        IF len=0 THEN
          RETURN NotValid
        END;
    ELSE
      neg:=FALSE
    END;

    status:=CardFromString(str,ind,len,res);
    IF status#OK THEN
      RETURN status;
    END;

    IF neg THEN
      IF res<NegMinLongLongInt THEN
        i:=-VAL(LONGLONGINT,res);
      ELSIF res=NegMinLongLongInt THEN
        i:=MIN(LONGLONGINT)
      ELSE
        RETURN OutOfRange
      END;
    ELSE
      IF res>MaxLongLongInt THEN
        RETURN OutOfRange
      END;
      i:=VAL(LONGLONGINT,res);
    END;
    RETURN OK;
  END FromString;

PROCEDURE DivMod10_32 (val:CARD; VAR div,mod:CARD);
  (* Optimization:
     div = val**0CCCCCCCDh DIV 2**35
     mod = val-10*div
     In XDS this optimization works slower than the simple implementation
 *)
  BEGIN
    div:=val DIV 10;
    mod:=val-div*10;
  END DivMod10_32;

PROCEDURE DivMod10_64 (val:LONGLONGCARD; VAR div:LONGLONGCARD; VAR mod:CARD);
  BEGIN
    div:=val DIV 10;
    mod:=SHORT(val-div*10);
  END DivMod10_64;

PROCEDURE CardToStringReversed (value:LONGLONGCARD; VAR str:String): LONGINT;
  (* Converts the given integer value to its decimal representation, reversed.
     The result is not null-terminated.
     The procedure returns number of written characters. *)
  VAR
    mod,ShortVal:CARD;
    pos:LONGINT;
  BEGIN
    ASSERT(LEN(str)>=MaxStringSize);

    pos:=0;
    WHILE value>MAX(CARD) DO
      DivMod10_64(value,value,mod);
      str[pos]:=CHR(ORD0+mod);
      INC(pos);
    END;

    ShortVal:=SHORT(value);
    WHILE ShortVal>0 DO
      DivMod10_32(ShortVal,ShortVal,mod);
      str[pos]:=CHR(ORD0+mod);
      INC(pos);
    END;
    (* Now str contains reversed result *)

    IF pos=0 THEN
      ASSERT(value=0);
      str[0]:='0';
      pos:=1;
    END;

    RETURN pos;
  END CardToStringReversed;

PROCEDURE Reverse (VAR s:String; len:LONGINT);
  VAR
    p1,p2:LONGINT;
    c:CHAR;
  BEGIN
    ASSERT(len>=0);
    ASSERT(len<=LEN(s));

    p1:=0;
    p2:=len-1;
    WHILE p1<p2 DO
      c:=s[p1];
      s[p1]:=s[p2];
      s[p2]:=c;
      INC(p1);
      DEC(p2);
    END;
  END Reverse;

PROCEDURE CardToString * (value:LONGLONGCARD; VAR str:String): LONGINT;
  (** Converts the given integer value to its decimal representation.
      The result is not null-terminated.
      The procedure returns number of written characters. *)
  VAR
    len:LONGINT;
  BEGIN
    len:=CardToStringReversed(value,str);
    Reverse(str,len);
    RETURN len;
  END CardToString;

PROCEDURE ToStringReversed (value:LONGLONGINT; VAR str:String): LONGINT;
  VAR
    neg:BOOLEAN;
    len:LONGINT;
    val:LONGLONGCARD;
  BEGIN
    neg:=(value<0);
    IF neg THEN
      IF value=MIN(LONGLONGINT) THEN
        val:=NegMinLongLongInt (* = -MIN(LONGLONGINT) *)
      ELSE
        val:=VAL(LONGLONGCARD,-value);
      END;
    ELSE
      val:=VAL(LONGLONGCARD,value);
    END;
    len:=CardToStringReversed(val,str);
    IF neg THEN
      str[len]:='-';
      INC(len)
    END;
    RETURN len;
  END ToStringReversed;

PROCEDURE ToString * (value:LONGLONGINT; VAR str:String): LONGINT;
  (** Converts the given integer value to its decimal representation.
      The result is not null-terminated.
      The procedure returns number of written characters. *)
  VAR
    len:LONGINT;
  BEGIN
    len:=ToStringReversed(value,str);
    Reverse(str,len);
    RETURN len;
  END ToString;

END msysLongLongIntConv.