(** This module provides tool to convert a string, formed char by char, to real,
    and back.

    Last changed: 12.04.2016 *)

<* O2EXTENSIONS + *>
<* NOOPTIMIZE - *>

MODULE msysLongLongRealConv;

IMPORT
  IntConv:=msysLongLongIntConv, Math:=mMath;

CONST
  SignificandAccuracy = 18;
  (* 18 digits of significand should be enough for LONGLONGREAL.
     Moreover there are restrictions of architecture at higher values *)

  Accuracy32 = 9;
  (* Numbers with number of digits <=Accuracy32 can be stored in 32-bit integers *)

  LenOfExponent = 4;
  MaxStringSize * = SignificandAccuracy+5+LenOfExponent;
  (** Size of a string, which is enough to present any LONGLONGREAL
      and a terminating 0X *)

  MinPower = -4951;
  MaxPower = 4932;
  ExpShift = SignificandAccuracy+1;
  (* Exponent shift is used sometimes to compensate contribution of mantissa
     digits to the exponent, at extremely low or extremely high exponents *)

  DefaultE = 'E';

VAR
  NegExpShiftPow10:LONGLONGREAL; (* 10^(-ExpShift) *)

TYPE
  String * = ARRAY MaxStringSize OF CHAR;
    (** This string is enough big to store any converted LONGLONGREAL
        and terminating 0X *)

  Section = (beg,sign,BeforeDot,dot,AfterDot,E,exponent);

  Scanner * = RECORD
    s:Section;
    is:IntConv.Scanner;
  END;

  Status * = (OK,OutOfRange,NotValid); (** Result of conversion procedures *)

VAR
  separator - : CHAR; (** Decimal separator *)

<* PUSH *>
<* ALIGNMENT = '2' *>
TYPE
  Inf = ARRAY 5 OF INTEGER;
  Tab0 = ARRAY 5*32 OF INTEGER;
  Tab1 = ARRAY 5*15 OF INTEGER;
  Tab2 = ARRAY 5*9 OF INTEGER;
<* POP *>

VAR
  inf:RECORD ["Modula"] (* inf.float = plus infinity *)
    CASE :INTEGER OF
      0:
        int:Inf;
    | 1:
        float:LONGLONGREAL;
    END;
  END;
  pow10:RECORD ["Modula"] (* pow10.float[ind] = 10^ind *)
    CASE :INTEGER OF
      0:
        int:Tab0;
    | 1:
        float:ARRAY 32 OF LONGLONGREAL;
    END;
  END;
  pow100:RECORD ["Modula"] (* pow100.float[ind] contains some powers of 10 in xx range *)
    CASE :INTEGER OF
      0:
        int:Tab1;
    | 1:
        float:ARRAY 15 OF LONGLONGREAL;
    END;
  END;
  pow1000:RECORD ["Modula"] (* pow1000.float[ind] contains some powers of 10 in xxx range *)
    CASE :INTEGER OF
      0:
        int:Tab2;
    | 1:
        float:ARRAY 9 OF LONGLONGREAL;
    END;
  END;

PROCEDURE Power10 (power:LONGINT): LONGLONGREAL;
  (* Returns 10**power *)
  VAR
    neg:BOOLEAN;
  PROCEDURE PosPower (): LONGLONGREAL;
    VAR
      result:LONGLONGREAL;
      ind:LONGINT;
    BEGIN
      result:=pow10.float[power & 01Fh];
      power:=power DIV 32;
      IF power=0 THEN
        RETURN result
      END;
      ind:=power & 0Fh -1;
      IF ind>=0 THEN
        result:=result*pow100.float[ind];
      END;
      power:=power DIV 16;
      IF power=0 THEN
        RETURN result;
      END;
      RETURN result*pow1000.float[power-1];
    END PosPower;
  BEGIN
    neg:=(power<0);
    IF neg THEN
      power:=-power;
    END;
    IF power>MaxPower THEN
      IF neg THEN
        IF power<=-MinPower THEN
          power:=power-(-MinPower-MaxPower);
          RETURN 1/PosPower()/pow10.float[-MinPower-MaxPower];
        ELSE
          RETURN 0
        END;
      ELSE
        RETURN inf.float
      END;
    END;
    IF neg THEN
      RETURN 1/PosPower()
    ELSE
      RETURN PosPower()
    END;
  END Power10;

PROCEDURE SetDecimalSeparator * (c:CHAR);
  (** Sets the decimal separator *)
  BEGIN
    separator:=c;
  END SetDecimalSeparator;

PROCEDURE Init * (VAR s:Scanner);
  (** Initiates the scanner. *)
  BEGIN
    s.s:=beg;
  END Init;

PROCEDURE IsNumber (c:CHAR): BOOLEAN;
  BEGIN
    RETURN (c>='0') & (c<='9')
  END IsNumber;

PROCEDURE IsE (c:CHAR): BOOLEAN;
  BEGIN
    RETURN (c='E') OR (c='e')
  END IsE;

PROCEDURE AddChar * (VAR s:Scanner; c:CHAR): BOOLEAN;
  (** If c is a valid next char of a number, the procedure
      changes the scanner correspondingly. Otherwise s is not changed.
      Returns if c is the valid next char of a number *)
  VAR
    res:BOOLEAN;
  PROCEDURE SetESection ();
    BEGIN
      s.s:=E;
      IntConv.Init(s.is);
    END SetESection;
  BEGIN
    CASE s.s OF
      beg:
        IF IsNumber(c) THEN
          s.s:=BeforeDot;
          res:=TRUE;
        ELSIF (c='-') OR (c='+') THEN
          s.s:=sign;
          res:=TRUE;
        ELSIF c=separator THEN
          s.s:=dot;
          res:=TRUE;
        ELSE
          res:=FALSE
        END;
    | sign:
        IF IsNumber(c) THEN
          s.s:=BeforeDot;
          res:=TRUE;
        ELSIF c=separator THEN
          s.s:=dot;
          res:=TRUE;
        ELSE
          res:=FALSE
        END;
    | BeforeDot:
        IF IsNumber(c) THEN
          res:=TRUE;
        ELSIF c=separator THEN
          s.s:=dot;
          res:=TRUE;
        ELSIF IsE(c) THEN
          SetESection;
          res:=TRUE;
        ELSE
          res:=FALSE
        END;
    | dot:
        IF IsNumber(c) THEN
          s.s:=AfterDot;
          res:=TRUE;
        ELSIF IsE(c) THEN
          SetESection;
          res:=TRUE;
        ELSE
          res:=FALSE
        END;
    | AfterDot:
        IF IsNumber(c) THEN
          res:=TRUE;
        ELSIF IsE(c) THEN
          SetESection;
          res:=TRUE;
        ELSE
          res:=FALSE
        END;
    | E:
        IF IntConv.AddChar(s.is,c) THEN
          s.s:=exponent;
          res:=TRUE;
        ELSE
          res:=FALSE;
        END;
    | exponent:
        res:=IntConv.AddChar(s.is,c);
    END;

    RETURN res;
  END AddChar;

TYPE
  Pow10int = ARRAY SignificandAccuracy-Accuracy32+1 OF LONGINT;
VAR
  pow10int:Pow10int;

PROCEDURE ExtractMantExp (str-:ARRAY OF CHAR; ind,len:LONGINT;
                          VAR neg:BOOLEAN; VAR mant:LONGLONGINT;
                          VAR exp:LONGINT): Status;
  (* 64-bit mantissa vs 32 integers *)
  VAR
    mlo,mhi,mcount:LONGINT;
    ExpAddition:LONGINT;
  PROCEDURE AddCharToMant (c:CHAR): BOOLEAN;
    (* returns if the symbol was added *)
    CONST
      ORD0 = ORD('0');
    VAR
      res:BOOLEAN;
    BEGIN
      res:=(mcount<SignificandAccuracy);
      IF res THEN
        IF mcount<Accuracy32 THEN
          mhi:=mhi*10 + ORD(c) - ORD0;
        ELSE
          mlo:=mlo*10 + ORD(c) - ORD0;
        END;
        INC(mcount);
      END;
      RETURN res;
    END AddCharToMant;
  PROCEDURE GetMant (): LONGLONGINT;
    BEGIN
      IF mcount<Accuracy32 THEN
        RETURN mhi
      ELSE
        RETURN VAL(LONGLONGINT,mhi)*pow10int[mcount-Accuracy32] + mlo
      END;
    END GetMant;

  VAR
    next:LONGINT;
    LongExp:LONGLONGINT;
    SepFound:BOOLEAN;
  BEGIN
    CASE str[ind] OF
      '+':
        neg:=FALSE;
        INC(ind);
        DEC(len);
        IF len=0 THEN
          RETURN NotValid
        END;
    | '-':
        neg:=TRUE;
        INC(ind);
        DEC(len);
        IF len=0 THEN
          RETURN NotValid
        END;
    ELSE
      neg:=FALSE
    END;

    next:=ind+len;
    ExpAddition:=0;

    (* remoning not meaning zeros *)
    WHILE (ind<next) & (str[ind]='0') DO
      INC(ind)
    END;
    IF (ind<next) & (str[ind]=separator) THEN
      INC(ind);
      WHILE (ind<next) & (str[ind]='0') DO
        INC(ind);
        DEC(ExpAddition);
      END;
      SepFound:=TRUE
    ELSE
      SepFound:=FALSE
    END;

    (* processing mantissa *)
    mant:=0;
    mlo:=0;
    mhi:=0;
    mcount:=0;
    WHILE (ind<next) & IsNumber(str[ind]) DO
      IF ~AddCharToMant(str[ind]) THEN
        INC(ExpAddition)
      END;
      IF SepFound THEN
        DEC(ExpAddition)
      END;
      INC(ind);
    END;
    IF (ind<next) & (str[ind]=separator) THEN
      INC(ind);
      IF SepFound THEN
        RETURN NotValid
      END;
      WHILE (ind<next) & IsNumber(str[ind]) DO
        IF AddCharToMant(str[ind]) THEN
          DEC(ExpAddition);
        END;
        INC(ind);
      END;
    END;

    (* Processing exponent *)
    IF (ind<next) & IsE(str[ind]) THEN
      INC(ind);
      IF ind=next THEN
        RETURN NotValid
      END;
      CASE IntConv.FromString(str,ind,next-ind,LongExp) OF
         IntConv.OK:
      |  IntConv.OutOfRange:
          IF str[ind]='-' THEN
            mant:=0;
            exp:=0;
            RETURN OK
          ELSE
            mant:=GetMant();
            IF mant=0 THEN
              exp:=0;
              RETURN OK
            ELSE
              RETURN OutOfRange
            END;
          END;
      | IntConv.NotValid:
          RETURN NotValid;
      END;
    ELSE
      IF ind#next THEN
        RETURN NotValid
      END;
      LongExp:=0;
    END;

    LongExp:=LongExp+ExpAddition;
    IF LongExp<MIN(LONGINT) THEN
      mant:=0;
      exp:=0;
      RETURN OK
    ELSE
      mant:=GetMant();
      IF LongExp>MAX(LONGINT) THEN
        IF mant=0 THEN
          exp:=0;
          RETURN OK
        ELSE
          RETURN OutOfRange
        END
      ELSE
        exp:=SHORT(LongExp);
        RETURN OK
      END;
    END;
  END ExtractMantExp;

PROCEDURE MantExpToNumberHigh (mant:LONGLONGINT; exp:LONGINT;
                               VAR x:LONGLONGREAL): Status;
  VAR
    pow,MantReal:LONGLONGREAL;
  BEGIN
    IF exp>MaxPower THEN
      IF mant=0 THEN
        x:=0;
        RETURN OK
      ELSE
        RETURN OutOfRange
      END;
    END;
    pow:=Power10(exp);
    MantReal:=mant;
    IF Math.MaxLONGLONGREAL/pow>=MantReal THEN
      x:=pow*MantReal;
      RETURN OK
    ELSE
      RETURN OutOfRange
    END;
  END MantExpToNumberHigh;

PROCEDURE MantExpToNumberLow (mant:LONGLONGINT; exp:LONGINT;
                              VAR x:LONGLONGREAL): Status;
  BEGIN
    x:=Power10(exp+ExpShift)*mant*NegExpShiftPow10;
    RETURN OK;
  END MantExpToNumberLow;

PROCEDURE MantExpToNumber (neg:BOOLEAN; mant:LONGLONGINT; exp:LONGINT;
                           VAR x:LONGLONGREAL): Status;
  (* Converts neg, mant, exp to the real value *)
  VAR
    s:Status;
  BEGIN
    IF exp>=MaxPower-ExpShift THEN
      s:=MantExpToNumberHigh(mant,exp,x)
    ELSIF exp<=MinPower+ExpShift THEN
      s:=MantExpToNumberLow(mant,exp,x)
    ELSE
      x:=mant*Power10(exp);
      s:=OK;
    END;
    IF (s=OK) & neg THEN
      x:=-x
    END;

    RETURN s;
  END MantExpToNumber;

PROCEDURE FromString * (str-:ARRAY OF CHAR; ind,len:LONGINT;
                        VAR x:LONGLONGREAL): Status;
  (** Converts the string[ind..ind+len-1] to the real value.
      Returns status of the conversion. *)
  VAR
    mant:LONGLONGINT;
    exp:LONGINT;
    status:Status;
    neg:BOOLEAN;
  BEGIN
    ASSERT(ind>=0);
    ASSERT(ind<LEN(str));
    ASSERT(len>0);
    ASSERT(ind+len<=LEN(str));

    status:=ExtractMantExp(str,ind,len,neg,mant,exp);
    IF status#OK THEN
      RETURN status
    END;
    RETURN MantExpToNumber(neg,mant,exp,x)
  END FromString;

PROCEDURE MantExp2 (z:LONGLONGREAL): LONGLONGREAL;
  (* Returns binary exponent of z multiplied by log_10 2 *)
  VAR
    e:LONGLONGREAL;
  BEGIN
    ASM
      fld z
      fxtract
      fstp e
      fldlg2
      fmulp
      fstp e
    END;
    RETURN e;
  END MantExp2;

PROCEDURE MantExp (z:LONGLONGREAL; VAR mantissa:LONGLONGREAL; VAR exp:LONGINT);
  (* Decays z to its decimal mantissa and exponent.
     Postcondition: 1<=sig<20 *)
  VAR
    exp2:LONGLONGREAL;
  BEGIN
    exp2:=MantExp2(z);
    exp:=ENTIER(exp2);
    IF exp<MinPower+(-MinPower-MaxPower) THEN
      mantissa:=z*Power10(-exp-(-MinPower-MaxPower))*pow10.float[-MinPower-MaxPower];
    ELSE
      mantissa:=z*Power10(-exp);
    END;
  END MantExp;

PROCEDURE ProduceZero (VAR str:String; SignificandDigits:LONGINT): LONGINT;
  VAR
    ind:LONGINT;
  PROCEDURE AddC (c:CHAR);
    BEGIN
      ASSERT(ind<LEN(str));
      str[ind]:=c;
      INC(ind);
    END AddC;
  VAR
    i:LONGINT;
  BEGIN
    ind:=0;
    AddC('0');
    IF SignificandDigits>1 THEN
      AddC(separator);
      FOR i:=2 TO SignificandDigits DO
        AddC('0');
      END;
    END;
    AddC(DefaultE);
    AddC('+');
    AddC('0');
    AddC('0');
    AddC('0');
    AddC('0');
    RETURN ind;
  END ProduceZero;

PROCEDURE PrepareMantString (x:LONGLONGREAL; SignificandDigits:LONGINT;
                             VAR MantInt:IntConv.String; VAR exp:LONGINT);
  (* Decays the x to the mantissa and the exponent.
     Mantissa is returned in MantInt as an integer number.
     For example, MantInt='1234' means mantissa = 1.234. *)
  VAR
    MantReal:LONGLONGREAL;
    mant:LONGLONGINT;
    i,len:LONGINT;
  BEGIN
    MantExp(x,MantReal,exp);
    IF MantReal>=10 THEN
      i:=SignificandDigits-2;
      IF i>=0 THEN
        mant:=Math.LongENTIER(MantReal*pow10.float[i] + 0.5);
      ELSE
        mant:=Math.LongENTIER(MantReal/10 + 0.5);
      END;
      INC(exp)
    ELSE
      mant:=Math.LongENTIER(MantReal*pow10.float[SignificandDigits-1] + 0.5);
    END;
    len:=IntConv.ToString(mant,MantInt);
      (* len is SignificandDigits or SignificandDigits+1 if rounding changes decimal order *)
    IF len=SignificandDigits+1 THEN
      INC(exp);
    END;
  END PrepareMantString;

PROCEDURE ToString * (x:LONGLONGREAL; SignificandDigits:LONGINT;
                      VAR str:String): LONGINT;
  (** Converts the given integer value to its decimal representation.
      The result is not null-terminated.
      The procedure returns number of written characters. *)
  VAR
    ind:LONGINT;
  PROCEDURE AddC (c:CHAR);
    BEGIN
      ASSERT(ind<LEN(str));
      str[ind]:=c;
      INC(ind);
    END AddC;
  VAR
    exp,len,i:LONGINT;
    buf:IntConv.String;
  BEGIN
    ASSERT(SignificandDigits>0);

    IF SignificandDigits>SignificandAccuracy THEN
      SignificandDigits:=SignificandAccuracy
    END;

    IF x=0 THEN
      RETURN ProduceZero(str,SignificandDigits);
    END;

    IF x<0 THEN
      str[0]:='-';
      ind:=1;
      x:=-x
    ELSE
      ind:=0
    END;

    PrepareMantString(x,SignificandDigits,buf,exp);

    AddC(buf[0]);
    IF SignificandDigits>1 THEN
      AddC(separator);
      FOR i:=1 TO SignificandDigits-1 DO
        AddC(buf[i]);
      END;
    END;

    AddC(DefaultE);
    IF exp<0 THEN
      exp:=-exp;
      AddC('-')
    ELSE
      AddC('+')
    END;
    len:=IntConv.ToString(exp,buf);
    FOR i:=len TO LenOfExponent-1 DO
      AddC('0');
    END;
    FOR i:=0 TO len-1 DO
      AddC(buf[i]);
    END;

    RETURN ind
  END ToString;

PROCEDURE ProduceZeroFixed(VAR str:ARRAY OF CHAR; SignificandDigits:LONGINT): LONGINT;
  VAR
    ind:LONGINT;
  PROCEDURE AddC (c:CHAR);
    BEGIN
      IF ind<LEN(str) THEN
        str[ind]:=c;
      END;
      INC(ind);
    END AddC;
  VAR
    i:LONGINT;
  BEGIN
    ind:=0;
    AddC('0');
    IF SignificandDigits>1 THEN
      AddC(separator);
      FOR i:=2 TO SignificandDigits DO
        AddC('0');
      END;
    END;
    RETURN ind
  END ProduceZeroFixed;

PROCEDURE ToStringFixed * (x:LONGLONGREAL; SignificandDigits:LONGINT;
                           VAR str:ARRAY OF CHAR): LONGINT;
  (** Converts the given integer value to its decimal representation in fixed form.
      SignificandDigits digits of the significand are alwais produced.
      Leading or trailing zeros are added if necessary.
      The result is not null-terminated.
      The procedure returns number of written characters.
      If the str has not enough length, the result is truncated.
      In this case the procedure returns length of the string, which is enough
      to contain the entier result without the terminating 0X. *)
  VAR
    ind:LONGINT;
  PROCEDURE AddC (c:CHAR);
    BEGIN
      IF ind<LEN(str) THEN
        str[ind]:=c;
      END;
      INC(ind);
    END AddC;
  VAR
    exp,i:LONGINT;
    buf:IntConv.String;
  BEGIN
    ASSERT(SignificandDigits>0);

    IF x=0 THEN
      RETURN ProduceZeroFixed(str,SignificandDigits);
    END;

    IF SignificandDigits>SignificandAccuracy THEN
      SignificandDigits:=SignificandAccuracy
    END;

    ind:=0;
    IF x<0 THEN
      AddC('-');
      x:=-x
    END;

    PrepareMantString(x,SignificandDigits,buf,exp);

    i:=0;
    IF exp<0 THEN
      AddC('0');
    END;
    WHILE (exp>=0) & (SignificandDigits>0) DO
      AddC(buf[i]);
      INC(i);
      DEC(exp);
      DEC(SignificandDigits);
    END;
    WHILE exp>=0 DO
      AddC('0');
      DEC(exp);
    END;
    IF SignificandDigits>0 THEN
      AddC(separator);
      WHILE exp<-1 DO
        AddC('0');
        INC(exp);
      END;
      WHILE SignificandDigits>0 DO
        AddC(buf[i]);
        INC(i);
        DEC(SignificandDigits);
      END;
    END;

    RETURN ind;
  END ToStringFixed;

BEGIN
  SetDecimalSeparator('.');

  inf.int:=Inf{00000H,00000H,00000H,08000H,07FFFH};
  pow10.int:=Tab0{
             00000H,00000H,00000H,08000H,03FFFH,        (* 10**0 *)
             00000H,00000H,00000H,0A000H,04002H,        (* 10**1 *)
             00000H,00000H,00000H,0C800H,04005H,        (* 10**2 *)
             00000H,00000H,00000H,0FA00H,04008H,        (* 10**3 *)
             00000H,00000H,00000H,09C40H,0400CH,        (* 10**4 *)
             00000H,00000H,00000H,0C350H,0400FH,        (* 10**5 *)
             00000H,00000H,00000H,0F424H,04012H,        (* 10**6 *)
             00000H,00000H,08000H,09896H,04016H,        (* 10**7 *)
             00000H,00000H,02000H,0BEBCH,04019H,        (* 10**8 *)
             00000H,00000H,02800H,0EE6BH,0401CH,        (* 10**9 *)
             00000H,00000H,0F900H,09502H,04020H,        (* 10**10 *)
             00000H,00000H,0B740H,0BA43H,04023H,        (* 10**11 *)
             00000H,00000H,0A510H,0E8D4H,04026H,        (* 10**12 *)
             00000H,00000H,0E72AH,09184H,0402AH,        (* 10**13 *)
             00000H,08000H,020F4H,0B5E6H,0402DH,        (* 10**14 *)
             00000H,0A000H,0A931H,0E35FH,04030H,        (* 10**15 *)
             00000H,00400H,0C9BFH,08E1BH,04034H,        (* 10**16 *)
             00000H,0C500H,0BC2EH,0B1A2H,04037H,        (* 10**17 *)
             00000H,07640H,06B3AH,0DE0BH,0403AH,        (* 10**18 *)
             00000H,089E8H,02304H,08AC7H,0403EH,        (* 10**19 *)
             00000H,0AC62H,0EBC5H,0AD78H,04041H,        (* 10**20 *)
             08000H,0177AH,026B7H,0D8D7H,04044H,        (* 10**21 *)
             09000H,06EACH,07832H,08786H,04048H,        (* 10**22 *)
             0B400H,00A57H,0163FH,0A968H,0404BH,        (* 10**23 *)
             0A100H,0CCEDH,01BCEH,0D3C2H,0404EH,        (* 10**24 *)
             084A0H,04014H,05161H,08459H,04052H,        (* 10**25 *)
             0A5C8H,09019H,0A5B9H,0A56FH,04055H,        (* 10**26 *)
             00F3AH,0F420H,08F27H,0CECBH,04058H,        (* 10**27 *)
             00984H,0F894H,03978H,0813FH,0405CH,        (* 10**28 *)
             00BE5H,036B9H,007D7H,0A18FH,0405FH,        (* 10**29 *)
             04EDFH,00467H,0C9CDH,0C9F2H,04062H,        (* 10**30 *)
             02296H,04581H,07C40H,0FC6FH,04065H         (* 10**31 *)
             };
  pow100.int:=Tab1{
             0B59EH,02B70H,0ADA8H,09DC5H,04069H,        (* 10**32 *)
             0A6D5H,0FFCFH,01F49H,0C278H,040D3H,        (* 10**64 *)
             014A3H,0C59BH,0AB16H,0EFB3H,0413DH,        (* 10**96 *)
             08CE0H,080E9H,047C9H,093BAH,041A8H,        (* 10**128 *)
             017AAH,07FE6H,0A12BH,0B616H,04212H,        (* 10**160 *)
             0556BH,03927H,0F78DH,0E070H,0427CH,        (* 10**192 *)
             0C930H,0E33CH,096FFH,08A52H,042E7H,        (* 10**224 *)
             0DE8EH,09DF9H,0EBFBH,0AA7EH,04351H,        (* 10**256 *)
             02F8CH,05C6AH,0FC19H,0D226H,043BBH,        (* 10**288 *)
             0E376H,0F2CCH,02F29H,08184H,04426H,        (* 10**320 *)
             00AD2H,0DB90H,02700H,09FA4H,04490H,        (* 10**352 *)
             0AA17H,0AEF8H,0E310H,0C4C5H,044FAH,        (* 10**384 *)
             09C59H,0E9B0H,09C07H,0F28AH,04564H,        (* 10**416 *)
             0F3D4H,0EBF7H,04AE1H,0957AH,045CFH,        (* 10**448 *)
             0A262H,00795H,0D8DCH,0B83EH,04639H         (* 10**480 *)
             };
  pow1000.int:=Tab2{
             091C7H,0A60EH,0A0AEH,0E319H,046A3H,        (* 10**512 *)
             00C17H,08175H,07586H,0C976H,04D48H,        (* 10**1024 *)
             0A7E4H,03993H,0353BH,0B2B8H,053EDH,        (* 10**1536 *)
             05DE5H,0C53DH,03B5DH,09E8BH,05A92H,        (* 10**2048 *)
             0F0A6H,020A1H,054C0H,08CA5H,06137H,        (* 10**2560 *)
             05A8BH,0D88BH,05D25H,0F989H,067DBH,        (* 10**3072 *)
             0F3F8H,0BF27H,0C8A2H,0DD5DH,06E80H,        (* 10**3584 *)
             0979BH,08A20H,05202H,0C460H,07525H,        (* 10**4096 *)
             059F0H,06ED5H,01162H,0AE35H,07BCAH         (* 10**4608 *)
             };

  NegExpShiftPow10:=Power10(-ExpShift);

  pow10int[0]:=1;
  pow10int[1]:=10;
  pow10int[2]:=100;
  pow10int[3]:=1000;
  pow10int[4]:=10000;
  pow10int[5]:=100000;
  pow10int[6]:=1000000;
  pow10int[7]:=10000000;
  pow10int[8]:=100000000;
  pow10int[9]:=1000000000;
END msysLongLongRealConv.