(** This module reads/writes text data from/to text files.

    Last changed: 12.04.2016 *)

<* O2EXTENSIONS + *>

MODULE msysTextData;

IMPORT
  In:=msysTextIn, DStr:=msysDynStr, SU:=msysUtils, COMPILER, Out:=msysTextOut,
  StrList:=msysDynStrList, CharSeq:=msysDynCharSeq, SF:=mhostStreamOfFile,
  Math:=mMath;

CONST
  ModuleName = COMPILER.EQUATION('module');

  Tab * = Out.Tab;

TYPE
  String = DStr.String;
  FloatArrayPtr = Math.ArrayPtr;

  Column * = StrList.List; (** List of strings *)
  ColumnPtr * = POINTER TO Column;
  Columns * = POINTER TO ARRAY MAX(SET)-MIN(SET)+1 OF ColumnPtr;

  Mode * = Out.Mode;

CONST
  Append * = Out.Append;
  Rewrite * = Out.Rewrite;

VAR
  separator * :DStr.String; (** Write procedures use this as a separator between the columns *)

PROCEDURE LinesCount * (FileName-:ARRAY OF CHAR): LONGINT;
  (** Returns number of lines in the given text file.
      All empty lines at the end of the file are ignored. *)
  CONST
    ln = 0AX; (* line separator *)
    len = 1024; (* length of a buffer *)
  VAR
    buf:ARRAY len OF CHAR;
    i,count,CountAtEnd,CharsRead:LONGINT;
    r:SF.Reader;
  BEGIN
    r:=SF.NewReader(FileName);
    ASSERT(r#NIL);
    count:=0;
    CountAtEnd:=0;
    REPEAT
      CharsRead:=r.Read(buf,LEN(buf));
      FOR i:=0 TO CharsRead-1 DO
        IF buf[i]=ln THEN
          INC(count);
          INC(CountAtEnd);
        ELSIF ~DStr.IsSpace(buf[i]) THEN
          CountAtEnd:=0;
        END;
      END;
    UNTIL CharsRead#len;
    SF.CloseReader(r);
    RETURN count-CountAtEnd+1
  END LinesCount;

PROCEDURE ReadString * (FileName-:ARRAY OF CHAR; SkipLines:LONGINT):DStr.String;
  (** Reads a string from the specified file after skiping the specified number
      of strings. *)
  VAR
    s:In.Scanner;
    i:LONGINT;
    result:DStr.String;
  BEGIN
    In.OpenFile(s,FileName);

    FOR i:=1 TO SkipLines DO
      In.SkipLine(s);
    END;
    In.DString(s,result);

    In.Close(s);

    RETURN result
  END ReadString;

PROCEDURE AppendString * (FileName-,s-:ARRAY OF CHAR);
  (** Appends the string s to the specified file.
      If the file doesn't exist, it is created. *)
  VAR
    f:Out.Formatter;
  BEGIN
    Out.OpenFile(f,FileName,Append);
    Out.DString(f,s,0);
    Out.Close(f)
  END AppendString;

PROCEDURE AppendLn * (FileName-:ARRAY OF CHAR);
  (** Appends a new line to the specified file.
      If the file doesn't exist, it is created. *)
  VAR
    f:Out.Formatter;
  BEGIN
    Out.OpenFile(f,FileName,Append);
    Out.Ln(f);
    Out.Close(f)
  END AppendLn;

PROCEDURE Length * (c:Columns): LONGINT;
  (** Returns length of each column in the specified columns *)
  VAR
    i:LONGINT;
    col:ColumnPtr;
  BEGIN
    ASSERT(c#NIL);

    i:=0;
    WHILE (i<LEN(c^)) & (c[i]=NIL) DO
      INC(i);
    END;

    IF i>=LEN(c^) THEN
      RETURN 0
    ELSE
      col:=c[i];
      ASSERT(col#NIL);
      RETURN col.count
    END;
  END Length;

PROCEDURE ReadUntilSeparator (VAR s:In.Scanner; separator:CHAR; VAR result:String);
  PROCEDURE InChar (VAR c:CHAR): BOOLEAN;
    BEGIN
      IF In.EndOfLine(s) THEN
        c:=0X; (* to suppress a warning message *)
        RETURN FALSE
      ELSE
        In.Char(s,c);
        RETURN TRUE
      END;
    END InChar;
  VAR
    seq:CharSeq.Sequence;
    c:CHAR;
  BEGIN
    CharSeq.Init(seq);
    result:=NIL;

    WHILE InChar(c) & (c#separator) DO
      CharSeq.AddChar(seq,c);
    END;

    result:=CharSeq.String(seq);
  END ReadUntilSeparator;

PROCEDURE ReadText * (FileName-:ARRAY OF CHAR; SkipLines:LONGINT;
                      indexes:SET; separator:CHAR): Columns;
  (** Reads data, which should be in columns in the specified file.
      The columns shuld be separated by the specified separator.
      Before reading data, SkipLines lines are skipped.
      Columns to be read are specified in indexes.
      The data read are returned as result of the procedure, the columns are
      the text ones.
      result[i] is read data if i in indexes, or NIL otherwise.
      All the resulting columns are text one *)
  VAR
    i,index,PointsCount:LONGINT;
    s:In.Scanner;
    result:Columns;
    col:ColumnPtr;
    str:String;
  BEGIN
    ASSERT(indexes#{});

    PointsCount:=LinesCount(FileName)-SkipLines;

    IF PointsCount<0 THEN
      RETURN NIL
    END;

    In.OpenFile(s,FileName);

    FOR i:=0 TO SkipLines-1 DO
      In.SkipLine(s);
    END;

    NEW(result);
    FOR i:=MIN(SET) TO MAX(SET) DO
      IF i IN indexes THEN
        NEW(col);
        StrList.Clear(col^);
        StrList.SetCapacity(col^,PointsCount);
        result[i]:=col;
      ELSE
        result[i]:=NIL
      END;
    END;

    FOR i:=0 TO PointsCount-1 DO
      FOR index:=MIN(SET) TO MAX(SET) DO
        IF index IN indexes THEN
          IF In.EndOfInput(s) THEN
            SU.RAISEError(DStr.Concat2(ModuleName+'.ReadText: too little lines in the file ',
                                       FileName)^);
          ELSIF In.EndOfLine(s) THEN
            SU.RAISEError(DStr.Concat2(ModuleName+'.ReadText: too little columns in the file ',
                                       FileName)^);
          END;
        END;
        ReadUntilSeparator(s,separator,str);
        IF index IN indexes THEN
          ASSERT(result[index]#NIL);
          col:=result[index];
          StrList.Add(col^,str^);
        END;
      END;
      In.SkipLine(s);
    END;

    In.Close(s);

    RETURN result
  END ReadText;

PROCEDURE ToFloat * (c-:Column): FloatArrayPtr;
  (** If all the strings of the column c contain real numbers, then the
      procedure converts the column to the array of REALs and returns it.
      Leading and trailing spaces are allowed.
      Otherwise it returns NIL. *)
  VAR
    i,count:LONGINT;
    res:FloatArrayPtr;
  BEGIN
    ASSERT(c.count>0);

    count:=c.count;
    NEW(res,count);
    i:=0;
    WHILE (i<count) & (DStr.TryToFloat(c.str[i]^,res[i])=DStr.OK) DO
      INC(i)
    END;

    IF i>=count THEN
      RETURN res
    ELSE
      RETURN NIL
    END;
  END ToFloat;

PROCEDURE WriteColumns * (c:Columns; FileName-:ARRAY OF CHAR; mode:Mode);
  (** Writes the given text columns to the specified file, using the specified
      separator between the columns. *)
  VAR
    f:Out.Formatter;
    i,j:LONGINT;
    col:ColumnPtr;
  BEGIN
    ASSERT(separator#NIL);
    ASSERT(c#NIL);

    Out.OpenFile(f,FileName,mode);

    FOR i:=0 TO Length(c)-1 DO
      FOR j:=0 TO LEN(c^)-1 DO
        IF c[j]#NIL THEN
          IF j#0 THEN
            Out.String(f,separator^,0);
          END;
          col:=c[j];
          Out.DString(f,col^.str[i]^,0);
        END;
      END;
      Out.Ln(f);
    END;

    Out.Close(f);
  END WriteColumns;

BEGIN
  separator:=DStr.Str(Tab);
END msysTextData.