(** This module contains a text scanner, which can read variables of different
    types from a byte stream, including text files.

    Last change: 14.04.2015 *)

<* O2EXTENSIONS + *>
<* NOOPTIMIZE - *>
MODULE msysTextIn;

IMPORT
  ByteStream:=msysStreamOfBytes, DStr:=msysDynStr, CSeq:=msysDynCharSeq,
  Math:=mMath, IntConv:=msysLongLongIntConv, RealConv:=msysLongLongRealConv,
  FileStream:=mhostStreamOfFile, SU:=msysUtils, COMPILER, RTS:=oberonRTS;

CONST
  CR * = 0DX; (** Carriage Return *)
  LF * = 0AX; (** Line Feed *)
  (** End of line mark is CR, LF, CR+LF, but not LF+CR *)

  ModuleName = COMPILER.EQUATION('module');

TYPE
  Reader = ByteStream.Reader;

  Scanner * = RECORD (** The scanner *)
    r - : Reader;       (** The reader of the scanner *)
    next : CHAR;         (* The next char in the stream, including end of line marks. *)
    NextActual: BOOLEAN; (* If the next char is actual *)
  END;

PROCEDURE ReadNextChar (VAR s:Scanner);
  BEGIN
    s.NextActual:= (s.r.Read(s.next,SIZE(s.next))=SIZE(s.next));
  END ReadNextChar;

PROCEDURE UpdateNextChar (VAR s:Scanner);
  BEGIN
    IF ~s.NextActual THEN
      ReadNextChar(s)
    END;
  END UpdateNextChar;

PROCEDURE IsSpace * (c:CHAR): BOOLEAN;
  (** Returns if the char is space, tab, end of line, or some other symbol
      which can be skipped until a number in the text stream *)
  BEGIN
    RETURN DStr.IsSpace(c)
  END IsSpace;

PROCEDURE ConnectTo * (VAR s:Scanner; r:Reader);
  (** Connects the scanner to the specified reader *)
  BEGIN
    ASSERT(r#NIL);

    s.r:=r;
    s.NextActual:=FALSE;
  END ConnectTo;

PROCEDURE EndOfInput * (VAR s:Scanner): BOOLEAN;
  (** Returns if there are no chars until end of input *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    RETURN ~s.NextActual;
  END EndOfInput;

PROCEDURE EOL (c:CHAR): BOOLEAN;
  (* Returns if c is end of line mark *)
  BEGIN
    RETURN (c=CR) OR (c=LF)
  END EOL;

PROCEDURE EndOfLine * (VAR s:Scanner): BOOLEAN;
  (** Returns if there are no chars until end of input or end of line mark *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    RETURN ~s.NextActual OR EOL(s.next)
  END EndOfLine;

PROCEDURE Assert (b:BOOLEAN; msg-:ARRAY OF CHAR);
  (* Asserts b, otherwise generates an exception with the specified message. *)
  BEGIN
    IF ~b THEN
      SU.RAISEError(msg);
    END;
  END Assert;

PROCEDURE SkipLine * (VAR s:Scanner);
  (** Skips rest of the line, if any, including the end of line mark *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    WHILE s.NextActual & ~EOL(s.next) DO
      ReadNextChar(s);
    END;
    IF s.NextActual THEN
      IF s.next=CR THEN
        ReadNextChar(s);
        IF s.NextActual & (s.next=LF) THEN
          s.NextActual:=FALSE
        END;
      ELSE
        s.NextActual:=FALSE
      END;
    END;
  END SkipLine;

PROCEDURE SkipSpaces * (VAR s:Scanner);
  (** Skips spaces (see IsSpace) until end of input or not space *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    WHILE s.NextActual & IsSpace(s.next) DO
      ReadNextChar(s)
    END;
  END SkipSpaces;

PROCEDURE SkipUntilSpaces * (VAR s:Scanner);
  (** Skips non-spaces until a space (see IsSpace) or end of line or
      end of input are reached. *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    WHILE s.NextActual & ~IsSpace(s.next) DO
      ReadNextChar(s)
    END;
  END SkipUntilSpaces;

PROCEDURE SkipSpacesUntilEOL * (VAR s:Scanner);
  (** Skips spaces (see IsSpace) until end of line mark or end of input or not space *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    WHILE s.NextActual & IsSpace(s.next) & ~EOL(s.next) DO
      ReadNextChar(s)
    END;
  END SkipSpacesUntilEOL;

PROCEDURE SkipUntilStr * (VAR s:Scanner; str-:ARRAY OF CHAR);
  (** Skips chars until str is met. If the str is met then it is also skipped.
      str is a dynamical string, see the module msysDynStr.
      str can not be empty. *)
  VAR
    i,len:LONGINT;
  BEGIN
    ASSERT(s.r#NIL);
    len:=DStr.Length(str);
    ASSERT(len>0);

    UpdateNextChar(s);
    REPEAT
      WHILE s.NextActual & (s.next#str[0]) DO
        ReadNextChar(s);
      END;
      i:=0;
      WHILE (i<len) & s.NextActual & (s.next=str[i]) DO
        ReadNextChar(s);
        INC(i);
      END;
    UNTIL ~s.NextActual OR (i=len)
  END SkipUntilStr;

PROCEDURE Char * (VAR s:Scanner; VAR c:CHAR);
  (** Reads the char *)
  BEGIN
    ASSERT(s.r#NIL);

    UpdateNextChar(s);
    Assert(s.NextActual,ModuleName+'.Char: end of input');
    c:=s.next;
    s.NextActual:=FALSE;
  END Char;

PROCEDURE NextChar * (VAR s:Scanner; VAR c:CHAR): BOOLEAN;
  (** Reads the char not removing it from the stream.
      Returns if the char was successfully read. *)
  BEGIN
    UpdateNextChar(s);
    IF s.NextActual THEN
      c:=s.next
    END;
    RETURN s.NextActual;
  END NextChar;

PROCEDURE String * (VAR s:Scanner; VAR str:ARRAY OF CHAR);
  (** Reads the string until end of line, or the str is full.
      If there is enough space in str, adds trailing 0X to the result.
      If the end of line is reached, the end of line mark is not skipped. *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(s.r#NIL);

    i:=0;
    UpdateNextChar(s);
    WHILE (i<LEN(str)) & s.NextActual & ~EOL(s.next) DO
      str[i]:=s.next;
      INC(i);
      ReadNextChar(s);
    END;
    IF i<LEN(str) THEN
      str[i]:=0X
    END;
  END String;

PROCEDURE DString * (VAR s:Scanner; VAR str:DStr.String);
  (** Reads the string until end of line.
      If the end of line is reached, the end of line mark is not skipped. *)
  VAR
    sm:CSeq.Sequence;
  BEGIN
    ASSERT(s.r#NIL);

    CSeq.Init(sm);
    UpdateNextChar(s);
    WHILE s.NextActual & ~EOL(s.next) DO
      CSeq.AddChar(sm,s.next);
      ReadNextChar(s);
    END;
    str:=CSeq.String(sm);
  END DString;

PROCEDURE ^ LongLongInt (VAR s:Scanner; VAR result:LONGLONGINT);

PROCEDURE Int * (VAR s:Scanner; VAR result:LONGINT);
  (** Reads the integer number, skeeping leading spaces (see IsSpace) *)
  VAR
    i:LONGLONGINT;
  BEGIN
    LongLongInt(s,i);
    Assert((i>=MIN(LONGINT)) & (i<=MAX(LONGINT)),ModuleName+'.Int: out of range');
    result:=SHORT(i);
  END Int;

TYPE
  IntProc = RECORD (CSeq.Proc)
    status:IntConv.Status;
    result:LONGLONGINT;
  END;

PROCEDURE (VAR p:IntProc) Do (s-:ARRAY OF CHAR; len:LONGINT);
  BEGIN
    IF len=0 THEN
      p.status:=IntConv.NotValid
    ELSE
      p.status:=IntConv.FromString(s,0,len,p.result);
    END;
  END Do;

PROCEDURE LongLongInt * (VAR s:Scanner; VAR result:LONGLONGINT);
  (** Reads the integer number, skeeping leading spaces (see IsSpace) *)
  VAR
    sc:IntConv.Scanner;
    seq:CSeq.Sequence;
    p:IntProc;
  BEGIN
    ASSERT(s.r#NIL);

    SkipSpaces(s); (* SkipSpaces updates the s.next *)
    Assert(s.NextActual,ModuleName+'.LongLongInt: end of input');
    IntConv.Init(sc);
    CSeq.Init(seq);

    WHILE s.NextActual & IntConv.AddChar(sc,s.next) DO
      CSeq.AddChar(seq,s.next);
      ReadNextChar(s)
    END;

    CSeq.FormString(seq,p);
    Assert(p.status#IntConv.OutOfRange,ModuleName+'.LongLongInt: out of range');
    Assert(p.status#IntConv.NotValid,ModuleName+'.LongLongInt: not valid number');
    result:=p.result;
  END LongLongInt;

PROCEDURE ^ LongLongReal (VAR s:Scanner; VAR result:LONGLONGREAL);

PROCEDURE Real * (VAR s:Scanner; VAR result:REAL);
  (** Reads the real number, skeeping leading spaces (see IsSpace) *)
  VAR
    x:LONGLONGREAL;
  BEGIN
    LongLongReal(s,x);
    Assert((x>=MIN(REAL)) & (x<=MAX(REAL)),ModuleName+'.Real: out of range');
    result:=VAL(REAL,x);
  END Real;

PROCEDURE LongReal * (VAR s:Scanner; VAR result:LONGREAL);
  (** Reads the real number, skeeping leading spaces (see IsSpace) *)
  VAR
    x:LONGLONGREAL;
  BEGIN
    LongLongReal(s,x);
    Assert((x>=MIN(LONGREAL)) & (x<=MAX(LONGREAL)),ModuleName+'.LongReal: out of range');
    result:=VAL(LONGREAL,x);
  END LongReal;

TYPE
  RealProc = RECORD (CSeq.Proc)
    status:RealConv.Status;
    result:LONGLONGREAL;
  END;

PROCEDURE (VAR p:RealProc) Do (s-:ARRAY OF CHAR; len:LONGINT);
  BEGIN
    IF len=0 THEN
      p.status:=RealConv.NotValid
    ELSE
      p.status:=RealConv.FromString(s,0,len,p.result);
    END;
  END Do;

PROCEDURE LongLongReal * (VAR s:Scanner; VAR result:LONGLONGREAL);
  (** Reads the real number, skeeping leading spaces (see IsSpace) *)
  VAR
    sc:RealConv.Scanner;
    seq:CSeq.Sequence;
    p:RealProc;
  BEGIN
    ASSERT(s.r#NIL);

    SkipSpaces(s); (* SkipSpaces updates the s.next *)
    Assert(s.NextActual,ModuleName+'.LongLongReal: end of input');
    RealConv.Init(sc);
    CSeq.Init(seq);

    WHILE s.NextActual & RealConv.AddChar(sc,s.next) DO
      CSeq.AddChar(seq,s.next);
      ReadNextChar(s)
    END;

    CSeq.FormString(seq,p);
    Assert(p.status#RealConv.OutOfRange,ModuleName+'.LongLongReal: out of range');
    Assert(p.status#RealConv.NotValid,ModuleName+'.LongLongReal: not valid number');
    result:=p.result;
  END LongLongReal;

PROCEDURE Float * (VAR s:Scanner; VAR result:Math.Float);
  (** Reads the real number, skeeping leading spaces (see IsSpace) *)
  BEGIN

   <* IF Float = "REAL" THEN *>
     Real(s,result)
   <* ELSIF Float = "LONGREAL" THEN *>
     LongReal(s,result)
   <* ELSIF Float = "LONGLONGREAL" THEN *>
     LongLongReal(s,result)
   <* ELSE *>
     'Equation Float is not defined
   <* END *>

  END Float;

PROCEDURE OpenFile * (VAR s:Scanner; FileName-:ARRAY OF CHAR);
  (** Opens the existing file and connects the scanner to it *)
  VAR
    r:Reader;
  BEGIN
    r:=FileStream.NewReader(FileName);
    Assert(r#NIL,DStr.Concat3(ModuleName+'.OpenFile: can not open the file "',FileName,'"')^);
    ConnectTo(s,r);
  END OpenFile;

PROCEDURE Close * (VAR s:Scanner);
  (** Closes the scanner *)
  BEGIN
    s.r:=NIL;
    RTS.Collect(); (* call of s.r.Finalizer, if possible *)
  END Close;

END msysTextIn.