(** This module contains a text formatter, which can write variables of different
    types to a byte stream, including text files.

    Last change: 27.05.2015 *)

<* O2EXTENSIONS + *>
MODULE msysTextOut;

IMPORT
  ByteStream:=msysStreamOfBytes, DStr:=msysDynStr, IntConv:=msysLongLongIntConv,
  RealConv:=msysLongLongRealConv, FileStream:=mhostStreamOfFile, RTS:=oberonRTS,
  SU:=msysUtils, COMPILER;

CONST
  Eoln * = ''+0DX+0AX; (** End of line mark *)
  Tab * = ''+09X;      (** Tab *)

  ModuleName = COMPILER.EQUATION('Module');

TYPE
  Writer = ByteStream.Writer;

  Formatter * = RECORD (** The formatter *)
    w - : Writer;      (** The connected writer *)
  END;

PROCEDURE Assert (b:BOOLEAN; msg-:ARRAY OF CHAR);
  (* Asserts b, otherwise generates an exception with the specified message. *)
  BEGIN
    IF ~b THEN
      SU.RAISEError(msg);
    END;
  END Assert;

PROCEDURE ConnectTo * (VAR f:Formatter; w:Writer);
  (** Connects the formater to the specifier writer *)
  BEGIN
    ASSERT(w#NIL);

    f.w:=w;
  END ConnectTo;

PROCEDURE Char * (VAR f:Formatter; c:CHAR);
  (** Writes the char *)
  VAR
    OK:BOOLEAN;
  BEGIN
    ASSERT(f.w#NIL);

    OK:= (f.w.Write(c,SIZE(CHAR))=SIZE(CHAR));
    Assert(OK,ModuleName+'.Char: can not output');
  END Char;

PROCEDURE WriteString (VAR f:Formatter; str-:ARRAY OF CHAR; count,MinLength:LONGINT);
  (* Writes count chars of the string. MinLength is minimal length;
     leading spaces are added if necessary. *)
  VAR
    OK:BOOLEAN;
  BEGIN
    ASSERT(count<=LEN(str));

    WHILE MinLength>count DO
      Char(f,' ');
      DEC(MinLength);
    END;
    OK:= (f.w.Write(str,count*SIZE(CHAR))=count*SIZE(CHAR));
    Assert(OK,ModuleName+'.WriteString: can not output');
  END WriteString;

PROCEDURE String * (VAR f:Formatter; str-:ARRAY OF CHAR; MinLength:LONGINT);
  (** Writes null-terminated string. MinLength is minimal length;
      leading spaces are added if necessary. *)
  VAR
    len:LONGINT;
  BEGIN
    ASSERT(f.w#NIL);

    len:=0;
    WHILE (len<LEN(str)) & (str[len]#0X) DO
      INC(len)
    END;

    WriteString(f,str,len,MinLength)
  END String;

PROCEDURE DString * (VAR f:Formatter; str-:ARRAY OF CHAR; MinLength:LONGINT);
  (** Writes dybamical string. MinLength is minimal length;
      leading spaces are added if necessary. *)
  BEGIN
    ASSERT(f.w#NIL);

    WriteString(f,str,DStr.Length(str),MinLength)
  END DString;

PROCEDURE Ln * (VAR f:Formatter);
  (** Writes end of line mark to the channel *)
  BEGIN
    WriteString(f,Eoln,LEN(Eoln)-1,0);
  END Ln;

PROCEDURE Int * (VAR f:Formatter; value:LONGLONGINT; MinLength:LONGINT);
  (** Writes the integer value. MinLength is minimal length;
      leading spaces are added if necessary. *)
  VAR
    buf:IntConv.String;
    size:LONGINT;
  BEGIN
    size:=IntConv.ToString(value,buf);
    WriteString(f,buf,size,MinLength);
  END Int;

PROCEDURE ^ RealEx (VAR f:Formatter; value:LONGLONGREAL; Digits,MinLength:LONGINT);

PROCEDURE Real * (VAR f:Formatter; value:LONGLONGREAL; MinLength:LONGINT);
  (** Writes real value. MinLength is minimal length;
      leading spaces are added if necessary. *)
  BEGIN
    RealEx(f,value,DStr.DefaultDigits,MinLength);
  END Real;

PROCEDURE RealEx * (VAR f:Formatter; value:LONGLONGREAL; digits,MinLength:LONGINT);
  (** Writes real value with specified digits of significand part.
      MinLength is minimal length; leading spaces are added if necessary. *)
  VAR
    buf:RealConv.String;
    len:LONGINT;
  BEGIN
    len:=RealConv.ToString(value,digits,buf);
    WriteString(f,buf,len,MinLength);
  END RealEx;

PROCEDURE RealFixed * (VAR f:Formatter; value:LONGLONGREAL; digits,MinLength:LONGINT);
  (** Writes real value with specified digits of significand part in fixed format.
      MinLength is minimal length; leading spaces are added if necessary. *)
  VAR
    buf:ARRAY 32 OF CHAR;
    res:DStr.String;
    len:LONGINT;
  BEGIN
    len:=RealConv.ToStringFixed(value,digits,buf);
    IF len<=LEN(buf) THEN
      WriteString(f,buf,len,MinLength);
    ELSE
      NEW(res,len);
      len:=RealConv.ToStringFixed(value,digits,res^);
      WriteString(f,res^,len,MinLength);
    END;
  END RealFixed;

PROCEDURE Bool * (VAR f:Formatter; value:BOOLEAN; MinLength:LONGINT);
  (** Writes boolean value. MinLength is minimal length;
      leading spaces are added if necessary. *)
  BEGIN
    IF value THEN
      DString(f,'TRUE',MinLength)
    ELSE
      DString(f,'FALSE',MinLength)
    END;
  END Bool;

PROCEDURE Set * (VAR f:Formatter; s:SET; MinLength:LONGINT);
  (** Writes set value to the channel. MinLength is minimal length;
      leading spaces are added if necessary. *)
  CONST
    LastBit = 31;
    MaxStrLen = 22*2 + 10 + 31 + 2 + 1;
  VAR
    i,ind:LONGINT;
    str:ARRAY MaxStrLen OF CHAR;
  PROCEDURE WriteChar (c:CHAR);
    BEGIN
      ASSERT(ind<LEN(str));

      str[ind]:=c;
      INC(ind);
    END WriteChar;
  PROCEDURE WriteInt (i:LONGINT);
    VAR
      buf:IntConv.String;
      size,j:LONGINT;
    BEGIN
      ASSERT(i>=0);
      ASSERT(i<=99);

      size:=IntConv.ToString(i,buf);
      FOR j:=0 TO size-1 DO
        WriteChar(buf[j]);
      END;
    END WriteInt;
  BEGIN
    ind:=0;

    WriteChar('{');
    i:=0;
    WHILE (i<=LastBit) & ~(i IN s) DO
      INC(i);
    END;
    IF i<=LastBit THEN
      WriteInt(i);
      INC(i);
      WHILE i<=LastBit DO
        IF i IN s THEN
          WriteChar(',');
          WriteInt(i);
        END;
        INC(i);
      END;
    END;
    WriteChar('}');

    WriteString(f,str,ind,MinLength);
  END Set;

TYPE
  Mode * = FileStream.Mode;

CONST
  Rewrite * = FileStream.Rewrite;
  Apend * = FileStream.Append;

PROCEDURE OpenFile * (VAR f:Formatter; FileName-:ARRAY OF CHAR; mode:Mode);
  (** Rewrites or appends to the specified file, and connects it to the formatter. *)
  VAR
    w:Writer;
  BEGIN
    w:=FileStream.NewWriter(FileName,mode);
    ConnectTo(f,w);
  END OpenFile;

PROCEDURE FlushFile * (VAR f:Formatter);
  (** Flushes the file. This procedure can only be used with FileStream writers. *)
  BEGIN
    ASSERT(f.w#NIL);
    ASSERT(f.w IS FileStream.Writer);

    FileStream.Flush(f.w(FileStream.Writer));
  END FlushFile;

PROCEDURE Close * (VAR f:Formatter);
  (** Closes the formatter *)
  BEGIN
    f.w:=NIL;
    RTS.Collect(); (* call f.w.Finalize if possible *)
  END Close;

END msysTextOut.