(** This module provides tools to organize "symmetrical" multithreading.
    "Symmetrical" means that a procedure is executed in threads simultaneously.

    Last changed: 27.03.2015 *)
    
<* O2EXTENSIONS + *>
<* O2ADDKWD + *>
MODULE mthrSym;

IMPORT
  Processes, SU:=msysUtils, HU:=mhostUtils, COMPILER;

IMPORT
  SEM:=Semaphores;

(*
IMPORT
  SEM:=mthrSpinlockSemaphores; (* For spinlock semaphores *)
*)

VAR
  Count - : LONGINT; (** Total count of used threads, including the main thread.
                         Default this is number of threads suppotred by the processor.
                         The count may be changed by procedure SetCount. *)

CONST
  StackSize * = 1024*1024; (** Stack size of each additional thread. *)
  Urgency * = 0;           (** Urgency of additional threads. *)

  ModuleName = COMPILER.EQUATION('module');

TYPE
  Proc * = RECORD
  END;
  (** This is a procedure type for execution. *)

PROCEDURE (VAR p:Proc) Do * ();
  BEGIN
    SU.RAISEError(ModuleName+'.Proc.Do: call of abstract method');
  END Do;

VAR

<* PUSH *>
<* VOLATILE + *>
  executing - : BOOLEAN; (** TRUE if the threads are executing *)
  s:POINTER TO ARRAY OF SEM.SEMAPHORE; (* Critical Sections. s[0] drives the main thread. *)
  procedure:POINTER TO Proc;
  finish:BOOLEAN;
<* POP *>

PROCEDURE^ Index (): LONGINT;

PROCEDURE Exec * (proc:POINTER TO Proc);
  (** This procedure executes ThreadBody in Count threads simultaneously,
      including the calling thread. *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(proc#NIL);
    ASSERT(~executing);

    IF Index()#0 THEN
      SU.RAISEError(ModuleName+'.Exec: the calling thread is not the main one.');
    END;

    executing:=TRUE;
    procedure:=proc;

    (* Start all additional threads *)
    FOR i:=1 TO Count-1 DO
      SEM.Release(s[i]);
    END;

    procedure.Do();

    (* Wait for all additional threads *)
    FOR i:=1 TO Count-1 DO
      SEM.Claim(s[0]);
    END;

    executing:=FALSE;
    procedure:=NIL;
  END Exec;

PROCEDURE MainBody;
  VAR
    ThrInd:LONGINT;
  BEGIN
    ThrInd:=Index();

    (* Suspent the thread after start *)
    SEM.Claim(s[ThrInd]);

    WHILE ~finish DO
      procedure.Do();
      SEM.Release(s[0]);    (* Signal to the main thread that I'm finished *)
      SEM.Claim(s[ThrInd]); (* Wait for the next task *)
    END;
  END MainBody;

PROCEDURE Index * (): LONGINT;
  (** Returns index of the calling thread (0..Count-1).
      The main thread has index 0, additional threads are
      indexed in 1..Count-1. *)
  BEGIN
    RETURN Processes.MyParam();
  END Index;

PROCEDURE StartThreads ();
  VAR
    p:Processes.ProcessId;
    i:LONGINT;
  BEGIN
    ASSERT(~executing);

    NEW(s,Count);
    FOR i:=0 TO Count-1 DO
      SEM.Create(s[i],0);
    END;

    finish:=FALSE;
    FOR i:=1 TO Count-1 DO
      Processes.Start(MainBody,StackSize,Urgency,i,p);
      (* Index of the thread is passed as parameter of the thread *)
    END;
  END StartThreads;

PROCEDURE StopThreads ();
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(~executing);
    ASSERT(s#NIL);

    finish:=TRUE;
    FOR i:=1 TO Count-1 DO
      SEM.Release(s[i]);
    END;
    FOR i:=0 TO Count-1 DO
      SEM.Destroy(s[i]);
    END;
    s:=NIL;
  END StopThreads;

PROCEDURE SetCount * (count:LONGINT);
  (** Sets count of threads *)
  BEGIN
    ASSERT(count>0);
    ASSERT(~executing);

    IF s#NIL THEN
      StopThreads;
    END;
    Count:=count;
    StartThreads;
  END SetCount;

BEGIN
  IF Index()#0 THEN
    SU.RAISEError(ModuleName+': the main thread is not zero-indexed.');
  END;

  executing:=FALSE;
  SetCount(HU.GetThreadNumber())
FINALLY
  StopThreads();
END mthrSym.