(** This module implements interpolation and derivatives of the 2D arrays
    taking into accound geometry of the source *)

<* O2EXTENSIONS + *>
MODULE GeoArea;

IMPORT
  Geo:=Geometry, Ar:=mArea2D, Math:=mMath, ComFunc:=mCommonFunctions;

TYPE
  Float = Math.Float;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;

  Area * = Ar.Area;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE XIn (a-:Area; x:Float): BOOLEAN;
  (* Asserts that x is in the area and returns TRUE *)
  BEGIN
    ASSERT(x>=a.MinX);
    ASSERT(x<=a.MaxX);

    RETURN TRUE
  END XIn;

PROCEDURE YIn (a-:Area; y:Float): BOOLEAN;
  (* Asserts that y is in the area and returns TRUE *)
  BEGIN
    ASSERT(y>=a.MinY);
    ASSERT(y<=a.MaxY);

    RETURN TRUE
  END YIn;

PROCEDURE XYIn (a-:Area; x,y:Float): BOOLEAN;
  (* Asserts that (x,y) is in the area and returns TRUE *)
  BEGIN
    ASSERT(XIn(a,x));
    ASSERT(YIn(a,y));

    RETURN TRUE
  END XYIn;

PROCEDURE X (a-:Area; iX:LONGINT): Float;
  BEGIN
    RETURN Ar.IndToX(a,iX);
  END X;

PROCEDURE Y (a-:Area; iY:LONGINT): Float;
  BEGIN
    RETURN Ar.IndToY(a,iY);
  END Y;

CONST
  PointsLen = 3;

TYPE
  IntPoints = ARRAY PointsLen OF LONGINT;
  FloatPoints = ARRAY PointsLen OF Float;

PROCEDURE InterpolateBy3Points (area-:Area; iX-,iY-:IntPoints; z-:FloatPoints;
                                x,y:Float): Float;
  VAR
    z1,z2,z3,a,b:Float;
    iX1,iY1,iX2,iY2,iX3,iY3,denom:LONGINT;
  BEGIN
    iX1:=iX[0];
    iY1:=iY[0];
    iX2:=iX[1];
    iY2:=iY[1];
    iX3:=iX[2];
    iY3:=iY[2];
    z1:=z[0];
    z2:=z[1];
    z3:=z[2];

    denom:=(iY2-iY1)*(iX3-iX1) - (iY3-iY1)*(iX2-iX1);
    ASSERT(denom#0); (* or else all the 3 points are in one line *)
    a:= -((z2-z1)*(iY3-iY1) - (z3-z1)*(iY2-iY1))/(area.dx*denom);
    b:=  ((z2-z1)*(iX3-iX1) - (z3-z1)*(iX2-iX1))/(area.dy*denom);
    RETURN z1 + a*(x-X(area,iX1)) + b*(y-Y(area,iY1))
  END InterpolateBy3Points;

PROCEDURE InterpolateBy2Points (area-:Area; VAR iX,iY:IntPoints; VAR z:FloatPoints;
                                x,y:Float): Float;
  BEGIN
    iX[2]:=iX[0]+iY[1]-iY[0];
    iY[2]:=iY[0]-(iX[1]-iX[0]);
    z[2]:=z[0];
    RETURN InterpolateBy3Points(area,iX,iY,z,x,y);
  END InterpolateBy2Points;

TYPE
  Analyzer = RECORD
    iX,iY:IntPoints;
    z:FloatPoints;
    a:Area;
    num:LONGINT;
    passed:PROCEDURE (x,y:Float): BOOLEAN;
  END;

PROCEDURE AddPoint (VAR a:Analyzer; iX,iY:LONGINT; field-:FloatArray2);
  VAR
    ind:LONGINT;
  BEGIN
    ASSERT(a.num>=0);
    ASSERT(a.passed#NIL);
    ASSERT(Ar.Correct(a.a));
    ASSERT(Ar.CorrectField(a.a,field));

    IF   (iX>=0) & (iX<=a.a.LastX) & (iY>=0) & (iY<=a.a.LastY)
       & a.passed(X(a.a,iX),Y(a.a,iY))
    THEN
      ASSERT(a.num<PointsLen);
      ind:=a.num;
      a.iX[ind]:=iX;
      a.iY[ind]:=iY;
      a.z[ind]:=field[iX,iY];
      INC(a.num);
    END;
  END AddPoint;

PROCEDURE AnalyzeAndInterpolate1Point (VAR a:Analyzer; iX0,iY0:LONGINT;
                                       field-:FloatArray2; x,y:Float): Float;
  VAR
    iX1,iY1:LONGINT;
  BEGIN
    ASSERT(a.num=1);
    iX1:=a.iX[0];
    iY1:=a.iY[0];
    AddPoint( a, iX1, 3*iY1-2*iY0-1, field );
    AddPoint( a, 3*iX1-2*iX0-1 , iY1, field);
    CASE a.num OF
      2: RETURN InterpolateBy2Points (a.a,a.iX,a.iY,a.z,x,y)
    | 3: RETURN InterpolateBy3Points (a.a,a.iX,a.iY,a.z,x,y)
    END;
  END AnalyzeAndInterpolate1Point;

PROCEDURE AnalyzeAndInterpolate2Points (VAR a:Analyzer; iX0,iY0:LONGINT;
                                        field-:FloatArray2; x,y:Float): Float;
  PROCEDURE Add (ind:LONGINT);
    BEGIN
      AddPoint( a, a.iX[ind]+a.iX[0]+a.iX[1]-2*iX0-1, a.iY[ind]+a.iY[0]+a.iY[1]-2*iY0-1, field );
    END Add;
  VAR
    ind:LONGINT;
    x1,y1,x2,y2:Float;
  BEGIN
    ASSERT(a.num=2);

    x1:=X(a.a,a.iX[0]);
    y1:=X(a.a,a.iY[0]);
    x2:=X(a.a,a.iX[1]);
    y2:=X(a.a,a.iY[1]);
    IF sqr(x-x1)+sqr(y-y1) > sqr(x-x2)+sqr(y-y2) THEN
      ind:=1
    ELSE
      ind:=0;
    END;
    Add(ind);
    IF a.num=2 THEN
      ind:=(ind+1) MOD 2;
      Add(ind);
    END;
    CASE a.num OF
      2: RETURN InterpolateBy2Points (a.a,a.iX,a.iY,a.z,x,y)
    | 3: RETURN InterpolateBy3Points (a.a,a.iX,a.iY,a.z,x,y)
    END;
  END AnalyzeAndInterpolate2Points;

PROCEDURE IsIn (x,y:Float): BOOLEAN;
  BEGIN
    RETURN Geo.GetSpaceType(x,y)=Geo.In;
  END IsIn;

PROCEDURE IsNotIn (x,y:Float): BOOLEAN;
  BEGIN
    RETURN Geo.GetSpaceType(x,y)#Geo.In;
  END IsNotIn;

PROCEDURE InterpolateWithAnalysis (area-:Area; field-:FloatArray2;
                                   iX,iY:LONGINT; x,y:Float): Float;
  VAR
    a:Analyzer;
  BEGIN
    a.a:=area;
    a.num:=0;
    IF Geo.GetSpaceType(x,y)=Geo.In THEN
      a.passed:=IsIn
    ELSE
      a.passed:=IsNotIn
    END;

    AddPoint(a,iX,iY,field);
    AddPoint(a,iX+1,iY,field);
    AddPoint(a,iX,iY+1,field);
    AddPoint(a,iX+1,iY+1,field);

    CASE a.num OF
      1: RETURN AnalyzeAndInterpolate1Point(a,iX,iY,field,x,y)
    | 2: RETURN AnalyzeAndInterpolate2Points(a,iX,iY,field,x,y)
    | 3: RETURN InterpolateBy3Points(area,a.iX,a.iY,a.z,x,y);
    END;
  END InterpolateWithAnalysis;

PROCEDURE CoordToGrid(a-:Area; VAR x,y:Float; VAR iX,iY:LONGINT);
  BEGIN
    x:=(x-a.MinX)*a.InvDx;
    y:=(y-a.MinY)*a.InvDy;
    iX:=ENTIER(x);
    iY:=ENTIER(y);
    x:=x-iX;
    y:=y-iY;
    IF iX=a.LastX THEN
      iX:=a.LastX-1;
      x:=1;
    END;
    IF iY=a.LastY THEN
      iY:=a.LastY-1;
      y:=1;
    END;
  END CoordToGrid;

PROCEDURE Interpolate * (a-:Area; field-:FloatArray2; x,y:Float): Float;
  VAR
    iX,iY:LONGINT;
    z0,z1:Float;
    x0,y0:Float;
    passed:PROCEDURE (x,y:Float): BOOLEAN;
  BEGIN
    ASSERT(Ar.Correct(a));
    ASSERT(Ar.CorrectField(a,field));
    ASSERT(XYIn(a,x,y));

    x0:=x;
    y0:=y;
    CoordToGrid(a,x,y,iX,iY);
    IF Geo.GetSpaceType(X(a,iX),Y(a,iY))=Geo.In THEN
      passed:=IsIn
    ELSE
      passed:=IsNotIn
    END;

    IF   passed(X(a,iX+1),Y(a,iY))
       & passed(X(a,iX),Y(a,iY+1))
       & passed(X(a,iX+1),Y(a,iY+1))
    THEN
      z0:=field[iX,iY] + x*(field[iX+1,iY] - field[iX,iY]);
      z1:=field[iX,iY+1] + x*(field[iX+1,iY+1] - field[iX,iY+1]);
      RETURN z0 + y*(z1-z0)
    ELSE
      RETURN InterpolateWithAnalysis(a,field,iX,iY,x0,y0);
    END;
  END Interpolate;

TYPE
  InterpolationProc * = RECORD (ComFunc.Float2)
    (** The interpolation takes place with specified area and fields *)
    area - :Area;
    field - :FloatArray2Ptr;
  END;

PROCEDURE (VAR d:InterpolationProc) f * (x,y:Float): Float;
  BEGIN
    IF x<d.area.MinX THEN
      x:=d.area.MinX
    END;
    IF x>d.area.MaxX THEN
      x:=d.area.MaxX
    END;
    IF y<d.area.MinY THEN
      y:=d.area.MinY
    END;
    IF y>d.area.MaxY THEN
      y:=d.area.MaxY
    END;
    RETURN Interpolate(d.area,d.field^,x,y);
  END f;

PROCEDURE Interpolation * (area-:Area; field:FloatArray2Ptr): POINTER TO InterpolationProc;
  VAR
    d:POINTER TO InterpolationProc;
  BEGIN
    ASSERT(Ar.Correct(area));
    ASSERT(field#NIL);
    ASSERT(Ar.CorrectField(area,field^));

    NEW(d);
    d.area:=area;
    d.field:=field;
    RETURN d;
  END Interpolation;

PROCEDURE IsInInd (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    RETURN Geo.GetSpaceType(X(a,iX),Y(a,iY))=Geo.In;
  END IsInInd;

PROCEDURE IsNotInInd (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    RETURN Geo.GetSpaceType(X(a,iX),Y(a,iY))#Geo.In;
  END IsNotInInd;

PROCEDURE DiffZInd * (a-:Area; field-:FloatArray2; iX,iY:LONGINT): Float;
  VAR
    passed:PROCEDURE (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    ASSERT(Ar.Correct(a));
    ASSERT(Ar.CorrectField(a,field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    IF IsInInd(a,iX,iY) THEN
      passed:=IsInInd
    ELSE
      passed:=IsNotInInd
    END;

    IF ~passed(a,iX+1,iY) THEN
      ASSERT(iX>=2);
      ASSERT(passed(a,iX-1,iY));
      ASSERT(passed(a,iX-2,iY));
      RETURN Ar.DiffXIndLeft(a,field,iX,iY)
    ELSIF ~passed(a,iX-1,iY) THEN
      ASSERT(iX<=a.LastX-2);
      ASSERT(passed(a,iX+1,iY));
      ASSERT(passed(a,iX+2,iY));
      RETURN Ar.DiffXIndRight(a,field,iX,iY)
    ELSE
      RETURN Ar.DiffXInd(a,field,iX,iY)
    END;
  END DiffZInd;

PROCEDURE DiffRInd * (a-:Area; field-:FloatArray2; iX,iY:LONGINT): Float;
  VAR
    passed:PROCEDURE (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    ASSERT(Ar.Correct(a));
    ASSERT(Ar.CorrectField(a,field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    IF IsInInd(a,iX,iY) THEN
      passed:=IsInInd
    ELSE
      passed:=IsNotInInd
    END;

    IF ~passed(a,iX,iY+1) THEN
      ASSERT(iY>=2);
      ASSERT(passed(a,iX,iY-1));
      ASSERT(passed(a,iX,iY-2));
      RETURN Ar.DiffYIndLeft(a,field,iX,iY)
    ELSIF ~passed(a,iX,iY-1) THEN
      ASSERT(iY<=a.LastY-2);
      ASSERT(passed(a,iX,iY+1));
      ASSERT(passed(a,iX,iY+2));
      RETURN Ar.DiffYIndRight(a,field,iX,iY)
    ELSE
      RETURN Ar.DiffYInd(a,field,iX,iY)
    END;
  END DiffRInd;

PROCEDURE Diff2ZInd * (a-:Area; field-:FloatArray2; iX,iY:LONGINT): Float;
  VAR
    passed:PROCEDURE (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    ASSERT(Ar.Correct(a));
    ASSERT(Ar.CorrectField(a,field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    IF IsInInd(a,iX,iY) THEN
      passed:=IsInInd
    ELSE
      passed:=IsNotInInd
    END;

    IF ~passed(a,iX+1,iY) THEN
      ASSERT(iX>=3);
      ASSERT(passed(a,iX-1,iY));
      ASSERT(passed(a,iX-2,iY));
      ASSERT(passed(a,iX-3,iY));
      RETURN Ar.Diff2XIndLeft(a,field,iX,iY)
    ELSIF ~passed(a,iX-1,iY) THEN
      ASSERT(iX<=a.LastX-3);
      ASSERT(passed(a,iX+1,iY));
      ASSERT(passed(a,iX+2,iY));
      ASSERT(passed(a,iX+3,iY));
      RETURN Ar.Diff2XIndRight(a,field,iX,iY)
    ELSE
      RETURN Ar.Diff2XInd(a,field,iX,iY)
    END;
  END Diff2ZInd;

PROCEDURE Diff2RInd * (a-:Area; field-:FloatArray2; iX,iY:LONGINT): Float;
  VAR
    passed:PROCEDURE (a-:Area; iX,iY:LONGINT): BOOLEAN;
  BEGIN
    ASSERT(Ar.Correct(a));
    ASSERT(Ar.CorrectField(a,field));
    ASSERT(iX>=0);
    ASSERT(iX<=a.LastX);
    ASSERT(iY>=0);
    ASSERT(iY<=a.LastY);

    IF IsInInd(a,iX,iY) THEN
      passed:=IsInInd
    ELSE
      passed:=IsNotInInd
    END;

    IF ~passed(a,iX,iY+1) THEN
      ASSERT(iY>=3);
      ASSERT(passed(a,iX,iY-1));
      ASSERT(passed(a,iX,iY-2));
      ASSERT(passed(a,iX,iY-3));
      RETURN Ar.Diff2YIndLeft(a,field,iX,iY)
    ELSIF ~passed(a,iX,iY-1) THEN
      ASSERT(iY<=a.LastY-3);
      ASSERT(passed(a,iX,iY+1));
      ASSERT(passed(a,iX,iY+2));
      ASSERT(passed(a,iX,iY+3));
      RETURN Ar.Diff2YIndRight(a,field,iX,iY)
    ELSE
      RETURN Ar.Diff2YInd(a,field,iX,iY)
    END;
  END Diff2RInd;

END GeoArea.