(** This module contains geometry of the simulations and all the parameters,
    which are specific for this geometry. *)

<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE SPECTROFasselGeometry;

IMPORT
  Math:=mMath, PHC:=msciPhConst, COMPILER, Protocol, TOut:=msysTextOut;

CONST
  pi = PHC.pi;
  cm = PHC.cm;
  mm = PHC.mm;

  TubeIntR0 * = 2.5*mm/2; (** internal radius of internal alumna tube *)
  TubeIntR1 * = 5*mm/2; (** external radius of internal alumna tube *)
  TubeIntLen * = 1*cm; (** length of internal alumna tube *)
  TubeMidR0 * = 13*mm/2; (** internal radius of middle quartz tube *)
  TubeMidR1 * = 15*mm/2; (** external radius of middle quartz tube *)
  TubeMidLen * = TubeIntLen + 1*mm; (** length of middle quartz tube *)
  TubeOutR0 * = 17*mm/2; (** internal radius of outer quartz tube *)
  TubeOutR1 * = 19*mm/2; (** external radius of outer quartz tube *)
  TubeOutLen * = TubeMidLen + 31*mm; (** length of outer quartz tube *)

  Coilr = 5*mm/2; (* radius of the coil wire *)
  CoilR = 30.5*mm/2; (* radius of the middle of the coil turns *)
  CoilPeriod = 2*Coilr + 2.3*mm; (* spatial period of coil turns *)
  CoilZ0 = TubeMidLen + 3*mm + Coilr; (* z of the center of the upstream turn *)
  CoilZ1 * = CoilZ0 + CoilPeriod; (** z of the center of the central turn *)
  CoilZ2 = CoilZ1 + CoilPeriod; (* z of the center of the downstream turn *)

  CoilCrossSectionArea * = pi*Coilr*Coilr; (** Area of cross-section of the coil *)
  Frequency * = 27.12; (** RF frequency in MHz *)

  (* Cone dimensions *)
  ConeHeight = 1*mm;
  ConeIntRad = 1.5*mm;
  ConeOutRad = 3.5*mm;

  (** Gabarit sizes of geometry and plasma *)
  TotalMinZ * = 0*cm;
  TotalMaxZ * = TubeOutLen + 7.5*mm;
  TotalRad * = CoilR + Coilr;

  (** Plasma is calculated in this region: *)
  PlasmaMinZ * = 0;
  PlasmaMaxZ * = TotalMaxZ;
  PlasmaRad * = TubeOutR0;

  MaxSigma = 2500 / (PHC.Ohm*PHC.meter);

  ModuleName = COMPILER.EQUATION('module');

TYPE
  SpaceType * = (In,Out,Glass,Coil); (** Type of spaces in the model *)
  (** In is a free space inside the torch (filled by Ar) *)
  (** Out is a free space outside the torch (filled by air) *)
  (** Glass is a space inside glass or alumna walls *)
  (** Coil is a space inside the coil *)

  Float = Math.Float;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE InRound (z,r,RoundZ,RoundR,RoundRad:Float): BOOLEAN;
  (* Returns if (z,r) is in a round with center (RoundZ,RoundR) and radius RoundRad *)
  BEGIN
    RETURN sqr(z-RoundZ) + sqr(r-RoundR) <= sqr(RoundRad)
  END InRound;

PROCEDURE CoilTurnIndex (z,r:Float): LONGINT;
  (* If (z,r) is inside the coil, then the procedure returns index of the
      coil's turn where lies (z,r).
      Else it returns -1.
      Indexing is zero-based. Turns with larger z have higher index. *)
  BEGIN
    IF (r<CoilR-Coilr) OR (r>CoilR+Coilr) OR
       (z<CoilZ0-Coilr) OR (z>CoilZ2+Coilr)
    THEN
      RETURN -1
    END;
    IF InRound(z,r,CoilZ0,CoilR,Coilr) THEN
      RETURN 0
    ELSIF InRound(z,r,CoilZ1,CoilR,Coilr) THEN
      RETURN 1
    ELSIF InRound(z,r,CoilZ2,CoilR,Coilr) THEN
      RETURN 2
    ELSE
      RETURN -1
    END;
  END CoilTurnIndex;

PROCEDURE GetSpaceType * (z,r:Float): SpaceType;
  (** Returns type of the given point.
      (z,r) may exceed dimensions of the cell. *)
  BEGIN
    IF r>=TubeOutR0 THEN
      IF (r<=TubeOutR1) & (z<=TubeOutLen) THEN
        RETURN Glass
      ELSE
        IF CoilTurnIndex(z,r)<0 THEN
          RETURN Out
        ELSE
          RETURN Coil
        END;
      END;
    ELSE
      IF z>TubeMidLen THEN
        IF z<TotalMaxZ-ConeHeight THEN
          RETURN In
        ELSIF z>=TotalMaxZ THEN
          RETURN Glass
        ELSIF r<=ConeOutRad-(z-TotalMaxZ)/ConeHeight*(ConeIntRad-ConeOutRad) THEN
          RETURN Glass
        ELSE
          RETURN In
        END;
      ELSE
        IF ((r>=TubeIntR0) & (r<=TubeIntR1) & (z<=TubeIntLen)) OR
           ((r>=TubeMidR0) & (r<=TubeMidR1))
        THEN
          RETURN Glass
        ELSE
          RETURN In
        END;
      END;
    END;
  END GetSpaceType;

PROCEDURE InitialElectricalConductivity * (z,r:Float): Float;
  (** Electrical conductivity at the 1st iteration *)
  BEGIN
    IF (z>=TubeMidLen) & (r<=TubeMidR0) THEN
      RETURN MaxSigma;
    ELSE
      RETURN 0;
    END;
  END InitialElectricalConductivity;

PROCEDURE ExportVal (VAR f:TOut.Formatter; prefix-:ARRAY OF CHAR; val:Float;
                     postfix-:ARRAY OF CHAR);
  BEGIN
    TOut.String(f,prefix,0);
    TOut.Real(f,val,0);
    TOut.String(f,postfix,0);
    TOut.Ln(f);
  END ExportVal;

PROCEDURE ExportForComsol * (FileName-:ARRAY OF CHAR);
  (** Exports the geometry parameters in the specified text file so that it
      can be imported by COMSOL *)
  VAR
    f:TOut.Formatter;
  BEGIN
    TOut.OpenFile(f,FileName,TOut.Rewrite);

    ExportVal(f,'TubeIntR0 ',TubeIntR0,'[cm]');
    ExportVal(f,'TubeIntR1 ',TubeIntR1,'[cm]');
    ExportVal(f,'TubeIntLen ',TubeIntLen,'[cm]');
    ExportVal(f,'TubeMidR0 ',TubeMidR0,'[cm]');
    ExportVal(f,'TubeMidR1 ',TubeMidR1,'[cm]');
    ExportVal(f,'TubeMidLen ',TubeMidLen,'[cm]');
    ExportVal(f,'TubeOutR0 ',TubeOutR0,'[cm]');
    ExportVal(f,'TubeOutR1 ',TubeOutR1,'[cm]');
    ExportVal(f,'TubeOutLen ',TubeOutLen,'[cm]');
    ExportVal(f,'ConeHeight ',ConeHeight,'[cm]');
    ExportVal(f,'ConeIntRad ',ConeIntRad,'[cm]');
    ExportVal(f,'ConeOutRad ',ConeOutRad,'[cm]');
    ExportVal(f,'CoilR ',CoilR,'[cm]');
    ExportVal(f,'Coilr ',Coilr,'[cm]');
    ExportVal(f,'CoilPeriod ',CoilPeriod,'[cm]');
    ExportVal(f,'CoilZ1 ',CoilZ1,'[cm]');
    ExportVal(f,'PlasmaMinZ ',PlasmaMinZ,'[cm]');
    ExportVal(f,'PlasmaMaxZ ',PlasmaMaxZ,'[cm]');
    ExportVal(f,'PlasmaRad ',PlasmaRad,'[cm]');

    TOut.Close(f);
  END ExportForComsol;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');
  Protocol.Real('Coilr = ',Coilr/cm,' cm');
  Protocol.Real('CoilR = ',CoilR/cm,' cm');
  Protocol.Real('CoilPeriod = ',CoilPeriod/cm,' cm');
  Protocol.Real('CoilZ0 = ',CoilZ0/cm,' cm');
  Protocol.Real('CoilZ1 = ',CoilZ1/cm,' cm');
  Protocol.Real('CoilZ2 = ',CoilZ2/cm,' cm');
  Protocol.Real('TubeIntR0 = ',TubeIntR0/cm,' cm');
  Protocol.Real('TubeIntR1 = ',TubeIntR1/cm,' cm');
  Protocol.Real('TubeIntLen = ',TubeIntLen/cm,' cm');
  Protocol.Real('TubeMidR0 = ',TubeMidR0/cm,' cm');
  Protocol.Real('TubeMidR1 = ',TubeMidR1/cm,' cm');
  Protocol.Real('TubeMidLen = ',TubeMidLen/cm,' cm');
  Protocol.Real('TubeOutR0 = ',TubeOutR0/cm,' cm');
  Protocol.Real('TubeOutR1 = ',TubeOutR1/cm,' cm');
  Protocol.Real('TubeOutLen = ',TubeOutLen/cm,' cm');
  Protocol.Real('ConeHeight = ',ConeHeight/cm,' cm');
  Protocol.Real('ConeIntRad = ',ConeIntRad/cm,' cm');
  Protocol.Real('ConeOutRad = ',ConeOutRad/cm,' cm');
  Protocol.Real('CoilCrossSectionArea = ',CoilCrossSectionArea/cm**2,' cm^2');
  Protocol.Real('Frequency = ',Frequency,' MHz');
  Protocol.Real('TotalMinZ = ',TotalMinZ/cm,' cm');
  Protocol.Real('TotalMaxZ = ',TotalMaxZ/cm,' cm');
  Protocol.Real('TotalRad = ',TotalRad/cm,' cm');
  Protocol.Real('PlasmaMinZ = ',PlasmaMinZ/cm,' cm');
  Protocol.Real('PlasmaMaxZ = ',PlasmaMaxZ/cm,' cm');
  Protocol.Real('PlasmaRad = ',PlasmaRad/cm,' cm');
  Protocol.Real('MaxSigma = ',MaxSigma *PHC.Ohm*PHC.meter,' Ohm^-1 * m^-1');
  Protocol.String('','','');

END SPECTROFasselGeometry.