(** This module contains geometry of the simulations and all the parameters,
    which are specific for this geometry. *)

<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE TulipGeometry;

IMPORT
  Protocol, COMPILER, PHC:=msciPhConst, Math:=mMath, Area, TOut:=msysTextOut,
  Areas;

CONST
  ModuleName = COMPILER.EQUATION('module');
  pi = PHC.pi;
  cm = PHC.cm;
  mm = PHC.mm;

  InputRad = 2.8/2*mm;  (* Internal radius of input tube of the SHIP torch *)
  MiddleRad = 11*mm;    (* Internal radius of the middle part of the torch *)
  FinalRad = MiddleRad+5*mm; (* Internal radius of the final part of the torch *)
  TransitionLen = 8*mm; (* Length of transitions between the input tube and the middle part of the torch *)
  MiddleLen = 13*mm; (* Length of the middle part of the torch *)
  FinalLen = 23*mm; (* Length of the final part of the torch *)

  FilletRad = 6*mm;     (* Radius of concave part of the fillets of the transitions *)
  WallThickness = 1*mm; (* Thickness of the SHIP wall *)
  CapillaryZ = 30*mm; (* Length of the capillaty *)
  CapillaryR0 = 0.4*mm; (* Internal radius of the capillary *)
  CapillaryR1 = 1*mm;   (* External radius of the capillary *)

  z0 = 0 * cm;
  r0 = InputRad;
  z1 = CapillaryZ;
  r1 = InputRad;
  z2 = z1+TransitionLen;
  r2 = MiddleRad;
  z3 = z2+MiddleLen;
  r3 = MiddleRad;
  z4 = z3+FinalLen;
  r4 = FinalRad;
  z5 = z4;
  r5 = r4;

  Coilr = 5*mm/2; (* radius of the coil wire *)
  CoilR = 30.5*mm/2; (* radius of the middle of the coil turns *)
  CoilPeriod = 2*Coilr + 2.3*mm; (* spatial period of coil turns *)
  CoilZ1 * = (z2+z3)/2; (** z of the center of the central turn *)
  CoilZ0 = CoilZ1 - CoilPeriod; (* z of the center of the upstream turn *)
  CoilZ2 = CoilZ1 + CoilPeriod; (* z of the center of the downstream turn *)

  CoilCrossSectionArea * = pi*Coilr*Coilr; (** Area of cross-section of the coil *)
  Frequency * = 27.12; (** RF frequency in MHz *)

  (* Gabarit sizes of geometry and plasma *)
  TotalMinZ * = z0;
  TotalMaxZ * = z5;
  TotalRad * = CoilR + Coilr;

  (** Plasma is calculated in this region: *)
  PlasmaMinZ * = z0;
  PlasmaMaxZ * = z5;
  PlasmaRad * = r2;

  MaxSigma = 2500 / (PHC.Ohm*PHC.meter);

TYPE
  Float = Math.Float;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE InRound * (z,r,RoundZ,RoundR,RoundRad:Float): BOOLEAN;
  (** Returns if (z,r) is in a round with center (RoundZ,RoundR) and radius RoundRad *)
  BEGIN
    RETURN sqr(z-RoundZ) + sqr(r-RoundR) <= sqr(RoundRad)
  END InRound;

PROCEDURE CoilTurnIndex * (z,r:Float): LONGINT;
  (** If (z,r) is inside the coil, then the procedure returns index of the
      coil's turn where lies (z,r).
      Else it returns -1.
      Indexing is zero-based. Turns with larger z have higher index. *)
  BEGIN
    IF (r<CoilR-Coilr) OR (r>CoilR+Coilr) OR
       (z<CoilZ0-Coilr) OR (z>CoilZ2+Coilr)
    THEN
      RETURN -1
    END;
    IF InRound(z,r,CoilZ0,CoilR,Coilr) THEN
      RETURN 0
    ELSIF InRound(z,r,CoilZ1,CoilR,Coilr) THEN
      RETURN 1
    ELSIF InRound(z,r,CoilZ2,CoilR,Coilr) THEN
      RETURN 2
    ELSE
      RETURN -1
    END;
  END CoilTurnIndex;

TYPE
  SpaceType * = (In,Out,Glass,Coil); (** Type of spaces in the model *)
  (** In is a free space inside the torch (filled by Ar) *)
  (** Out is a free space outside the torch (filled by air) *)
  (** Glass is a space inside glass or alumna walls *)
  (** Coil is a space inside the coil *)

PROCEDURE GetSpaceType * (z,r:Float): SpaceType;
  (** Returns type of the given point.
      (z,r) may exceed dimensions of the cell. *)
  BEGIN
    IF Areas.Index(z,r)>=0 THEN
      RETURN In
    ELSIF r<r0 THEN
      RETURN Glass
    ELSIF CoilTurnIndex(z,r)<0 THEN
      RETURN Out
    ELSE
      RETURN Coil
    END;
  END GetSpaceType;

PROCEDURE InitialElectricalConductivity * (z,r:Float): Float;
  (** Electrical conductivity at the 1st iteration *)
  BEGIN
    IF GetSpaceType(z,r)=In THEN
      RETURN MaxSigma
    ELSE
      RETURN 0
    END;
  END InitialElectricalConductivity;

PROCEDURE Min (x,y:Float): Float;
  BEGIN
    IF x<y THEN
      RETURN x
    ELSE
      RETURN y
    END;
  END Min;

<* PUSH *>
<* WOFF311 + *>
PROCEDURE SetAreas;
  VAR
    z, r, SecZ1, SecR1, SecZ2, SecR2: Float;
  BEGIN
(* ======================== Here areas are set ===============================*)
    Areas.Add(Area.Cylinder( CapillaryZ, z5, r0 ));
    Areas.Add(Area.Cylinder( z0, CapillaryZ, CapillaryR0 ));
    Areas.Add(Area.HollowCylinder( z0, CapillaryR1, CapillaryZ, r0 ));
    Areas.Add(Area.FilletToAxis( z0, r0, z1, r1, z2, r2, FilletRad+WallThickness, SecZ1, SecR1, SecZ2, SecR2 ));

    z:=SecZ2;
    r:=SecR2;
    Areas.Add(Area.FilletToAxis( z1, r1, z2, r2, z3, r3, FilletRad, SecZ1, SecR1, SecZ2, SecR2 ));
    Areas.Add(Area.IntCone( z, r, SecZ1, SecR1 ));
    Areas.Add(Area.HollowCylinder( z, Min(r,SecR1), SecZ1, r0 ));
    Areas.Add(Area.HollowCylinder( SecZ1, Min(SecR1,SecR2), SecZ2, r0 ));

    z:=SecZ2;
    r:=SecR2;
    Areas.Add(Area.FilletToAxis( z2, r2, z3, r3, z4, r4, FilletRad, SecZ1, SecR1, SecZ2, SecR2 ));
    Areas.Add(Area.IntCone( z, r, SecZ1, SecR1 ));
    Areas.Add(Area.HollowCylinder( z, Min(r,SecR1), SecZ1, r0 ));
    Areas.Add(Area.HollowCylinder( SecZ1, Min(SecR1,SecR2), SecZ2, r0 ));

    z:=SecZ2;
    r:=SecR2;
    Areas.Add(Area.FilletToAxis( z3, r3, z4, r4, z5, r5, FilletRad+WallThickness, SecZ1, SecR1, SecZ2, SecR2 ));
    Areas.Add(Area.IntCone( z, r, SecZ1, SecR1 ));
    Areas.Add(Area.HollowCylinder( z, Min(r,SecR1), SecZ1, r0 ));
    Areas.Add(Area.HollowCylinder( SecZ1, Min(SecR1,SecR2), SecZ2, r0 ));

    Areas.Add(Area.IntCone( SecZ2, SecR2, z5, r5 ));
    Areas.Add(Area.HollowCylinder( SecZ2, Min(SecR2,r5), z5, r0 ));
  END SetAreas;
<* POP *>

PROCEDURE ExportVal (VAR f:TOut.Formatter; prefix-:ARRAY OF CHAR; val:Float;
                     postfix-:ARRAY OF CHAR);
  BEGIN
    TOut.String(f,prefix,0);
    TOut.Real(f,val,0);
    TOut.String(f,postfix,0);
    TOut.Ln(f);
  END ExportVal;

PROCEDURE ExportForComsol * (FileName-:ARRAY OF CHAR);
  (** Exports the geometry parameters in the specified text file so that it
      can be imported by COMSOL *)
  VAR
    f:TOut.Formatter;
  BEGIN
    TOut.OpenFile(f,FileName,TOut.Rewrite);

    ExportVal(f,'z0 ',z0,'[cm]');
    ExportVal(f,'r0 ',r0,'[cm]');
    ExportVal(f,'z1 ',z1,'[cm]');
    ExportVal(f,'r1 ',r1,'[cm]');
    ExportVal(f,'z2 ',z2,'[cm]');
    ExportVal(f,'r2 ',r2,'[cm]');
    ExportVal(f,'z3 ',z3,'[cm]');
    ExportVal(f,'r3 ',r3,'[cm]');
    ExportVal(f,'z4 ',z4,'[cm]');
    ExportVal(f,'r4 ',r4,'[cm]');
    ExportVal(f,'FilletRad ',FilletRad,'[cm]');
    ExportVal(f,'WallThickness ',WallThickness,'[cm]');
    ExportVal(f,'CapillaryZ ',CapillaryZ,'[cm]');
    ExportVal(f,'CapillaryR0 ',CapillaryR0,'[cm]');
    ExportVal(f,'CapillaryR1 ',CapillaryR1,'[cm]');
    ExportVal(f,'Coilr ',Coilr,'[cm]');
    ExportVal(f,'CoilR ',CoilR,'[cm]');
    ExportVal(f,'CoilZ0 ',CoilZ0,'[cm]');
    ExportVal(f,'CoilPeriod ',CoilPeriod,'[cm]');
    ExportVal(f,'PlasmaMinZ ',PlasmaMinZ,'[cm]');
    ExportVal(f,'PlasmaMaxZ ',PlasmaMaxZ,'[cm]');
    ExportVal(f,'PlasmaRad ',PlasmaRad,'[cm]');

    TOut.Close(f);
  END ExportForComsol;

BEGIN

  Protocol.String('MODULE ',ModuleName,':');

  SetAreas;

  Protocol.Real('InputRad = ',InputRad/cm,' cm');
  Protocol.Real('MiddleRad = ',MiddleRad/cm,' cm');
  Protocol.Real('FinalRad = ',FinalRad/cm,' cm');
  Protocol.Real('TransitionLen = ',TransitionLen/cm,' cm');
  Protocol.Real('MiddleLen = ',MiddleLen/cm,' cm');
  Protocol.Real('FinalLen = ',FinalLen/cm,' cm');
  Protocol.Real('FilletRad = ',FilletRad/cm,' cm');
  Protocol.Real('WallThickness = ',WallThickness/cm,' cm');
  Protocol.Real('CapillaryZ = ',CapillaryZ/cm,' cm');
  Protocol.Real('CapillaryR0 = ',CapillaryR0/cm,' cm');
  Protocol.Real('CapillaryR1 = ',CapillaryR1/cm,' cm');
  Protocol.Real('z0 = ',z0/cm,' cm');
  Protocol.Real('r0 = ',r0/cm,' cm');
  Protocol.Real('z1 = ',z1/cm,' cm');
  Protocol.Real('r1 = ',r1/cm,' cm');
  Protocol.Real('z2 = ',z2/cm,' cm');
  Protocol.Real('r2 = ',r2/cm,' cm');
  Protocol.Real('z3 = ',z3/cm,' cm');
  Protocol.Real('r3 = ',r3/cm,' cm');
  Protocol.Real('z4 = ',z4/cm,' cm');
  Protocol.Real('r4 = ',r4/cm,' cm');
  Protocol.Real('Coilr = ',Coilr/cm,' cm');
  Protocol.Real('CoilR = ',CoilR/cm,' cm');
  Protocol.Real('CoilPeriod = ',CoilPeriod/cm,' cm');
  Protocol.Real('CoilZ0 = ',CoilZ0/cm,' cm');
  Protocol.Real('CoilZ1 = ',CoilZ1/cm,' cm');
  Protocol.Real('CoilZ2 = ',CoilZ2/cm,' cm');
  Protocol.Real('CoilCrossSectionArea = ',CoilCrossSectionArea/cm**2,' cm^2');
  Protocol.Real('Frequency = ',Frequency,' MHz');
  Protocol.Real('TotalMinZ = ',TotalMinZ/cm,' cm');
  Protocol.Real('TotalMaxZ = ',TotalMaxZ/cm,' cm');
  Protocol.Real('TotalRad = ',TotalRad/cm,' cm');
  Protocol.Real('PlasmaMinZ = ',PlasmaMinZ/cm,' cm');
  Protocol.Real('PlasmaMaxZ = ',PlasmaMaxZ/cm,' cm');
  Protocol.Real('PlasmaRad = ',PlasmaRad/cm,' cm');
  Protocol.Real('MaxSigma = ',MaxSigma *PHC.Ohm*PHC.meter,' Ohm^-1 * m^-1');

  Protocol.String('','','');

END TulipGeometry.