(** Implements Ar and He properties like in the model of Lindner *)
<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE LiData;

IMPORT
  Math:=mMath, PHC:=msciPhConst, LiGas, Distr:=LiLTEDistribution,
  Out:=msysTextOut;

CONST
  meter = PHC.meter;
  amu = PHC.amu;

TYPE
  Float = Math.Float;
  Arr6 = ARRAY 6 OF Float;

VAR
  ArAdded,HeAdded:BOOLEAN;

PROCEDURE ArRad * (T:Float): Float;
  (** Ar radius as function of temperature *)
  VAR
    res:Float;
  BEGIN
    IF T<273 THEN
      T:=273
    ELSIF T>12000 THEN
      T:=12000
    END;

    IF T<=9300 THEN
      res:=5.2603E-10/Math.power(T-154.46485,0.0826) - 7.0465E-16*T + 1.2167E-11
    ELSE
      res:=3.3E-28*Math.power(T-7700,4.7)+2.5275E-10
    END;
    RETURN res*meter/2; (* /2 - because this is radius rather than diameter *)
  END ArRad;

PROCEDURE HeRad * (T:Float): Float;
  (** Ar radius as function of temperature *)
  VAR
    res:Float;
  BEGIN
    IF T<273 THEN
      T:=273
    ELSIF T>11250 THEN
      T:=11250
    END;

    res:=9.5066E-6/Math.power(T+1133.33407,1.66528) - 2.6575E-15*T + 1.6499E-10;
    RETURN res*meter/2; (* /2 - because this is radius rather than diameter *)
  END HeRad;

PROCEDURE PolyLn (T:Float; a-:Arr6): Float;
  VAR
    TInPower,res:Float;
    i:LONGINT;
  BEGIN
    IF T<1000 THEN
      T:=1000
    ELSIF T>16000 THEN
      T:=16000
    END;

    T:=Math.ln(T);
    TInPower:=1;
    res:=0;
    FOR i:=0 TO LEN(a)-1 DO
      res:=res+a[i]*TInPower;
      TInPower:=TInPower*T;
    END;
    RETURN Math.exp(res);
  END PolyLn;

PROCEDURE PartitionFunctionOfAr0 * (T:Float): Float;
  (** Partition function of Ar atoms as function of temperature *)
  CONST
    a = Arr6{-1.50860761E2,9.44834641E1,-2.36105115E1,2.94261084,-1.82910244E-1,4.53641770E-3};
  BEGIN
    RETURN PolyLn(T,a);
  END PartitionFunctionOfAr0;

PROCEDURE PartitionFunctionOfAr1 * (T:Float): Float;
  (** Partition function of Ar ions as function of temperature *)
  CONST
    a = Arr6{8.00614439E1,-4.51603962E1,1.02073231E1,-1.13868546,6.29814970E-2,-1.38557483E-3};
  BEGIN
    RETURN PolyLn(T,a);
  END PartitionFunctionOfAr1;

PROCEDURE PartitionFunctionOfAr2 * (T:Float): Float;
  (** Partition function of doubly ionized Ar particles as function of temperature *)
  CONST
    a = Arr6{3.08147365E2,-1.8771751E2,4.56112646E1,-5.50134735,3.29750197E-1,-7.85881328E-3};
  BEGIN
    RETURN PolyLn(T,a);
  END PartitionFunctionOfAr2;

PROCEDURE PartitionFunctionOfHe0 (T:Float): Float;
  CONST
    a = Arr6{-3.76575219E-1,2.33951687E-1,-5.79755525E-2,7.16333160E-3,-4.41302573E-4,1.08442997E-5};
  BEGIN
    RETURN PolyLn(T,a);
  END PartitionFunctionOfHe0;

PROCEDURE PartitionFunctionOfHe1 (T:Float): Float;
  CONST
    a = Arr6{6.93147179E-1,9.29636701E-10,-2.30049742E-10,2.83829746E-11,-1.74590774E-12,4.28355287E-14};
  BEGIN
    RETURN PolyLn(T,a);
  END PartitionFunctionOfHe1;

TYPE
  Element = LiGas.Element;
  PartitionFunction = Distr.PartitionFunction;
  Level = Distr.Level;
  Levels = Distr.Levels;

PROCEDURE InsertLevel (VAR levels:Levels; VAR index:LONGINT;
                       IonizationEnergy:Float; pf:PartitionFunction);
  VAR
    l:POINTER TO Level;
  BEGIN
    ASSERT(pf#NIL);
    ASSERT(index<LEN(levels));

    NEW(l);
    l.pf:=pf;
    l.IonizationEnergy:=IonizationEnergy;
    levels[index]:=l;
    INC(index);
  END InsertLevel;

PROCEDURE AddAr * (nrel:Float);
  (** Adds argon to the LTE gas model. nrel is relative number density of Ar. *)
  VAR
    i:LONGINT;
    el:Element;
  BEGIN
    el.nrel:=nrel;
    NEW(el.l,3);
    i:=0;
    InsertLevel(el.l^,i,0,PartitionFunctionOfAr0);
    InsertLevel(el.l^,i,15.7596,PartitionFunctionOfAr1);
    InsertLevel(el.l^,i,27.629,PartitionFunctionOfAr2);
    ASSERT(i=LEN(el.l^));

    LiGas.Add(el,40*amu,ArRad,'Ar');
    ArAdded:=TRUE;
  END AddAr;

PROCEDURE AddHe * (nrel:Float);
  (** Adds helium to the LTE gas model. nrel is relative number density of Ar. *)
  VAR
    i:LONGINT;
    el:Element;
  BEGIN
    el.nrel:=nrel;
    NEW(el.l,2);
    i:=0;
    InsertLevel(el.l^,i,0,PartitionFunctionOfHe0);
    InsertLevel(el.l^,i,24.5874,PartitionFunctionOfHe1);
    ASSERT(i=LEN(el.l^));

    LiGas.Add(el,4*amu,HeRad,'He');
    HeAdded:=TRUE;
  END AddHe;

PROCEDURE ExportTemperature * (FileName-:ARRAY OF CHAR);
  (** Exports gas properties as function of temperature at the current
      composition of elements *)
  CONST
    p = 1*PHC.Atmosphere;
    MinT = 1000;
    MaxT = 11000;
    dT = 100;
    LenT = ENTIER((MaxT-MinT)/dT+0.5)+1;

    Tab = Out.Tab;
    MinRelConc = 1.0E-10;
  VAR
    f:Out.Formatter;
    i:LONGINT;
    n,T:Float;
  BEGIN
    Out.OpenFile(f,FileName,Out.Rewrite);

    Out.String(f, 'Temperature(K)'+Tab+'Viscosity(kg/(m*s))'
                 +Tab+'ThermalConductivity(W/(m*K))'
                 +Tab+'SpecificHeatCapacity(J/(kg*K))'
                 +Tab+'ElectricalConductivity(1/(Ohm*m))'
                 +Tab+'RadiationLoss(W/m^3)',0);
    IF ArAdded THEN
      Out.String(f,Tab+'ArDiffusion(m^2/s)',0);
    END;
    IF HeAdded THEN
      Out.String(f,Tab+'HeDiffusion(m^2/s)',0);
    END;
    Out.Ln(f);
    FOR i:=0 TO LenT-1 DO
      T:=MinT+(MaxT-MinT)*i/(LenT-1);
      n:=p/(PHC.k*T);
      LiGas.Calculate(n,T);
      Out.Real(f,T,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.Viscosity()*PHC.meter*PHC.second/PHC.kg,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.ThermalConductivity()*PHC.meter*PHC.Kelvin/PHC.Watt,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.SpecificHeatCapacity()*PHC.kg*PHC.Kelvin/PHC.Joule,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.ElectricalConductivity()*PHC.meter*PHC.Ohm,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.RadiationLoss()*PHC.meter**3/PHC.Watt,0);
      IF ArAdded THEN
        Out.Char(f,Tab);
        Out.Real(f,LiGas.Diffusion('Ar')*PHC.second/PHC.meter**2,0);
      END;
      IF HeAdded THEN
        Out.Char(f,Tab);
        Out.Real(f,LiGas.Diffusion('He')*PHC.second/PHC.meter**2,0);
      END;
      Out.Ln(f);
    END;

    Out.Close(f);
  END ExportTemperature;

PROCEDURE Export * (FileName-:ARRAY OF CHAR);
  (** Exports the gas properties with different temperatures and compositions.
      The current composition is changed. *)
  CONST
    p = 1*PHC.Atmosphere;
    MinT = 1000;
    MaxT = 11000;
    dT = 100;
    LenT = ENTIER((MaxT-MinT)/dT+0.5)+1;

    (* w is a mass fraction of He *)
    MinW = 0;
    MaxW = 1;
    dW = 0.05;
    LenW =ENTIER((MaxW-MinW)/dW+0.5)+1;

    MassAr = 40*PHC.amu;
    MassHe = 4*PHC.amu;
    Tab = Out.Tab;

    MinRelConc = 1.0E-10;
  VAR
    iW,iT:LONGINT;
    w,n,T,ArRelConc,HeRelConc:Float;
    f:Out.Formatter;
  BEGIN
    Out.OpenFile(f,FileName,Out.Rewrite);

    Out.String(f, 'HeMassFraction'+Tab+'Temperature(K)'+Tab+'Viscosity(kg/(m*s))'
                 +Tab+'ThermalConductivity(W/(m*K))'
                 +Tab+'SpecificHeatCapacity(J/(kg*K))'
                 +Tab+'ElectricalConductivity(1/(Ohm*m))'
                 +Tab+'RadiationLoss(W/m^3)'
                 +Tab+'ArDiffusion(m^2/s)'
                 +Tab+'HeDiffusion(m^2/s)',0);
    Out.Ln(f);
    FOR iW:=0 TO LenW-1 DO
      w:=MinW+(MaxW-MinW)*iW/(LenW-1);
      HeRelConc:=w*MassAr/(MassHe+w*(MassAr-MassHe));
      ArRelConc:=1-HeRelConc;
      IF iW=0 THEN
        HeRelConc:=MinRelConc;
        ArRelConc:=1-MinRelConc;
      END;
      IF iW=LenW-1 THEN
        ArRelConc:=MinRelConc;
        HeRelConc:=1-MinRelConc;
      END;
      LiGas.Clear;
      AddAr(ArRelConc);
      AddHe(HeRelConc);
      FOR iT:=0 TO LenT-1 DO
        T:=MinT+(MaxT-MinT)*iT/(LenT-1);
        n:=p/(PHC.k*T);
        LiGas.Calculate(n,T);

        Out.Real(f,w,0);
        Out.Char(f,Tab);
        Out.Real(f,T,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.Viscosity()*PHC.meter*PHC.second/PHC.kg,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.ThermalConductivity()*PHC.meter*PHC.Kelvin/PHC.Watt,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.SpecificHeatCapacity()*PHC.kg*PHC.Kelvin/PHC.Joule,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.ElectricalConductivity()*PHC.meter*PHC.Ohm,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.RadiationLoss()*PHC.meter**3/PHC.Watt,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.Diffusion('Ar')*PHC.second/PHC.meter**2,0);
        Out.Char(f,Tab);
        Out.Real(f,LiGas.Diffusion('He')*PHC.second/PHC.meter**2,0);
        Out.Ln(f);
      END;
    END;

    Out.Close(f);
  END Export;

PROCEDURE Install * ();
  (** Installs properties of the gas *)
  BEGIN
    LiGas.Clear;
    AddAr(1);
  END Install;

BEGIN
  ArAdded:=FALSE;
  HeAdded:=FALSE;
END LiData.