(** This module calculates properties of the gas, based on LTE ionization at
    given n (total gas number density) and T.
    First the elements should be added by the procedure Add.
    Than the calculation should be carried out by the procedure Calculate.
    Then the gas properties like viscosity, thermal conductivity, etc.
    can be calculated. *)
<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE LiGas;

IMPORT
  Math:=mMath, PHC:=msciPhConst, Distr:=LiLTEDistribution, DStr:=msysDynStr;

CONST
  meter = PHC.meter;
  pi = Math.pi;
  e = PHC.e;
  k = PHC.k;
  amu = PHC.amu;
  Me = PHC.Me;
  eV = PHC.eV;

TYPE
  Float = Math.Float;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

TYPE
  Element * = Distr.Element;
  Elements = Distr.Elements;
  Level = Distr.Level;
  String = DStr.String;

TYPE
  ElementsPtr = Distr.ElementsPtr;

  RadProc = PROCEDURE (T:Float): Float;
  ElementProp = RECORD
    mass:Float;
    rad:RadProc; (* radius as function of temperature *)
  END;
  ElementData = RECORD
    p:ElementProp;
    nrel:Float; (* relative number density *)
    name:String;
  END;

  Solution = RECORD
    els:ElementsPtr;
    n,T:Float;
    ld:Float; (* Debye radius *)
    nerel:Float; (* relative concentration of electrons *)
  END;

CONST
  DeltaT = 1.0E1;
  MaxElements = 10;

VAR
  ElementCount:LONGINT; (* number of elements *)
  sol,ShiftedSol:Solution; (* solutions at given (n,T) and T+DeltaT at constent pressure *)
  props:ARRAY MaxElements OF ElementData; (* Elements *)
  TmpElements:ARRAY MaxElements OF Element;
  el:ElementProp; (* electrons *)
  MeanMass:Float; (* mean mass of heavy particles in the mixture *)

PROCEDURE ClearTmpElements ();
  VAR
    i:LONGINT;
  BEGIN
    FOR i:=0 TO MaxElements-1 DO
      TmpElements[i].l:=NIL;
    END;
  END ClearTmpElements;

PROCEDURE Clear * ();
  (** Clears all the internal structures of the module *)
  BEGIN
    sol.els:=NIL;
    ShiftedSol.els:=NIL;
    ElementCount:=0;
    ClearTmpElements;
  END Clear;

PROCEDURE Add * (el-:Element; mass:Float; rad:RadProc; name-:ARRAY OF CHAR);
  (** Adds the element to the module *)
  VAR
    i:LONGINT;
  BEGIN
    ASSERT(el.nrel>=0);
    IF el.nrel=0 THEN
      RETURN
    END;

    ASSERT(ElementCount<MaxElements);
    ASSERT(rad#NIL);

    i:=ElementCount;
    TmpElements[i]:=el;
    props[i].p.mass:=mass;
    props[i].p.rad:=rad;
    props[i].nrel:=el.nrel;
    props[i].name:=DStr.Str(name);

    INC(ElementCount);
  END Add;

PROCEDURE Calc (VAR s:Solution; n,T:Float);
  BEGIN
    s.n:=n;
    s.T:=T;
    Distr.WithEnergyCorrection(s.els,n,T);
    s.nerel:=Distr.nerel(s.els^);
    s.ld:=Distr.DebyeRad(s.els^,n,T);
  END Calc;

PROCEDURE CopyElement (source-:Element; VAR dest:Element);
  VAR
    i:LONGINT;
    l:POINTER TO Level;
  BEGIN
    ASSERT(source.l#NIL);

    dest.nrel:=source.nrel;
    NEW(dest.l,LEN(source.l^));
    FOR i:=0 TO LEN(source.l^)-1 DO
      NEW(l);
      l^:=source.l[i]^;
      dest.l[i]:=l;
    END;
  END CopyElement;

PROCEDURE TmpToSolution (VAR s:Solution);
  VAR
    i:LONGINT;
  BEGIN
    NEW(s.els,ElementCount);
    FOR i:=0 TO ElementCount-1 DO
      CopyElement(TmpElements[i],s.els[i]);
    END;
  END TmpToSolution;

PROCEDURE Calculate * (n,T:Float);
  (** Calculates the gas properties as function of n and T *)
  VAR
    mass:Float;
    i:LONGINT;
    p:Float;
    n1,T1:Float;
    sum:Float;
  BEGIN
    ASSERT(ElementCount>0);
    IF (sol.els=NIL) OR (LEN(sol.els^)#ElementCount) THEN
      (* it means that sol and ShiftedSol are not initialized yet *)
      sum:=0;
      FOR i:=0 TO ElementCount-1 DO
        sum:=sum+TmpElements[i].nrel;
      END;
      FOR i:=0 TO ElementCount-1 DO
        TmpElements[i].nrel:=TmpElements[i].nrel/sum;
        props[i].nrel:=TmpElements[i].nrel;
      END;
      TmpToSolution(sol);
      TmpToSolution(ShiftedSol);
    END;

    p:=n*T;
    T1:=T+DeltaT;
    n1:=p/T1;

    Calc(ShiftedSol,n1,T1);
    Calc(sol,n,T);

    mass:=0;
    FOR i:=0 TO LEN(sol.els^)-1 DO
      mass:=mass + props[i].nrel*props[i].p.mass;
    END;
    MeanMass:=mass;
  END Calculate;

PROCEDURE CrossSectionNeutral (rad1,rad2:Float): Float;
  BEGIN
    RETURN 2**(1/2)*pi*sqr(rad1+rad2)
  END CrossSectionNeutral;

PROCEDURE CrossSectionCharged (s-:Solution; Z1,Z2:LONGINT): Float;
  (* from Raizer rather than from Lindner *)
  VAR
    ne,lnl,b0:Float;
  BEGIN
    ne:=s.n*s.nerel;
    IF ne=0 THEN
      RETURN Math.MaxFloat
    ELSE
      lnl:=Math.ln(12*pi*s.ld*s.ld*s.ld*ne);
      b0:=2*e*e/(3*k*s.T);
      RETURN Z1*Z1*Z2*Z2*pi*sqr(b0)*lnl
    END;
  END CrossSectionCharged;

PROCEDURE CrossSection (s-:Solution; rad1,rad2:RadProc; Z1,Z2:LONGINT): Float;
  BEGIN
    ASSERT(rad1#NIL);
    ASSERT(rad2#NIL);

    IF (Z1=0) OR (Z2=0) THEN
      RETURN CrossSectionNeutral(rad1(s.T),rad2(s.T))
    ELSE
      RETURN CrossSectionCharged(s,Z1,Z2);
    END;
  END CrossSection;

PROCEDURE Sum (s-:Solution; p-:ElementProp; Z:LONGINT): Float;
  VAR
    i,Ze:LONGINT;
    SqrtMass,sum:Float;
  BEGIN
    SqrtMass:=sqrt(p.mass*el.mass/(p.mass+el.mass));
    sum:=SqrtMass*s.nerel*CrossSection(s,p.rad,el.rad,Z,1);
      (* Contribution of electrons *)

    FOR i:=0 TO LEN(s.els^)-1 DO
      SqrtMass:=sqrt(p.mass*props[i].p.mass/(p.mass+props[i].p.mass));
      FOR Ze:=0 TO LEN(s.els[i].l^)-1 DO
        sum:=sum+SqrtMass*s.els[i].l[Ze].degree*CrossSection(s,p.rad,props[i].p.rad,Z,Ze);
      END;
    END;
    RETURN sum;
  END Sum;

PROCEDURE ViscosityOfParticle (s-:Solution; p-:ElementProp; degree:Float; Z:LONGINT): Float;
  VAR
    sum:Float;
  BEGIN
    sum:=Sum(s,p,Z);
    RETURN sqrt(k*s.T/pi)*degree*p.mass/sum
  END ViscosityOfParticle;

PROCEDURE Viscosity * (): Float;
  (** Returns viscosity *)
  VAR
    res:Float;
    i,Z:LONGINT;
  BEGIN
    ASSERT(sol.els#NIL);

    res:=ViscosityOfParticle(sol,el,sol.nerel,1); (* viscosity of electrons *)
    FOR i:=0 TO LEN(sol.els^)-1 DO
      FOR Z:=0 TO LEN(sol.els[i].l^)-1 DO
        res:=res+ViscosityOfParticle(sol,props[i].p,sol.els[i].l[Z].degree,Z);
      END;
    END;
    RETURN res;
  END Viscosity;

PROCEDURE ThermalConductivity * (): Float;
  (** Returns thermal conductivity *)
  VAR
    i,Z:LONGINT;
    sum:Float;
  BEGIN
    ASSERT(sol.els#NIL);

    sum:=ViscosityOfParticle(sol,el,sol.nerel,1)/Me; (* electrons *)
    FOR i:=0 TO LEN(sol.els^)-1 DO
      FOR Z:=0 TO LEN(sol.els[i].l^)-1 DO
        sum:=sum+ViscosityOfParticle(sol,props[i].p,sol.els[i].l[Z].degree,Z)/props[i].p.mass;
      END;
    END;
    RETURN 15/4*k*sum
  END ThermalConductivity;

PROCEDURE Hion (s-:Solution): Float;
  (* Contribution of ionization to enthalpy at the given solution *)
  VAR
    sum,energy:Float;
    i,Z:LONGINT;
    l:POINTER TO Level;
  BEGIN
    ASSERT(s.els#NIL);

    sum:=5/2*s.nerel*k*s.T; (* electrons *)
    FOR i:=0 TO LEN(s.els^)-1 DO
      energy:=0;
      FOR Z:=1 TO LEN(s.els[i].l^)-1 DO
        l:=s.els[i].l[Z];
        energy:=energy+l.IonizationEnergy;
        sum:=sum + energy*eV*l.degree
      END;
    END;

    RETURN sum/MeanMass;
  END Hion;

PROCEDURE SpecificHeatCapacity * (): Float;
  (** Returns mass specific heat capacity *)
  BEGIN
    ASSERT(sol.els#NIL);

    RETURN 5/2*PHC.R/MeanMass*PHC.amu (* Specific heat capacity of ideal gas *)
    + (Hion(ShiftedSol)-Hion(sol))/DeltaT
  END SpecificHeatCapacity;

PROCEDURE SigmaSum (p-:ElementProp; Z:LONGINT): Float;
  VAR
    i,Zs:LONGINT;
    res,InvSqrtMass:Float;
  BEGIN
    InvSqrtMass:=sqrt((p.mass+Me)/(p.mass*Me));
    res:=sol.nerel*CrossSection(sol,p.rad,el.rad,Z,1)*InvSqrtMass; (* electrons *)
    FOR i:=0 TO LEN(sol.els^)-1 DO
      InvSqrtMass:=sqrt((p.mass+props[i].p.mass)/(p.mass*props[i].p.mass));
      FOR Zs:=0 TO LEN(sol.els[i].l^)-1 DO
        res:=res + sol.els[i].l[Zs].degree*CrossSection(sol,p.rad,props[i].p.rad,Z,Zs)*InvSqrtMass
      END;
    END;
    RETURN res;
  END SigmaSum;

PROCEDURE ElectricalConductivity * (): Float;
  (** Returns electrical conductivity *)
  VAR
    i,Zs:LONGINT;
    res,sum:Float;
  BEGIN
    sum:=SigmaSum(el,1);
    IF sum=0 THEN
      ASSERT(sol.nerel=0);
      res:=0
    ELSE
      res:=sol.nerel/Me/SigmaSum(el,1); (* electrons *)
    END;
    FOR i:=0 TO LEN(sol.els^)-1 DO
      FOR Zs:=1 TO LEN(sol.els[i].l^)-1 DO
        sum:=SigmaSum(props[i].p,Zs);
        IF sum=0 THEN
          ASSERT(sol.nerel=0)
        ELSE
          res:=res + Zs*Zs*sol.els[i].l[Zs].degree/props[i].p.mass/SigmaSum(props[i].p,Zs)
        END;
      END;
    END;
    RETURN 1.5*e*e*sqrt(pi/(4*k*sol.T))*res
  END ElectricalConductivity;

PROCEDURE RadiationLoss * (): Float;
  (** Returns radiation loss pf energy *)
  CONST
    C1 = -6.932E-41;
    C2 = 4.753E-39;
    C3 = -5.808E-38;
    C4 = -2.843E-45;
    C5 = 8.629E-47;
    C6 = 6.706E4;
  VAR
    T,ne:Float;
  BEGIN
    T:=sol.T;
    ne:=sol.n*sol.nerel*meter**3;
    RETURN 4*pi*sqr(ne)*(C1+C2/sqrt(T)+C3/T+C4*T+C5*T*sqrt(T))*Math.exp(C6/T) *PHC.Watt/meter**3;
  END RadiationLoss;

PROCEDURE Diffusion * (name-:ARRAY OF CHAR): Float;
  (** Returns diffusion of the specified element
      If the element is not found then \ero is returned. *)
  VAR
    res,sum:Float;
    index,Z:LONGINT;
  BEGIN
    index:=0;
    WHILE (index<LEN(sol.els^)) & (props[index].name^#name) DO
      INC(index);
    END;

    ASSERT(index>=0);
    ASSERT(index<LEN(sol.els^));
    ASSERT(props[index].nrel#0);

    res:=0;
    FOR Z:=0 TO LEN(sol.els[index].l^)-1 DO
      sum:=Sum(sol,props[index].p,Z);
      res:=res+sol.els[index].l[Z].degree/sum;
    END;
    RETURN 6/5*sqrt(PHC.k*sol.T/pi)*res /props[index].nrel/sol.n
  END Diffusion;

PROCEDURE Ne * (): Float;
  (** Returnc concentration of electrons *)
  BEGIN
    RETURN sol.n*sol.nerel
  END Ne;

<* PUSH *>
<* WOFF301 + *>
PROCEDURE Zero (T:Float): Float;
  BEGIN
    RETURN 0;
  END Zero;
<* POP *>

BEGIN
  Clear;
  el.rad:=Zero;
  el.mass:=Me;
END LiGas.