(** This module calculates distribution of ionized levels as function of
    temperature and atomic composition of the gas, as described in [1,2].
    LTE approachis employed.

    1. M.R. Zaghloul, M.A; Bourham, J.M. Doster, "A simple formulation and
       solution strategy of the Saha equation for ideal and nonideal plasmas",
       J. Phys. D: Appl. Phys., 2000, 33, 977984

    2. M.R. Zaghloul, "Reduced formulation and efficient algorithm for the
       determination of equilibrium composition and partition functions of
       ideal and nonideal complex plasma mixtures",
       Phys. Rev. E, 2004, 69, 026702 *)
<* O2EXTENSIONS + *>
MODULE LiLTEDistribution;

IMPORT
  PHC:=msciPhConst, Math:=mMath, AlgEqu:=mAlgEqu;

CONST
  k = PHC.k;
  me = PHC.Me;
  h = PHC.h;
  pi = Math.pi;
  eV = PHC.eV;
  e = PHC.e;

TYPE
  Float = Math.Float;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

PROCEDURE PowerOneAndHalf (x:Float): Float;
  BEGIN
    RETURN x*sqrt(x);
  END PowerOneAndHalf;

TYPE
  PartitionFunction * = PROCEDURE (T:Float): Float;
    (** Partition function as function of temperature in K *)

  Level * = RECORD (** This is a level of ionization of an element *)
    IonizationEnergy * :Float;
      (** Energy which is necessary to convert the atom from the lower ionization
          level to the current one, in eV *)
    pf * :PartitionFunction;
      (** Partition function of the level *)
    degree - :Float;
      (** The result is returned here. This is degree of ionization of this level,
          which means that degree=[number density of particles in this level]/[number density of all atoms and ions of all elements] *)
    K:Float;
      (* RHS of Saha equation with this level and one level lower *)
  END;
  Levels * = ARRAY OF POINTER TO Level;

  Element * = RECORD (** This is information about the element. *)
    nrel * : Float;
      (** relative number density of the element:
          nrel=[number density of atoms and ions of this element]/[number density of all atoms and ions of all elements] *)
    l * : POINTER TO Levels;
      (** This is an array of levels. l[0] should be ground level, l[1] the ion. l[2] the doubly ionized ion, ets. *)
  END;
  Elements * = ARRAY OF Element;
  ElementsPtr * = POINTER TO Elements;
    (** This structure should be correctly formed by the user for the calculations *)

PROCEDURE K (UpperPF,LowerPF:PartitionFunction;
             EnergyDifference,T:Float): Float;
  (* RHSs of Saha equations for two levels. EnergyDifference is in eV, T is in K *)
  BEGIN
    ASSERT(UpperPF#NIL);
    ASSERT(LowerPF#NIL);

    RETURN 2 * PowerOneAndHalf(2*pi*me*k*T/(h*h)) * UpperPF(T)/LowerPF(T)
             * Math.exp(-EnergyDifference*eV/(k*T))
  END K;

PROCEDURE LoweringOfEnergy (ld:Float; Z:LONGINT; T:Float): Float;
  (* ld is Debye radius, result is in eV *)
  VAR
    lb:Float;
  BEGIN
    lb:=h/sqrt(2*pi*me*k*T);
    RETURN Z*e*e/(ld+lb/8) /eV
  END LoweringOfEnergy;

PROCEDURE CalculateKOfElement (el-:Element; T,ld:Float);
  (* Fills all K in the el *)
  VAR
    energy:Float;
    Z:LONGINT;
    l:POINTER TO Level;
    pf:PartitionFunction;
  BEGIN
    ASSERT(el.l#NIL);

    el.l[0].K:=0;
    pf:=el.l[0].pf;
    FOR Z:=1 TO LEN(el.l^)-1 DO
      l:=el.l[Z];
      energy:=l.IonizationEnergy - LoweringOfEnergy(ld,Z,T);
      l.K:=K(l.pf,pf,energy,T);
      pf:=l.pf;
    END;
  END CalculateKOfElement;

PROCEDURE CalculateK (els-:Elements; T,ld:Float);
  VAR
    i:LONGINT;
  BEGIN
    FOR i:=0 TO LEN(els)-1 DO
      CalculateKOfElement(els[i],T,ld)
    END;
  END CalculateK;

PROCEDURE GetSums (el-:Element; n,Zav:Float; VAR SumWithZ,SumWithoutZ:Float);
  (* Returns a special sum. All Ks should be already set *)
  VAR
    l:POINTER TO Level;
    prod,nzInPowerZ,nz:Float;
    Z:LONGINT;
  BEGIN
    ASSERT(el.l#NIL);

    IF Zav=0 THEN
      SumWithZ:=0;
      SumWithoutZ:=0;
      RETURN;
    END;

    prod:=1;
    SumWithZ:=0;
    SumWithoutZ:=0;
    nz:=n*Zav;
    nzInPowerZ:=1;

    FOR Z:=1 TO LEN(el.l^)-1 DO
      l:=el.l[Z];
      prod:=prod*l.K;
      nzInPowerZ:=nzInPowerZ*nz;
      IF nzInPowerZ#0 THEN (* ionization degree is lower than possible for variables *)
        SumWithZ:=SumWithZ + Z*prod/nzInPowerZ;
        SumWithoutZ:=SumWithoutZ + prod/nzInPowerZ;
      ELSE
        ASSERT(prod=0);
      END;
    END;
  END GetSums;

PROCEDURE CalculateDegreeOfElement (el-:Element; n,Zav:Float);
  (* Fills all degree in the el. All Ks should be already set *)
  VAR
    SumWithZ,SumWithoutZ,degree,Zel:Float;
    l:POINTER TO Level;
    Z:LONGINT;
  BEGIN
    ASSERT(el.l#NIL);

    GetSums(el,n,Zav,SumWithZ,SumWithoutZ);
    Zel:=el.nrel*SumWithZ/(1+SumWithoutZ);

    IF (Zel=0) OR (SumWithZ=0) THEN
      ASSERT(Zel=0);
      degree:=el.nrel
    ELSE
      degree:=Zel/SumWithZ;
    END;
    el.l[0].degree:=degree;

    FOR Z:=1 TO LEN(el.l^)-1 DO
      l:=el.l[Z];
      IF Zav=0 THEN
        degree:=0
      ELSE
        degree:=degree*l.K/(Zav*n);
      END;
      l.degree:=degree;
    END;
  END CalculateDegreeOfElement;

PROCEDURE CalculateDegree (els-:Elements; n,Zav:Float);
  VAR
    i:LONGINT;
  BEGIN
    FOR i:=0 TO LEN(els)-1 DO
      CalculateDegreeOfElement(els[i],n,Zav);
    END;
  END CalculateDegree;

PROCEDURE Debye (els-:Elements; n,T,ld,Zav:Float): Float;
  (* This function is Deby radius as function of the specified variables,
     including ld, which is Debye redius :) *)
  VAR
    sum:Float;
    i,Z:LONGINT;
  BEGIN
    CalculateK(els,T,ld);
    CalculateDegree(els,n,Zav);

    sum:=Zav;
    FOR i:=0 TO LEN(els)-1 DO
      FOR Z:=1 TO LEN(els[i].l^)-1 DO
        sum:=sum+Z*Z*els[i].l[Z].degree;
      END;
    END;
    IF sum=0 THEN
      RETURN Math.MaxFloat
    ELSE
      RETURN 1/sqrt(4*pi*e*e*n*sum/(k*T));
    END;
  END Debye;

TYPE
  F1Func = RECORD (AlgEqu.BiFunc)
    els:ElementsPtr;
    n:Float;
  END;

PROCEDURE (VAR f:F1Func) f (Zav:Float): Float;
  (* All Ks hsould already be calculated *)
  VAR
    sum,SumWithZ,SumWithoutZ:Float;
    i:LONGINT;
  BEGIN
    ASSERT(f.els#NIL);

    sum:=0;
    FOR i:=0 TO LEN(f.els^)-1 DO
      GetSums(f.els[i],f.n,Zav,SumWithZ,SumWithoutZ);
      sum:=sum+f.els[i].nrel*SumWithZ/(1+SumWithoutZ);
    END;

    RETURN Zav - sum
  END f;

PROCEDURE Zav (els:ElementsPtr; n,T,ld:Float): Float;
  (* Calculates Zav as function of ld and other parameters *)
  CONST
    x1 = 1.0E-35;
    x2 = 10;
    RelErr = 1.0E-6;
  VAR
    f:F1Func;
    res:Float;
  BEGIN
    ASSERT(els#NIL);

    CalculateK(els^,T,ld);

    f.els:=els;
    f.n:=n;
    IF f.f(x1)*f.f(x2)>0 THEN (* it means that ionization degree is too small *)
      res:=0
    ELSE
      res:=AlgEqu.Bisection(f,x1,x2,RelErr,-1);
    END;
    RETURN res;
  END Zav;

TYPE
  Func = RECORD (AlgEqu.BiFunc)
    n,T:Float;
    els:ElementsPtr;
  END;

PROCEDURE (VAR f:Func) f (ld:Float): Float;
  BEGIN
    RETURN ld-Debye(f.els^,f.n,f.T,ld,Zav(f.els,f.n,f.T,ld))
  END f;

PROCEDURE WithEnergyCorrection * (els:ElementsPtr; n,T:Float);
  (** Calculates the degrees of ionization and returns them in els[i].l[Z].degree.
      n is total number density of all the elements, T is temperature in K *)
  CONST
    x1 = 0;
    x2 = 2.0E-5;
    RelErr = 1.0E-5;
  VAR
    f:Func;
    ld,zav:Float;
  BEGIN
    ASSERT(els#NIL);
    ASSERT(n>0);
    ASSERT(T>0);

    f.els:=els;
    f.n:=n;
    f.T:=T;
    IF f.f(x1)*f.f(x2)>0 THEN
      ld:=Math.MaxFloat;
    ELSE
      ld:=AlgEqu.Bisection(f,x1,x2,RelErr,-1);
    END;

    zav:=Zav(els,n,T,ld);
    CalculateDegree(els^,n,zav);
  END WithEnergyCorrection;

PROCEDURE WithoutEnergyCorrection * (els:ElementsPtr; n,T:Float);
  (** The same as CalculateWithEnergyCorrection but without correction
      of ionization potentials. Faster than CalculateWithEnergyCorrection *)
  VAR
    zav:Float;
  BEGIN
    ASSERT(els#NIL);
    ASSERT(n>0);
    ASSERT(T>0);

    zav:=Zav(els,n,T,Math.MaxFloat);
    CalculateDegree(els^,n,zav);
  END WithoutEnergyCorrection;

PROCEDURE DebyeRad * (els-:Elements; n,T:Float): Float;
  (** Returns Debye radius of the elemens with the specified total number density
      n and temperature T. els should already be calculated. *)
  VAR
    sum:Float;
    i,Z:LONGINT;
  BEGIN
    sum:=0;
    FOR i:=0 TO LEN(els)-1 DO
      FOR Z:=1 TO LEN(els[i].l^)-1 DO
        sum:=sum+(Z*Z+1)*els[i].l[Z].degree;
      END;
    END;
    IF sum=0 THEN
      RETURN Math.MaxFloat
    ELSE
      RETURN 1/sqrt(4*pi*e*e*n*sum/(k*T));
    END;
  END DebyeRad;

PROCEDURE nerel * (els-:Elements): Float;
  (** Returns relative electron number density of the specified distribution,
      i.e. [electron number density]/[number density of all atoms and ions of all elements] *)
  VAR
    i,Z:LONGINT;
    sum:Float;
  BEGIN
    sum:=0;
    FOR i:=0 TO LEN(els)-1 DO
      FOR Z:=1 TO LEN(els[i].l^)-1 DO
        sum:=sum+els[i].l[Z].degree;
      END;
    END;
    RETURN sum
  END nerel;

END LiLTEDistribution.