<* O2EXTENSIONS + *>
<* O2NUMEXT + *>
MODULE Scenario;

IMPORT
  Math:=mMath, Protocol, COMPILER, DStr:=msysDynStr, Geo:=Geometry,
  Area:=mArea2D, GasData:=LiData, TeEqu, Comsol, ComFunc:=mCommonFunctions,
  EMM, TeEquData, PHC:=msciPhConst, Out:=mhostStdOut, InOut, GeoArea,
  TextData:=msysTextData, TeAndEmm, LiData, LiGas, TOut:=msysTextOut,
  Abel:=mAbelPoly, Com:=Common, HU:=mhostUtils;

TYPE
  Float = Math.Float;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;

  String = DStr.String;
  Func = ComFunc.Float2;
  FuncPtr = POINTER TO Func;

CONST
  ModuleName = COMPILER.EQUATION('module');

  RelErr = 1.0E-3;

  PIFileName = Com.PIFileName;
  ArPropFileName = 'Ar properties.txt';

  ComsolFzFileName = 'COMSOL_Fz.txt';
  ComsolFrFileName = 'COMSOL_Fr.txt';
  ComsolPowerFileName = 'COMSOL_Power.txt';

  TFileName = 'COMSOL_T.txt'; (* Temperature file name *)
  VzFileName = 'COMSOL_Vz.txt'; (* Vz file name *)
  VrFileName = 'COMSOL_Vr.txt'; (* Vr file name *)

  TeFileName = 'Te.txt';
  DEFileName = 'DeltaE.txt';

  (* Default values for areas out of plasma: *)
  DefaultT = 300;
  DefaultVz = 0;
  DefaultVr = 0;

  Tab = Out.Tab;

(* ====================================== Service procedures ================ *)

TYPE
  EnergyExchangeFunc = RECORD (ComFunc.Float2) (* energy exchange as function of T and Te *)
    T:FuncPtr;
    Te:FloatArray2Ptr;
  END;

PROCEDURE (VAR f:EnergyExchangeFunc) f (z,r:Float): Float;
  VAR
    T,Te,ne,na:Float;
  BEGIN
    ASSERT(f.T#NIL);
    ASSERT(f.Te#NIL);

    r:=ABS(r);
    IF (z>=TeEqu.area.MinX) & (z<=TeEqu.area.MaxX) &
       (r>=TeEqu.area.MinY) & (r<=TeEqu.area.MaxY) &
       (Geo.GetSpaceType(z,r)=Geo.In)
    THEN
      T:=f.T.f(z,r);
      Te:=GeoArea.Interpolate(TeEqu.area,f.Te^,z,r);
      TeEquData.NeNa(T,Te,ne,na);
      RETURN TeEquData.EnergyExchange(T,Te,ne);
    ELSE
      RETURN 0
    END;
  END f;

PROCEDURE SaveForCOMSOL (VAR power:Func);
  (* Saves power density and Lorentz forces in the COMSOL format. *)
  CONST
    MinZ = Comsol.MinZ;
    MaxZ = Comsol.MaxZ;
    Rad = Comsol.Rad;
    dz = 0.01; (* cm *)
    dr = dz; (* cm *)
    LastZ = ENTIER((MaxZ-MinZ)/dz+0.5);
    LastR = ENTIER(Rad/dr+0.5);
  VAR
    a:Area.Area;
  BEGIN
    Out.String('Saving data for COMSOL... ',0);

    Area.Init(a,LastZ,LastR,MinZ,MaxZ,0,Rad);
    Comsol.SaveSimpleFunc(a,EMM.Fz,ComsolFzFileName,1/(PHC.Newton/PHC.meter**3));
    Comsol.SaveSimpleFunc(a,EMM.Fr,ComsolFrFileName,1/(PHC.Newton/PHC.meter**3));
    Comsol.SaveFunc(a,power,ComsolPowerFileName,1/(PHC.Watt/PHC.meter**3));

    Out.Ln;

  END SaveForCOMSOL;

PROCEDURE LoadFromComsol (FileName-:ARRAY OF CHAR; factor,default:Float): POINTER TO GeoArea.InterpolationProc;
  (** Loads the gas dynamics data from the last COMSOL iteration.
      Should be called before all the other peocedures *)
  CONST
    MinZ = Comsol.MinZ;
    MaxZ = Comsol.MaxZ;
    Rad = Comsol.Rad;
    dz = 0.01; (* cm *)
    dr = 0.01; (* cm *)
    LastZ = ENTIER((MaxZ-MinZ)/dz+0.5);
    LastR = ENTIER(Rad/dr+0.5);
  VAR
    a:Area.Area;
  BEGIN
    Area.Init(a,LastZ,LastR,MinZ,MaxZ,0,Rad);
    RETURN Comsol.LoadFunc(a,FileName,factor,default);
  END LoadFromComsol;

PROCEDURE SaveEMM (factor:Float);
  (* Saves all the data which are results of the current magnetic field *)
  CONST
    MinZ = EMM.MinZ;
    MaxZ = EMM.MaxZ;
    Rad = EMM.Rad;
    dz = 0.02; (* cm *)
    dr = dz; (* cm *)
    LastZ = ENTIER((MaxZ-MinZ)/dz+0.5);
    LastR = ENTIER(Rad/dr+0.5);
  VAR
    a:Area.Area;
  BEGIN
    Out.String('Saving EMM results... ',0);

    Area.Init(a,LastZ,LastR,MinZ,MaxZ,0,Rad);

    InOut.SaveSimpleFuncC(a,EMM.Ephi,'EPhi.txt',1/(PHC.Volt/PHC.meter) *factor);
    InOut.SaveSimpleFuncC(a,EMM.Bz,'Bz.txt',1/(PHC.AmperPerMeter) *factor);
    InOut.SaveSimpleFuncC(a,EMM.Br,'Br.txt',1/(PHC.AmperPerMeter) *factor);
    InOut.SaveSimpleFunc(a,EMM.Power,'Power.txt',1/(PHC.Watt/PHC.meter**3) *factor*factor);
    InOut.SaveSimpleFuncC(a,EMM.GetA,'A.txt',1/(PHC.Tesla*PHC.meter) *factor);
    InOut.SaveSimpleFunc(a,EMM.Fz,'Fz.txt',1/(PHC.Newton/PHC.meter**3) *factor*factor);
    InOut.SaveSimpleFunc(a,EMM.Fr,'Fr.txt',1/(PHC.Newton/PHC.meter**3) *factor*factor);
    InOut.SaveSimpleFunc(a,EMM.Sigma,'Sigma.txt',1/(1/(PHC.Ohm*PHC.meter)));

    Out.Ln;
  END SaveEMM;

PROCEDURE SaveTemperatures (area-:Area.Area; VAR T:Func; Te-:FloatArray2);
  PROCEDURE SaveLine (fm:TOut.Formatter; z,r,val:Float);
    CONST
      tab = TOut.Tab;
    BEGIN
      TOut.Real(fm,z,0); TOut.Char(fm,tab);
      TOut.Real(fm,r,0); TOut.Char(fm,tab);
      TOut.Real(fm,val,0); TOut.Ln(fm);
    END SaveLine;
  VAR
    iZ,iR:LONGINT;
    z,r,TT,TTe,ne,na,rad:Float;
    TeFile,RadFile:TOut.Formatter;
    AbelRad:FloatArray2Ptr;
  CONST
    RadFileName = 'Radiation.txt';
    AbelRadFileName = 'AbelRadiation.txt';
  BEGIN
    ASSERT(Area.Correct(area));
    ASSERT(Area.CorrectField(area,Te));

    TOut.OpenFile(TeFile,TeFileName,TOut.Rewrite);
    TOut.OpenFile(RadFile,RadFileName,TOut.Rewrite);

    NEW(AbelRad,area.LastX+1,area.LastY+1);
    FOR iZ:=0 TO area.LastX DO
      z:=Area.IndToX(area,iZ);
      FOR iR:=0 TO area.LastY DO
        r:=Area.IndToY(area,iR);
        TT:=T.f(z,r);
        TTe:=Te[iZ,iR];
        TeEquData.NeNa(TT,TTe,ne,na);
        ASSERT(TTe#0);
        rad:=ne*ne/Math.sqrt(TTe);
        AbelRad[iZ,iR]:=rad;

        SaveLine(TeFile,z,r,TTe);
        SaveLine(RadFile,z,r,rad);
      END;
    END;

    TOut.Close(TeFile);
    TOut.Close(RadFile);

    FOR iZ:=0 TO area.LastX DO
      Abel.Direct(AbelRad[iZ]);
    END;
    Area.SaveArray(area,AbelRad^,AbelRadFileName,1);

    InOut.CopyFile(TeFileName);
    InOut.CopyFile(RadFileName);
    InOut.CopyFile(AbelRadFileName);
  END SaveTemperatures;

(* ====================================== 2-temperature model =============== *)

PROCEDURE TwoTemp (input:Float; CalcWithPower:BOOLEAN);
  (* Calculates the plasma using 2-temperature model.
     If CalcWithPower then input is the desired plasma power in Watts,
     else the input is the coil current in Ampers. *)
  VAR
    T,vz,vr:FuncPtr;
    Te:FloatArray2Ptr;
    i:LONGINT;
    power,current,val,OldVal:Float;
    tEqu:TeEqu.Equ;
    ee:EnergyExchangeFunc;
  BEGIN
    IF CalcWithPower THEN
      power:=input;
      current:=0;
      Protocol.Real(ModuleName+'.TwoTemp  DesiredPower = ',power/PHC.Watt,' W');
      val:=0;
      TextData.AppendString(PIFileName,'Current (A)');
      TextData.AppendLn(PIFileName);
    ELSE
      current:=input;
      power:=0;
      Protocol.Real(ModuleName+'.TwoTemp  Current = ',current/PHC.Amper,' A');
      val:=0;
      TextData.AppendString(PIFileName,'Power (W)');
      TextData.AppendLn(PIFileName);
    END;

    TeEquData.ExportArProperties(ArPropFileName);

    T:=ComFunc.Constant2(300);
    vz:=ComFunc.Constant2(0);
    vr:=vz;
    NEW(Te,TeEqu.LastZ+1,TeEqu.LastR+1);
    TeEqu.Init(tEqu);
    i:=1;
    REPEAT
      OldVal:=val;
      Protocol.Int('Iteration ',i,'');
      InOut.NextIteration;

      IF CalcWithPower THEN
        current:=TeAndEmm.CalculateWithPower(tEqu,T,vz,vr,Te,power);
        Protocol.Real('  Current = ',current/PHC.Amper,' A');
        TextData.AppendString(PIFileName,DStr.FromReal(current/PHC.Amper)^);
        TextData.AppendLn(PIFileName);
        val:=current;
      ELSE
        power:=TeAndEmm.CalculateWithCurrent(tEqu,T,vz,vr,Te,current);
        Protocol.Real('  Power = ',power/PHC.Watt,' W');
        TextData.AppendString(PIFileName,DStr.FromReal(power/PHC.Watt)^);
        TextData.AppendLn(PIFileName);
        val:=power/2; (* /2 - to reduce DeltaVal in 2 times, because power is proportional to current^2 *)
      END;

      InOut.SaveArray(TeEqu.area,Te^,TeFileName,1);
      SaveEMM(1);
      ee.T:=T;
      ee.Te:=Te;
      InOut.SaveFunc(TeEqu.area,ee,DEFileName,1);

      SaveForCOMSOL(ee);
      Comsol.Launch('');
      InOut.CopyFile('Torch temperature.txt');
      InOut.CopyFile(Comsol.OutputFile);
      InOut.CopyFile(ArPropFileName);

      Out.String('Loading data calculated by COMSOL... ',0);
      T:=LoadFromComsol(TFileName,1,DefaultT);
      vz:=LoadFromComsol(VzFileName,PHC.meter/PHC.second,DefaultVz);
      vr:=LoadFromComsol(VrFileName,PHC.meter/PHC.second,DefaultVr);
      SaveTemperatures(TeEqu.area,T^,Te^);
      Out.Ln;

      INC(i);
    UNTIL ABS(val-OldVal)<=RelErr*ABS(val)
  END TwoTemp;

PROCEDURE TwoTempWithPower * (DesiredPower:Float);
  (** Calculates the plasma using 2-temperature model
      with the desired power (in Watts). *)
  BEGIN
    TwoTemp(DesiredPower,TRUE);
  END TwoTempWithPower;

PROCEDURE TwoTempWithCurrent * (current:Float);
  (** Calculates the plasma using 2-temperature model
      with the desired coil current (in Ampers). *)
  BEGIN
    TwoTemp(current,FALSE);
  END TwoTempWithCurrent;

(* ====================================== LTE =============================== *)

TYPE
  PowerDensity = RECORD (Func)
  END;

PROCEDURE (VAR p:PowerDensity) f (z,r:Float): Float;
  BEGIN
    RETURN EMM.Power(z,r)
  END f;

TYPE
  Sigma = RECORD (Func) (* electrical conductivity *)
    T:FuncPtr;
  END;

PROCEDURE (VAR s:Sigma) f (z,r:Float): Float;
  (* Returns electrical conductivity *)
  CONST
    p = PHC.Atmosphere;
  VAR
    n,T:Float;
  BEGIN
    ASSERT(s.T#NIL);

    r:=ABS(r);
    IF (z<Geo.PlasmaMinZ) OR (z>Geo.PlasmaMaxZ) OR (r>Geo.PlasmaRad)
       OR (Geo.GetSpaceType(z,r)#Geo.In)
    THEN
      RETURN 0
    ELSE
      T:=s.T.f(z,r);
      n:=p/(PHC.k*T);
      LiGas.Calculate(n,T);
      RETURN LiGas.ElectricalConductivity();
    END;
  END f;

TYPE
  InitialSigma = RECORD (Func) (* initial electrical conductivity *)
  END;

PROCEDURE (VAR s:InitialSigma) f (z,r:Float): Float;
  (* Returns electrical conductivity *)
  BEGIN
    RETURN Geo.InitialElectricalConductivity(z,r)
  END f;

PROCEDURE LTE (input:Float; CalcWithPower:BOOLEAN);
  (* Calculates the plasma using LTE model.
     If CalcWithPower then input is the desired plasma power in Watts,
     else the input is the coil current in Ampers. *)
  VAR
    T:POINTER TO GeoArea.InterpolationProc;
    i,j:LONGINT;
    power,current,val,OldVal:Float;
    pd:PowerDensity;
    ms:POINTER TO Sigma; (* main sigma *)
    is:POINTER TO InitialSigma; (* initial sigma *)
    s:POINTER TO Func; (* sigma *)

  PROCEDURE ActionsAfterComsol();
    BEGIN
      InOut.CopyFile('Torch temperature.txt');
      InOut.CopyFile(Comsol.OutputFile);
      InOut.CopyFile(ArPropFileName);

      Out.String('Loading data calculated by COMSOL... ',0);
      T:=LoadFromComsol(TFileName,1,DefaultT);

      SaveTemperatures(T.area,T^,T.field^);
      Out.Ln;

      ms.T:=T;
      s:=ms;

      INC(i);
    END ActionsAfterComsol;

  BEGIN
    i:=Com.LastIteration(val);
    IF i=0 THEN
      LiData.ExportTemperature(ArPropFileName);

      IF CalcWithPower THEN
        power:=input;
        current:=0;
        Protocol.Real(ModuleName+'.LTE  DesiredPower = ',power/PHC.Watt,' W');
        TextData.AppendString(PIFileName,'Current (A)');
        TextData.AppendLn(PIFileName);
      ELSE
        current:=input;
        power:=0;
        Protocol.Real(ModuleName+'.LTE  current = ',current/PHC.Amper,' A');
        TextData.AppendString(PIFileName,'Power (W)');
        TextData.AppendLn(PIFileName);
      END;

      NEW(is);
      NEW(ms);
      s:=is;

      i:=1;
      val:=0;
    ELSE
      Protocol.allowed:=TRUE;
      Protocol.ProgramInfo();

      IF CalcWithPower THEN
        power:=input;
        current:=val;
      ELSE
        current:=input;
        power:=val;
        val:=power/2; (* /2 - to reduce DeltaVal in 2 times, because power is proportional to current^2 *)
      END;

      FOR j:=1 TO i DO
        InOut.NextIteration();
      END;

      IF HU.DeleteFile(Comsol.StatusFileName) THEN END;
      IF HU.DeleteFile(Protocol.ScreenFileName) THEN END;

      NEW(ms);

      ActionsAfterComsol();
    END;

    REPEAT
      OldVal:=val;
      Protocol.Int('Iteration ',i,'');
      InOut.NextIteration;
      EMM.SetSigma(s^);
      EMM.Calculate(1);

      IF CalcWithPower THEN
        current:=Math.sqrt(power/EMM.TotalPower());
        EMM.Multiply(current);
        val:=current;
        Protocol.Real('  Current = ',current/PHC.Amper,' A');
        TextData.AppendString(PIFileName,DStr.FromReal(current/PHC.Amper)^);
        TextData.AppendLn(PIFileName);
      ELSE
        EMM.Multiply(current);
        power:=EMM.TotalPower();
        val:=power/2; (* /2 - to reduce DeltaVal in 2 times, because power is proportional to current^2 *)
        Protocol.Real('  Power = ',power/PHC.Watt,' W');
        TextData.AppendString(PIFileName,DStr.FromReal(power/PHC.Watt)^);
        TextData.AppendLn(PIFileName);
      END;
      SaveEMM(1);

      SaveForCOMSOL(pd);
      Comsol.Launch('');
      ActionsAfterComsol();
    UNTIL ABS(OldVal-val)<=RelErr*ABS(val)
  END LTE;

PROCEDURE LTEWithPower * (DesiredPower:Float);
  (** Calculates the plasma using LTE model
      with the desired power (in Watts). *)
  BEGIN
    LTE(DesiredPower,TRUE);
  END LTEWithPower;

PROCEDURE LTEWithCurrent * (current:Float);
  (** Calculates the plasma using LTE model
      with the coil current (in Ampers). *)
  BEGIN
    LTE(current,FALSE);
  END LTEWithCurrent;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');

  GasData.Install();

  Protocol.Real('RelErr = ',RelErr,'');
  Protocol.Real('DefaultT = ',DefaultT,' K');
  Protocol.Real('DefaultVz = ',DefaultVz/(PHC.cm/PHC.second),' cm/s');
  Protocol.Real('DefaultVr = ',DefaultVr/(PHC.cm/PHC.second),' cm/s');
  Protocol.Real('RelErr = ',RelErr,'');
  Protocol.String('','','');
END Scenario.