(** Solves EMM and Te equations consistently as function of gas temperature T,
    gas velocity and the desired plasma power *)
<* O2EXTENSIONS + *>
MODULE TeAndEmm;

IMPORT
  TeEqu, EMM, Math:=mMath, ComFunc:=mCommonFunctions, Geo:=Geometry,
  Area:=mCylArea, TeEquData, PHC:=msciPhConst, Arrays:=mArrays,
  Out:=mhostStdOut, PEqu:=mCylParabolicEqu, COMPILER, Protocol, GeoArea;

CONST
  RelErr = 1.0E-3; (* smaller RelErr does not work because of stabilization in TeEqu :( *)
  tau = 0.1;
  IterationLimit = 100;

  ModuleName = COMPILER.EQUATION('module');

TYPE
  Float = Math.Float;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;

  Func = ComFunc.Float2;
  FuncPtr = POINTER TO Func;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

TYPE
  Sigma = RECORD (Func) (* electrical conductivity *)
    T:FuncPtr;
    Te:FloatArray2Ptr;
  END;

PROCEDURE (VAR s:Sigma) f (z,r:Float): Float;
  (* Returns electrical conductivity *)
  VAR
    T,Te:Float;
  BEGIN
    ASSERT(s.T#NIL);
    ASSERT(s.Te#NIL);

    IF (z<TeEqu.area.MinX) OR (z>TeEqu.area.MaxX) OR (r<TeEqu.area.MinY) OR (r>TeEqu.area.MaxY)
       OR (Geo.GetSpaceType(z,r)#Geo.In)
    THEN
      RETURN 0
    ELSE
      T:=s.T.f(z,r);
      Te:=GeoArea.Interpolate(TeEqu.area,s.Te^,z,r);
      RETURN TeEquData.ElectricalConductivity(T,Te);
    END;
  END f;

TYPE
  P = RECORD (Func) (* power density *)
  END;

PROCEDURE (VAR p:P) f (z,r:Float): Float;
  BEGIN
    r:=ABS(r);
    RETURN EMM.Power(z,r);
  END f;

VAR
  p:POINTER TO P; (* power density *)

PROCEDURE SetMax (VAR field:FloatArray2; a-:Area.Area; VAR F:Func);
  (* field[iZ,iR]:=Max(Field[iZ,iR],F.f(z,r)) *)
  VAR
    iZ,iR:LONGINT;
    v,z,r:Float;
  BEGIN
    ASSERT(Area.Correct(a));
    ASSERT(Area.CorrectField(a,field));

    FOR iZ:=0 TO a.LastX DO
      z:=Area.IndToZ(a,iZ);
      FOR iR:=0 TO a.LastY DO
        r:=ABS(Area.IndToR(a,iR));
        v:=F.f(z,r);
        IF field[iZ,iR]<v THEN
          field[iZ,iR]:=v
        END;
      END;
    END;
  END SetMax;

PROCEDURE CombineSolutions (VAR new:FloatArray2; old-:FloatArray2);
  (* new:=combination(new,old) *)
  VAR
    iZ,iR:LONGINT;
  BEGIN
    ASSERT(LEN(old)=LEN(new));
    ASSERT(LEN(old,1)=LEN(new,1));

    FOR iZ:=0 TO LEN(old)-1 DO
      FOR iR:=0 TO LEN(old,1)-1 DO
        new[iZ,iR]:= tau*new[iZ,iR] + (1-tau)*old[iZ,iR];
      END;
    END;
  END CombineSolutions;

PROCEDURE Calculate (VAR tEqu:TeEqu.Equ; T,vz,vr:FuncPtr; Te:FloatArray2Ptr;
                     input:Float; VAR output:Float; CalcWithPower:BOOLEAN);
  (* Calculates EMM and Te consistently as function of gas temperature T,
     gas velocity vz and vr.
     Depending on CalcWithPower, power or current are used as input.
     Then current or power are returned in the output, respectively. *)
  VAR
    s:Sigma;
    current,power:Float;

  PROCEDURE CalcEmm;
    VAR
      factor:Float;
    BEGIN
      EMM.SetSigma(s);
      EMM.Calculate(current);
      IF CalcWithPower THEN
        factor:=sqrt(power/EMM.TotalPower());
        EMM.Multiply(factor);
        current:=current*factor;
      ELSE
        power:=EMM.TotalPower();
      END;
    END CalcEmm;

  VAR
    disc,MaxF:Float;
    OldTe:FloatArray2Ptr;
    it:LONGINT;
  BEGIN
    ASSERT(TeEqu.Initialized(tEqu));
    ASSERT(T#NIL);
    ASSERT(vz#NIL);
    ASSERT(vr#NIL);
    ASSERT(Te#NIL);
    ASSERT(Area.CorrectField(TeEqu.area,Te^));

    IF CalcWithPower THEN
      power:=input;
      ASSERT(power>0);
      current:=30*PHC.Amper;
    ELSE
      current:=input;
      power:=0;
    END;

    Arrays.Fill2(Te^,5000);
    s.T:=T;
    s.Te:=Te;
    CalcEmm;
    TeEqu.Local(T^,p^,Te^);
    OldTe:=Arrays.Copy2(Te^);
    it:=1; (* one iteration is already done *)

    REPEAT
      CalcEmm;
      TeEqu.SetCoeffs(tEqu,T,vz,vr,p,Te);
      TeEqu.Calculate(tEqu,Te);
      SetMax(Te^,TeEqu.area,T^);
      MaxF:=Arrays.MaxAbs2(tEqu.equ.f^);
      disc:=PEqu.NormOfRHS(tEqu.equ^,Te^)/MaxF;
      Out.String(', RHS_RelErr = ',0); Out.Real(disc,0); Out.Ln;

      CombineSolutions(Te^,OldTe^);
      Arrays.Assign2(OldTe^,Te^);
      disc:=PEqu.NormOfRHS(tEqu.equ^,Te^)/MaxF;
      IF CalcWithPower THEN
        Out.String('    Current = ',0); Out.Real(current/PHC.Amper,0);
      ELSE
        Out.String('    Power = ',0); Out.Real(power/PHC.Watt,0);
      END;
      Out.String(', RHS_RelErr = ',0); Out.Real(disc,0); Out.Ln;
      INC(it);
    UNTIL (disc<=RelErr) OR (it>IterationLimit);
    Out.Int(it,11); Out.String(' iterations',0); Out.Ln;
    Protocol.Int('  ',it,' iterations');
    IF it>IterationLimit THEN
      Protocol.String('','    Iteration limit is reached','');
    END;

    IF CalcWithPower THEN
      output:=current
    ELSE
      output:=power
    END;
  END Calculate;

PROCEDURE CalculateWithPower * (VAR tEqu:TeEqu.Equ; T,vz,vr:FuncPtr; Te:FloatArray2Ptr;
                                DesiredPower:Float): Float;
  (** Calculates EMM and Te consistently as function of gas temperature T,
      gas velocity vz and vr, and the desired total power.
      Returns the resulting coil current. *)
  VAR
    current:Float;
  BEGIN
    Calculate(tEqu,T,vz,vr,Te,DesiredPower,current,TRUE);
    RETURN current;
  END CalculateWithPower;

PROCEDURE CalculateWithCurrent * (VAR tEqu:TeEqu.Equ; T,vz,vr:FuncPtr; Te:FloatArray2Ptr;
                                  current:Float): Float;
  (** Calculates EMM and Te consistently as function of gas temperature T,
      gas velocity vz and vr, and the coil current.
      Returns the resulting total dissipated power. *)
  VAR
    power:Float;
  BEGIN
    Calculate(tEqu,T,vz,vr,Te,current,power,FALSE);
    RETURN power;
  END CalculateWithCurrent;

BEGIN
  Protocol.String('MODULE ',ModuleName,':');

  NEW(p);

  Protocol.Real('RelErr = ',RelErr,'');
  Protocol.Real('tau = ',tau,'');
  Protocol.Int('IterationLimit = ',IterationLimit,'');
  Protocol.String('','','');
END TeAndEmm.