(** Solves elliptical equation for Te.
    T is gas temperature, P is density of electromagnetic power,
    vr and vr are z- and r- components of the gas speed *)
<* O2EXTENSIONS + *>
MODULE TeEqu;

IMPORT
  Data:=TeEquData, Geo:=Geometry, PHC:=msciPhConst,
  PEqu:=mCylParabolicEqu, Math:=mMath, COMPILER, Area:=mCylArea,
  AlgEqu:=mAlgEqu, Protocol, Out:=mhostStdOut, ComFunc:=mCommonFunctions,
  Arrays:=mArrays, Steps:=mEvolutionSteps, GeoArea;

CONST
  cm = PHC.cm;

  MinZ * = Geo.PlasmaMinZ;
  MaxZ * = Geo.PlasmaMaxZ;
  Rad * = Geo.PlasmaRad;
  dz = 0.02*cm;
  dr = dz;
  LastZ * = ENTIER((MaxZ-MinZ)/dz+0.5); (** z-dimension of the Te field *)
  LastR * = ENTIER(Rad/dz+0.5);         (** r-dimension of the Te field *)

  RelErr = 1.0E-5; (* Relative error for Te calculation *)
  ThermalConductivityThreshold = 1.0E-5*PHC.Watt/(PHC.meter*PHC.Kelvin);
    (* Thermal conductivity is limites from down by the ThermalConductivityThreshold
        to improve stability *)
  l_min_factor = 10000;

  eps = 15.7596*PHC.eV; (* ionization energy of Ar *)
  k = PHC.k;
  Me = PHC.Me; (* mass of electron *)
  Ma = 40*PHC.amu; (* mass of Ar atom *)
  ModuleName = COMPILER.EQUATION('module');

TYPE
  Float = Math.Float;
  FloatArray2 = Math.Array2;
  FloatArray2Ptr = Math.Array2Ptr;
  Func = ComFunc.Float2;
  SpaceTypeArrayPtr * = POINTER TO ARRAY OF ARRAY OF Geo.SpaceType;

VAR
  area - : Area.Area; (** Area of the Te equation *)
  ReducedArea - :Area.Area; (* Area without outer points *)
  spaceType - :SpaceTypeArrayPtr; (* tabulated space type in points of the grid *)

TYPE
  BiFunc = RECORD (AlgEqu.BiFunc)
    T,P0:Float;
  END;

PROCEDURE (VAR f: BiFunc) f (Te:Float): Float;
  VAR
    T,ne,na:Float;
  BEGIN
    T:=f.T;
    Data.NeNa(T,Te,ne,na);
    RETURN f.P0 - Data.RadiationLoss(Te,ne) - Data.EnergyExchange(T,Te,ne);
  END f;

PROCEDURE Calculate0 (T,P0:Float): Float;
  (* Returns Te, which is result of the specified T (gas temperature)
     and P (power consumption density) in local approximation *)
  VAR
    f:BiFunc;
    res:Float;
  BEGIN
    f.T:=T;
    f.P0:=P0;
    res:=AlgEqu.Bisection(f,100,15000,RelErr,-1);
    IF res<T THEN
      RETURN T
    ELSE
      RETURN res
    END;
  END Calculate0;

PROCEDURE JustifyBoundary (VAR Te:FloatArray2; VAR T:Func);
  (* Sets boundary values of the array to T *)
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
  BEGIN
    ASSERT(LEN(Te)=LastZ+1);
    ASSERT(LEN(Te,1)=LastR+1);

    FOR iZ:=1 TO LastZ-1 DO
      z:=Area.IndToZ(area,iZ);
      Te[iZ,0]:=Te[iZ,1];
      Te[iZ,LastR]:=T.f(z,Rad);
    END;
    FOR iR:=1 TO LastR-1 DO
      r:=Area.IndToR(area,iR);
      Te[0,iR]:=T.f(MinZ,r);
      Te[LastZ,iR]:=T.f(MaxZ,r);
    END;
    Te[0,0]:=(Te[0,1]+Te[1,0])/2;
    Te[LastZ,LastR]:=(Te[LastZ-1,LastR]+Te[LastZ,LastR-1])/2;
    Te[LastZ,0]:=(Te[LastZ-1,0]+Te[LastZ,1])/2;
    Te[0,LastR]:=(Te[1,LastR]+Te[0,LastR-1])/2;
  END JustifyBoundary;

PROCEDURE Local * (VAR T,P:Func; VAR Te:FloatArray2);
  (** Returns Te as function of T and P in local approximation, i.e. diffusion
      and transfer by gas flow are neglected *)
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
  BEGIN
    ASSERT(Area.CorrectField(area,Te));

    Out.String('Local Te... ',0);

    FOR iZ:=1 TO LastZ-1 DO
      z:=Area.IndToZ(area,iZ);
      FOR iR:=1 TO LastR-1 DO
        r:=Area.IndToR(area,iR);
        IF spaceType[iZ,iR]=Geo.In THEN
          Te[iZ,iR]:=Calculate0(T.f(z,r),P.f(z,r));
        ELSE
          Te[iZ,iR]:=T.f(z,r);
        END;
      END;
    END;
    JustifyBoundary(Te,T);

    Out.Ln;
  END Local;

(* Derivatives for the setting of the coefficients: *)

TYPE
  DDDz = RECORD (PEqu.DerFunc)
  END;

PROCEDURE (VAR d:DDDz) f (iZ,iR:LONGINT; D-:FloatArray2): Float;
  BEGIN
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;
    DEC(iZ); (* adaptation of indexes to the reduces area field D *)
    DEC(iR);
    RETURN GeoArea.DiffZInd(ReducedArea,D,iZ,iR)
  END f;

TYPE
  DDDr = RECORD (PEqu.DerFunc)
  END;

PROCEDURE (VAR d:DDDr) f (iZ,iR:LONGINT; D-:FloatArray2): Float;
  BEGIN
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;
    DEC(iZ); (* adaptation of indexes to the reduces area field D *)
    DEC(iR);
    RETURN GeoArea.DiffRInd(ReducedArea,D,iZ,iR)
  END f;

TYPE
  DVzDz = RECORD (PEqu.DerFunc)
  END;

PROCEDURE (VAR d:DVzDz) f (iZ,iR:LONGINT; D-:FloatArray2): Float;
  BEGIN
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    ELSE
      RETURN GeoArea.DiffZInd(ReducedArea,D,iZ-1,iR-1)
    END;
  END f;

TYPE
  DVrDr = RECORD (PEqu.DerFunc)
  END;

PROCEDURE (VAR d:DVrDr) f (iZ,iR:LONGINT; D-:FloatArray2): Float;
  BEGIN
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    ELSE
      RETURN GeoArea.DiffRInd(ReducedArea,D,iZ-1,iR-1)
    END;
  END f;

TYPE
  Equ * = RECORD (** Contains data for solution of the Te elliptical equation *)
    P,T,ne:FloatArray2Ptr; (* pool of already deposited fields *)
    equ - :POINTER TO PEqu.Equ; (** The equation *)
  END;

PROCEDURE Init * (VAR equ:Equ);
  (** Initiates the equ *)
  VAR
    dddz:POINTER TO DDDz;
    dddr:POINTER TO DDDr;
    dvzdz:POINTER TO DVzDz;
    dvrdr:POINTER TO DVrDr;
  BEGIN
    NEW(equ.P,LastZ+1,LastR+1);
    NEW(equ.T,LastZ+1,LastR+1);
    NEW(equ.ne,LastZ+1,LastR+1);

    NEW(equ.equ);
    PEqu.Init(equ.equ,area);

    NEW(dddz);
    equ.equ.DDDz:=dddz;
    NEW(dddr);
    equ.equ.DDDr:=dddr;
    NEW(dvzdz);
    equ.equ.DVzDz:=dvzdz;
    NEW(dvrdr);
    equ.equ.DVrDr:=dvrdr;
  END Init;

PROCEDURE Initialized * (equ-:Equ): BOOLEAN;
  (** Returns if the equ is already initialized *)
  BEGIN
    ASSERT(equ.equ#NIL);
    RETURN equ.equ#NIL
  END Initialized;

PROCEDURE IndexesIn (iZ,iR:LONGINT): BOOLEAN;
  BEGIN
    ASSERT(iZ>=0);
    ASSERT(iZ<=LastZ);
    ASSERT(iR>=0);
    ASSERT(iR<=LastR);
    RETURN TRUE
  END IndexesIn;

(* Following are coefficients of the Te elliptical equation *)

TYPE
  D = RECORD (Func)
    Te,ne,na:FloatArray2Ptr;
  END;

PROCEDURE (VAR d:D) f (z,r:Float): Float;
  VAR
    iZ,iR:LONGINT;
    res:Float;
  BEGIN
    ASSERT(d.Te#NIL);
    ASSERT(d.ne#NIL);
    ASSERT(d.na#NIL);

    iZ:=Area.ZToInd(area,z);
    iR:=Area.RToInd(area,r);
    ASSERT(IndexesIn(iZ,iR));
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;

    res:=Data.ThermalConductivityOfElectrons(d.Te[iZ,iR],d.ne[iZ,iR],d.na[iZ,iR]);
    IF res<ThermalConductivityThreshold THEN
      RETURN ThermalConductivityThreshold
    ELSE
      RETURN res
    END;
  END f;

TYPE
  V = RECORD (Func)
    ne:FloatArray2Ptr;
    v:POINTER TO Func;
  END;

PROCEDURE (VAR d:V) f (z,r:Float): Float;
  VAR
    iZ,iR:LONGINT;
  BEGIN
    ASSERT(d.ne#NIL);
    ASSERT(d.v#NIL);

    iZ:=Area.ZToInd(area,z);
    iR:=Area.RToInd(area,r);
    ASSERT(IndexesIn(iZ,iR));
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;

    RETURN 5/2*k*d.ne[iZ,iR]*d.v.f(z,r);
  END f;

TYPE
  E = RECORD (Func)
    T,Te,ne:FloatArray2Ptr;
  END;

PROCEDURE (VAR d:E) f (z,r:Float): Float;
  VAR
    iZ,iR:LONGINT;
  BEGIN
    ASSERT(d.T#NIL);
    ASSERT(d.Te#NIL);
    ASSERT(d.ne#NIL);

    iZ:=Area.ZToInd(area,z);
    iR:=Area.RToInd(area,r);
    ASSERT(IndexesIn(iZ,iR));
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;

    RETURN -3*Me/Ma*d.ne[iZ,iR]*Data.CollFreq(d.T[iZ,iR],d.Te[iZ,iR])*k;
  END f;

TYPE
  F = RECORD (Func)
    vz,vr:POINTER TO Func;
    P,T,Te,ne:FloatArray2Ptr;
  END;

PROCEDURE DivZ (ne-:FloatArray2; VAR vz:Func; z,r:Float; iZ,iR:LONGINT): Float;
  PROCEDURE func (z,r:Float; iZ,iR:LONGINT): Float;
    BEGIN
      RETURN eps*ne[iZ,iR]*vz.f(z,r);
    END func;
  BEGIN
    RETURN (func(z+area.dx,r,iZ+1,iR) - func(z-area.dx,r,iZ-1,iR))/(2*area.dx)
  END DivZ;

PROCEDURE DivR (ne-:FloatArray2; VAR vr:Func; z,r:Float; iZ,iR:LONGINT): Float;
  PROCEDURE func (z,r:Float; iZ,iR:LONGINT): Float;
    BEGIN
      RETURN eps*ne[iZ,iR]*vr.f(z,ABS(r)) *r;
    END func;
  BEGIN
    RETURN (func(z,r+area.dy,iZ,iR+1) - func(z,r-area.dy,iZ,iR-1))/(2*area.dy) /r
  END DivR;

PROCEDURE (VAR d:F) f (z,r:Float): Float;
  VAR
    iZ,iR:LONGINT;
    T,Te,ne,P:Float;
  BEGIN
    ASSERT(d.T#NIL);
    ASSERT(d.Te#NIL);
    ASSERT(d.ne#NIL);
    ASSERT(d.P#NIL);
    ASSERT(d.vz#NIL);
    ASSERT(d.vr#NIL);

    iZ:=Area.ZToInd(area,z);
    iR:=Area.RToInd(area,r);
    ASSERT(IndexesIn(iZ,iR));
    IF spaceType[iZ,iR]#Geo.In THEN
      RETURN 0
    END;

    T:=d.T[iZ,iR];
    Te:=d.Te[iZ,iR];
    ne:=d.ne[iZ,iR];
    P:=d.P[iZ,iR];
    RETURN   (P - Data.RadiationLoss(Te,ne) + 3*Me/Ma*ne*Data.CollFreq(T,Te)*k*T
           - DivZ( d.ne^, d.vz^, z, r, iZ, iR ) - DivR( d.ne^, d.vr^, z, r, iZ, iR ) ) ;
  END f;

TYPE
  TAlongZ = RECORD (ComFunc.Float1)
    r:Float;
    T:POINTER TO Func;
  END;

PROCEDURE (VAR d:TAlongZ) f (z:Float): Float;
  BEGIN
    ASSERT(d.T#NIL);
    RETURN d.T.f(z,d.r)
  END f;

TYPE
  TAlongR = RECORD (ComFunc.Float1)
    z:Float;
    T:POINTER TO Func;
  END;

PROCEDURE (VAR d:TAlongR) f (r:Float): Float;
  BEGIN
    ASSERT(d.T#NIL);
    RETURN d.T.f(d.z,ABS(r))
  END f;

VAR
  zero,unity:POINTER TO ComFunc.Float1;

PROCEDURE SetCoeffs * (VAR equ:Equ; T,vz,vr,P:POINTER TO Func;
                       Te:FloatArray2Ptr);
  (** Sets coefficients of the Te equation based on the specified Te
      and other parameters. equ should be already initialized. *)
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
    TT,na,ne:FloatArray2Ptr;
    d:D;
    Vz,Vr:V;
    e:E;
    f:F;
    tAlongZ:TAlongZ;
    tAlongR:TAlongR;
  BEGIN
    ASSERT(Initialized(equ));
    ASSERT(T#NIL);
    ASSERT(vz#NIL);
    ASSERT(vr#NIL);
    ASSERT(P#NIL);
    ASSERT(Te#NIL);
    ASSERT(Area.CorrectField(area,Te^));

    Out.String('Te coeffs... ',0);

    TT:=equ.T;
    ne:=equ.ne;
    na:=equ.P; (* equ.P is not used yet *)
    FOR iZ:=0 TO LastZ DO
      z:=Area.IndToZ(area,iZ);
      FOR iR:=0 TO LastR DO
        r:=ABS(Area.IndToR(area,iR));
        TT[iZ,iR]:=T.f(z,r);
        IF Geo.GetSpaceType(z,r)=Geo.In THEN
          Data.NeNa(TT[iZ,iR],Te[iZ,iR],ne[iZ,iR],na[iZ,iR]);
        END;
      END;
    END;

    tAlongZ.T:=T;
    tAlongZ.r:=Rad;
    PEqu.SetRad(equ.equ^,unity^,zero^,tAlongZ);
    tAlongR.z:=MinZ;
    tAlongR.T:=T;
    PEqu.SetMinZ(equ.equ^,unity^,zero^,tAlongR);
    tAlongR.z:=MaxZ;
    tAlongR.T:=T;
    PEqu.SetMaxZ(equ.equ^,unity^,zero^,tAlongR);

    d.Te:=Te;
    d.ne:=ne;
    d.na:=na;
    Vz.ne:=ne;
    Vz.v:=vz;
    Vr.ne:=ne;
    Vr.v:=vr;
    e.T:=TT;
    e.Te:=Te;
    e.ne:=ne;
    PEqu.SetCoeffs(equ.equ^,d,Vz,Vr,e);

    f.T:=TT;
    f.Te:=Te;
    f.ne:=ne;
    f.vz:=vz;
    f.vr:=vr;
    f.P:=equ.P;
    Area.Fill(P^,area,equ.P^);
    PEqu.SetF(equ.equ^,f);
    (* At the end equ.P contains tabulated P, equ.T contains tabulated T *)
  END SetCoeffs;

(* Interface for evolution steps: *)

TYPE
  Step = RECORD (Steps.Step)
    equ:POINTER TO PEqu.Equ;
    old,new:FloatArray2Ptr;
    iterations:LONGINT;
  END;

PROCEDURE (VAR s:Step) Step (dt:Float);
  BEGIN
    ASSERT(s.equ#NIL);
    ASSERT(s.old#NIL);
    ASSERT(s.new#NIL);

    Arrays.Swap2(s.old,s.new);
    PEqu.Step(s.equ,s.old,s.new,dt);
    INC(s.iterations);
  END Step;

PROCEDURE (VAR s:Step) Norm (): Float;
  BEGIN
    ASSERT(s.new#NIL);

    RETURN Arrays.Norm2(s.new^);
  END Norm;

PROCEDURE (VAR s:Step) NormOfDifference (): Float;
  BEGIN
    ASSERT(s.old#NIL);
    ASSERT(s.new#NIL);

    RETURN Arrays.NormOfDifference2(s.old^,s.new^);
  END NormOfDifference;

PROCEDURE ModifyTe (VAR equ:Equ; VAR Te:FloatArray2);
  VAR
    iZ,iR:LONGINT;
  BEGIN
    FOR iZ:=0 TO LastZ DO
      FOR iR:=0 TO LastR DO
        IF spaceType[iZ,iR]#Geo.In THEN
          Te[iZ,iR]:=equ.T[iZ,iR];
        END;
      END;
    END;
  END ModifyTe;

VAR
  FirstCalculation:BOOLEAN;
  InvRangeOfEigenvalues:Float;

PROCEDURE Calculate * (VAR equ:Equ; Te:FloatArray2Ptr);
  (** Solves the equation and returns the result in Te^.
      Te^ is used also as initial conditions of Te.
      Coefficients of the equ should be already set. *)
  VAR
    s:Step;
    ev:Steps.Eigenvalues;
    l_min,l_max,CurrRelErr,InvMaxF:Float;
  BEGIN
    (* At entrance equ.P contains tabulated P, equ.T contains tabulated T: *)

    ASSERT(Initialized(equ));
    ASSERT(Te#NIL);
    ASSERT(Area.CorrectField(area,Te^));

    Out.String('Te... ',0);

    PEqu.StabilizeByMigration(equ.equ^,1);
    ModifyTe(equ,Te^);

    IF FirstCalculation THEN
      PEqu.EigenvaluesExacter(equ.equ^,ev);
      Steps.LogEigenvalues(ev,l_min,l_max);

      l_min:=l_min*l_min_factor;

      InvRangeOfEigenvalues:=l_min/l_max;
      Protocol.Real('  Range of Eigenvalues = ',1/InvRangeOfEigenvalues,'');
      FirstCalculation:=FALSE;
    ELSE
      PEqu.Eigenvalues(equ.equ^,ev);
      Steps.LogEigenvalues(ev,l_min,l_max);
      l_min:=l_max*InvRangeOfEigenvalues;
    END;

    s.equ:=equ.equ;
    s.new:=Te;
    s.old:=equ.ne; (* ne is not used here *)
    s.iterations:=0;

    InvMaxF:=1/Arrays.MaxAbs2(equ.equ.f^);
    CurrRelErr:=PEqu.NormOfRHS(equ.equ^,s.new^)*InvMaxF;
    WHILE CurrRelErr>RelErr DO
      Steps.Log(s,l_min,l_max,RelErr/CurrRelErr);
      CurrRelErr:=PEqu.NormOfRHS(equ.equ^,s.new^)*InvMaxF;
    END;
    IF s.new#Te THEN
      Arrays.Assign2(Te^,s.new^);
    END;

    Out.Int(s.iterations,0);
    Out.String(' iterations',0);
  END Calculate;

PROCEDURE Initialize;
  VAR
    iZ,iR:LONGINT;
    z,r:Float;
  BEGIN
    Area.Init(area,LastZ,LastR,MinZ,MaxZ,Rad);
    Area.Reduced(area,ReducedArea);

    NEW(spaceType,LastZ+1,LastR+1);
    FOR iZ:=0 TO LastZ DO
      z:=Area.IndToZ(area,iZ);
      FOR iR:=0 TO LastR DO
        r:=Area.IndToR(area,iR);
        spaceType[iZ,iR]:=Geo.GetSpaceType(z,r)
      END;
    END;

    zero:=ComFunc.Constant1(0);
    unity:=ComFunc.Constant1(1);
    FirstCalculation:=TRUE;
  END Initialize;

BEGIN
  Initialize;

  Protocol.String('MODULE ',ModuleName,':');
  Protocol.Real('MinZ = ',MinZ/cm,' cm');
  Protocol.Real('MaxZ = ',MaxZ/cm,' cm');
  Protocol.Real('Rad = ',Rad/cm,' cm');
  Protocol.Real('approximate dz = ',dz/cm,' cm');
  Protocol.Real('approximate dr = ',dr/cm,' cm');
  Protocol.Real('dz = ',area.dx/cm,' cm');
  Protocol.Real('dr = ',area.dy/cm,' cm');
  Protocol.Int('LastZ = ',LastZ,'');
  Protocol.Int('LastR = ',LastR,'');
  Protocol.Real('RelErr = ',RelErr,'');
  Protocol.Real('ThermalConductivityThreshold = ',ThermalConductivityThreshold/(PHC.Watt/(PHC.meter*PHC.Kelvin)),' W/(m*K)');
  Protocol.Real('l_min_factor = ',l_min_factor,'');
  Protocol.Real('eps = ',RelErr/PHC.eV,' eV');
  Protocol.String('','','');
END TeEqu.
