(** This module provides data, which are used in the Te equation at
    2-temperature model. *)

<* O2NUMEXT + *>
<* O2EXTENSIONS + *>
MODULE TeEquData;

IMPORT
  Math:=mMath, PHC:=msciPhConst, LiData, Inter2D:=mLinearInterpolation2D,
  LI:=mLinearInterpolation, Out:=msysTextOut, LiGas;

CONST
  pi = Math.pi;
  k = PHC.k;
  Me = PHC.Me;
  h = PHC.h;
  e = PHC.e;

  p = PHC.Atmosphere; (* pressure in the plasma *)

TYPE
  Float = Math.Float;

PROCEDURE sqr (x:Float): Float;
  BEGIN
    RETURN x*x
  END sqr;

PROCEDURE sqrt (x:Float): Float;
  BEGIN
    RETURN Math.sqrt(x)
  END sqrt;

PROCEDURE NeNa * (T,Te:Float; VAR ne,na:Float);
  (** Returns concentration of lectrons ne and neutral atoms na of argon as
      function of gas temperature T and electron temperature Te. *)
  CONST
    eps = 15.7596*PHC.eV; (* ionization energy of Ar *)
  VAR
    L,s,SumT,Ui,Ua:Float;
  BEGIN
    ASSERT(T>0);

    IF Te<263 THEN
      ne:=0;
      na:=p/(k*T);
      RETURN
    END;

    Ui:=LiData.PartitionFunctionOfAr1(Te); (* partition function of Ar ions *)
    Ua:=LiData.PartitionFunctionOfAr0(Te); (* partition function of Ar atoms *)
    L:=(h*h/(2*pi*Me*k*Te))**(1/2); (* lambda-constant at Sahah equation *)
    s:=2/(L*L*L)*Ui/Ua*Math.exp(-eps/(k*Te)); (* RHS in the Sahah equation *)
    SumT:=T+Te;
    ne:=2*p/k/ ( sqrt( sqr(SumT) + 4*p*T/(k*s) ) + SumT ) ;
    na:=ne*ne/s;
  END NeNa;

PROCEDURE CrossSectionNeutral (rad1,rad2:Float): Float;
  BEGIN
    RETURN 2**(1/2)*pi*sqr(rad1+rad2)
  END CrossSectionNeutral;

PROCEDURE CrossSectionCharged (Te,ne:Float): Float;
  (* from Raizer rather than from Lindner *)
  VAR
    lnl:Float;
  BEGIN
    IF ne=0 THEN
      lnl:=0
    ELSE
      lnl:=Math.ln( 3/(2*pi**(1/2)) * (k*Te)**(3/2) / e**3 / ne**(1/2) );
    END;

    RETURN 4*pi/9 * e**4 * lnl / (k*Te)**2 ;
  END CrossSectionCharged;

PROCEDURE ViscosityOfElectrons (Te,ne,na:Float): Float;
  CONST
    Ma = 40*PHC.amu; (* mass of argon *)
    Mr = Me*Ma/(Me+Ma); (* reduced mass of Ar and electron *)
  VAR
    sum:Float;
  BEGIN
    sum:=   Mr**(1/2) *na*CrossSectionNeutral(0,LiData.ArRad(Te))
          + Mr**(1/2) *ne*CrossSectionCharged(Te,ne)
          + (Me/2)**(1/2) *ne*CrossSectionCharged(Te,ne);
    RETURN (k*Te/pi)**(1/2)*ne*Me/sum
  END ViscosityOfElectrons;

PROCEDURE ThermalConductivityOfElectrons * (Te,ne,na:Float): Float;
  (** Returns thermal conductivity of electrons as function of electron
      temperature Te, electron concentration ne, and neutral ar atoms
      concentration na. *)
  BEGIN
    RETURN 15/4*k/Me*ViscosityOfElectrons(Te,ne,na);
  END ThermalConductivityOfElectrons;

VAR
  inp:Inter2D.Interpolation;

PROCEDURE Init;
  TYPE
    arr = ARRAY OF Float;
    arr2 = ARRAY 15,5 OF Float;
  CONST
    Te = arr{6,8,10,12,14};
    rel = arr{1,1.5,2,2.5,3,3.5,4,4.5,5,5.5,6,6.5,7,7.5,8};
    nu = arr2{
              {10.23301,10.78641,11.45631,11.93204,12.30097},
              {10.39806,10.91262,11.5534,12.02913,12.36893},
              {10.5534,11,11.62136,12.09709,12.42718},
              {10.65049,11.06796,11.6699,12.15534,12.48544},
              {10.72816,11.1165,11.69903,12.18447,12.52427},
              {10.78641,11.16505,11.73786,12.2233,12.5534},
              {10.83495,11.19417,11.7767,12.25243,12.59223},
              {10.8835,11.23301,11.81553,12.28155,12.62136},
              {10.92233,11.27184,11.83495,12.30097,12.64078},
              {10.97087,11.31068,11.86408,12.3301,12.66019},
              {11.00971,11.33981,11.8932,12.34951,12.67961},
              {11.03883,11.35922,11.92233,12.36893,12.70874},
              {11.08738,11.39806,11.94175,12.38835,12.71845},
              {11.1068,11.41748,11.96117,12.39806,12.72816},
              {11.13592,11.4466,11.97087,12.41748,12.73786}
             };
  BEGIN
    Inter2D.Init(inp,rel,Te,nu);
  END Init;

PROCEDURE CollFreq * (T,Te:Float): Float;
  (** Returns frequency of collisions of electrons with Ar neutral atoms as
      function of gas temperature T and electron temperature Te,
      as described by Mostaghimi. The collision is calculated as an
      interpolation of the picture presented in his article and book of
      Montaser. *)
  VAR
    rel:Float;
  BEGIN
    rel:=Te/T;
    Te:=Te/1000; (* the tabulation is in kilokelvins *)
    IF rel<inp.x[0] THEN
      rel:=inp.x[0];
    END;
    IF rel>inp.x[LEN(inp.x^)-1] THEN
      rel:=inp.x[LEN(inp.x^)-1];
    END;
    IF Te<inp.y[0] THEN
      Te:=inp.y[0];
    END;
    IF Te>inp.y[LEN(inp.y^)-1] THEN
      Te:=inp.y[LEN(inp.y^)-1];
    END;
    RETURN Math.power(10,Inter2D.Calculate(inp,rel,Te));
  END CollFreq;

PROCEDURE EnergyExchange * (T,Te,ne:Float): Float;
  (** Returnd energy exchange term *)
  CONST
    Me = PHC.Me;
    Ma = 40*PHC.amu;
  BEGIN
    RETURN 3*Me/Ma*ne*CollFreq(T,Te)*k*(Te-T)
  END EnergyExchange;

PROCEDURE RadiationLoss * (Te,ne:Float): Float;
  (** Returns radiation loss of energy *)
  CONST
    meter = PHC.meter;
    C1 = -6.932E-41;
    C2 = 4.753E-39;
    C3 = -5.808E-38;
    C4 = -2.843E-45;
    C5 = 8.629E-47;
    C6 = 6.706E4;
  BEGIN
    IF ne=0 THEN
      RETURN 0
    ELSE
      ne:=ne*meter**3;
      RETURN 4*pi*sqr(ne)*(C1+C2/sqrt(Te)+C3/Te+C4*Te+C5*Te*sqrt(Te))*Math.exp(C6/Te) *PHC.Watt/meter**3;
    END;
  END RadiationLoss;

TYPE
  Arr = ARRAY OF Float;

VAR
  ec:LI.Interpolation; (* electrical conductivity *)

PROCEDURE ElectricalConductivity * (T,Te:Float): Float;
  (** Electrcial conductivity as function of electron temperature *)
  (* from Prolux *)
  VAR
    LastInd:LONGINT;
    res:Float;
    tc1,tc2,ne,na:Float;
  BEGIN
    IF Te<T THEN
      Te:=T;
    END;

    IF Te<=ec.x[0] THEN
      Te:=ec.x[0];
    END;
    LastInd:=LEN(ec.x^)-1;
    IF Te>=ec.x[LastInd] THEN
      res:=ec.y[LastInd]
    ELSE
      res:=LI.Calculate(ec,Te);
    END;

    NeNa(T,Te,ne,na);
    tc1:=ThermalConductivityOfElectrons(Te,ne,na);
    NeNa(Te,Te,ne,na);
    tc2:=ThermalConductivityOfElectrons(Te,ne,na);

    RETURN  res*PHC.Amper/(PHC.Volt*PHC.meter) *tc1/tc2;
  END ElectricalConductivity;

PROCEDURE ExportArProperties * (FileName-:ARRAY OF CHAR);
  (** Exports properties of Ar for COMSOL *)
  CONST
    MinT = 1000;
    MaxT = 11000;
    dT = 100;
    LenT = ENTIER((MaxT-MinT)/dT+0.5)+1;

    Tab = Out.Tab;
    MinRelConc = 1.0E-10;
  VAR
    f:Out.Formatter;
    i:LONGINT;
    n,T:Float;
  BEGIN
    LiGas.Clear;
    LiData.Install;

    Out.OpenFile(f,FileName,Out.Rewrite);

    Out.String(f, 'Temperature(K)'+Tab+'Viscosity(kg/(m*s))'
                 +Tab+'ThermalConductivity(W/(m*K))'
                 +Tab+'SpecificHeatCapacity(J/(kg*K))'
                 +Tab+'ElectricalConductivity(1/(Ohm*m))'
                 +Tab+'RadiationLoss(W/m^3)',0);
    Out.Ln(f);
    FOR i:=0 TO LenT-1 DO
      T:=MinT+(MaxT-MinT)*i/(LenT-1);
      n:=p/(PHC.k*T);
      LiGas.Calculate(n,T);
      Out.Real(f,T,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.Viscosity()*PHC.meter*PHC.second/PHC.kg,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.ThermalConductivity()*PHC.meter*PHC.Kelvin/PHC.Watt,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.SpecificHeatCapacity()*PHC.kg*PHC.Kelvin/PHC.Joule,0);
      Out.Char(f,Tab);
      Out.Real(f,LiGas.ElectricalConductivity()*PHC.meter*PHC.Ohm,0);
      Out.Char(f,Tab);
(*      Out.Real(f,LiGas.RadiationLoss()*PHC.meter**3/PHC.Watt,0); *)
      Out.Real(f,0,0); (* Radiation loss is taken into account in the Te equation rather than in the COMSOL *)
      Out.Ln(f);
    END;

    Out.Close(f);

    LiGas.Clear;
  END ExportArProperties;

BEGIN
  Init;
  LI.Init(ec,Arr{300,400,500,600,700,800,900,1000,1100,1200,1300,1400,1500,1600,1700,1800,1900,2000,2100,2200,2300,2400,2500,2600,2700,2800,2900,3000,3100,3200,3300,3400,3500,3600,3700,3800,3900,4000,4100,4200,4300,4400,4500,4600,4700,4800,4900,5000,5100,5200,5300,5400,5500,5600,5700,5800,5900,6000,6100,6200,6300,6400,6500,6600,6700,6800,6900,7000,7100,7200,7300,7400,7500,7600,7700,7800,7900,8000,8100,8200,8300,8400,8500,8600,8700,8800,8900,9000,9100,9200,9300,9400,9500,9600,9700,9800,9900,10000,10100,10200,10300,10400,10500,10600,10700,10800,10900,11000,11100,11200,11300,11400,11500,11600,11700,11800,11900,12000,12100,12200,12300,12400,12500,12600,12700,12800,12900,13000,13100,13200}
            ,Arr{0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,2.501,4.897,8.847,14.84,23.22,34.14,47.41,62.55,78.86,95.59,112,134.3,157,179.6,201.4,222.2,241.7,260.1,277.5,294.6,311.9,348,388.1,432.7,481.9,535.9,594.9,659,728.2,802.3,881,955.3,1032,1112,1193,1277,1363,1450,1539,1630,1722,1818,1915,2014,2113,2214,2316,2418,2521,2625,2729,2835,2941,3048,3154,3261,3368,3474,3581,3687,3793,3900,4006,4111,4217,4322,4427,4531,4635,4738,4842,4947,5051,5156,5260,5363,5466,5569,5672,5773,5875,5977,6078});
END TeEquData.